// util.cc: various utility methods related to ACIS.
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: util.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (util, "$Id: util.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"
#include "xstring.hxx"
#include "acis/acis.hxx"
#include "acis/silapi.hxx"

// The following section contains a number of components required to
// produce a suitable environment for the construction and analysis of
// solid models using the ACIS solid modeling kernel.
//
// They were originally contributed by Bruce I. Mills <B.I.Mills@cs.cf.ac.uk>
// and later modified by Frank C Langbein <frank@langbein.org>.

// Given a vector, find another that is orthogonal to it.
unit_vector ortho (vector v)
{
  vector v1 (v.y (), -v.x (), 0);
  vector v2 (-v.z (), 0, v.x ());
  vector v3 (0, v.z (), -v.y ());

  if (v1.len () > v2.len ()) {
    if (v1.len () > v3.len ()) 
      return normalise (v1);
    return normalise (v3);
  }
  if (v2.len () > v3.len ()) 
    return normalise (v2);
  return normalise (v3);
}

// Compute the volume of a simplex.
double volume (position a, position b, position c, position d)
{
  vector u (b.x () - a.x (), b.y () - a.y (), b.z () - a.z ());
  vector v (c.x () - a.x (), c.y () - a.y (), c.z () - a.z ());
  vector w (d.x () - a.x (), d.y () - a.y (), d.z () - a.z ());
  vector p = vector (u.y () * v.z () - u.z () * v.y (),
		     u.z () * v.x () - u.x () * v.z (),
		     u.x () * v.y () - u.y () * v.x ());
  return p.x () * w.x () + p.y () * w.y () + p.z () * w.z ();
}


// Various routines for dealing with convex polyhedra, and related issues.
//
// They were originally contributed by Bruce I. Mills <B.I.Mills@cs.cf.ac.uk>
// and later modified by Frank C Langbein <frank@langbein.org>.

// Split a BODY into two pieces, and select one of those pieces.  We
// are given a plane defined as that plane that goes through the
// points p,p1 and p2, this plane cuts space into two parts.  The
// given solid B is also cut into two parts by this plane.  We wish to
// get back the part of the solid wich is on the same side of the
// plane as the point q. For this operation to make sense the four
// points p,p1,p2 and q must form a non degenerate simplex, however at
// the moment no check for this is performed.
void section (position p, position p1, position p2, position q, BODY *b)
{
  // Get two vectors that lie in the plane
  vector v1 = p1 - p;
  vector v2 = p2 - p;

  // Compute a unit normal to the plane
  unit_vector u = normalise (v1 * v2);
  if ((u.x () * (q.x () - p.x ()) +
       u.y () * (q.y () - p.y ()) +
       u.z () * (q.z () - p.z ())) < 0)
    u = -u;

  // Produce a vector in the plane to act as `radius' of the circle
  vector v = (1000.0 * v1) / v1.len ();

  // Construct circular wire loop that lies in the plane through p
  // with normal u, and has p as centre, and contains the point p+u.
  // The radius of this circle just has to be `big enough'.
  EDGE *slicer_e;
  sil_acis_check (api_mk_ed_ellipse (p, u, v, 1, 0, 0, slicer_e),
		  "acis:section", "api_mk_ed_ellipse");

  // Change this wire loop into a flat solid object
  EDGE *array[1] = { slicer_e };
  BODY *slicer_b;
  sil_acis_check (api_make_ewire (1, array, slicer_b),
		  "acis:section", "api_make_ewire");
  ENTITY_LIST slicer_f;
  sil_acis_check (api_cover_wires (slicer_b, *(surface *)NULL, slicer_f),
		  "acis:section", "api_cover_wires");

  // Remove the part of the object on u side of the plane
  sil_acis_check (api_boolean (slicer_b, b, SUBTRACTION),
		  "acis:section", "api_boolean");

  return;
}

// The idea here is to create a flat face that is a convex polygon.
// FIXME: But it still needs a lot of work.
BODY *convex_polygon (position p, position p1, position p2, position *pts,
		      int n)
{
  vector v1 (p1.x () - p.x (), p1.y () - p.y (), p1.z () - p.z ());
  vector v2 (p2.x () - p.x (), p2.y () - p.y (), p2.z () - p.z ());

  // Compute a unit normal to the plane
  unit_vector u (v1.y () * v2.z () - v1.z () * v2.y (),
		 v1.z () * v2.x () - v1.x () * v2.z (),
		 v1.x () * v2.y () - v1.y () * v2.x ());

  //t = u.x()*(q.x()-p.x()) +
  //    u.y()*(q.y()-p.y()) +
  //    u.z()*(q.z()-p.z()) ;
  //t = (t>0?1:-1);
  //u = unit_vector(u.x()*t, u.y()*t, u.z()*t);

  // generate the centroid
  position c = pts[0];
  for(int i = 1; i < n; ++i) {
    c += vector (pts[i].x (),  pts[i].y (), pts[i].z ());
    //position (c.x () + pts[i].x (),
    //c.y () + pts[i].y (),
    //c.z () + pts[i].z ());
  }

  c = position (c.x () / (double)n, c.y () / (double)n, c.z () / (double)n);

  // Produce a vector in the plane to act as `radius' of the circle
  vector v = 40.0 * v / v1.len ();

  // Construct circular wire loop that lies in the plane through p
  // with normal u, and has p as centre, and contains the point p+u
  // The radius of this circle just has to be `big enough'.
  EDGE *slicer_e;
  sil_acis_check (api_mk_ed_ellipse (p, u, v, 1, 0, 0, slicer_e),
		  "sil:convex_polygon", "api_mk_ed_ellipse");

  // Change this wire loop into a flat solid object
  EDGE *array[1] = { slicer_e };
  BODY *slicer_b;
  sil_acis_check (api_make_ewire (1, array, slicer_b),
		  "sil:convex_polygon", "api_make_ewire");
  ENTITY_LIST slicer_f;
  sil_acis_check (api_cover_wires (slicer_b, *(surface *)NULL, slicer_f),
		  "sil_convex_polygon", "api_cover_wires");

  for(int i = 0; i < 1; ++i) {
    position x (pts[i].x () + u.x (),
		pts[i].y () + u.y (),
		pts[i].z () + u.z ());
    printf ("pts[i]   = (%2.1f,%2.1f,%2.1f)\n", pts[i].x (), pts[i].y (),
	    pts[i].z ());
    printf ("pts[i+1] = (%2.1f,%2.1f,%2.1f)\n", pts[i+1].x (), pts[i+1].y (),
	    pts[i+1].z ());
    printf ("x        = (%2.1f,%2.1f,%2.1f)\n", x.x (), x.y (), x.z ());
    printf ("c        = (%2.1f,%2.1f,%2.1f)\n", c.x (), c.y (), c.z ());
    section (pts[i], pts[i+1], x, c, slicer_b);
  }

  return slicer_b;
}

// A rather inefficient convex hull generator, the basic idea is to
// intersect B with the convex hull of the set of points listed in the
// array, this could do with some cleaning up at some time, and
// replacement with a faster algorithm.
void cut_convex (position *point, int n, BODY *B)
{
  if (n < 4) 
    return;

  // generate the centroid
  position c = point[0];
  for (int i = 1; i < n; ++i) 
    c += vector (point[i].x (), point[i].y (), point[i].z ());
  c = position (c.x () / (double)n, c.y () / (double)n, c.z () / (double)n);

  // just go through every three points and check the plane
  for (int i = 0; i < n - 2; ++i) {
    for (int j = i + 1; j < n - 1; ++j) {
      for (int k = j + 1; k < n; ++k) {
	int in = 0;
	int ex = 0;
	for (int a = 0; a < n; ++a) {
	  if (volume (point[a], point[i], point[j], point[k]) > 0.0) 
	    in = 1;
	  if (volume (point[a], point[i], point[j], point[k]) < 0.0)
	    ex = 1;
	}
	if (in && !ex) 
	  section (point[i], point[j], point[k], c, B);
	if (!in && ex)
	  section (point[i], point[j], point[k], c, B);
      }
    }
  }

  return;
}


// Various routines for creating standard objects
//
// They were originally contributed by Bruce I. Mills <B.I.Mills@cs.cf.ac.uk>
// and later modified by Frank C Langbein <frank@langbein.org>.

// Create a Torus
void create_torus (position p,          // centroid of torus
		   unit_vector u,       // direction of axis
		   double a,            // radius of circle of rotation
		   double b,            // radius of cross section
		   BODY **B)
{
  // Initial axis is (0,0,1)
  sil_acis_check (api_solid_torus (position (0, 0, 0), a, b, *B),
		  "ais:create_torus", "api_solid_torus");

  double angle = acos (u.z ());
  if (angle * angle < sil_get_epsilon2 ()) 
    return; // no rotation required

  vector axis (-u.y (), u.x (), 0.0);
  sil_acis_check (api_apply_transf (*B, rotate_transf (angle, axis)),
		  "acis:create_torus", "api_apply_transf");
  sil_acis_check (api_apply_transf (*B, 
				    translate_transf (vector (p.x (), p.y (), 
							      p.z ()))),
		  "acis:create_torus", "api_apply_transf");

  return;
}

// Create a cylinder Cone
//   position of first end of axis
//   position of other end of axis
//   one axis of first face
//   other axis of first face
//   one axis of second face
//   direction
//   BODY pointer      
void create_cylinder (position p,         // centroid of cylinder
		      unit_vector u,      // direction of axis
		      double a,           // length of cylinder
		      double r,           // radius of cylinder
		      BODY **B)
{
  position p1 (p.x () - a * u.x () / 2,
	       p.y () - a * u.y () / 2,
	       p.z () - a * u.z () / 2);
  position p2 (p.x () + a * u.x () / 2,
	       p.y () + a * u.y () / 2,
	       p.z () + a * u.z () / 2);
  sil_acis_check (api_solid_cylinder_cone (p1, p2, r, r, r, NULL, *B),
		  "acis:create_cylinder", "api_solid_cylinder_cone");
  return;
}

// Create a Sphere
void create_sphere(position c,    // centre of sphere
		   position p,    // a point on the sphere
		   BODY **B)
{
  sil_acis_check (api_solid_sphere (c, (p - c) . len (), *B),
		  "acis:create_sphere", "api_solid_sphere");
  return;
}

// Create a Convex Polyhedron
void create_convex (position *point, int n, BODY **B)
{
  if (n < 4)
    return;

  // first figure out a bounding box

  double max_x = point[0].x ();
  double max_y = point[0].y ();
  double max_z = point[0].z ();
  double min_x = point[0].x ();
  double min_y = point[0].y ();
  double min_z = point[0].z ();
 
  for(int i = 1; i < n; ++i) {
    if (max_x < point[i].x ())
      max_x = point[i].x ();
    if (max_y < point[i].y ())
      max_y = point[i].y ();
    if (max_z < point[i].z ())
      max_z = point[i].z ();
    if (min_x > point[i].x ())
      min_x = point[i].x ();
    if (min_y > point[i].y ())
      min_y = point[i].y ();
    if (min_z > point[i].z ())
      min_z = point[i].z ();
  }

  max_x += 10.0;
  max_y += 10.0;
  max_z += 10.0;
  min_x -= 10.0;
  min_y -= 10.0;
  min_z -= 10.0;

  // make the bounding box
  sil_acis_check (api_solid_block (position (min_x, min_y, min_z),
				   position (max_x, max_y, max_z),
				   *B),
		  "acis:create_convex", "api_solid_block");

  // Now just cut it as a convex object
  cut_convex (point, n, *B);

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
