// gui.hxx: GUI class.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: gui.hxx,v 1.36 2001/04/19 15:23:39 scmfcl Exp $

#ifndef GUI_HXX
#define GUI_HXX 1

class App;
class GUI;
class Renderer;
class PluginInterface;
class Symbols;
class Symbol;
class ProgressBar;

typedef enum {
  SIL_NON_INTERACTIVE,
  SIL_INTERACTIVE
} InterMode;

class Plugin {
public:
  Plugin (void (*fnc) (void), const char *n, const char *v,
	  const char *a, const char *c, const char *d, const char *cc,
	  const char *m) {
    f.noninteractive = fnc;
    intrct = SIL_NON_INTERACTIVE;
    nm = n;
    vrsn = v;
    athr = a;
    cpyrght = c;
    dsc = d;
    cls = cc;
    mdls = m;
  }
  Plugin (void (*fnc) (SimpleList<Symbol const> *), const char *n,
	  const char *v, const char *a, const char *c, const char *d,
	  const char *cc, const char *m) {
    f.interactive = fnc;
    intrct = SIL_INTERACTIVE;
    nm = n;
    vrsn = v;
    athr = a;
    cpyrght = c;
    dsc = d;
    cls = cc;
    mdls = m;
  }
  void call () {
    assert (interactive () == SIL_NON_INTERACTIVE);
    (*f.noninteractive) ();
    return;
  }
  void call (SimpleList<Symbol const> *s) {
    assert (interactive () == SIL_INTERACTIVE);
    (*f.interactive) (s);
    return;
  }
  const char *name () const {
    return nm;
  }
  const char *version () const {
    return vrsn;
  }
  const char *author () const {
    return athr;
  }
  const char *copyright () const {
    return cpyrght;
  }
  const char *description () const {
    return dsc;
  }
  const InterMode interactive () const {
    return intrct;
  }
  const char *module_class () const {
    return cls;
  }
  const char *modules () const {
    return mdls;
  }
private:
  union {
    void (*interactive) (SimpleList<Symbol const> *);
    void (*noninteractive) (void);
  } f;
  const char *nm;
  const char *vrsn;
  const char *athr;
  const char *cpyrght;
  const char *dsc;
  InterMode intrct;
  const char *cls;
  const char *mdls;
};

class ToolWindow : public Gtk::Window {
public:
  ToolWindow ();
  ~ToolWindow ();
  void update ();
  void rescan ();
  void closeit ();
  void destroyit ();
private:
  void make_sym_list (Gtk::CList *&list, char const *n);
  void make_buttons (Gtk::Table *&ti, Gtk::Table *&tni, char const *n);
  void scan_plugins (const char *p);
  void register_plugin (const char *n);
  void call (Plugin *p);
  Gtk::VBox vbox;
  Gtk::HBox hbox;
  Gtk::Notebook nb;
  Gtk::Button b_rescan;
  Gtk::Button b_close;
  Gtk::CList **sym_list;
  Gtk::Table **i_t;
  Gtk::Table **ni_t;
}; 

class OGLDisplay : public Gtk::GLArea
{
public:
  OGLDisplay ();
  ~OGLDisplay ();
protected:
  Renderer *renderer;
  friend GUI;
  friend PluginInterface;
  friend Symbols;
private:
  gdouble mouse_last_x;
  gdouble mouse_last_y;
  void realize_impl();
  gint motion_notify_event_impl (GdkEventMotion*);
  gint button_press_event_impl (GdkEventButton*);
  gint button_release_event_impl (GdkEventButton*);
  gint expose_event_impl (GdkEventExpose* expose);
  gint configure_event_impl (GdkEventConfigure *event);
};

class GUI : public Gtk::Window
{
public:
  GUI ();
  ~GUI();
  void status_msg (string const &text);
  void status_msg (string const &name, string const &text);
  void status_msg_warn (string const &name, string const &text);
  void status_msg_error (string const &name, string const &text);
  void progress_bar (bool flag, bool cancel_flag);
  void progress_bar_percentage (gfloat p);
  void progress_bar_text (string const &text);
  bool progress_cancel ();
  void close_tool_window ();
  gint delete_event_impl (GdkEventAny*) {
    Gtk::Main::quit(); 
    return 0;
  }
protected:
  bool status_available ();
  void update_tool_window ();
  void closed_tool_window () {
    tool_win = NULL;
  }
  void refacet ();
  void redraw ();
  void refresh ();
  void pcancel_handler ();
  bool win_visible;
  Gtk::VBox vbox;
  Gtk::Toolbar toolbar;
  OGLDisplay display;
  Gtk::Text status_text;
  Gtk::HBox status_scroll;
  Gtk::VScrollbar status_scrollbar;
  Gdk_Font status_font;
  Gtk::HBox hb_progress;
  Gtk::ProgressBar pb_progress;
  Gtk::Button b_pcancel;
  friend PluginInterface;
  friend App;
  friend Symbols;
  friend ToolWindow;
  friend ProgressBar;
private:
  ToolWindow *tool_win;
  Gtk::Pixmap *new_pix;
  Gtk::Pixmap *open_pix;
  Gtk::Pixmap *save_pix;
  Gtk::Pixmap *config_pix;
  Gtk::Pixmap *init_view_pix;
  Gtk::Pixmap *plugin_pix;
  Gtk::Pixmap *exit_pix;
  Gtk::FileSelection *file_sel;
  Gtk::CheckButton *file_replace_cb;
  Gtk::CheckButton *file_sil_cb;
  Gtk::Window *configwin;
  Gtk::Button *config_ok_button;
  Gtk::Button *config_cancel_button;
  Gtk::Button *config_apply_button;
  Gtk::HBox *config_hbox;
  Gtk::Table *config_table;
  Gtk::CheckButton *cb_persp;
  Gtk::CheckButton *cb_ortho;
  Gtk::Label *label_proj[6];
  Gtk::SpinButton *sb_proj[6];
  Gtk::Adjustment *adj_proj[8];
  Gtk::CheckButton *cb_local_viewer;
  Gtk::CheckButton *cb_solid;
  Gtk::CheckButton *cb_sub_parts;
  Gtk::Label *label_ref[6];
  Gtk::SpinButton *sb_ref[6];
  Gtk::Adjustment *adj_ref[6];
  Gtk::RadioButton *rb_tri_ref[6];
  Gtk::RadioButton *rb_grid_ref[3];
  Gtk::RadioButton *rb_adj_ref[3];
  Gtk::Label *label_gl[2];
  Gtk::SpinButton *sb_gl[2];
  Gtk::Adjustment *adj_gl[2];
  int mi_n;
  Gtk::MenuItem **mi_filetypes;
  Gtk::Menu *m_filetypes;
  Gtk::OptionMenu *om_filetypes;
  int filetypes_id;
  void sel_filetypes (int i);
  void config_ok ();
  void config_cancel ();
  void config_apply ();
  gint config_close (GdkEventAny *e);
  void config_persp ();
  void config_ortho ();
  bool replace_sym;
  bool sil_file_ext;
  bool progress_flag;
  bool progress_cancel_flag;
  char *last_file_dir;
  void open_ok();
  void open_cancel();
  gint open_close(GdkEventAny *e);
  void save_ok();
  void save_cancel();
  gint save_close(GdkEventAny *e);
  void newlist ();
  void open ();
  void save ();
  void config ();
  void init_view ();
  void plugin ();
  void win_dest ();
  void exit ();
};

#endif // !GUI_HXX

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
