// interface.cc: Interface implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: interface.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (interface, "$Id: interface.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "interface.hxx"
#include "renderer.hxx"
#include "gui.hxx"
#include "symbols.hxx"
#include "app.hxx"
#include "modules.hxx"
#include "xmath.hxx"

extern char *tmp_dir;

PluginInterface *_sil_interface = NULL;

PluginInterface::PluginInterface ()
{
  app = App::instance ();
  if (_sil_interface) {
    cerr << app->name () << ": FATAL: tried to define two interfaces\n";
    exit (1);
  }
  _sil_interface = this;
}

PluginInterface::~PluginInterface ()
{
  _sil_interface = NULL;
  app = NULL;
}

bool PluginInterface::add_symbol (char *n, void *d, char *t, bool v)
{
  return app->symbols->add (n, d, t, v);
}

const Symbol *PluginInterface::find_symbol (const char *n)
{
  return app->symbols->find (n);
}

void PluginInterface::del_symbol (const char *n)
{
  assert (n);
  app->symbols->del (n);
  return;
}

const Symbol *PluginInterface::first_symbol ()
{
  return app->symbols->first ();
}

const Symbol *PluginInterface::next_symbol ()
{
  return app->symbols->next ();
}

void PluginInterface::clear_symbols ()
{
  app->symbols->clear ();
  return;
}

void PluginInterface::symbols_updated () const
{
  app->symbols->updated ();
  return;
}

void *PluginInterface::symbol_data (const Symbol *s) const
{
  return s->data ();
}

const char *PluginInterface::symbol_type (const Symbol *s) const
{
  return s->type ();
}

const char *PluginInterface::symbol_name (const Symbol *s) const
{
  return s->name ();
}

bool PluginInterface::symbol_is (const Symbol *s, const char *type) const
{
  return s->is (type);
}

bool PluginInterface::symbol_is_visible (const Symbol *s) const 
{
  return s->is_visible ();
}

void PluginInterface::show_symbol (Symbol *s)
{
  s->show ();
  return;
}

void PluginInterface::hide_symbol (Symbol *s)
{
  s->hide ();
  return;
}

bool PluginInterface::draw_sub_parts () const
{
  return app->gui->display.renderer->draw_sub_parts();
}

void PluginInterface::set_draw_sub_parts (bool f)
{
  app->gui->display.renderer->set_draw_sub_parts (f);
  return;
}

bool PluginInterface::draw_solid () const
{
  return app->gui->display.renderer->solid ();
}

void PluginInterface::render_box (double &a, double &b, double &c,
				  double &d, double &e, double &f) const
{
  app->gui->display.renderer->render_box (a, b, c, d, e, f);
  return;
}

void PluginInterface::refinement (double &surf_tol,
				  double &normal_tol,
				  SiL_TriangMode &t_mode,
				  SiL_GridMode &g_mode,
				  int &max_l,
				  int &min_l,
				  double &max_ll,
				  double &ar,
				  SiL_AdjustMode &am) const
{
  app->gui->display.renderer->refinement (surf_tol, normal_tol,
					  t_mode, g_mode,
					  max_l, min_l, max_ll,
					  ar, am);
  return;
}

Module *PluginInterface::module (char *name) const
{
  return app->modules->module (name);
}

void PluginInterface::refacet ()
{
  app->gui->refacet ();
  return;
}

void PluginInterface::redraw ()
{
  app->gui->redraw ();
  return;
}

void PluginInterface::refresh ()
{
  app->gui->refresh ();
  return;
}

void PluginInterface::hilite (bool hi, char const *name, void *ref)
{
  app->symbols->hilite (hi, name, ref);
  return;
}

bool PluginInterface::is_hilited (const Symbol *s, const void *r)
{
  return s->is_hilited (r);
}

void PluginInterface::msg (string const &text)
{
  if (app && app->gui && app->gui->status_available ())
    app->gui->status_msg (text);
#ifndef DEBUG
  else
#endif !DEBUG
    cout << app->name () << ": " << text << endl;
  return;
}

void PluginInterface::msg (string const &name, string const &text)
{
  if (app && app->gui && app->gui->status_available ())
    app->gui->status_msg (name, text);
#ifndef DEBUG
  else
#endif !DEBUG
    cout << app->name () << ": " << name << ": " << text << endl;
  return;
}

void PluginInterface::msg_warn (string const &name, string const &text) 
{
  if (app && app->gui && app->gui->status_available ())
    app->gui->status_msg_warn (name, text);
#ifndef DEBUG
  else
#endif !DEBUG
    cerr << app->name () << ": " << name << ": " << text << endl;
  return;
}

void PluginInterface::msg_error (string const &name, string const &text)
{
  if (app && app->gui && app->gui->status_available ())
    app->gui->status_msg_error (name, text);
#ifndef DEBUG
  else
#endif // !DEBUG
    cerr << app->name () << ": " << name << ": " << text << endl;
  return;
}

unsigned long PluginInterface::tmp_name ()
{
  return app->tmp_name ();
}

string const PluginInterface::tmp_filename () const
{
  char *s = new char[strlen (tmp_dir) + 12];
  strcpy (s, tmp_dir);
  strcat (s, "/sil_XXXXXX");
  string str = static_cast<string> (mktemp (s));
  delete s;
  return str;
}

const double PluginInterface::epsilon () const
{
  return EPSILON;
}

const double PluginInterface::epsilon2 () const
{
  return EPSILON2;
}

void PluginInterface::start_time (unsigned long t)
{
  assert (t < SIL_TIMERS);
#if defined (HAVE_GETTIMEOFDAY) && defined (HAVE_TIMEVAL)
  // Get time and remember it.
  struct timezone dtz;
  gettimeofday (&time_before[t], &dtz);
#endif // defined (HAVE_GETTIMEOFDAY) && defined (HAVE_TIMEVAL)
  return;
}

void PluginInterface::measure_time (unsigned long &secs,
				    unsigned int &msecs,
				    unsigned long t) const
{
  assert (t < SIL_TIMERS);
#if defined (HAVE_GETTIMEOFDAY) && defined (HAVE_TIMEVAL)
  // Get time.
  struct timezone dtz;
  struct timeval after, diff;
  gettimeofday (&after, &dtz);
  // Measure time.
  diff_timeval (diff, time_before[t], after);
  timeval_to_secs (&diff, secs, msecs);
#else // defined (HAVE_GETTIMEOFDAY) && defined (HAVE_TIMEVAL)
  secs = 0;
  msecs = 0;
#endif // defined (HAVE_GETTIMEOFDAY) && defined (HAVE_TIMEVAL)
  return;
}

#ifdef HAVE_TIMEVAL

void PluginInterface::diff_timeval (struct timeval &d,
				    struct timeval const &t1,
				    struct timeval const &t2) const
{
  // Measure difference between timevals: d = t2 - t1.
  d.tv_sec = t2.tv_sec - t1.tv_sec;
  d.tv_usec = t2.tv_usec - t1.tv_usec;
  if (d.tv_usec < 0) {
    d.tv_usec += 1000000;
    d.tv_sec -= 1;
    if (d.tv_sec < 0) { // Sanity check.
      d.tv_sec = 0;
      d.tv_usec = 0;
    }
  }
  return;
}

void PluginInterface::timeval_to_secs (struct timeval const *tv,
				       unsigned long &secs,
				       unsigned int &msecs) const
{
  // Convert a timeval structure into seconds and thousands of a
  // second using rounding, not truncation.
  secs = tv->tv_sec;
  msecs = tv->tv_usec % 1000000; // Basically that's a NOP.
  unsigned int rest = msecs % 1000;
  msecs = (msecs * 1000) / 1000000;
  if (rest >= 500)
    msecs += 1;
  // Sanity check.
  while (msecs >= 1000) {
    secs += 1;
    msecs -= 1000;
  }
  return;
}

#endif // HAVE_TIMEVAL

void PluginInterface::progress_bar (bool flag, bool cancel_flag)
{
  pb_b = 0.0;
  pb_r = 1.0;
  if (app && app->gui && app->gui->status_available ())
    app->gui->progress_bar (flag, cancel_flag);
  return;
}

void PluginInterface::progress_bar_assign (gfloat base, gfloat range)
{
  pb_b = base;
  pb_r = range;
  if (app && app->gui && app->gui->status_available ())
    app->gui->progress_bar_percentage (base);
  return;
}

void PluginInterface::progress_bar_percentage (gfloat p)
{
  if (app && app->gui && app->gui->status_available ())
    app->gui->progress_bar_percentage (pb_b + pb_r * p);
  return;
}

void PluginInterface::progress_bar_text (string const &str)
{
  if (app && app->gui && app->gui->status_available ())
    app->gui->progress_bar_text (str);
#ifndef DEBUG
  else
#endif // !DEBUG
    cerr << app->name () << ": " << str << endl;
  return;
}

bool PluginInterface::progress_cancel ()
{
  return app->gui->progress_cancel ();
}

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
