// analyse.cc: Analyse a SiL model
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse, "$Id: analyse.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "kernel/kernel.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"

char *one_id = "";
char *inch_id = "Inch";
char *pi_id = "Pi";
char *deg_id = "Deg";
char *tan_id = "arctan";

string const Aspect::tid = "Aspect";
string const Regularity::tid = "Regularity";
string const SubRegularity::tid = "SubRegularity";
string const Cluster::tid = "Cluster";
string const SpecialValue::tid = "SpecialValue";
string const SpecialRatio::tid = "SpecialRatio";
string const ParallelDirection::tid = "ParallelDirection";
string const OrthogonalDirection::tid = "OrthogonalDirection";
string const DirectionAngle::tid = "DirectionAngle";
string const PlanarAngleRegular::tid = "PlanarAngleRegular";
string const ConicalAngleRegular::tid = "ConicalAngleRegular";
string const PointLine::tid = "PointLine";
string const RegularPointPosition::tid = "RegularPointPosition";
string const RegularPointCircle::tid = "RegularPointCircle";
string const AltGeometry::tid = "AltGeometry";
string const Analyse::tid = "Analyse";

string const Regularity::description () const
{
  string desc = name () + "(" + this->describe ();
  if (sub_regs) {
    desc += " = ";
    Regularity *sr = sub_regs->first ();
    while (sr) {
      desc += sr->description ();
      sr = sub_regs->next ();
      if (sr)
	desc += "; ";
    }
  }
  if (!isnan (tol))
    desc += ", " + as_string (tol);
  desc += ")";
  return desc;
}

string const Cluster::name () const
{
  return val->type_id ();
}

string const Cluster::describe () const
{
  string str = "";
  if (num_aspcts != 0) {
    str += as_string (num_aspcts);
    if (aspcts && sub_c) {
      str += " [" + as_string (aspcts->length ()) + "/" + 
	as_string (sub_c->length ()) +"]";
    } else if (sub_c) {
      str += " [0/" + as_string (sub_c->length ()) + "]";
    }
    str += ", ";
  }
  str += val->description ();
  return str;
}

void Cluster::hilite (bool hi) const
{
  if (aspcts) {
    Aspect *a = aspcts->first ();
    while (a) {
      Entity *e = a->owners () . first ();
      while (e) {
	val->hilite (hi, analyse () -> hilite_name (), e);
	e = a->owners () . next ();
      }
      a = aspcts->next ();
    }
  }
  if (sub_c) {
    Cluster *c = sub_c->first ();
    while (c) {
      c->hilite (hi);
      c = sub_c->next ();
    }
  }
  return;
}

void Cluster::combine (Cluster *c, Feature *v)
{
  delete val;
  val = v;
  num_aspcts += c->num_aspcts;
  if (c->aspects ()) {
    if (!aspcts)
      aspcts = new SimpleList<Aspect>;
    aspcts->combine (c->aspects ());
  }
  if (c->sub_c) {
    if (!sub_c)
      sub_c = new SimpleList<Cluster>;
    Cluster *sc = c->sub_c->unlink_first ();
    while (sc) {
      sub_c->add_first (sc);
      sc->parent_c = this;
      sc = c->sub_c->unlink_first ();
    }
  }
  return;
}

void Cluster::combine (Cluster *c)
{
  num_aspcts += c->num_aspcts;
  if (c->aspects ()) {
    if (!aspcts)
      aspcts = new SimpleList<Aspect>;
    aspcts->combine (c->aspects ());
  }
  if (c->sub_c) {
    if (!sub_c)
      sub_c = new SimpleList<Cluster>;
    Cluster *sc = c->sub_c->unlink_first ();
    while (sc) {
      sub_c->add_first (sc);
      sc->parent_c = this;
      sc = c->sub_c->unlink_first ();
    }
  }
  return;
}

void Cluster::collapse (Real t0, Real t1)
{
  if (!sub_c || tolerance () <= t0)
    return;
  
  if (tolerance () > t1) {
    Cluster *sc = sub_c->first ();
    while (sc) {
      sc->collapse (t0, t1);
      sc = sub_c->next ();
    }
    return;
  }

  // Collapse clusters between t0 and t1.
  SimpleList<Cluster> *sub_n = new SimpleList<Cluster>;
  Cluster *sc = sub_c->unlink_first ();
  while (sc) {
    if (sc->tolerance () > t0) {
      if (sc->aspects ()) {
	if (!aspcts)
	  aspcts = new SimpleList<Aspect>;
	Aspect *a = sc->aspects () -> unlink_first ();
	while (a) {
	  aspcts->add_first (a);
	  a = sc->aspects () -> unlink_first ();
	}
      }
      if (sc->sub_c && sc->sub_c->length () > 0) {
	Cluster *cc = sc->sub_c->first ();
	while (cc) {
	  cc->parent_cluster () = this;
	  cc = sc->sub_c->next ();
	}
	sub_c->combine (sc->sub_c);
      }
      delete sc;
    } else {
      sub_n->add_first (sc);
    }
    sc = sub_c->unlink_first ();
  }
  delete sub_c;
  if (sub_n->length () > 0) {
    sub_c = sub_n;
  }  else {
    sub_c = NULL;
    delete sub_n;
  }
  return;
}

bool Cluster::collapse ()
{
  if (parent_c) {
    // Combine with parent.
    parent_c->sub_c->del (this);
    if (sub_c) {
      Cluster *sc = sub_c->first ();
      while (sc) {
	sc->parent_cluster () = parent_c;
	sc = sub_c->next ();
      }
      parent_c->sub_c->combine (sub_c);
    }
    if (aspcts) {
      if (!parent_c->aspcts)
	parent_c->aspcts = new SimpleList<Aspect>;
      parent_c->aspcts->combine (aspcts);
    }
    if (parent_c->sub_c->length () == 0) {
      delete parent_c->sub_c;
      parent_c->sub_c = NULL;
    }
    return true;
  }
  return false;
}

void Cluster::get_owners (Entity **owners, unsigned long &n)
{
  // Collect the owners from all aspects in this cluster and its sub
  // clusters.
  assert (owners);
  if (aspects ()){
    Aspect *a = aspects () -> first ();
    while (a){
      Entity *e = a->owners () . first ();
      while (e) {
        bool found = false;
        for (unsigned long l = 0; l < n; ++l) {
          if (owners[l] == e) {
            found = true;
            break;
          }
        }
        if (!found)
          owners[n++] = e;
        e = a->owners () . next ();
      }
      a = aspects () -> next ();
    }
  }
  if (sub_clusters ()) {
    Cluster *c = sub_clusters () -> first ();
    while (c) {
      c->get_owners (owners, n);
      c = sub_clusters () -> next ();
    }
  }
  return;
}

void Cluster::get_owners (SimpleList<Entity> *owners)
{
  // Collect the owners from all aspects in this cluster and its sub
  // clusters.
  assert (owners);
  if (aspects ()) {
    Aspect *a = aspects () -> first ();
    while (a) {
      Entity *e = a->owners () . first ();
      while (e) {
	owners->add_first_check (e);
        e = a->owners () . next ();
      }
      a = aspects () -> next ();
    }
  }
  if (sub_clusters ()) {
    Cluster *c = sub_clusters () -> first ();
    while (c) {
      c->get_owners (owners);
      c = sub_clusters () -> next ();
    }
  }
  return;
}

void Cluster::get_aspects (SimpleList<Aspect> *asp)
{
  // Collect the aspects in this cluster and its sub clusters.
  assert (asp);
  if (aspects ()) {
    Aspect *a = aspects () -> first ();
    while (a) {
      asp->add_first (a);
      a = aspects () -> next ();
    }
  }
  if (sub_clusters ()) {
    Cluster *c = sub_clusters () -> first ();
    while (c) {
      c->get_aspects (asp);
      c = sub_clusters () -> next ();
    }
  }
  return;
}

void Cluster::get_aspects (List<Aspect> *asp)
{
  // Collect the aspects in this cluster and its sub clusters.
  assert (asp);
  if (aspects ()) {
    Aspect *a = aspects () -> first ();
    while (a) {
      asp->add_first (a);
      a = aspects () -> next ();
    }
  }
  if (sub_clusters ()) {
    Cluster *c = sub_clusters () -> first ();
    while (c) {
      c->get_aspects (asp);
      c = sub_clusters () -> next ();
    }
  }
  return;
}

void Cluster::get_all_clusters (List<Cluster> *clusters)
{
  assert (clusters);
  clusters->add_first (this);
  if (sub_c) {
    Cluster *c = sub_c->first ();
    while (c) {
      c->get_all_clusters (clusters);
      c = sub_c->next ();
    }
  }
  return;
}

void Cluster::get_all_base_clusters (List<Cluster> *clusters)
{
  assert (clusters);
  if (aspcts && aspcts->length () > 0)
    clusters->add_first (this);
  if (sub_c) {
    Cluster *c = sub_c->first ();
    while (c) {
      c->get_all_base_clusters (clusters);
      c = sub_c->next ();
    }
  }
  return;
}

bool Cluster::cluster_is_member (Cluster *c)
{
  if (this == c)
    return true;
  if (sub_c) {
    Cluster *sc = sub_c->first ();
    while (sc) {
      if (sc->cluster_is_member (c))
	return true;
      sc = sub_c->next ();
    }
  }
  return false;
}

string const SpecialValue::describe () const 
{
  string str = "";
  bool prev = false;
  for (unsigned long l = 0; l < b_nums; ++l) {
    if (rats[l]) {
      if (prev) 
	str += "; ";
      str += static_cast<string> (b_id[l]) + ": ";
      prev = true;
      Ratio *r = rats[l]->first ();
      while (r) {
	str += r->description ();
	r = rats[l]->next ();
	if (r)
	  str += ", ";
      }
    }
  }
  return str;
}

string const SpecialRatio::describe () const
{
  assert (dynamic_cast<Scalar *> (c1->value ()) &&
	  dynamic_cast<Scalar *> (c2->value ()));
  string str = as_string (static_cast<Scalar *> (c1->value ()) -> value ()) +
    " / " + as_string (static_cast<Scalar *> (c2->value ()) -> value ()) +
    " = ";
  Ratio *r = rats->first ();
  while (r) {
    str += r->description ();
    r = rats->next ();
    if (r)
      str += ", ";
  }
  return str;
}

bool SpecialValue::is_equal (SubRegularity const *s) const
{
  if (s->is (this)) {
    SpecialValue *r = (SpecialValue *)s;
    if (r->b_id != b_id)
      return false;
    for (unsigned long l = 0; l < b_nums; ++l) {
      if (rats[l] && r->rats[l]) {
	if (rats[l]->length () != r->rats[l]->length ()) 
	  return false;
	Ratio *r1 = rats[l]->first ();
	while (r1) {
	  bool f = false;
	  Ratio *r2 = r->rats[l]->first ();
	  while (r2) {
	    if (r1->is_equal (r2)) {
	      f = true;
	      break;
	    }
	    r2 = r->rats[l]->next ();
	  }
	  if (!f)
	    return false;
	  r1 = rats[l]->next ();
	}
      } else if (rats[l] || r->rats[l]) {
	return false;
      }
    }
    return true;
  }
  return false;
}

string const ParallelDirection::describe () const
{
  return as_string  (*static_cast<Direction const *> (dir->value ()));
}

string const OrthogonalDirection::describe () const
{
  return as_string (*(static_cast<Direction *> (dir->value ())));
}

string const DirectionAngle::describe () const
{
  return as_string (angl / PI * 180) + " deg: " +
    as_string (*static_cast<Direction*> (dir1->value ())) + " - " + 
    as_string (*static_cast<Direction*> (dir2->value ()));
}

void DirectionAngle::hilite (bool hi) const 
{
  dir1->hilite (hi);
  dir2->hilite (hi);
  return;
}

string const PlanarAngleRegular::describe () const
{
  string str = as_string (dirs->length ()) + ", " +
    as_string (plane_normal ()) + ": " + 
    as_string (angl_n) + " / " + as_string (180.0 / (Real)angl_n) + " deg";
  if (fu)
    str += " (all)";
  else if (ra && ra->length () > 0) {
    str += " (";
    unsigned long *l = ra->first ();
    while (l) {
      str += as_string (*l);
      l = ra->next ();
      if (l) {
	str += ",";
      }
    }
    str += ")";
  }
  return str;
}

void PlanarAngleRegular::hilite (bool hi) const 
{
  Cluster *c = dirs->first ();
  while (c) {
    c->hilite (hi);
    c = dirs->next ();
  }
  return;
}

string const ConicalAngleRegular::describe () const
{
  string str = as_string (dirs->length ()) + ", " +
    as_string (*cone_dir) + " / " +
    as_string (cone_angl / PI * 180.0) + " deg: " +
    as_string (an) + " / " + as_string (360.0 / an) + " deg";
  if (ortho)
    str += " (orthogonal)";
  else if (f)
    str += " (all)";
  else if (ra && ra->length () > 0) {
    str += " (";
    unsigned long *l = ra->first ();
    while (l) {
      str += as_string (*l);
      l = ra->next ();
      if (l) {
	str += ",";
      }
    }
    str += ")";
  }
  return str;
}

void ConicalAngleRegular::hilite (bool hi) const
{
  Cluster *c = dirs->first ();
  while (c) {
    c->hilite (hi);
    c = dirs->next ();
  }
  return;
}

string const PointLine::describe () const
{
  return as_string (pts->length ()) + ", " + as_string (pos) + " / " +
    as_string (dir);
}

void PointLine::hilite (bool hi) const
{
  Cluster *c = pts->first ();
  while (c) {
    c->hilite (hi);
    c = pts->next ();
  }
  return;
}

string const RegularPointPosition::describe () const
{
  string str = as_string (pts->length ()) + ", " + as_string (p) + " - " + 
    as_string (dir1) + " / " +  as_string (d1);
  if (dir2.norm_sq () > sil_get_epsilon ())
    str += " - " + as_string (dir2) + " / " + as_string (d2);
  return str;
}

void RegularPointPosition::hilite (bool hi) const
{
  Cluster *c = pts->first ();
  while (c) {
    c->hilite (hi);
    c = pts->next ();
  }
  return;
}

bool RegularPointPosition::match_pts (RegularPointPosition *grid, Real &tol)
{
  // Check if it can be combined be checking if the points are within
  // the grid tolerance.
  Real t = tolerance ();
  Cluster *c = grid->pts->first ();
  while (c) {
    if (!pts->is_member (c)) {
      if (c->value () -> is (SIL_ALIGNEDAXES_TYPE)) {
	Vector3 pos = *static_cast<AlignedAxes*> (c->value ()) - p;
	Real f1 = pos % dir1;
	f1 = rint (f1 / d1) * d1 - f1;
	Real f2 = pos % dir2;
	f2 = rint (f2 / d2) * d2 - f2;
	f1 = sqrt (f1 * f1 + f2 * f2);
	if (f1 > tol) {
	  return false;
	} else if (f1 > t) {
	  t = f1;
	}
      } else {
	sil_msg_warn ("RegularPositionType", "Unkown point cluster type");
      }
    }
    c = grid->pts->next ();
  }
  tol = t;
  return true;
}

bool RegularPointPosition::match_dirs (RegularPointPosition *grid, Real tol)
{
  // Check if directions / steps match.

  Vector3 d = grid->dir1 * grid->d1;
  Real f1 = d % dir1;
  f1 = rint (f1 / d1) * d1 - f1;
  Real f2 = d % dir2;
  f2 = rint (f2 / d2) * d2 - f2;
  f1 = sqrt (f1 * f1 + f2 * f2);
  
  if (f1 < tol) {
    d = grid->dir2 * grid->d2;
    f1 = d % dir1;
    f1 = rint (f1 / d1) * d1 - f1;
    f2 = d % dir2;
    f2 = rint (f2 / d2) * d2 - f2;
    f1 = sqrt (f1 * f1 + f2 * f2);
    return (f1 < tol);
  }

  return false;
}

bool RegularPointPosition::match_dir (RegularPointPosition *grid, Real tol)
{
  // Check if first direction of grid matches.
  
  Vector3 d = grid->dir1 * grid->d1;
  Real f1 = d % dir1;
  f1 = rint (f1 / d1) * d1 - f1;
  Real f2 = d % dir2;
  f2 = rint (f2 / d2) * d2 - f2;
  f1 = sqrt (f1 * f1 + f2 * f2);

  return (f1 < tol);
}

void RegularPointPosition::combine (RegularPointPosition *grid)
{
  // Add the points of the grid.
  Cluster *c = grid->pts -> first ();
  while (c) {
    pts->add_first_check (c);
    c = grid->pts->next ();
  }
  return;
}

string const RegularPointCircle::describe () const
{
  string str = as_string (pts->length ()) + ", " +
    as_string (c) + ", " +
    as_string (r) + ", " + 
    as_string (an) + " / " +
    as_string (360.0 / an) + " deg";
  if (f)
    str += " (all)";
  else if (ra && ra->length () > 0) {
    str += " (";
    unsigned long *l = ra->first ();
    while (l) {
      str += as_string (*l);
      l = ra->next ();
      if (l) {
	str += ",";
      }
    }
    str += ")";
  }
  str += ", " + as_string (norm);
  return str;
}

void RegularPointCircle::hilite (bool hi) const
{
  Cluster *e = pts->first ();
  while (e) {
    e->hilite (hi);
    e = pts->next ();
  }
  return;
}

string const AltGeometry::describe () const
{
  return c->geometry () -> type_id () + " -> " + ag->type_id ();
}

void AltGeometry::hilite (bool hi) const
{
  if (c->reference ())
    sil_hilite (hi, analyse () -> hilite_name () . c_str (), c->reference ());
  else
    sil_hilite (hi, analyse () -> hilite_name () . c_str (), c);
  return;
}

Analyse::Analyse (Solid *s, string const &n, string const &hn,
		  Real t_eq, Real t_len_dist, Real t_angle_dist,
		  Real t_ratio_approx, Real t_ratio_min, Real t_loop_dist,
		  unsigned long reg_angle_max_n,
		  unsigned long reg_fourier_coeff,
		  unsigned long reg_p_max_p, unsigned long reg_p_max_f,
		  unsigned long reg_p_min_f,
		  unsigned long reg_r_max_p, unsigned long reg_r_max_f,
		  unsigned long reg_r_min_f,
		  char **len_base, char **angle_base,
		  unsigned long max_pcbs,
		  bool reg_edges,
		  bool fa, bool fas, bool far, bool fl, bool fls, 
		  bool flr, bool fd, bool fdp, bool fpar, bool fcar,
		  bool fai, bool fap, bool farp, bool farc, bool fp,
		  bool fpp, bool flos, bool fg) : Entity ()
{
  Entity::type (1, SIL_ANALYSE_TYPE);
  // Solid
  sld = s;
  take (sld);
  // Symbol name for solid
  nm = n;
  // Hilite symbol name for solid
  hnm = hn;
  // Tolerances
  tol_eq = t_eq;
  if (t_len_dist < tol_eq)
    tol_len_dist = tol_eq;
  else
    tol_len_dist = t_len_dist;
  if (t_angle_dist < tol_eq)
    tol_angle_dist = tol_eq;
  else
    tol_angle_dist = t_angle_dist;
  if (tol_ratio_approx < tol_eq)
    tol_ratio_approx = tol_eq;
  else
    tol_ratio_approx = t_ratio_approx;
  if (tol_ratio_min < tol_eq)
    tol_ratio_min = tol_eq;
  else
    tol_ratio_min = t_ratio_min;
  if (t_loop_dist < tol_eq)
    tol_loop_dist = tol_eq;
  else
    tol_loop_dist = t_loop_dist;
  // Regularity parameters
  regular_angle_max_n = reg_angle_max_n;
  loop_fourier_coefficients = reg_fourier_coeff;
  para_max_power = reg_p_max_p;
  para_max_frac = reg_p_max_f;
  para_min_frac = reg_p_min_f;
  ratio_max_power = reg_r_max_p;
  ratio_max_frac = reg_r_max_f;
  ratio_min_frac = reg_r_min_f;
  len_base_ids = len_base;
  len_base_nums = 0;
  len_base_vals = NULL;
  if (len_base) {
    while (len_base[len_base_nums++]);
    --len_base_nums;
    if (len_base_nums > 0) {
      len_base_vals = new Real[len_base_nums];
      for (unsigned long l = 0; l < len_base_nums; ++l) {
	if (len_base[l] == one_id) {
	  len_base_vals[l] = 1.0;
	} else if (len_base[l] == pi_id) {
	  len_base_vals[l] = PI;
	} else if (len_base[l] == inch_id) {
	  len_base_vals[l] = 2.54;
	} else if (len_base[l] == deg_id) {
	  len_base_vals[l] = 180.0 / PI;
	} else {
	  len_base_vals[l] = atof (len_base[l]);
	}
      }
    }
  }
  angle_base_ids = angle_base;
  angle_base_nums = 0;
  angle_base_vals = NULL;
  if (angle_base) {
    while (angle_base[angle_base_nums++]);
    --angle_base_nums;
    if (angle_base_nums > 0) {
      angle_base_vals = new Real[angle_base_nums];
      for (unsigned long l = 0; l < angle_base_nums; ++l) {
	if (angle_base[l] == one_id) {
	  angle_base_vals[l] = 1.0;
	} else if (angle_base[l] == pi_id) {
	  angle_base_vals[l] = PI;
	} else if (angle_base[l] == inch_id) {
	  angle_base_vals[l] = 2.54;
	} else if (len_base[l] == deg_id) {
	  len_base_vals[l] = 180.0 / PI;
	} else if (angle_base[l] == tan_id) {
	angle_base_vals[l] = 0.0;
	} else {
	  angle_base_vals[l] = atof (angle_base[l]);
	}
      }
    }
  }
  max_pcircle_bucket_size = max_pcbs;
  // Regularity selection parameters
  use_edges = reg_edges;
  f_angles = fa;
  f_angles_special = fas;
  f_angles_ratios = far;
  f_lengths = fl;
  f_lengths_special = fls;
  f_lengths_ratios = flr;
  f_directions = fd;
  f_direction_planes = fdp;
  f_planar_angle_regular = fpar;
  f_conical_angle_regular = fcar;
  f_axis_inter = fai;
  f_axis_pos = fap;
  f_axis_regpos = far;
  f_axis_regcirc = farc;
  f_positions = fp;
  f_positions_partial = fpp;
  f_loops = flos;
  f_geometry = fg;
  // Other initialisations
  reg_list = new SimpleList<List<Regularity> >;
  parallel_directions = NULL;
  intersection_positions = NULL;
  direction_angles = NULL;
  conical_angle_regularities = NULL;
  avg_len = max_len = NAN;
}

Analyse::~Analyse ()
{
  if (len_base_vals) 
    delete len_base_vals;
  if (len_base_ids) {
    for (unsigned long l = 0; l < len_base_nums; ++l) {
      if (len_base_ids[l] != one_id &&
	  len_base_ids[l] != inch_id &&
	  len_base_ids[l] != pi_id &&
	  len_base_ids[l] != deg_id &&
	  len_base_ids[l] != tan_id)
	delete len_base_ids[l];
    }
    delete len_base_ids;
  }
  if (angle_base_vals)
    delete angle_base_vals;
  if (angle_base_ids) {
    for (unsigned long l = 0; l < angle_base_nums; ++l) {
      if (angle_base_ids[l] != one_id &&
	  angle_base_ids[l] != inch_id &&
	  angle_base_ids[l] != pi_id &&
	  angle_base_ids[l] != deg_id &&
	  angle_base_ids[l] != tan_id)
	delete angle_base_ids[l];
    }
    delete angle_base_ids;
  }
  List<Regularity> *rl = reg_list->unlink_first ();
  while (rl) {
    rl->purge ();
    delete rl;
    rl = reg_list->unlink_first ();
  }
  delete reg_list;
  drop (sld);
}

void Analyse::eval ()
{
  if (!sld)
    return;

  sil_progress_bar_text ("Initializing analyser");

  // Initilize data structures.
  position_aspects.clear ();
  direction_aspects.clear ();
  length_aspects.clear ();
  angle_aspects.clear ();
  loop_aspects.clear ();
  List<Regularity> *rl = reg_list->unlink_first ();
  while (rl) {
    rl->purge ();
    delete rl;
    rl = reg_list->unlink_first ();
  }

  parallel_directions = NULL;
  extra_parallel_directions = NULL;
  intersection_positions = NULL;
  direction_angles = NULL;
  conical_angle_regularities = NULL;

  for (unsigned long l = 0; l < 13; ++l) {
    secs[l] = 0;
    msecs[l] = 0;
  }

  // Total time.
  sil_start_time (1);

  // Get features.
  sil_progress_bar_percentage (0.0);
  // Lump features (just for consistency).
  sil_progress_bar_text ("Extracting lump features");
  XListNode<Lump> *lump = sld->lumps () . first ();
  while (lump) {
    SimpleList<Feature> fl;
    lump->data () -> features (&fl);
    sort_features (&fl, lump->data ());
    lump = lump->next ();
  }
  // Face features.
  sil_progress_bar_text ("Extracting face features");
  ListNode<Face> *face = sld->faces () . first ();
  while (face) {
    SimpleList<Feature> fl;
    face->data () -> features (&fl);
    sort_features (&fl, face->data ());
    face = face->next ();
  }
  // Edge features (if wanted).
  if (use_edges) {
    sil_progress_bar_text ("Extracting edge features");
    ListNode<Edge> *edge = sld->edges () . first ();
    while (edge) {
    SimpleList<Feature> fl;
      edge->data () -> features (&fl);
      sort_features (&fl, edge->data ());
      edge = edge->next ();
    }
  }
  // Vertex features.
  sil_progress_bar_text ("Extracting vertex features");
  ListNode<Vertex> *vertex = sld->vertices () . first ();
  while (vertex) {
    SimpleList<Feature> fl;
    vertex->data () -> features (&fl);
    sort_features (&fl, vertex->data ());
    vertex = vertex->next ();
  }

  // Determine the maximal length in the model (sort of max).
  sil_progress_bar_text ("Maximum length");
  avg_len = 0.0;
  unsigned long count = 0;
  max_len = 1.0; // At least 1.0!
  // Maximum distance between position features in model
  ListNode<Aspect> *a1 = position_aspects.first ();
  while (a1) {
    Position *p1 = static_cast<Position*> (a1->data () -> feature ());
    ListNode<Aspect> *a2 = a1->next ();
    while (a2) {
      Position *p2 = static_cast<Position*> (a2->data () -> feature ());
      Real d = (*p1 - *p2) . norm ();
      avg_len += d;
      ++count;
      if (d > max_len)
	max_len = d;
      a2 = a2->next ();
    }
    a1 = a1->next ();
  }
  // Maximum of length features in model.
  a1 = length_aspects.first ();
  while (a1) {
    Real l = static_cast<Length*> (a1->data () -> feature ()) -> value ();
    avg_len += l;
    ++count;
    if (l > max_len)
      max_len = l;
    a1 = a1->next ();
  }
  avg_len /= static_cast<Real> (count);

  // Look for approximately equal angles and lengths, special values
  // for them and special ratios between them.
  sil_progress_bar_text ("Checking scalar features");
  sil_progress_bar_percentage (0.0);
  sil_start_time ();
  eval_scalar ();
  sil_measure_time (secs[1], msecs[1]);

  // Check for parallel directions.
  if (f_directions) {
    sil_progress_bar_text ("Checking " + 
			   as_string (direction_aspects . length ()) +
			   " directions");
    sil_progress_bar_percentage (0.1);
    sil_start_time ();
    parallel_directions = eval_parallel_directions ();
    sil_measure_time (secs[2], msecs[2]);
  }

  if (parallel_directions) {

    // Check for direction planes.
    List<Cluster> *direction_planes = NULL;
    if (f_direction_planes) {
      sil_progress_bar_text ("Checking for direction planes");
      sil_progress_bar_percentage (0.2);
      sil_start_time ();
      direction_planes = eval_direction_planes (parallel_directions);
      sil_measure_time (secs[3], msecs[3]);
    }

    // Check for planar angle regular directions.
    if (f_planar_angle_regular && direction_planes) {
      sil_progress_bar_text ("Checking for planar angle regular directions");
      sil_progress_bar_percentage (0.3);
      sil_start_time ();
      eval_planar_angle_regular (direction_planes);
      sil_measure_time (secs[4], msecs[4]);
    }

    // Check for direction cones.
    if (f_conical_angle_regular) {
      sil_progress_bar_text ("Checking for direction cones");
      sil_progress_bar_percentage (0.4);
      sil_start_time ();
      List<Cluster> *direction_cones = 
	eval_direction_cones (parallel_directions);
      sil_measure_time (secs[5], msecs[5]);
      
      // Check for conical angle regular directions.
      if (direction_cones) {
	sil_progress_bar_text 
	  ("Checking for conical angle regular directions");
	sil_progress_bar_percentage (0.5);
	sil_start_time ();
	eval_conical_angle_regular (direction_cones);
	sil_measure_time (secs[6], msecs[6]);
      }
    }

    if (extra_parallel_directions) {
      if (parallel_directions) {
	sil_progress_bar_text ("Adding extra directions");
	// Add extra parallel directions which should only be considered
	// as directions, but not for a direction plane or cone to the
	// parallel directions list here.
	sil_start_time ();
	add_extra_directions (extra_parallel_directions);
	delete extra_parallel_directions;
	extra_parallel_directions = NULL;
	// Add time to parallel direction time.
	unsigned long s;
	unsigned int m;
	sil_measure_time (s, m);
	secs[2] += s;
	msecs[2] += m;
	secs[2] += msecs[1] / 1000;
	msecs[2] %= 1000;
      } else {
	// We do not have parallel dxirections, so just get rid of the
	// extra directions.
	delete extra_parallel_directions;
	extra_parallel_directions = NULL;
      }
    }

    // Check for axis intersections.
    if (f_axis_inter) {
      sil_progress_bar_text ("Checking for axis intersections");
      sil_progress_bar_percentage (0.6);
      sil_start_time ();
      eval_axis_intersections (parallel_directions);
      sil_measure_time (secs[7], msecs[7]);
    }

    if (f_axis_pos) {
      // Check for axis positions.
      sil_progress_bar_text ("Checking axis positions");
      sil_progress_bar_percentage (0.7);
      sil_start_time ();
      eval_axis_positions (parallel_directions);
      sil_measure_time (secs[8], msecs[8]);
    }

  }

  // Check position features.
  if (f_positions) {
    sil_progress_bar_text ("Checking position features");
    sil_progress_bar_percentage (0.8);
    sil_start_time ();
    eval_positions ();
    sil_measure_time (secs[9], msecs[9]);
  }

  // Check for similar loop features.
  if (f_loops) {
    sil_progress_bar_text ("Checking loop features");
    sil_progress_bar_percentage (0.9);
    sil_start_time ();
    eval_loops ();
    sil_measure_time (secs[10], msecs[10]);
  }

  if (f_geometry) {
    sil_progress_bar_text ("Checking geometry types");
    sil_progress_bar_percentage (0.9);
    sil_start_time ();
    eval_geometry ();
    sil_measure_time (secs[11], msecs[11]);
  }

  // Total time.
  sil_measure_time (secs[0], msecs[0], 1);

  return;
}

void Analyse::sort_features (SimpleList<Feature> *fl, Entity *owner)
{
  // Sort features depending on type
  Feature *f = fl->unlink_first ();
  while (f) {
    if (f->is (SIL_POSITION_TYPE)) {
      position_aspects.add_first (new Aspect (f, owner));
    } else if (f->is (SIL_DIRECTION_TYPE)) {
      direction_aspects.add_first (new Aspect (f, owner));
    } else if (f->is (SIL_LENGTH_TYPE)) {
      length_aspects.add_first (new Aspect (f, owner));
    } else if (f->is (SIL_ANGLE_TYPE)) {
      angle_aspects.add_first (new Aspect (f, owner));
    } else if (f->is (SIL_LOOPPOLYGON_TYPE)) {
      loop_aspects.add_first (new Aspect (f, owner));
    } else {
      sil_msg_warn ("Analyse", "unkown feature type " + f->type_id ());
      // Delete feature, if not used elsewhere.
      if (!f->used ())
	delete f;
    }
    f = fl->unlink_first ();
  }
  return;
}

Analyse *sil_analyse (Solid *s, string const &n, string const &hn,
		      Real t_eq, Real t_len_dist, Real t_angle_dist,
		      Real t_ratio_approx, Real t_ratio_min, Real t_loop_dist,
		      unsigned long reg_angle_max_n,
		      unsigned long reg_fourier_coeff,
		      unsigned long reg_p_max_p, unsigned long reg_p_max_f,
		      unsigned long reg_p_min_f,
		      unsigned long reg_r_max_p, unsigned long reg_r_max_f,
		      unsigned long reg_r_min_f,
		      char **len_base, char **angle_base,
		      unsigned long max_pcbs,
		      bool reg_edges,
		      bool fa, bool fas, bool far, bool fl, bool fls, 
		      bool flr, bool fd, bool fdp, bool fpar, bool fcar,
		      bool fai, bool fap, bool farp, bool farc, bool fp,
		      bool fpp, bool flos, bool fg)
{
  Analyse *a = new Analyse (s, n, hn,
			    t_eq, t_len_dist, t_angle_dist,t_ratio_approx,
			    t_ratio_min, t_loop_dist,
			    reg_angle_max_n, reg_fourier_coeff,
			    reg_p_max_p, reg_p_max_f, reg_p_min_f,
			    reg_r_max_p, reg_r_max_f, reg_r_min_f,
			    len_base, angle_base,
			    max_pcbs,
			    reg_edges,
			    fa, fas, far, fl, fls, flr, fd, fdp, fpar, fcar,
			    fai, fap, farp, farc, fp, fpp, flos, fg);
  a->eval ();
  return a;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
