// analyse.hxx: analyse classes.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse.hxx,v 1.127.2.1 2001/04/19 18:36:30 scmfcl Exp $

#ifndef ANALYSE_HXX
#define ANALYSE_HXX 1

#include "xmath.hxx"
#include "kernel/memory.hxx"
#include "kernel/vector.hxx"
#include "kernel/geometry.hxx"

class Cluster;
class SubRegularity;
class Analyse;
class Edge;
class Loop;
class Solid;
class RegularPointPosition;

// Max value for any tolerance
#define TOL_MAX 10e4

extern char *one_id;
extern char *inch_id;
extern char *pi_id;
extern char *deg_id;
extern char *tan_id;

class Aspect : public Entity {
public:
  Aspect (Feature *f, Entity *o = NULL) : Entity () {
    Entity::type (1, SIL_ASPECT_TYPE);
    feat = f;
    take (f);
    if (o)
      owns.add_first (o);
  }
  virtual ~Aspect () {
    drop (feat);
  }
  virtual string const type_id () const {
    return tid;
  }
  Feature *feature () const {
    return feat;
  }
  SimpleList<Entity> owners () const {
    return owns;
  }
  SimpleList<Entity> &owners () {
    return owns;
  }
private:
  static string const tid;
  SimpleList<Entity> owns;
  Feature *feat;
};

class Regularity : public Entity {
public:
  explicit Regularity (Analyse *a) : Entity () {
    Entity::type (1, SIL_REGULARITY_TYPE);
    ana = a;
    tol = 0.0;
    sub_regs = NULL;
  }
  virtual ~Regularity () {
    if (sub_regs) {
      sub_regs->purge ();
      delete sub_regs;
    }
  }
  virtual string const type_id () const {
    return tid;
  }
  Analyse *analyse () const {
    return ana;
  }
  Real tolerance () const {
    return tol;
  }
  Real &tolerance () {
    return tol;
  }
  SimpleList<Regularity> *sub_regularities () const {
    return sub_regs;
  }
  void add_sub_regularity (Regularity *sr) {
    if (!sub_regs)
      sub_regs = new SimpleList<Regularity>;
    sub_regs->add_last (sr);
    return;
  }
  virtual SimpleList<Regularity> *child_regularities () const {
    return NULL;
  }
  string const description () const;
  virtual string const name () const = 0;
  virtual string const describe () const = 0;
  virtual void hilite (bool hi) const = 0;
  virtual bool is_hierarchical () {
    return false;
  }
  virtual bool is_sub_regularity () {
    return false;
  }
private:
  static string const tid;
  Analyse *ana;
  Real tol;
  SimpleList<Regularity> *sub_regs;
};

class SubRegularity : public Regularity {
public:
  explicit SubRegularity (Analyse *a) : Regularity (a) {
    Entity::type (2, SIL_SUBREGULARITY_TYPE);
  }
  virtual ~SubRegularity () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual void hilite (bool hi) const {
    return;
  }
  virtual bool is_sub_regularity () {
    return true;
  }
protected:
  virtual bool is_equal (SubRegularity const *s) const = 0;
  friend bool operator== (SubRegularity const &s1, SubRegularity const &s2);
private:
  static string const tid;
};

inline bool operator== (SubRegularity const &s1, SubRegularity const &s2) {
  return s1.is (&s2) && s1.is_equal (&s2);
}

class Cluster : public Regularity {
public:
  Cluster (Analyse *an, Feature *v) : Regularity (an) {
    Entity::type (2, SIL_CLUSTER_TYPE);
    val = v->copy ();
    aspcts = NULL;
    sub_c = NULL;
    num_aspcts = 0;
    parent_c = NULL;
  }
  Cluster (Analyse *an, Aspect *a) : Regularity (an) {
    Entity::type (2, SIL_CLUSTER_TYPE);
    val = a->feature () -> copy ();
    aspcts = new SimpleList<Aspect>;
    aspcts->add_first (a);
    sub_c = NULL;
    num_aspcts = 1;
    parent_c = NULL;
  }
  virtual ~Cluster () {
    delete val;
    if (aspcts) {
      aspcts->purge ();
      delete aspcts;
    }
    if (sub_c) {
      sub_c->purge ();
      delete sub_c;
    }
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual SimpleList<Regularity> *child_regularities () const {
    return (SimpleList<Regularity> *)sub_c;
  }
  virtual string const name () const;
  virtual string const describe () const;
  virtual void hilite (bool hi) const;
  virtual bool is_hierarchical () {
    return true;
  }
  Feature *value () const {
    return val;
  }
  void set_value (Feature *v) {
    delete val;
    val = v;
    return;
  }
  void combine (Cluster *c, Feature *v);
  void combine (Cluster *c);
  SimpleList<Aspect> *aspects () const {
    return aspcts;
  }
  void add_aspect (Aspect *a) {
    if (!aspcts)
      aspcts = new SimpleList<Aspect>;
    aspcts->add_first (a);
    ++num_aspcts;
  }
  unsigned long &num_aspects () {
    return num_aspcts;
  }
  unsigned long num_aspects () const {
    return num_aspcts;
  }
  SimpleList<Cluster> *sub_clusters () const {
    return sub_c;
  }
  void add_sub_cluster (Cluster *c) {
    if (!sub_c)
      sub_c = new SimpleList<Cluster>;
    sub_c->add_first (c);
    num_aspcts += c->num_aspcts;
    c->parent_c = this;
  }
  Cluster *&parent_cluster () {
    return parent_c;
  }
  Cluster *parent_cluster () const {
    return parent_c;
  }
  Real sim (Cluster const *c) const {
    return val->sim (c->val);
  }
  Real sim (Feature const *v) const {
    return val->sim (v);
  }
  void collapse (Real t0, Real t1);
  bool collapse ();
  void get_owners (Entity **owners, unsigned long &n);
  void get_owners (SimpleList<Entity> *owners);
  void get_aspects (SimpleList<Aspect> *asp);
  void get_aspects (List<Aspect> *asp);
  void get_all_clusters (List<Cluster> *clusters);
  void get_all_base_clusters (List<Cluster> *clusters);
  bool cluster_is_member (Cluster *c);
private:
  static string const tid;
  Feature *val;
  unsigned long num_aspcts;
  SimpleList<Aspect> *aspcts;
  SimpleList<Cluster> *sub_c;
  Cluster *parent_c;
};

class SpecialValue : public SubRegularity {
public:
  SpecialValue (Analyse *an, SimpleList<Ratio> **rs,
		char **base_id, Real *base_vals, unsigned long base_nums)
    : SubRegularity (an) {
    Entity::type (3, SIL_SPECIALVALUE_TYPE);
    Real t = 0.0;
    for (unsigned long l = 0; l < base_nums; ++l) {
      if (rs[l]) {
	Ratio *r = rs[l]->first ();
	while (r) {
	  if (r->tolerance () > t)
	    t = r->tolerance ();
	  r = rs[l]->next ();
	}
      }
    }
    tolerance () = t;
    rats = rs;
    b_id = base_id;
    b_vals = base_vals;
    b_nums = base_nums;
  }
  virtual ~SpecialValue () {
    for (unsigned long l = 0; l < b_nums; ++l) {
      if (rats[l]) {
	rats[l]->purge ();
	delete rats[l];
      }
    }
    delete rats;
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  SimpleList<Ratio> **ratios () const {
    return rats;
  }
  char **base_ids () const {
    return b_id;
  }
  Real *base_values () const {
    return b_vals;
  }
  unsigned long base_num () const {
    return b_nums;
  }
protected:
  virtual bool is_equal (SubRegularity const *s) const;
private:
  static string const tid;
  SimpleList<Ratio> **rats;
  char **b_id;
  Real *b_vals;
  unsigned long b_nums;
};

class SpecialRatio : public Regularity {
public:
  SpecialRatio (Analyse *an,
		Cluster *cc1, Cluster *cc2, 
		SimpleList<Ratio> *rs) : Regularity (an) {
    Entity::type (2, SIL_SPECIALRATIO_TYPE);
    tolerance () = NAN;
    c1 = cc1;
    c2 = cc2;
    rats = rs;
  }
  virtual ~SpecialRatio () {
    if (rats) {
      rats->purge ();
      delete rats;
    }
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  virtual void hilite (bool hi) const {
    c1->hilite (hi);
    c2->hilite (hi);
    return;
  }
  SimpleList<Ratio> *ratios () const {
    return rats;
  }
  Cluster *cluster1 () const {
    return c1;
  }
  Cluster *cluster2 () const {
    return c2;
  }
private:
  static string const tid;
  Cluster *c1, *c2;
  SimpleList<Ratio> *rats;
};

class ParallelDirection : public SubRegularity {
public:
  ParallelDirection (Analyse *an, Real t, Cluster *d)
    : SubRegularity (an) {
    Entity::type (3, SIL_PARALLELDIRECTION_TYPE);
    tolerance () = t;
    dir = d;
  }
  virtual ~ParallelDirection () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  Direction *direction () const {
    return static_cast<Direction*> (dir->value ());
  }
  Cluster *direction_cluster () const {
    return dir;
  }
protected:
  virtual bool is_equal (SubRegularity const *s) const {
    return ((ParallelDirection *)s) -> dir == dir;
  }
private:
  static string const tid;
  Cluster *dir;
};

class OrthogonalDirection : public SubRegularity {
public:
  OrthogonalDirection (Analyse *an, Real t, Cluster *d)
    : SubRegularity (an) {
    Entity::type (3, SIL_ORTHOGONALDIRECTION_TYPE);
    tolerance () = t;
    dir = d;
  }
  virtual ~OrthogonalDirection () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  Direction *direction () const {
    return static_cast<Direction*> (dir->value ());
  }
  Cluster *direction_cluster () const {
    return dir;
  }
protected:
  virtual bool is_equal (SubRegularity const *s) const {
    return ((OrthogonalDirection *)s) -> dir == dir;
  }
private:
  static string const tid;
  Cluster *dir;
};

class DirectionAngle : public Regularity {
public:
  DirectionAngle (Analyse *an, Real t, Cluster *d1, Cluster *d2, Real a)
    : Regularity (an) {
    Entity::type (2, SIL_DIRECTIONANGLE_TYPE);
    tolerance () = t;
    dir1 = d1;
    dir2 = d2;
    angl = a;
  }
  virtual ~DirectionAngle () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  virtual void hilite (bool hi) const;
  Real angle () const {
    return angl;
  }
  Cluster *direction1 () const {
    return dir1;
  }
  Cluster *direction2 () const {
    return dir2;
  }
private:
  static string const tid;
  Real angl;
  Cluster *dir1, *dir2;
};

class PlanarAngleRegular : public Regularity {
public:
  PlanarAngleRegular (Analyse *an, Real t, SimpleList<Cluster> *ds,
		      Cluster *p, unsigned long a_n)
    : Regularity (an) {
    Entity::type (2, SIL_PLANARANGLEREGULAR_TYPE);
    tolerance () = t;
    dirs = ds;
    pln = p;
    angl_n = a_n;
    fu = false;
    ra = NULL;
  }
  virtual ~PlanarAngleRegular () {
    if (ra) {
      ra->purge ();
      delete ra;
    }
    if (dirs)
      delete dirs;
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  virtual void hilite (bool hi) const;
  UnitVector3 plane_normal () const {
    return static_cast<UnitVector3> 
      (*(static_cast<DirectionPlane *> (pln->value ())));
  }
  Cluster *direction_plane () const {
    return pln;
  }
  Real angle () const {
    return PI / (Real)angl_n;
  }
  unsigned long &angle_n () {
    return angl_n;
  }
  unsigned long angle_n () const {
    return angl_n;
  }
  bool &all_directions () {
    return fu;
  }
  bool all_directions () const {
    return fu;
  }
  SimpleList<Cluster> *directions () const {
    return dirs;
  }
  void set_regular_arrangements (SimpleList<unsigned long> *nra) {
    if (ra) {
      ra->purge ();
      delete ra;
    } 
    ra = nra;
    return;
  }
  SimpleList<unsigned long> *regular_arrangements () const {
    return ra;
  }
private:
  static string const tid;
  SimpleList<Cluster> *dirs;
  Cluster *pln;
  unsigned long angl_n;
  bool fu;
  SimpleList<unsigned long> *ra;
};

class ConicalAngleRegular : public Regularity {
public:
  ConicalAngleRegular (Analyse *ana, Real t, Real ct, SimpleList<Cluster> *ds,
		       Vector3 const &nn, Real aan, 
		       unsigned long a) : Regularity (ana) {
    Entity::type (2, SIL_CONICALANGLEREGULAR_TYPE);
    tolerance () = t;
    c_tol = ct;
    dirs = ds;
    cone_dir = new Direction (nn);
    take (cone_dir);
    cone_angl = aan;
    an = a;
    f = false;
    ortho = false;
    ra = NULL;
  }
  virtual ~ConicalAngleRegular () {
    drop (cone_dir);
    if (ra) {
      ra->purge ();
      delete ra;
    }
    if (dirs)
      delete dirs;
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  virtual void hilite (bool hi) const;
  Direction *cone_direction () const {
    return cone_dir;
  }
  Real cone_angle () const {
    return cone_angl;
  }
  Real angle () const {
    return 2.0 * PI / (Real)an;
  }
  unsigned long &angle_n () {
    return an;
  }
  unsigned long angle_n () const {
    return an;
  }
  SimpleList<Cluster> *directions () const {
    return dirs;
  }
  Real cone_tolerance () const {
    return c_tol;
  }
  bool &all_directions () {
    return f;
  }
  bool all_directions () const {
    return f;
  }
  bool &orthogonal () {
    return ortho;
  }
  bool orthogonal () const {
    return ortho;
  }
  void set_regular_arrangements (SimpleList<unsigned long> *nra) {
    if (ra) {
      ra->purge ();
      delete ra;
    }
    ra = nra;
    return;
  }
  SimpleList<unsigned long> *regular_arrangements () const {
    return ra;
  }
private:
  static string const tid;
  SimpleList<Cluster> *dirs;
  Direction *cone_dir;
  Real cone_angl, c_tol;
  unsigned long an;
  bool f, ortho;
  SimpleList<unsigned long> *ra;
};

class PointLine : public Regularity {
public:
  PointLine (Analyse *ana, Real t, Vector3 const &p, UnitVector3 const &d,
	     SimpleList<Cluster> *ps) : Regularity (ana) {
    Entity::type (2, SIL_POINTLINE_TYPE);
    tolerance () = t;
    pos = p;
    dir = d;
    pts = ps;
    raps = NULL;
  }
  virtual ~PointLine () {
    if (raps) {
      raps->purge ();
      delete raps;
    }
    if (pts)
      delete pts;
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual SimpleList<Regularity> *child_regularities () const {
    return (SimpleList<Regularity> *)raps;
  }
  virtual string const describe () const;
  virtual void hilite (bool hi) const;
  Vector3 position () const {
    return pos;
  }
  UnitVector3 direction () const {
    return dir;
  }
  SimpleList<Cluster> *points () const {
    return pts;
  }
  void add_regular_dist (RegularPointPosition *rap) {
    if (!raps)
      raps = new SimpleList<RegularPointPosition>;
    raps->add_first (rap);
    return;
  }
  void delete_regular_dist () {
    if (raps) {
      delete raps;
      raps = NULL;
    }
    return;
  }
  SimpleList<RegularPointPosition> *regular_dist () const {
    return raps;
  }
private:
  static string const tid;
  SimpleList<Cluster> *pts;
  Vector3 pos;
  UnitVector3 dir;
  SimpleList<RegularPointPosition> *raps;
};

class RegularPointPosition : public Regularity {
public:
  RegularPointPosition (Analyse *ana, Real t, Vector3 const &pos,
			UnitVector3 const &v1, Real s1,
			UnitVector3 const &v2, Real s2,
			SimpleList<Cluster> *points) : Regularity (ana) {
    Entity::type (2, SIL_REGULARPOINTPOSITION_TYPE);
    tolerance () = t;
    p = pos;
    dir1 = v1;
    d1 = s1;
    dir2 = v2;
    d2 = s2;
    pts = points;
    dir1_reg = dir2_reg = NULL;
    d1_reg = d2_reg = NULL;
    u = false;
    dpts = 0;
  }
  virtual ~RegularPointPosition () {
    if (pts)
      delete pts;
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  virtual void hilite (bool hi) const;
  Vector3 &position () {
    return p;
  }
  Vector3 position () const {
    return p;
  }
  Real &distance1 () {
    return d1;
  }
  Real distance1 () const {
    return d1;
  }
  UnitVector3 &direction1 () {
    return dir1;
  }
  UnitVector3 direction1 () const {
    return dir1;
  }
  SubRegularity *&direction1_reg () {
    return dir1_reg;
  }
  SubRegularity *direction1_reg () const {
    return dir1_reg;
  }
  Real &distance2 () {
    return d2;
  }
  Real distance2 () const {
    return d2;
  }
  SpecialValue *&distance1_reg () {
    return d1_reg;
  }
  SpecialValue *distance1_reg () const {
    return d1_reg;
  }
  UnitVector3 &direction2 () {
    return dir2;
  }
  UnitVector3 direction2 () const {
    return dir2;
  }
  SubRegularity *&direction2_reg () {
    return dir2_reg;
  }
  SubRegularity *direction2_reg () const {
    return dir2_reg;
  }
  SpecialValue *&distance2_reg () {
    return d2_reg;
  }
  SpecialValue *distance2_reg () const {
    return d2_reg;
  }
  SimpleList<Cluster> *points () const {
    return pts;
  }
  unsigned long &different_points () {
    return dpts;
  }
  unsigned long different_points () const {
    return dpts;
  }
  bool &used () {
    return u;
  }
  bool used () const {
    return u;
  }
  bool match_pts (RegularPointPosition *grid, Real &tol);
  bool match_dirs (RegularPointPosition *grid, Real tol);
  bool match_dir (RegularPointPosition *grid, Real tol);
  void combine (RegularPointPosition *grid);
private:
  static string const tid;
  SimpleList<Cluster> *pts;
  Vector3 p;
  UnitVector3 dir1, dir2;
  SubRegularity *dir1_reg, *dir2_reg;
  Real d1, d2;
  SpecialValue *d1_reg, *d2_reg;
  unsigned long dpts;
  bool u;
};

class RegularPointCircle : public Regularity {
public:
  RegularPointCircle (Analyse *ana, Real t, Vector3 cc, Real rr,
		      unsigned long nn, UnitVector3 const &nnorm,
		      SimpleList<Cluster> *ps) : Regularity (ana) {
    Entity::type (2, SIL_REGULARPOINTCIRCLE_TYPE);
    tolerance () = t;
    c = cc;
    r = rr;
    an = nn;
    norm = nnorm;
    pts = ps;
    f = false;
    ra = NULL;
  }
  virtual ~RegularPointCircle () {
    if (pts)
      delete pts;
    if (ra) {
      ra->purge ();
      delete ra;
    }
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  virtual void hilite (bool hi) const;
  Vector3 &centre () {
    return c;
  }
  Vector3 centre () const {
    return c;
  }
  Real &radius () {
    return r;
  }
  Real radius () const {
    return r;
  }
  unsigned long &angle_n () {
    return an;
  }
  unsigned long angle_n () const {
    return an;
  }
  Real angle () const {
    return 2.0 * PI / static_cast<Real> (an);
  }
  UnitVector3 &normal () {
    return norm;
  }
  UnitVector3 normal () const {
    return norm;
  }
  SimpleList<Cluster> *points () const {
    return pts;
  }
  bool &all_directions () {
    return f;
  }
  bool all_directions () const {
    return f;
  }
  void set_regular_arrangements (SimpleList<unsigned long> *nra) {
    if (ra) {
      ra->purge ();
      delete ra;
    }
    ra = nra;
    return;
  }
  SimpleList<unsigned long> *regular_arrangements () const {
    return ra;
  }
private:
  static string const tid;
  SimpleList<Cluster> *pts;
  Vector3 c;
  Real r;
  unsigned long an;
  UnitVector3 norm;
  bool f;
  SimpleList<unsigned long> *ra;
};

class AltGeometry : public Regularity {
public:
  AltGeometry (Analyse *ana, Real k, Cell *C, Geometry *AG) :
    Regularity (ana) {
    Entity::type (2, SIL_ALTGEOMETRY_TYPE);
    tolerance () = k;
    c = C;
    ag = AG;
    take (AG);
  }
  virtual ~AltGeometry () {
    drop (ag);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const name () const {
    return tid;
  }
  virtual string const describe () const;
  virtual void hilite (bool hi) const;
  Cell *cell () {
    return c;
  }
  Geometry *alt_geometry () {
    return ag;
  }
private:
  static string const tid;
  Cell *c;
  Geometry *ag;
};

class Analyse : public Entity {
public:
  Analyse (Solid *s, string const &n, string const &hn,
	   Real t_eq, Real t_len_dist, Real t_angle_dist,
	   Real t_ratio_approx, Real t_ratio_min,
	   Real t_loop_dist,
	   unsigned long reg_angle_max_n,
	   unsigned long reg_fourier_coeff,
	   unsigned long reg_p_max_p, unsigned long reg_p_max_f,
	   unsigned long reg_p_min_f,
	   unsigned long reg_r_max_p, unsigned long reg_r_max_f,
	   unsigned long reg_p_min_f,
	   char **len_base, char **angle_base,
	   unsigned long max_pcbs,
	   bool reg_edges,
	   bool fa, bool fas, bool far, bool fl, bool fls, 
	   bool flr, bool fd, bool fdp, bool fpar, bool fcar,
	   bool fai, bool fap, bool farp, bool farc, bool fp,
	   bool fpp, bool flos, bool fg);
  virtual ~Analyse ();
  virtual string const type_id () const {
    return tid;
  }
  void eval ();
  Solid *solid () const {
    return sld;
  }
  const string &name () const {
    return nm;
  }
  const string &hilite_name () const {
    return hnm;
  }
  SimpleList< List<Regularity> > *regularities () const {
    return reg_list;
  }
  unsigned long *time_secs () {
    return secs;
  }
  unsigned int *time_msecs () {
    return msecs;
  }
  Real maximum_length () const {
    return max_len;
  }
  Real average_length () const {
    return avg_len;
  }
  Real tolerance_angle_dist () {
    return tol_angle_dist;
  }
  Real tolerance_length_dist () {
    return tol_len_dist;
  }
private:
  // General methods.
  void sort_features (SimpleList<Feature> *fl, Entity *owner);
  // Clustering.
  void cluster_aspects (List<Aspect> *aspects,
			Real tol_eq, Real tol_dist, Real tol_approx,
			bool hier, List<Cluster> *clusters);
  void cluster_aspects (List<Cluster> *cls, Real tol_eq,
			Real tol_dist, Real tol_approx, bool hier);
  void tolerance_levels (List<Cluster> *clusters,
			 SimpleList<Real> *tols, Real tol_dist, Real tol_max);
  // Methods for scalar evaluation.
  void eval_scalar ();
  void eval_ratios (List<Cluster> *clusters);
  void check_ratio (Cluster *c1, Cluster *c2,
		    SimpleList<Ratio> *a_ratios,
		    List<SpecialRatio> *ratios);
  SimpleList<Ratio> *ratio (Real v, Real tol_min, Real tol_approx,
			    unsigned long min_frac, unsigned long max_frac,
			    unsigned long max_power, bool real_ratio);
  void integer_rel (bool w_neg, Real const &w,
		    unsigned long const &n, unsigned long const &m,
		    bool const &pm, bool &N_neg, 
		    unsigned long const &N, unsigned long const &p,
		    unsigned long const &max_m,
		    unsigned long const &max_frac,
		    Real const &tol_min, Real const &tol_approx,
		    SimpleList<Ratio> *&rs);
  // Methods to find special scalars.
  SpecialValue *check_special_scalar (Real v, Real tol_approx,
				      bool angle_flag);
  // Methods for direction evaluation.
  List<Cluster> *eval_parallel_directions ();
  List<Cluster> *eval_direction_planes (List<Cluster> *directions);
  List<Cluster> *eval_direction_cones (List<Cluster> *directions);
  List<Cluster> *cluster_merge_direction_cones (Cluster *angle);
  UnitVector3 solve_cone (UnitVector3 const &d1,
			  UnitVector3 const &d2,
			  UnitVector3 const &d3) const;
  // Methods to find parallel and/or orthogonal directions.
  Cluster *find_parallel (UnitVector3 const &n, List<Cluster> *directions,
			  Real &tol) const;
  Cluster *find_orthogonal (UnitVector3 const &n, List<Cluster> *directions,
			    Real &tol) const;
  Cluster *find_parallel_orthogonal (UnitVector3 const &n,
				     UnitVector3 const *p,
				     List<Cluster> *directions,
				     Real &tol, bool &par) const;
  Cluster *find_direction_pair (Direction *d1, Direction *d2, 
				List<Cluster> *directions);
  // Methods for extra direction handling.
  void add_extra_directions (List<Cluster> *extra_dirs);
  Cluster *add_extra_parallel_direction (Direction *d);
  // Methods for angle regular evaluations.
  void eval_planar_angle_regular (List<Cluster> *direction_planes);
  void eval_conical_angle_regular (List<Cluster> *direction_cones);
  void check_direction_plane (Cluster *ap, List<PlanarAngleRegular> *regs,
			      List<DirectionAngle> *da_regs);
  void check_direction_cone (Cluster *ac, List<ConicalAngleRegular> *regs);
  // Methods for evaluating axis intersections.
  void eval_axis_intersections (List<Cluster> *dirs);
  Real intersect_axes (Position *p1, Direction *d1,
		       Position *p2, Direction *d2,
		       Vector3 &ip) const;
  bool solve_lineq_3 (Vector3 const &m0, Vector3 const &m1, Vector3 const &m2,
		      Real &x0, Real &x1, Real &x2, Vector3 const &c) const;
  // Methods for evaluating regular axis arrangements.
  void eval_axis_positions (List<Cluster> *dirs);
  List<Cluster> *check_parallel_axes (Cluster *dirs);
  void check_aligned_axes (List<Cluster> *aligned_axes,
			   List<RegularPointPosition> *raps,
			   List<Regularity> *rpls,
			   List<Regularity> *pcls,
			   List<RegularPointCircle> *rpcs);
  void check_point_line (PointLine *pl);
  void check_grid (SimpleList<PointLine> *lines1,
		   SimpleList<PointLine> *lines2,
		   List<RegularPointPosition> *raps);
  void group_lines (SimpleList<PointLine> *lines,
		    SimpleList<SimpleList<RegularPointPosition> > *groups);
  void form_grids (SimpleList<RegularPointPosition> *group1,
		   SimpleList<RegularPointPosition> *group2,
		   List<RegularPointPosition> *raps);
  List<Cluster> *cluster_merge_point_circles (Cluster *circle);
  void check_point_circle (Cluster *pc, UnitVector3 const &norm,
			   List<RegularPointCircle> *regs);
  Vector3 ls_circle_centre (SimpleList<Vector3> &pts, Real r,
			    UnitVector3 const &norm,
			    Vector3 const &centre0);
  // Methods for evaluating regular position arrangmenets.
  void eval_positions ();
  void check_plane_partially_equal (Cluster *dir, List<Cluster> *pts,
				    List<Cluster> *regs);
  void check_line_partially_equal (Cluster *dir, List<Cluster> *pts,
				   List<Cluster> *regs);
  // Methods for loop polygon evaluation.
  void eval_loops ();
  // Methods for geometry evaluation.
  void eval_geometry ();
  Geometry *fit_plane_pos (SimpleList<Position> *pos, Real &tol, Real &dist);
  Geometry *fit_cylinder_circ (SimpleList<Circle> *circs, bool conv,
			       Real &tol, Real &dp);
  // Type id, data and symbol names.
  static string const tid;
  Solid *sld;
  string nm, hnm;
  // Tolerance values.
  Real tol_eq, tol_len_dist, tol_angle_dist, tol_ratio_approx, tol_ratio_min;
  Real tol_loop_dist;
  // Special values configuration, parameters.
  char **len_base_ids;
  Real *len_base_vals;
  unsigned long len_base_nums;
  char **angle_base_ids;
  Real *angle_base_vals;
  unsigned long angle_base_nums;
  unsigned long para_max_power;
  unsigned long para_max_frac;
  unsigned long para_min_frac;
  unsigned long ratio_max_power;
  unsigned long ratio_max_frac;
  unsigned long ratio_min_frac;
  unsigned long regular_angle_max_n;
  unsigned long loop_fourier_coefficients;
  unsigned long max_pcircle_bucket_size;
  // Flags indicating which regularities should be evaluated.
  bool f_angles, f_angles_special, f_angles_ratios;
  bool f_lengths, f_lengths_special, f_lengths_ratios;
  bool f_directions, f_direction_planes, f_planar_angle_regular;
  bool f_conical_angle_regular, f_axis_inter, f_axis_pos, f_axis_regpos;
  bool f_axis_regcirc, f_positions, f_positions_partial, f_loops, f_geometry;
  bool use_edges;
  // Lengths in model
  Real max_len, avg_len;
  // Aspect lists connecting features to the cells in the model.
  List<Aspect> position_aspects;
  List<Aspect> direction_aspects;
  List<Aspect> length_aspects;
  List<Aspect> angle_aspects;
  List<Aspect> loop_aspects;
  // Regularity lists.
  SimpleList< List<Regularity> > *reg_list;
  // References to special regularities exchanged between the analyse
  // methods.
  List<Cluster> *parallel_directions, *extra_parallel_directions;
  List<Cluster> *intersection_positions;
  List<DirectionAngle> *direction_angles;
  List<ConicalAngleRegular> *conical_angle_regularities;
  // Timings.
  //    0: total
  //    1: scalar features
  //    2: parallel directions
  //    3: direction planes
  //    4: planar angle regular
  //    5: direction cones
  //    6: conical angle regular
  //    7: axis intersections
  //    8: axis positions
  //    9: positions
  //   10: loops
  //   11: geometry types
  unsigned long secs[12];
  unsigned int msecs[12];
};

#endif // !ANALYSE_HXX

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
