// analyse_angleregs.cc: Analyse a SiL model / angle regularities
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_angleregs.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_angleregs,
       "$Id: analyse_angleregs.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"
#include "list.hxx"
#include "xstring.hxx"
#include "kernel/kernel.hxx"
#include "kernel/matrix.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"
#include "kernel/geometry.hxx"

void Analyse::eval_planar_angle_regular (List<Cluster> *direction_planes)
{
  // Find planar angle regular arrangements for each main direction
  // plane regularity. Those that are present in sub regularities
  // should be found over the main direction plane as well.

  // Get the main clusters.
  SimpleList<Cluster> dp;
  ListNode<Cluster> *p = direction_planes->first ();
  while (p) {
    dp.add_first (p->data ());
    p = p->next ();
  }
  
  List<PlanarAngleRegular> *regs = new List<PlanarAngleRegular>;
  List<DirectionAngle> *da_regs = new List<DirectionAngle>;
  Cluster *plane = dp.unlink_first ();
  while (plane) {
    check_direction_plane (plane, regs, da_regs);
    if (plane->sub_clusters ()) {
      Cluster *c = plane->sub_clusters () -> first ();
      while (c) {
	dp.add_first (c);
	c = plane->sub_clusters () -> next ();
      }
    }
    plane = dp.unlink_first ();
  }

  // Add regularities.
  if (regs->length () > 0)
    reg_list->add_first ((List<Regularity> *)regs);
  else
    delete regs;
  if (da_regs->length () > 0) {
    reg_list->add_first ((List<Regularity> *)da_regs);
    direction_angles = da_regs;
  } else {
    direction_angles = NULL;
    delete da_regs;
  }

  return;
}

void Analyse::check_direction_plane (Cluster *ap,
				     List<PlanarAngleRegular> *regs,
				     List<DirectionAngle> *da_regs)
{
  // Angular tolerance.
  Real tol = ap->tolerance () + tol_angle_dist;

  // The owners of the direction plane are the direction clusters
  // lying in it. We collect the direction clusters according to the
  // tolerance of the direction plane.
  SimpleList<Entity> owners;
  ap->get_owners (&owners);
  // Check if we have to increase the tolerance for one of the
  // direction clusters has a larger tolerance and no children.
  Entity *e = owners.first ();
  while (e) {
    Entity *ne = owners.next ();
    Cluster *c = dynamic_cast<Cluster*> (e);
    if (c && c->value () -> is (SIL_DIRECTION_TYPE) &&
	c->tolerance () > tol) {
      if (c->sub_clusters () && c->sub_clusters () -> length () > 0) {
	Cluster *s = c->sub_clusters () -> first ();
	while (s) {
	  owners.add_last_check (s);
	  s = c->sub_clusters () -> next ();
	}
	owners.del (e);
      } else {
	tol = c->tolerance ();
      }
    }
    e = ne;
  }
  // Find the appropriate direction clusters.
  SimpleList<Cluster> dirs;
  e = owners.unlink_first ();
  while (e) {
    Cluster *c = dynamic_cast<Cluster*> (e);
    if (c && c->value () -> is (SIL_DIRECTION_TYPE)) {
      if (c->parent_cluster () &&
	  c->parent_cluster () -> tolerance () < tol) {
	// Only use parent if parent is part of direction plane,
	// otherwise still use cluster. Do not use paretns that are
	// not in the direction plane.
	if (!owners.is_member (c->parent_cluster ()))
	  dirs.add_last_check (c);
      } else {
	dirs.add_last_check (c);
      }
    } else {
      sil_msg_warn ("Analyse",
		    "direction plane has non-direction cluster owner");
    }
    e = owners.unlink_first (); 
  }
  // The maximal base angle for which we can still find different
  // arrangements is Pi / N with N = (Pi / (4t) where t is the
  // tolerance [as 2N = 2Pi / (4t)].
  unsigned long N;
  if (tol < tol_eq) {
    // tol is too small, so use upper limit (divided by two as the
    // upper limit refers to the complete circle).
    N = regular_angle_max_n >> 1;
  } else {
    N = static_cast<unsigned long> (floor (PI / (4.0 * tol)));
    if (N > (regular_angle_max_n >> 1))
      N = regular_angle_max_n >> 1;
  }
  if (N < 3)
    return;

  // Store the directions in an array to access them.
  unsigned long n = dirs.length ();
  // This works for n == 2 as well (it just checks a few things that
  // are not really required). Note that at the end we check for
  // special angles between two directions that are not in any of the
  // other regularities. But n smaller 2 is of course not valid.
  if (n < 2)
    return;
  Cluster **pdirs = new Cluster*[n];
  unsigned long l = 0;
  Cluster *c = dirs.unlink_first ();
  while (c) {
    pdirs[l++] = c;
    c = dirs.unlink_first ();
  }

  sil_progress_bar_text ("Checking " + as_string (n) + 
			 " directions in plane " + 
			 as_string (*static_cast<DirectionPlane*> (ap->
								   value ())));
 
  // Direction parallel to direction plane normal or this normal so
  // that the normal can become a extra direction.
  Cluster *direction = NULL;

  // List of regularities for this plane
  SimpleList<PlanarAngleRegular> par_regs;

  // Find regular angle sets between the directions, by comparing the
  // angles between an initial direction and the other directions. We
  // have to check all directions as intial directions and compare the
  // angles between the initial directions and all following
  // directions in the list.  At least two more directions should be
  // in the list or we cannot get an angular regular relation and the
  // angles between two directions are checked later.
  HalfMatrix<Real> angles (n);
  // List of regular angles for each initial direction.
  SimpleList<unsigned long> **valid_angles = new SimpleList<unsigned long>*[n];
  for (unsigned long k = 0 ; k < n - 1; ++k) {
    // Initialize the valid angles list for k as initial direction.
    valid_angles[k] = new SimpleList<unsigned long>;
    // Compute all angles between the initial direction and the
    // following directions in the list and collect the regular
    // angles.
    //
    // Compute a direction in the direction plane as orhtogonal to the
    // direction plane normal and the initial direction which is later
    // used to indicate if another direction is left or right to the
    // initial direction. We have to be consistent in selecting either
    // the left or the reight direction (opposite directions are
    // identified and without a consisten choice there is a change
    // that the arrangement is mixed up).
    Vector3 regdir = *((Direction *)pdirs[k]->value ()) * 
      *((DirectionPlane *)ap->value ());
    for (unsigned long l = k + 1; l < n; ++l) {
      // Note that the angle is not based on projection to plane,
      // which might actually lead to a better result as all is an
      // approximation. But we get the angle with the direction which
      // is right of the initial direction with respect to regdir from
      // the two possible choices of identified directions (well, any
      // consistent choice will do).
      Real a = *((Direction *)pdirs[k]->value ()) %
	*((Direction *)pdirs[l]->value ());
      if (a >= 1.0)
	a = 0.0;
      else if (a <= -1.0)
	a = PI;
      else
	angles (l, k) = acos (a);
      if (regdir % *((Direction *)pdirs[l]->value ()) < 0.0)
	angles (l, k) = PI - angles (l, k);
      // Find the regular angles suggested by this angle:
      if (angles (l, k) > tol) {
	unsigned long a_min = (unsigned long)floor (PI / (angles (l, k) +
							  tol));
	if (a_min < N) {
	  if (a_min < 3)
	      a_min = 3;
	  unsigned long a_max = (unsigned long)ceil (PI / (angles (l, k) -
							   tol)) + 1;
	  if (a_max > N)
	    a_max = N;
	  for (unsigned long aa = a_min; aa < a_max; ++aa) {
	    // Check if angle is already listed as a valid angle and
	    // also check for multiples.
	    bool add = true;
	    unsigned long *an = valid_angles[k]->first ();
	    while (an) {
	      unsigned long *next_an = valid_angles[k]->next ();
	      if (aa < *an) {
		Real a = (Real)*an / (Real)aa;
		if (fabs (a - rint (a)) < tol_eq) {
		  add  = false;
		  break;
		}
	      } else {
		Real a = (Real)aa / (Real)*an;
		if (fabs (a - rint (a)) < tol_eq) {
		  valid_angles[k]->del (an);
		  delete an;
		}
	      }
	      an = next_an;
	    }
	    // If the angle is not listed as a valid angle, then add
	    // it to the list.
	    if (add)
	      valid_angles[k]->add_first (new unsigned long (aa));
	  }
	}
      }
    }
  }

  // Increase angular tolerance by tol_angle_dist
  tol = ap->tolerance () + tol_angle_dist;
  // NOTE: The maximal tolerance depends on the *an below. We could
  //       set the tolerance for a valid direction to PI/(4(*an))
  //       which is the maximal tolerance before the arrangements become
  //       ambiguous.

  // Now check for all initial directions if one of the suggested
  // regular angles are actually really there.
  for (unsigned long k = 0 ; k < n - 1; ++k) {

    // Check all angles with the initial direction k if they are
    // multiples of the suggested regular angles.
    unsigned long *an = valid_angles[k]->unlink_first ();
    while (an) {

      // The value of the regular angle.
      Real angl = PI / (Real)*an;
      // Array keeping track which direction in the list is actually
      // occupying a direction suggested buy the regular angle.
      unsigned long directions[*an + 1];
      // Tolerances for the directions list.
      Real tols[*an + 1];
      // Number of directions which are in the current regular subset.
      unsigned long num = 1;
      // Direction 0 (0 * PI / n) is always occupied by the initial
      // direction k.
      directions[0] = k;
      tols[0] = 0.0;
      // Mark the other regular directions as not-occupied.
      for (unsigned long j = 1; j <= *an; ++j)
	directions[j] = n;
      // Check the angles between initial direction and all following
      // directions. If the angle suggest a regular arrangement add
      // the direction to the current regular subset.
      for (unsigned long j = k + 1; j < n; ++j) {
	// Compute the tolerance between the actual angle and the
	// angle suggested by the regular angle.
	unsigned long ta = (unsigned long)rint (fabs (angles (j, k)) / angl);
	Real t = fabs ((Real)ta * angl - fabs (angles (j, k)));
	if (t < tol && ta <= *an) {
	  // The angle is approximately regular, check if we have to
	  // add the direction to the curent subset.
	  if (directions[ta] < n) {
	    // Direction already occupied by other direction, but
	    // parallel directions, etc.  have been determined before,
	    // so we take the one which is closer.
	    if (t < tols[ta]) {
	      directions[ta] = j;
	      tols[ta] = t;
	    }
	  } else {
	    // Direction not occupied, add it as new one.
	    ++num;
	    directions[ta] = j;
	    tols[ta] = t;
	  }
	}
      }

      if (num > 2) {
	// The current regular subset has more than two members, so it
	// could be a regularity.

	// First check if any of the directions is already a member of
	// a planar angle regular set with the same base angle.
	PlanarAngleRegular *ipar = NULL;
	PlanarAngleRegular *par = par_regs.first ();
	while (par && !ipar) {
	  // Check if angles are compatible.
	  unsigned long angl_n = par->angle_n ();
	  bool match = false;
	  if (angl_n == *an) {
	    match = true;
	  } else if (angl_n < *an) {
	    // check if: integer = *an / angl_n
	    Real r = rint ((Real)*an / (Real)angl_n);
	    match = (fabs (r * (Real)angl_n - (Real)(*an)) < tol_eq);
	  } else {
	    // check if: integer =  angl_n / *an
	    Real r = rint ((Real)angl_n / (Real)*an);
	    match = (fabs (r * (Real)(*an) - (Real)angl_n) < tol_eq);
	  }
	  if (match) {
	    for (unsigned long j = 0; j <= *an; ++j) {
	      if (directions[j] < n && 
		  par->directions () -> is_member (pdirs[directions[j]])) {
		ipar = par;
		break;
	      }
	    }
	  }
	  par = par_regs.next ();
	}
	if (ipar) {
	  // Combine current regular set with the one found.
	  for (unsigned long j = 0; j <= *an; ++j) {
	    if (directions[j] < n) {
	      ipar->directions () -> add_last_check (pdirs[directions[j]]);
	      // Disable angles for special angle check below
	      for (unsigned long l = 0; l < j; ++l) {
		unsigned long j2 = directions[l];
		if (j2 < n) {
		  unsigned long j1 = directions[j];
		  if (j1 > j2) 
		    angles (j1, j2) =  -fabs (angles (j1, j2));
		  else
		    angles (j2, j1) =  -fabs (angles (j2, j1));
		}
	      }
	    }
	  }
	  // And set the angle.
	  if (ipar->angle_n () > *an)
	    ipar->angle_n () = *an;

	} else {
	  // New regularity. 

	  // Generate direction list.
	  SimpleList<Cluster> *dirlist = new SimpleList<Cluster>;
	  unsigned long cnt = 0;
	  Real t = 0.0;
	  for (unsigned long j = 0; j <= *an; ++j) {
	    if (directions[j] < n) {
	      dirlist->add_first (pdirs[directions[j]]);
	      t += fabs (tols[j]);
	      ++cnt;
	    }
	  }
	  if (cnt > 0)
	    t /= (Real)cnt;

	  // Generate and add the current subset as a new regularity.
	  PlanarAngleRegular *nr = 
	    new PlanarAngleRegular (this, t, dirlist, ap, *an);
	  // Add normal direction of direction plane to extra parallel
	  // directions.
	  Direction *v = new Direction (nr->plane_normal ());
	  if(!direction) {
	    if (extra_parallel_directions) {
	      Real t = tol_angle_dist;
	      direction = find_parallel (*(static_cast<UnitVector3*> (v)),
					 extra_parallel_directions, t);
	    }
	    if (!direction) {
	      // Plane direction is not parallel to any existing
	      // direction. Generate a new direction and note it as
	      // an extra direction which is not considered for
	      // direction planes and cones, but as a direction
	      // alone.
	      direction = add_extra_parallel_direction (v);
	    }
	  }
	  // Parallel direction exists or has been created, add
	  // plane direction aspect to it.
	  Aspect *asp = new Aspect (v, nr);
	  // Update aspect counter and parent aspect counter.
	  direction->add_aspect (asp);
	  Cluster *pa = direction->parent_cluster ();
	  while (pa) {
	    pa->num_aspects () += 1;
	    pa = pa->parent_cluster ();
	  }
	  par_regs.add_first (nr);
	}
      }

      delete an;
      an = valid_angles[k]->unlink_first ();
    }

    delete valid_angles[k];
  }

  delete valid_angles;

  // Add regularities to reg list
  PlanarAngleRegular *r = par_regs.unlink_first ();
  while (r) {

    // Check for m-regular arrangements of the directions in the plane,
    // where m is a divisor of angel_n.
    //
    // Regular direction for orientation.
    Direction *bd = static_cast<Direction *> (r->directions () -> 
					      first () -> value ());
    Vector3 regdir = (*bd) * *((DirectionPlane *)ap->value ());
    // Create a boolean array to list occupied directions and maximum
    // tolerance.
    Real max_tol_a = tol;
    bool od[r->angle_n ()];
    od[0] = true;
    for (unsigned long l = 1; l < r->angle_n (); ++l)
      od[l] = false;
    Cluster *c = r->directions () -> first ();
    while (c) {
      Real a = *static_cast<Direction*> (c->value ()) % (*bd);
      if (a >= 1.0)
	a = 0.0;
      else if (a <= -1.0)
	a = PI;
      else
	a = acos (a);
      if (regdir % *((Direction *)c->value ()) < 0.0)
	a = PI - a;
      unsigned long l = (unsigned long) rint (a / r->angle ());
      if (l < r->angle_n ())
	od[l] = true;
      Real t = fabs (static_cast<Real> (l) * r->angle () - a);
      if (t > max_tol_a)
	t = max_tol_a;
      // If it's larger than angle_n, then it's on the other side, but
      // we check for this. If it's equal, then it's at position 0 and
      // this one is anyway true.
      c = r->directions () -> next ();
    }
    bool all = true;
    for (unsigned long l = 1; all && l < r->angle_n (); ++l)
      all = (all && od[l]);
    // If all directions are set, we do not have to check anything
    // else.
    if (all) {
      r->all_directions () = true;
    } else {
      // In this half matrix we store which regularities have been
      // found.  Row l, column k is true if an l regularity starting
      // with position k has been found (- means not present, + means
      // present and * means redundant).
      unsigned long a_n = r->angle_n () >> 1;
      HalfMatrix<char> regs (a_n + 1);
      for (unsigned long l = 1; l < a_n; ++l) {
	for (unsigned long k = 0; k < l; ++k)
	  regs (l, k) = '-';
      }
      // Now check the divisors.
      unsigned long m = 2;
      while (m <= a_n) {
	Real t = (Real)n / (Real)m;
	if (fabs (t - rint (t)) < tol_eq) {
	  // m is a divisor.
	  //
	  // Find divisors of m in order to check if regualrity has not
	  // already been verified earlier (with smaller m).
	  SimpleList<unsigned long> div;
	  unsigned long M = 2;
	  while (M < (m >> 1)) {
	    t = (Real)m / (Real)M;
	    if (fabs (t - rint (t)) < tol_eq)
	      div.add_first (new unsigned long (M));
	    ++M;
	  }
	  // Start checking if an m regularity starting with position l
	  // is there.
	  unsigned long l = 0;
	  while (l < m) {
	    // First we check if the regularity has not already been
	    // verified earlier.
	    bool reg_check = true;
	    unsigned long *MM = div.first ();
	    while (MM) {
	      unsigned long k = l % *MM;
	      if (regs (*MM, k) != '-') {
		reg_check = false;
		break;
	      }
	      MM = div.next ();
	    }
	    if (reg_check) {
	      // Check if m regularity stating with direction l is present.
	      bool reg = true;
	      unsigned long j = l;
	      while (reg && j < r->angle_n ()) {
		reg = (reg && od[j]);
		j += m;
	      }
	      if (reg) {
		// Regularity is present, note it.
		regs (m, l) = '+';
	      }
	    } else {
	      // Mark as redundant as already checked earlier.
	      regs (*MM, l) = '*';
	    }
	    ++l;
	  }
	  // Delete divisors of m.
	  div.purge ();
	}
	++m;
      }

      // Note the regularities found.
      SimpleList<unsigned long> *ra = new SimpleList<unsigned long>;
      for (unsigned long l = 1; l <= a_n; ++l) {
	for (unsigned long k = 0; k < l; ++k) {
	  if (regs (l, k) == '+') {
	    ra->add_first (new unsigned long (l));
	    break;
	  }
	}
      }
      if (ra->length () > 0) {
	r->set_regular_arrangements (ra);
      } else {
	delete ra;
      }

    }

    // Use the maximum angular tolerance to replace child directions
    // with parent directions if they are within tolerance.
    c = r->directions () -> first ();
    while (c) {
      Cluster *nc = r->directions () -> next ();
      if (c->parent_cluster () &&
	  c->parent_cluster () -> tolerance () < max_tol_a) {
	r->directions () -> add_last_check (c->parent_cluster ());
	r->directions () -> del (c);
	if (!nc)
	  nc = c->parent_cluster ();
      }
      c = nc;
    }
    
    regs->add_first (r);
    r = par_regs.unlink_first ();
  }

  // Check interesting angles between two directions, which have not been
  // considered yet.

  // To avoid check angles more than once we also have to ensure that
  // the direction plane combining the two angles is actually an
  // aspect of the current direction cluster.

  if (ap->aspects ()) {

    for (unsigned long k = 0 ; k < n; ++k) {
      for (unsigned long l = k + 1; l < n; ++l) {
	// Do only accept angles that are neither suggesting an
	// orthogonal nor a parallel relation as these are checked
	// elsewhere. Also only use positive angles, i.e. angles that
	// where not yet considered for an planar angle regular set.
	if (angles (l, k) > tol_angle_dist && 
	    fabs (angles (l, k) - PI) > tol_angle_dist ) {
	  // For the remaining angles we are looking for the smaller one
	  // again and not for the one on the correct side of an initial
	  // direction, so if it is larger than PI/2, we have to take
	  // the other angle.
	  //
	  // Check if direction plane cluster has the aspect generating
	  // the plane of the two directions (otherwise we'd check angles
	  // more than once.
	  bool found = false;
	  Aspect *a = ap->aspects () -> first ();
	  while (a) {
	    if (a->owners () . is_member (pdirs[l]) &&
		a->owners () . is_member (pdirs[k])) {
	      found = true;
	      a = NULL;
	    } else {
	      a = ap->aspects () -> next ();
	    }
	  }
	  if (found) {
	    if (angles (l, k) > PI / 2.0)
	      angles (l, k) = PI - angles (l, k);
	    // Check if the angle is regular
	    SpecialValue *rv = check_special_scalar (angles (l, k),
						     tol_angle_dist, true);
	    if (rv) {
	      // We store the regularity as a DirectionAngle
	      // Regularity, but the regular value is stored as a sub
	      // regularity. We also have to remove PI and 0 as a
	      // regular value as it might still be found
	      // approximately.
	      bool reg = false;
	      SimpleList<Ratio> **rats = rv->ratios ();
	      for (unsigned long j = 0; j < rv->base_num (); ++j) {
		if (rats[j]) {
		  if (strcmp (rv->base_ids () [j], tan_id) == 0) {
		    // Regular values based on tangens.
		    Ratio *r = rats[j]->first ();
		    while (r) {
		      Ratio *nr = rats[j]->next ();
		      Real v = atan (r->eval ());
		      if (fabs (v) < tol_eq ||
			  fabs (v - PI) < tol_eq) {
			rats[j]->del (r);
			delete r;
		      } else {
			reg = true;
		      }
		      r = nr;
		    }
		  } else {
		    // Regular values based on base values.
		    Ratio *r = rats[j]->first ();
		    while (r) {
		      Ratio *nr = rats[j]->next ();
		      Real v = r->eval () * rv->base_values () [j];
		      if (fabs (v) < tol_eq ||
			  fabs (v - PI) < tol_eq ) {
			rats[j]->del (r);
			delete r;
		      } else {
			reg = true;
		      }
		      r = nr;
		    }
		  }
		  if (rats[j]->length () == 0) {
		    // Nothing left for current regularity set, so delete it
		    // completely.
		    delete rats[j];
		    rats[j] = NULL;
		  }
		}
	      }
	      if (reg) {
		// There are regular values left, so add it as a
		// regularity.
		DirectionAngle *nr = new DirectionAngle (this,
							 rv->tolerance (), 
							 pdirs[l], pdirs[k],
							 angles (l, k));
		nr->add_sub_regularity (rv);
		da_regs->add_first (nr);
	      } else {
		// No regular values left, so skip it.
		delete rv;
	      }
	    }
	  }
	}
      }
    }
  }

  delete pdirs;

  return;
}

void Analyse::eval_conical_angle_regular (List<Cluster> *direction_cones)
{
  // Find conical angle regular arrangements for each main direction
  // cone regularity. Those that are present in sub regularities
  // should be found over the main direction cone as well.

  // Get the main clusters.
  SimpleList<Cluster> dc;
  ListNode<Cluster> *p = direction_cones->first ();
  while (p) {
    dc.add_first (p->data ());
    p = p->next ();
  }

  // Check all cone clusters, not just ones with small angle
  // variations, but consider the error in the cone for the maximal
  // number of different directions, which can be detected 
  // (n = (2 Pi) / (4 t) where t is the angle tolerance and n 
  // specifies the base angle for which we have maximal overlap, i.e.
  // B < 2 Pi / n) [see check_direction_cone].
  List<ConicalAngleRegular> *regs = new List<ConicalAngleRegular>;
  Cluster *cone = dc.unlink_first ();
  while (cone) {
    check_direction_cone (cone, regs);
    if (cone->sub_clusters ()) {
      Cluster *c = cone->sub_clusters () -> first ();
      while (c) {
	dc.add_first (c);
	c = cone->sub_clusters () -> next ();
      }
    }
    cone = dc.unlink_first ();
  }

  // Add regularities.
  if (regs->length () > 0) {
    reg_list->add_first ((List<Regularity> *)regs);
    conical_angle_regularities = regs;
  } else {
    delete regs;
    conical_angle_regularities = NULL;
  }

  return;
}

void Analyse::check_direction_cone (Cluster *ac,
				    List<ConicalAngleRegular> *regs)
{
  Real tol = ac->tolerance () + tol_angle_dist;

  // The owners of the direction cone are the direction clusters lying
  // in it. We collect the direction clusters according to the
  // tolerance of the direction cone.
  SimpleList<Entity> owners;
  ac->get_owners (&owners);
  // Check if we have to increase the tolerance for one of the
  // direction clusters has a larger tolerance and no children.
  Entity *e = owners.first ();
  while (e) {
    Entity *ne = owners.next ();
    Cluster *c = dynamic_cast<Cluster*> (e);
    if (c && c->value () -> is (SIL_DIRECTION_TYPE) &&
	c->tolerance () > tol) {
      if (c->sub_clusters () && c->sub_clusters () -> length () > 0) {
	Cluster *s = c->sub_clusters () -> first ();
	while (s) {
	  owners.add_last_check (s);
	  s = c->sub_clusters () -> next ();
	}
	owners.del (e);
      } else {
	tol = c->tolerance ();
      }
    }
    e = ne;
  }
  // Find the appropriate direction clusters.
  SimpleList<Cluster> dirs;
  e = owners.unlink_first ();
  while (e) {
    Cluster *c = dynamic_cast<Cluster*> (e);
    if (c && c->value () -> is (SIL_DIRECTION_TYPE)) {
      if (c->parent_cluster () &&
	  c->parent_cluster () -> tolerance () < tol) {
	// Only use parent if parent is part of direction plane,
	// otherwise still use cluster. Do not use paretns that are
	// not in the direction plane (would break consistency even if
	// tolerane suggests something else).
	if (!owners.is_member (c->parent_cluster ()))
	  dirs.add_last_check (c);
      } else {
	dirs.add_last_check (c);
      }
    } else {
      sil_msg_warn ("Analyse",
		    "direction cone has non-direction cluster owner");
    }
    e = owners.unlink_first (); 
  }

  // The maximal base angle for which we can still find different
  // arrangements is 2Pi / N with N = (2Pi / (4t) where t is the
  // tolerance.
  unsigned long N;
  if (tol < tol_eq) {
    // tol is too small, so use upper limit.
    N = regular_angle_max_n;
  } else {
    N = static_cast<unsigned long> (floor (PI / (2.0 * tol)));
    if (N > regular_angle_max_n)
      N = regular_angle_max_n;
  }
  if (N < 3)
    return;

  // Store the directions in an array to access them.
  unsigned long n = dirs.length ();
  // For conical angle regular directions we need at least 3
  // directions.
  if (n < 3)
    return;
  Cluster **pdirs = new Cluster*[n];
  unsigned long l = 0;
  Cluster *c = dirs.unlink_first ();
  while (c) {
    pdirs[l++] = c;
    c = dirs.unlink_first ();
  }

  sil_progress_bar_text ("Checking " + as_string (n) + 
			 " directions in cone " + 
			 as_string (*static_cast<DirectionCone*> (ac->
								  value ())) +
			 " / " + 
			 as_string (static_cast<DirectionCone*> 
				    (ac->value ()) -> angle () / PI * 180.0));

  // Directions projected in plane.
  UnitVector3 dirs_proj[n];
  // Direction for projection.
  Vector3 projdir = *((DirectionCone *)ac->value ());
  // Project directions onto plane considering the `orientation'.
  for (unsigned long k = 0; k < n; ++k) {
    Real v = projdir % *((Direction *)pdirs[k]->value ());
    dirs_proj[k] = (*((Direction *)pdirs[k]->value ()) - 
		    v * projdir) . normalise ();
    if (v < 0.0)
      dirs_proj[k] = -dirs_proj[k];
  }

  // Direction parallel to the direction of the current cone. Used to
  // detect parallel directions for the set of conical angle regular
  // arrangements.
  Cluster *direction = NULL;

  // List of regularities for this cone.
  SimpleList<ConicalAngleRegular> car_regs;

  // Find regular angle sets between the directions, by comparing the
  // angles between an initial direction and the other directions. We
  // have to check all directions as intial directions and compare the
  // angles between the initial directions and all following
  // directions in the list.  At least three more directions should be
  // in the list or we cannot get an angular regular relation and the
  // angles between two directions are checked later.
  HalfMatrix<Real> angles (n);
  // List of regular angles for each initial direction.
  SimpleList<unsigned long> **valid_angles = new SimpleList<unsigned long>*[n];
  for (unsigned long k = 0; k < n - 2; ++k) {
    // Initialize the valid angles list for k as initial direction.
    valid_angles[k] = new SimpleList<unsigned long>;
    // Compute all angles between the initial direction and the
    // following directions in the list and collect the regular
    // angles.
    //
    // A regular direction such that we can tell on which side of the
    // initial direction we are. Note that it doesn't matter whcih
    // side is which as long as we make the distinction.
    Vector3 regdir = dirs_proj[k] * (*((DirectionCone *)ac->value ()));
    for (unsigned long l = k + 1; l < n; ++l) {
      Real a = dirs_proj[k] % dirs_proj[l];
      if (a >= 1.0)
	angles (l, k) = 0.0;
      else if (a <= -1.0)
	angles (l, k) = PI;
      else
	angles (l, k) = acos (a);
      if (regdir % dirs_proj[l] < 0.0)
	angles (l, k) = 2.0 * PI - angles (l, k);
      // Find the regular angles suggested by this angle:
      if (angles (l, k) > tol) {
	unsigned long a_min = (unsigned long)floor (2.0 * PI / 
						    (angles (l, k) +
						     tol));
	if (a_min < N) {
	  if (a_min < 3)
	    a_min = 3;
	  unsigned long a_max = (unsigned long)ceil (2.0 * PI /
						     (angles (l, k) -
						      tol)) + 1;
	  if (a_max > N)
	    a_max = N;
	  for (unsigned long aa = a_min; aa < a_max; ++aa) {
	    // Check if the angle is alrady listed as a valid angle and
	    // also check for multiples.
	    bool add = true;
	    unsigned long *an = valid_angles[k]->first ();
	    while (an) {
	      unsigned long *next_an = valid_angles[k]->next ();
	      if (aa < *an) {
		Real a = (Real)*an / (Real)aa;
		if (fabs (a - rint (a)) < tol_eq) {
		  add = false;
		  break;
		}
	      } else {
		Real a = (Real)aa / (Real)*an;
		if (fabs (a - rint (a)) < tol_eq) {
		  valid_angles[k]->del (an);
		  delete an;
		}
	      }
	      an = next_an;
	    }
	    // If the angle is not listed as a valid angle, then add
	    // it to the list.
	    if (add)
	      valid_angles[k]->add_first (new unsigned long (aa));
	  }
	}
      }
    }
  }

  // Increase angular tolerance by tol_angle_dist
  tol = ac->tolerance () + tol_angle_dist;
  // NOTE: The maximal tolerance depends on the *an below. We could
  //       set the tolerance for a valid direction to PI/(2(*an))
  //       which is the maximal tolerance before the arrangements become
  //       ambiguous.

  // Now check for all initial directions if one of the suggested
  // regular angles are actually really there.
  for (unsigned long k = 0 ; k < n - 2; ++k) {

    // Check all angles with the initial direction k if they are
    // multiples of the suggested regular angles.
    unsigned long *an = valid_angles[k]->unlink_first ();
    while (an) {

      // The value of the regular angle.
      Real beta = 2.0 * PI / (Real)*an;
      // Array keeping track which direction in the list is actually
      // occupying a direction suggested buy the regular angle.
      unsigned long directions[*an + 1];
      // Tolerances for the directions list.
      Real tols[*an + 1];
      // Number of directions which are in the current regular subset.
      unsigned long num = 1;
      // Direction 0 (0 * 2PI / n) is always occupied by the initial
      // direction k.
      directions[0] = k;
      // Mark the other regular directions as not-occupied.
      for (unsigned long j = 1; j <= *an; ++j)
	directions[j] = n;
      // Check the angles between initial direction and all following
      // directions. If the angle suggest a regular arrangement add
      // the direction to the current regular subset.
      for (unsigned long j = k + 1; j < n; ++j) {
	// Compute the tolerance between the actual angle and the
	// angle suggested by the regular angle.
	unsigned long ta = (unsigned long)rint (angles (j, k) / beta);
	Real t = fabs ((Real)ta * beta - angles (j, k));
	if (t < tol && ta <= *an) {
	  // The angle is approximately regular, check if we have to
	  // add the direction to the curent subset.
	  if (directions[ta] < n) {
	    // Direction already occupied by other direction, but
	    // parallel directions, etc.  have been determiend
	    // before, so we take the one which is closer.
	    if (t < tols[ta]) {
	      directions[ta] = j;
	      tols[ta] = t;
	    }
	  } else {
	    // Direction not occupied, add it as new one.
	    ++num;
	    directions[ta] = j;
	    tols[ta] = t;
	  }
	}
      }

      if (num > 2) {
	// The current regular subset has more than two members, so
	// it could be a regularity.
	
	// First check if any of the directions is already a member of
	// a conical angle regular set with the same base angle.
	ConicalAngleRegular *icar = NULL;
	ConicalAngleRegular *car = car_regs.first ();
	while (car && !icar) {
	  // Check if angles are compatible.
	  unsigned long angl_n = car->angle_n ();
	  bool match = false;
	  if (angl_n == *an) {
	    match = true;
	  } else if (angl_n < *an) {
	    // check if: integer = *an / angl_n
	    Real r = rint ((Real)*an / (Real)angl_n);
	    match = (fabs (r * (Real)angl_n - (Real)(*an)) < tol_eq);
	  } else {
	    // check if: integer =  angl_n / *an
	    Real r = rint ((Real)angl_n / (Real)*an);
	    match = (fabs (r * (Real)(*an) - (Real)angl_n) < tol_eq);
	  }
	  if (match) {
	    for (unsigned long j = 0; j <= *an; ++j) {
	      if (directions[j] < n && 
		  car->directions () -> is_member (pdirs[directions[j]])) {
		icar = car;
		break;
	      }
	    }
	  }
	  car = car_regs.next ();
	}
	if (icar) {
	  // Combine current regular set with the one found.
	  for (unsigned long j = 0; j <= *an; ++j) {
	    if (directions[j] < n)
	      icar->directions () -> add_last_check (pdirs[directions[j]]);
	  }
	  // And set the angle.
	  if (icar->angle_n () < *an)
	    icar->angle_n () = *an;
	} else {

	  // New regularity. 

	  // Generate direction list and recompute tolerances
	  // considering distance from regular direction and from cone
	  // angle.
	  SimpleList<Cluster> *dirlist = new SimpleList<Cluster>;
	  unsigned long cnt = 0;
	  Real t = 0.0;
	  for (unsigned long j = 0; j <= *an; ++j) {
	    if (directions[j] < n) {
	      dirlist->add_first (pdirs[directions[j]]);
	      Real a = fabs (*static_cast<DirectionCone*> (ac->value ()) %
			     *static_cast<Direction*> (pdirs[directions[j]] -> 
						       value ()));
	      if (a >= 1.0)
		a = static_cast<DirectionCone*> (ac->value ()) -> angle ();
	      else
		a = fabs (acos (a) - 
			  static_cast<DirectionCone*> (ac->value ()) -> 
			  angle ());
	      t += sqrt (tols[j] * tols[j] + a * a);
	      ++cnt;
	    }
	  }
	  if (cnt > 0)
	    t /= (Real)cnt;
	  // Generate and add the current subset as a new regularity.
	  ConicalAngleRegular *nr = 
	    new ConicalAngleRegular (this, tol, ac->tolerance (), dirlist,
				     *((DirectionCone *)ac->value ()), 
				     ((DirectionCone *)ac->value ()) -> 
				     angle (), *an);
	  // Check if angle has a regular value.
	  SpecialValue *rv = 
	    check_special_scalar (((DirectionCone *)ac->value ()) ->
				  angle (), ac->tolerance (), true);
	  if (rv)
	    nr->add_sub_regularity (rv);
	  // Check if cone direction is parallel to some other extra
	  // direction.
	  Direction *v = nr->cone_direction ();
	  if(!direction) {
	    if (extra_parallel_directions) {
	      Real t = tol_angle_dist;
	      direction = find_parallel (*(static_cast<UnitVector3*> (v)),
					 extra_parallel_directions, t);
	    }
	    if (!direction) {
	      // Cone direction is not parallel to any existing
	      // direction. Generate a new direction and note it as
	      // an extra direction which is not considered for
	      // direction planes and cones, but as a direction
	      // alone.
	      direction = add_extra_parallel_direction (v);
	    }
	  }
	  // Parallel direction exists or has been created, add cone
	  // direction aspect to it.
	  Aspect *asp = new Aspect (v, nr);
	  // Update aspect counter and parent aspect counter.
	  direction->add_aspect (asp);
	  Cluster *pa = direction->parent_cluster ();
	  while (pa) {
	    pa->num_aspects () += 1;
	    pa = pa->parent_cluster ();
	  }
	  car_regs.add_first (nr);
	}
      }

      delete an;
      an = valid_angles[k]->unlink_first ();
    }

    delete valid_angles[k];
  }

  delete valid_angles;
  delete pdirs;
  
  // Add regularities to reg list
  ConicalAngleRegular *r = car_regs.unlink_first ();
  while (r) {
    
    // Check for m-regular arrangements of the directions in the cone,
    // where m is a divisor of angel_n.
    //
    // Projected base direction.
    Direction *b = static_cast<Direction*> (r->directions () -> 
					     first () -> value ());
    Real v = projdir % (*b);
    UnitVector3 bd = ((*b) - v * projdir) . normalise ();
    if (v < 0.0)
      bd = -bd;
    // Regular direction for orientation.
    Vector3 regdir = bd * projdir;
    // Create a boolean array to list occupied directions.
    bool od[r->angle_n ()];
    od[0] = true;
    for (unsigned long l = 1; l < r->angle_n (); ++l)
      od[l] = false;
    Cluster *c = r->directions () -> first ();
    while (c) {
      // Project direction on plane.
      v = projdir % *((Direction *)c->value ());
      UnitVector3 d = (*((Direction *)c->value ()) - 
		       v * projdir) . normalise ();
      if (v < 0.0)
	d = -d;
      // Compute angle.
      Real a = d % bd;
      if (a >= 1.0)
	a = 0.0;
      else if (a <= -1.0)
	a = PI;
      else
	a = acos (a);
      if (regdir % d < 0.0)
	a = 2.0 * PI - a;
      unsigned long l = (unsigned long) rint (a / r->angle ());
      if (l < r->angle_n ())
	od[l] = true;
      // If it's larger than angle_n, then it's not really possible
      // (>2Pi!)  If it's equal, then it's at position 0 and this one
      // is anyway true.
      c = r->directions () -> next ();
    }
    bool all = true;
    for (unsigned long l = 1; all && l < r->angle_n (); ++l)
      all = (all && od[l]);
    // If all directions are set, we do not have to check anything
    // else.
    if (all) {
      r->all_directions () = true;
      // Check if it is an orthogonal system.
      r->orthogonal () = (r->angle_n () == 3) &&
	(fabs (r->cone_angle () - asin (sqrt (2.0 / 3.0))) < tol);
    } else if (r->angle_n () > 3) {
      // For angle_n == 3 we can actually only check if all directions
      // are there. The check is just for seucrity as we should
      // usually have all directions in this case anyway.
      //
      // In this half matrix we store which regularities have been
      // found.  Row l, column k is true if an l regularity starting
      // with position k has been found (- means not present, + means
      // present and * means redundant).
      unsigned long a_n = r->angle_n () >> 1;
      HalfMatrix<char> regs (a_n + 1);
      for (unsigned long l = 1; l < a_n; ++l) {
	for (unsigned long k = 0; k < l; ++k)
	  regs (l, k) = '-';
      }
      // Now check the divisors.
      unsigned long m = 2;
      while (m < a_n) {
	Real t = (Real)n / (Real)m;
	if (fabs (t - rint (t)) < tol_eq) {
	  // m is a divisor.
	  //
	  // Find divisors of m in order to check if regualrity has not
	  // already been verified earlier (with smaller m).
	  SimpleList<unsigned long> div;
	  unsigned long M = 2;
	  while (M < (m >> 1)) {
	    t = (Real)m / (Real)M;
	    if (fabs (t - rint (t)) < tol_eq)
	      div.add_first (new unsigned long (M));
	    ++M;
	  }
	  // Start checking if an m regularity starting with position l
	  // is there.
	  unsigned long l = 0;
	  while (l < m) {
	    // First we check if the regularity has not already been
	    // verified earlier.
	    bool reg_check = true;
	    unsigned long *MM = div.first ();
	    while (MM) {
	      unsigned long k = l % *MM;
	      if (regs (*MM, k) != '-') {
		reg_check = false;
		break;
	      }
	      MM = div.next ();
	    }
	    if (reg_check) {
	      // Check if m regularity stating with direction l is present.
	      bool reg = true;
	      unsigned long j = l;
	      while (reg && j < r->angle_n ()) {
		reg = (reg && od[j]);
		j += m;
	      }
	      if (reg) {
		// Regularity is present, note it.
		regs (m, l) = '+';
	      }
	    } else {
	      // Mark as redundant as already checked earlier.
	      regs (*MM, l) = '*';
	    }
	    ++l;
	  }
	  // Delete divisors of m.
	  div.purge ();
	}
	++m;
      }

      // Note the regularities found.
      SimpleList<unsigned long> *ra = new SimpleList<unsigned long>;
      for (unsigned long l = 1; l <= a_n; ++l) {
	for (unsigned long k = 0; k < l; ++k) {
	  if (regs (l, k) == '+') {
	    ra->add_first (new unsigned long (l));
	    break;
	  }
	}
      }
      if (ra->length () > 0) {
	r->set_regular_arrangements (ra);
      } else {
	delete ra;
      }

    }
    
    regs->add_first (r);
    r = car_regs.unlink_first ();
  }

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
