// analyse_axes.cc: Analyse a SiL model / Axes
//
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_axes.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_axes,
       "$Id: analyse_axes.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "kernel/kernel.hxx"
#include "kernel/matrix.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"
#include "kernel/geometry.hxx"
#include "kernel/pair.hxx"
#include "kernel/lapack.h"

void Analyse::eval_axis_intersections (List<Cluster> *dirs)
{
  // Get base direction clusters.
  List<Cluster> dcl;
  ListNode<Cluster> *an1 = dirs->first ();
  while (an1) {
    an1->data () -> get_all_base_clusters (&dcl);
    an1 = an1->next ();
  }

  // At least two different directions
  if (dcl.length () < 2)
    return;

  // Get the owners for all directions (grouping them by base directions).
  SimpleList<Entity> *owners[dcl.length ()];
  unsigned long num = 0;
  unsigned long l = 0;
  ListNode<Cluster> *dir = dcl.first ();
  while (dir) {
    owners[l] = new SimpleList<Entity>;
    dir->data () -> get_owners (owners[l]);
    num += owners[l]->length ();
    dir = dir->next ();
    ++l;
  }

  sil_progress_bar_text ("Intersecting " + as_string (num) + " axes");

  // Create all intersection points of directions and positions from
  // owners, but only between directions belonging to different base
  // direction clusters, i.e. which are not approximately parallel.
  List<Aspect> intersections;
  for (l = 0; l < dcl.length () - 1; ++l) {
    Entity *e1 = owners[l]->first ();
    while (e1) {
      // Get direction and positions of owner of first base direction.
      SimpleList<Position> pos1;
      Direction *d1;
      bool straight1 = false; // have to check vertices of straight edges
      if (e1->is (SIL_FACE_TYPE)) {
	d1 = ((Face *)e1)->get_axes (&pos1);
      } else if (e1->is (SIL_EDGE_TYPE)) {
	d1 = ((Edge *)e1)->get_axes (&pos1);
	if (d1)
	  straight1 = ((Edge *)e1)->geometry () -> is (SIL_STRAIGHT_TYPE);
      } else if (e1->is (SIL_CONICALANGLEREGULAR_TYPE) ||
		 e1->is (SIL_PLANARANGLEREGULAR_TYPE)) {
	// NOTE: We ignore conical and planar angle regular arrangements 
	//       even if their directions might be associated with a 
	//       position if the directions in the plane/cone relate to
	//       intersecting axes. It does not appear to be an important
	//       regularity and complicates the logic for the intersection
	//       algorithm. There is no reason to avoid this in a later
	//       version.
	d1 = NULL;
      } else {
	sil_msg_warn ("Analyse", "Unkown owner " + e1->type_id () +
		      " for axis intersection check");
	d1 = NULL;
      }
      if (d1) {
	// Intersect with remaining axes.
	for (unsigned long k = l + 1; k < dcl.length (); ++k) {
	  Entity *e2 = owners[k]->first ();
	  while (e2) {
	    // Get direction and positions of owner of second base direction.
	    SimpleList<Position> pos2;
	    Direction *d2;
	    bool straight2 = false; // have to check vertices of straight edges
	    if (e2->is (SIL_FACE_TYPE)) {
	      d2 = ((Face *)e2)->get_axes (&pos2);
	    } else if (e2->is (SIL_EDGE_TYPE)) {
	      d2 = ((Edge *)e2)->get_axes (&pos2);
	      if (d2)
		straight2 = ((Edge *)e2)->geometry () -> 
		  is (SIL_STRAIGHT_TYPE);
	    } else {
	      // We ignore conical and planar angle regular
	      // arrangements even if their directions (see above). As
	      // we already warn above, we do not warn again if there
	      // is an unkown owner.
	      d2 = NULL;
	    }
	    if (d2) {
	      // Check if not straight edges with identical vertices.
	      if (straight1 && straight2) {
		Vertex *sv1 = ((Edge *)e1) -> start_vertex ();
		Vertex *ev1 = ((Edge *)e1) -> end_vertex ();
		Vertex *sv2 = ((Edge *)e2) -> start_vertex ();
		Vertex *ev2 = ((Edge *)e2) -> end_vertex ();
		if (sv1 == sv2 || sv1 == ev2 || ev1 == sv2 || ev1 == ev2)
		  d2 = NULL;
	      }
	      if (d2) {
		// Some of the directions could still be parallel as the
		// extra directions are added later and parallel
		// directions for this are noted only by a
		// sub-regularity. So we still have to check the angle
		// between the directions.
		Real a = fabs ((*d1) % (*d2));
		if (a >= 1.0)
		  a = 0.0;
		else
		  a = acos (a);
		if (a >= tol_angle_dist) {
		  // Compute approx. intersection points between positions.
		  Position *p1 = pos1.first ();
		  while (p1) {
		    Position *p2 = pos2.first ();
		    while (p2) {
		      Vector3 ip;
		      Real dist = intersect_axes (p1, d1, p2, d2, ip);
		      if (dist >= 0.0) {
			// Found a possible intersection.
			IntersectionPosition *ipos =
			  new IntersectionPosition (ip, dist);
			Aspect *a = new Aspect (ipos, p1);
			a->owners () . add_last (d1);
			a->owners () . add_last (p2);
			a->owners () . add_last (d2);
			intersections.add_first (a);
		      }
		      p2 = pos2.next ();
		    }
		    p1 = pos1.next ();
		  }
		}
	      }
	    }
	    e2 = owners [k]->next ();
	  }
	}
      }
      e1 = owners[l]->next ();
    }
  }

  if (intersections.length () > 0) {

    sil_progress_bar_text ("Clustering " + 
			   as_string (intersections.length ()) + 
			   " intersection points");

    List<Cluster> *clusters = new List<Cluster>;
    // Cluster the intersection points.
    cluster_aspects (&intersections, tol_eq, tol_len_dist, NAN, true, 
		     clusters);

    // Find tolerances.
    SimpleList<Real> tols;
    tols.add_first (new Real (tol_len_dist));
    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      if (cn->data () -> tolerance () > tol_len_dist)
	tols.add_first (new Real (cn->data () -> tolerance ()));
      cn = cn->next ();
    }

    tolerance_levels (clusters, &tols, tol_len_dist, 0.0);

    // Find regular intersection point clusters.
    List<Cluster> bcls;
    cn = clusters->first ();
    while (cn) {
      cn->data () -> get_all_base_clusters (&bcls);
      cn = cn->next ();
    }
    cn = bcls.first ();
    while (cn) {
      Cluster *c = cn->data ();

      // Count number of different axes. We have to use the aspects as an
      // axis is a combiantion of the direction and the position, both
      // owners of the aspect.
      SimpleList<Aspect> aspects;
      c->get_aspects (&aspects);
      // Collect unique position/direction pairs.
      List<Direction> dirs;
      SimpleList<Position> pos;
      Aspect *a = aspects.first ();
      while (a) {
	// Owners are added as p1, d1, p2, d2.
	Position *p1 = dynamic_cast<Position*> (a->owners () . first ());
	Direction *d1 = dynamic_cast<Direction*> (a->owners () . next ());
	Position *p2 = dynamic_cast<Position*> (a->owners () . next ());
	Direction *d2 = dynamic_cast<Direction*> (a->owners () . next ());
	if (p1 && d1 && p2 && d2) {
	  ListNode<Direction> *d = dirs.first ();
	  Position *p = pos.first ();
	  while (d && (d1 || d2)) {
	    if (d->data () == d1 && p == p1)
	      d1 = NULL;
	    if (d->data () == d2 && p == p2)
	      d2 = NULL;
	    d = d->next ();
	    p = pos.next ();
	  }
	  if (d1) {
	    dirs.add_first (d1);
	    pos.add_first (p1);
	  }
	  if (d2) {
	    dirs.add_first (d2);
	    pos.add_first (p2);
	  }
	} else {
	  sil_msg_warn ("Analyse", 
			"intersection position cluster owners mismatch");
	}
	a = aspects.next ();
      }
      // If we have n different positions, then n(n-1)/2 axes would
      // have to be present for the intersection if we have a
      // transitive structure in the cluster. If the cluster is not
      // transitive, it is not regular.
      unsigned long n = dirs.length ();
      n = ((n * (n - 1)) >> 1);

      if (n != c->num_aspects ()) {
	// Note that we do not have intersection points for parallel
	// axes, so if n is the number of different positions and thus
	// axes, the maximum number of intersection points for these
	// axes is n(n-1)/2 - num_par where num_par is the number of
	// parallel axes pairs. If we have so many intersection
	// points, then the cluster is transitive.
	
	// Count number of parallel directions as we do not have
	// intersection points for them.
	unsigned long num_par = 0;
	ListNode<Direction> *d1 = dirs.first ();
	while (d1) {
	  ListNode<Direction> *d2 = d1->next ();
	  while (d2) {
	    Real a = fabs ((*d1->data ()) % (*d2->data ()));
	    if (a >= 1.0)
	      a = 0.0;
	    else
	      a = acos (a);
	    if (a < c->analyse () -> tolerance_angle_dist ())
	      ++num_par;
	    d2 = d2->next ();
	  }
	  d1 = d1->next ();
	}

	n -= num_par;
      }


      if (c->num_aspects () == n) {
	// Cluster is regular

	// Check if distance is regular.
	SubRegularity *sr = check_special_scalar 
	  (static_cast<IntersectionPosition*> (c->value ()) -> 
	   intersection_tolerance (), c->tolerance (), false);
	if (sr)
	  c->add_sub_regularity (sr);
	
	if (c->parent_cluster ())
	  bcls.add_last_check (c->parent_cluster ());

	if (!sr && c->num_aspects () < 2) {
	  // Only keep single intersection points with regular
	  // distance values or at least two approximately equal
	  // intersection points.
	  if (!c->collapse ()) {
	    // c is top level cluster, so remove it, but add it's sub
	    // clusters.
	    clusters->del (c);
	    if (c->sub_clusters ()) {
	      Cluster *sc = c->sub_clusters () -> unlink_first ();
	      while (sc) {
		clusters->add_first (sc);
		sc->parent_cluster () = NULL;
		sc = c->sub_clusters () -> unlink_first ();
	      }
	    }
	  }
	  delete c;
	}

      } else {

	// Cluster is not regular, collapse it (i.e. move everyting to
	// parent if possible).
	if (c->collapse ()) {
	  // Cluster collpase, so remove it and it's parent to the list.
	  bcls.add_last_check (c->parent_cluster ());
	} else {
	  // c is top level cluster, so remove it, but add it's sub
	  // clusters.
	  clusters->del (c);
	  if (c->sub_clusters ()) {
	    Cluster *sc = c->sub_clusters () -> unlink_first ();
	    while (sc) {
	      clusters->add_first (sc);
	      sc->parent_cluster () = NULL;
	      sc = c->sub_clusters () -> unlink_first ();
	    }
	  }
	}
	delete c;

      }

      cn = cn->next ();
    }

    // Add regularities.
    if (clusters->length () > 0) {
      reg_list->add_first ((List<Regularity> *)clusters);
      intersection_positions = clusters;
    } else {
      delete clusters;
    }

  }

  // Delete owner lists.
  l = 0;
  while (l < dcl.length ())
    delete owners[l++];

  return;
}

Real Analyse::intersect_axes (Position *p1, Direction *d1,
			      Position *p2, Direction *d2,
			      Vector3 &ip) const
{
  // Approximate intersection.

  // Normal of plane the two directions are in:
  
  // computes c2 on line 2 closest to line 1
  // computes c1 on line 1 closest to line 2

  // connecting line is orthogonal to both
  Vector3 n = (*d1) * (*d2);

  // check for near-parallel lines
  // should actually never happen, but for security...
  if (n.norm () > tol_eq) {

    // Use the plane with point p1 and directions d1, n.
    // Intersecting this plane with the line 2 will yield c2:
    //   p1 + t2 * d1 + s2 * n = p2 - r2 * d2
    //   (p1 - p2) + t2 * d1 + s2 * n + r2 * d2 = 0
    //   c2 = p2 - r2 * d2
    Real t2, r2, s2;
    if (solve_lineq_3 (static_cast<Vector3> (*d1), static_cast<Vector3> (*d2),
		       n, t2, r2, s2, (*p2) - (*p1))) {
      Vector3 c2 = (*p2) - r2 * (*d2);
      
      // Use the plane with point p2 and directions d2, n.
      // Intersecting this plane with the line 1 will yield c1:
      //   p2 + t1 * d2 + s1 * n = p1 - r1 * d1
      //   (p2 - p1) + t1 * d2 + s1 * n + r1 * d1 = 0
      //   c1 = p1 - r1 * d1
      Real t1, r1, s1;
      if (solve_lineq_3 (static_cast<Vector3> (*d1), static_cast<Vector3>(*d2),
			 n, r1, t1, s1, (*p1) - (*p2))) {
	Vector3 c1 = (*p1) - r1 * (*d1);
	
	// Approximate intersection point:
	ip = (c1 + c2) / 2.0;
	// Distance:
	return (c1 - c2) . norm ();
      }
    }
  }
  // No intersection.
  return -1.0;
}

bool Analyse::solve_lineq_3 (Vector3 const &m0, Vector3 const &m1,
			     Vector3 const &m2, Real &x0, Real &x1,
			     Real &x2, Vector3 const &c) const
{
  // Solve: [m0, m1, m2] [x0; x1; x2] = c
  
  // Pivot-Gauss algorithm.

  Real *R1 = new Real[4];
  Real *R2 = new Real[4];
  Real *R3 = new Real[4];
  Real *HR;

  *R1 = m0.x ();
  R1[1] = m1.x ();
  R1[2] = m2.x ();
  R1[3] = c.x ();
  *R2 = m0.y ();
  R2[1] = m1.y ();
  R2[2] = m2.y ();
  R2[3] = c.y ();
  *R3 = m0.z ();
  R3[1] = m1.z ();
  R3[2] = m2.z ();
  R3[3] = c.z ();

  if (fabs (*R1) < fabs (*R2)) {
    HR = R2;
    R2 = R1;
    R1 = HR;
  }
  if (fabs (*R1) < fabs (*R3)) {
    HR = R3;
    R3 = R1;
    R1 = HR;
  }
  if (fabs (*R1) < sil_get_epsilon ()) {
    delete R1;
    delete R2;
    delete R3;
    return false;
  }

  Real f = *R2 / *R1;
  R2[1] -=  R1[1] * f;
  R2[2] -=  R1[2] * f;
  R2[3] -=  R1[3] * f;
  f = *R3 / *R1;
  R3[1] -= R1[1] * f;
  R3[2] -= R1[2] * f;
  R3[3] -= R1[3] * f;

  if (fabs (R2[1]) < fabs (R3[1])) {
    HR = R3;
    R3 = R2;
    R2 = HR;
  }
  if (fabs (R2[1]) < sil_get_epsilon ()) {
    delete R1;
    delete R2;
    delete R3;
    return false;
  }

  f = R3[1] / R2[1];
  R3[2] -= R2[2] * f;
  if (fabs (R3[2]) < sil_get_epsilon ()) {
    delete R1;
    delete R2;
    delete R3;
    return false;
  }
  R3[3] -= R2[3] * f;

  x2 = R3[3] / R3[2];
  x1 = (R2[3] - R2[2] * x2) / R2[1];
  x0 = (R1[3] - R1[1] * x1 - R1[2] * x2) / *R1;

  delete R1;
  delete R2;
  delete R3;
  
  return true;
}

void Analyse::eval_axis_positions (List<Cluster> *dirs)
{
  // Get base direction clusters.
  List<Cluster> dcl;
  ListNode<Cluster> *an1 = dirs->first ();
  while (an1) {
    an1->data () -> get_all_base_clusters (&dcl);
    an1 = an1->next ();
  }

  // Each cluster represents a set of parallel axes with respect to
  // the cluster tolerance. Starting with the base direction clusters
  // move upwards in the cluster hierarchy and check for aligned axis
  // on each tolerance level. Only accept it as a new aligned axis
  // tolerance if it is not completely part of a sub-cluster
  // (otherwise it should have been found earlier).

  List<Regularity> *rpls = new List<Regularity>;
  List<Regularity> *pcls = new List<Regularity>;
  List<RegularPointPosition> *raps = new List<RegularPointPosition>;
  List<RegularPointCircle> *rpcs = new List<RegularPointCircle>;
  List<Cluster> *aligned_axes = new List<Cluster>;

  ListNode<Cluster> *dn = dcl.unlink_first ();
  while (dn) {

    // Check parallel axis cluster for aligned axes if angle between
    // average direction and directions is smaller than 45
    // degree. Larger angles cannot be considered for aligned axes in
    // a consistent way.

    if (dn->data () -> tolerance () < PI / 4.0 - tol_angle_dist) {

      List<Cluster> *res = check_parallel_axes (dn->data ());
      
      if (res) {

	// Check for regular arrangements.
	check_aligned_axes (res, raps, rpls, pcls, rpcs);

	// Aligned axes list (we also keep single axes in the list as
	// they might be on one of the grids and it doesn't hurt to
	// keep them).
	aligned_axes->combine (res);

	delete res;
      }

    }
    
    if (dn->data () -> parent_cluster ())
      dcl.add_last_check (dn->data () -> parent_cluster ());
    delete dn;
    dn = dcl.unlink_first ();
  }

  if (aligned_axes->length () > 0)
    reg_list->add_first ((List<Regularity> *)aligned_axes);
  else
    delete aligned_axes;

  if (rpls->length () > 0)
    reg_list->add_first (rpls);
  else
    delete rpls;

  if (raps->length () > 0)
    reg_list->add_first ((List<Regularity> *)raps);
  else
    delete raps;

  if (pcls->length () > 0)
    reg_list->add_first (pcls);
  else
    delete pcls;

  if (rpcs->length () > 0)
    reg_list->add_first ((List<Regularity> *)rpcs);
  else
    delete rpcs;

  return;
}

List<Cluster> *Analyse::check_parallel_axes (Cluster *dirs)
{
  // Find aligned axes.

  sil_progress_bar_text ("Collecting axis positions for direction " +
			 as_string (*static_cast<Direction*> (dirs->
							      value ())));

  // Get the owners of the axes.
  SimpleList<Entity> owners;
  dirs->get_owners (&owners);
  if (owners.length () < 2)
    return NULL;

  // Projection direction is average of direction cluster.
  UnitVector3 n = *static_cast<UnitVector3*> (static_cast<Direction*>
					      (dirs->value ()));

  // Get all axis positions.
  Real avg_d = 0.0;
  List<Aspect> proj_pts;
  SimpleList<Direction> proj_dirs;
  Entity *e = owners.first ();
  while (e) {
    // Get direction and positions of owner of first base direction.
    SimpleList<Position> pos;
    Direction *d;
    if (e->is (SIL_FACE_TYPE)) {
      d = ((Face *)e)->get_axes (&pos);
    } else if (e->is (SIL_EDGE_TYPE)) {
      d = ((Edge *)e)->get_axes (&pos);
    } else if (e->is (SIL_CONICALANGLEREGULAR_TYPE)) {
      d = NULL;
      if (intersection_positions) {
	ConicalAngleRegular *car = (ConicalAngleRegular *)e;
	// If the axes that form a concial angle regularity intersect,
	// then find the intersection point in the axis intersection 
	// point list and add it here as an axes for regular arrangements.
	//
	// Get the owners of the cone directions sorted by the direction they
	// occupy.
	SimpleList<Entity> *cone_owners[car->angle_n ()];
	for (unsigned long l = 0; l < car->angle_n (); ++l)
	  cone_owners[l] = NULL;
	unsigned long num = 0;
	Cluster *c = car->directions () -> first ();
	UnitVector3 bd = *static_cast<Direction*> (c->value ());
	UnitVector3 projdir = (*car->cone_direction ());
	Vector3 regdir = bd * projdir;
	while (c) {
	  Direction *dd = static_cast<Direction*> (c->value ());
	  Real v = projdir % (*dd);
	  UnitVector3 pd = (*dd - v * projdir) . normalise ();
	  if (v < 0.0)
	    pd = -pd;
	  Real a = bd % pd;
	  if (a >= 1.0)
	    a = 0.0;
	  else if (a <= -1.0)
	    a = PI;
	  else
	    a = acos (a);
	  if (regdir % pd < 0.0)
	    a = 2.0 * PI - a;
	  unsigned long m = static_cast<unsigned long> 
	    (rint (static_cast<Real> (car->angle_n ()) * a / (2.0 * PI)));
	  if (m >= car->angle_n ())
	    m = 0;
	  if (!cone_owners[m]) {
	    cone_owners[m] = new SimpleList<Entity>;
	    ++num;
	  }
	  c->get_owners (cone_owners[m]);
	  c = car->directions () -> next ();
	}
	// Find the intersection points of the axes.
	ListNode<Cluster> *ip = intersection_positions->first ();
	while (ip) {
	  // Check if axes of intersection position contain the axes
	  // of the cone as a subset. At least one owner per occupied
	  // cone direction has to be present.
	  SimpleList<Entity> ip_owners;
	  ip->data () -> get_owners (&ip_owners);
	  if (ip_owners.length () >= num) {
	    // Check if cone_owners are a subset of intersection point
	    // owners.
	    bool subset = true;
	    for (unsigned long l = 0; l < car->angle_n (); ++l) {
	      if (cone_owners[l]) {
		bool found = false;
		Entity *e = cone_owners[l]->first ();
		while (e) {
		  if (ip_owners.is_member (e)) {
		    found = true;
		    break;
		  }
		  e = cone_owners[l]->next ();
		}
		if (!found) {
		  subset = false;
		  break;
		}
	      }
	    }
	    if (subset) {
	      // Yes, it's a subset. We have the direction from the
	      // angle regularity and we have the position from the
	      // intersection, so we can use the axis to generate
	      // another aspect here.
	      d = car->cone_direction ();
	      pos.add_first (static_cast<Position*> 
			     ((IntersectionPosition *)ip->data () -> 
			      value ()));
	    }
	  }
	  ip = ip->next ();
	}

	// Delete cone owner lists.
	for (unsigned long l = 0; l < car->angle_n (); ++l) {
	  if (cone_owners[l])
	    delete cone_owners[l];
	}
      }
    } else if (e->is (SIL_PLANARANGLEREGULAR_TYPE)) {
      d = NULL;
      PlanarAngleRegular *par = (PlanarAngleRegular *)e;
      if (intersection_positions &&
	  par->angle_n () > 0) {
	// If the axes that form a planar angle regularity intersect,
	// then find the intersection point in the axis intersection 
	// point list and add it here as an axes for regular arrangements.
	//
	// Get the owners of the plane directions.
	unsigned long num = 0;
	SimpleList<Entity> *plane_owners[par->angle_n () + 1];
	for (unsigned long j = 0; j <= par->angle_n (); ++j)
	  plane_owners[j] = NULL;
	UnitVector3 bd = *static_cast<Direction*> (par->directions () ->
						   first () -> value ());
	Cluster *D = par->directions () -> first ();
	while (D) {
	  Real a = fabs (*static_cast<Direction*> (D->value ()) % bd);
	  if (a >= 1.0)
	    a = 0.0;
	  else
	    a = acos (a);
	  unsigned long j = 
	    static_cast<unsigned long> (rint ((a * par->angle_n ()) / PI));
	  if (!plane_owners[j]) {
	    plane_owners[j] = new SimpleList<Entity>;
	    ++num;
	  }
	  D->get_owners (plane_owners[j]);
	  D = par->directions () -> next ();
	}
	// Find the intersection points of the axes.
	ListNode<Cluster> *ip = intersection_positions->first ();
	while (ip) {
	  // Check if axes of intersection position contain the axes
	  // of the plane as a subset. At least one owner per occupied
	  // plane direction has to be present.
	  SimpleList<Entity> ip_owners;
	  ip->data () -> get_owners (&ip_owners);
	  if (ip_owners.length () >= num) {
	    // Check each direction:
	    bool subset = true;
	    for(unsigned long l = 0; l < par->angle_n () + 1; ++l) {
	      if (plane_owners[l]) {
		bool found = false;
		Entity *e = plane_owners[l]->first ();
		while (e) {
		  if (ip_owners.is_member (e)) {
		    found = true;
		    break;
		  }
		  e = plane_owners[l]->next ();
		}
		if (!found) {
		  subset = false;
		  break;
		}
	      }
	    }
	    if (subset) {
	      // Yes, it's a subset. We have the direction from the
	      // angle regularity and we have the position from the
	      // intersection, so we can use the axis to generate
	      // another aspect here.
	      d = static_cast<Direction *> (par->direction_plane () 
					    -> value ());
	      pos.add_first (static_cast<Position*> 
			     ((IntersectionPosition *)ip->data () ->
			      value ()));
	    }
	  }
	  ip = ip->next ();
	}
	// Delete cone direction owners.
	for (unsigned long j = 0; j <= par->angle_n (); ++j) {
	  if (plane_owners[j])
	    delete plane_owners[j];
	}
      }
    } else {
      sil_msg_warn ("Analyse", "Unkown owner " + e->type_id () +
		    " for aligned axes check");
      d = NULL;
    }
    if (d) {
      // Skip directions orthogonal to plane normal (only for
      // security, detailed check earlier).
      Real nd = fabs (n % (*d));
      if (nd > tol_eq) {
	// Collect the positions (the projection is done later).
	Position *p = pos.first ();
	while (p) {
	  Real f = (*p) % n;
	  avg_d += f;
	  proj_pts.add_first (new Aspect (new AlignedAxes (*p, dirs), e));
	  proj_dirs.add_first (d);
	  p = pos.next ();
	}
      } else {
	// There are orthogonal directions in the direction cluster,
	// so drop the whole thing as we can't possibly get a
	// consistent result. Something went wrong...
	proj_pts.purge ();
	return NULL;
      }
    }
    e = owners.next ();
  }

  // Less than two points left, so nothing left to do.
  if (proj_pts.length () < 2) {
    proj_pts.purge ();
    return NULL;
  }

  // Project the collected axis positions onto a plane defined by the
  // projection direction n.
  avg_d /= static_cast<Real> (proj_pts.length ());
  ListNode<Aspect> *A = proj_pts.first ();
  Direction *d = proj_dirs.first ();
  while (d) {
    AlignedAxes *a = static_cast<AlignedAxes*> (A->data () -> feature ());
    // Compute the intersection point between the axis and the average
    // plane.
    //
    //   <n;x> = avg_d; x = pos + td  =>  t = (avg_d - <n;pos>) / <n;d>
    //
    Real t = (avg_d - (n % (*a))) / (n % (*d));
    a->set ((*a) + t * (*d));
    A = A->next ();
    d = proj_dirs.next ();
  }

  // Cluster points on the plane to find aligned axes.
  sil_progress_bar_text ("Clustering " + as_string (proj_pts.length ()) +
			 " axes");
  List<Cluster> *clusters = new List<Cluster>;
  cluster_aspects (&proj_pts, tol_eq, tol_len_dist, NAN, true,
		   clusters);

  if (dirs->sub_clusters ()) {

    // Check the resulting clusters. An aligned axes cluster is
    // considered to be regular if it contains axes from different
    // sub-clusters of the direction clusters dirs. Otherwise it
    // should have been detected earlier or it is not at all valid. Of
    // course the direction cluster has to have sub-clusters in order
    // for this check to be required at all.

    SimpleList<Cluster> dcl;
    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      dcl.add_first (cn->data ());
      cn = cn->next ();
    }
    Cluster *c = dcl.unlink_first ();
    while (c) {
      SimpleList<Entity> owners;
      c->get_owners (&owners);

      Cluster *cl1 = NULL;
      // Find direction sub cluster for first owner.
      Entity *e = owners.first ();
      Cluster *sc = dirs->sub_clusters () -> first ();
      while (sc) {
	SimpleList<Entity> d_owners;
	sc->get_owners (&d_owners);
	if (d_owners.is_member (e)) {
	  cl1 = sc;
	  break;
	}
	sc = dirs->sub_clusters () -> next ();
      }
      // Now we need a subcluster different from cl1. If cl1 is NULL,
      // any sub-cluster will do. Otherwise we look for a really
      // different one.
      bool different_sub = false;
      e = owners.next ();
      while (e) {
	Cluster *cl2 = NULL;
	Cluster *sc = dirs->sub_clusters () -> first ();
	while (sc) {
	  SimpleList<Entity> d_owners;
	  sc->get_owners (&d_owners);
	  if (d_owners.is_member (e)) {
	    cl2 = sc;
	    break;
	  }
	  sc = dirs->sub_clusters () -> next ();
	}
	if (cl2 != cl1) {
	  different_sub = true;
	  break;
	}
	e = owners.next ();
      }
      if (different_sub) {
	// Found a different sub--cluster, so cluster is valid. Add it's
	// sub clusters for further checking.
	if (c->sub_clusters ()) {
	  Cluster *sc = c->sub_clusters () -> first ();
	  while (sc) {
	    dcl.add_first (sc);
	    sc = c->sub_clusters () -> next ();
	  }
	}
      } else {
	// All owners belong to the same direction sub cluster. Eliminate the
	// cluster.
	if (!c->collapse ()) {
	  // Cluster is top cluster, so we have to remove it completely.
	  clusters->del (c);
	}
	delete c;
      }
      c = dcl.unlink_first ();
    }
  }

  if (clusters->length () == 0) {
    delete clusters;
    return NULL;
  }

  return clusters;
}

void Analyse::check_aligned_axes (List<Cluster> *aligned_axes,
				  List<RegularPointPosition> *raps,
				  List<Regularity> *rpls,
				  List<Regularity> *pcls,
				  List<RegularPointCircle> *rpcs)
{
  if (f_axis_regpos) {

    // Find axis positions on lines and 2D grids.

    // Get the points in the plane.
    List<Cluster> points;
    ListNode<Cluster> *cn = aligned_axes->first ();
    // Normal of the plane (should be the same for all points, so the
    // first one should be fine).
    Cluster *plane_dir = static_cast<AlignedAxes*> (cn->data () -> value ())
      -> direction ();
    Direction *plane_normal = static_cast<Direction*> (plane_dir->value ());
    while (cn) {
      cn->data () -> get_all_base_clusters (&points);
      cn = cn->next ();
    }

    if (points.length () < 3)
      return;

    sil_progress_bar_text ("Checking for linear arrangements of " +
			   as_string (points.length ()) + " axes");

    // For each point generate the lines to the remaining points and
    // cluster them with respect to direction considering their
    // distance from the point.
    List<Aspect> aspects;
    ListNode<Cluster> *pn1 = points.first ();
    while (pn1) {
      ListNode<Cluster> *pn2 = pn1->next ();
      while (pn2) {
	Vector3 p1 = *static_cast<AlignedAxes*> (pn1->data () -> value ());
	Vector3 p2 = *static_cast<AlignedAxes*> (pn2->data () -> value ());
	Aspect *a =  new Aspect (new ParallelLine (p2 - p1), pn1->data ());
	a->owners () . add_first (pn2->data ());
	aspects.add_first (a);
	pn2 = pn2->next ();
      }
      pn1 = pn1->next ();
    }

    // Cluster lines into parallel lines.
    List<Cluster> parallel;
    cluster_aspects (&aspects, tol_eq, tol_len_dist, NAN, true, &parallel);

    // Get the tolerances and recompute average direction using a
    // least squares approximation.
    SimpleList<Real> tols;
    tols.add_first (new Real (tol_len_dist));
    SimpleList<Cluster> cls;
    cn = parallel.first ();
    while (cn) {
      cls.add_first (cn->data ());
      cn = cn->next ();
    }
    Cluster *c = cls.unlink_first ();
    while (c) {

      // Solve system in least squares sense using SVD.
      SimpleList<Aspect> aspects;
      c->get_aspects (&aspects);
      
      ParallelLine *dd = static_cast<ParallelLine*> (c->value ());
      UnitVector3 d (dd->x (), dd->y (), dd->z ());

      if (aspects.length () > 2) {
	
	// If we have more than two aspects do a least squares fit.
	
	char JOBU = 'S'; // only compute the left singular vectors
	char JOBVT = 'O'; // only compute the right singular vectors in A
	long M = aspects.length (); // number of rows
	long N = 3; // number of columns
	Real A[M * N]; // direction matrix
	Real U[M * 3]; // U matrix (only the first three vectors)
	Real S[3]; // singular values
	long LWORK = 50 + M; // work space
	Real WORK[LWORK];
	long INFO = 0; // result
	
	// Initialize A
	Aspect *a = aspects.first ();
	long l1 = 0;
	long l2 = M;
	long l3 = l2 + M;
	while (a) {
	  ParallelLine *DD = dynamic_cast<ParallelLine*> (a->feature ());
	  if (DD) {
	    UnitVector3 D (DD->x (), DD->y (), DD->z ());
	    if (D % d >= 0.0) {
	      A[l1++] = D.x ();
	      A[l2++] = D.y ();
	      A[l3++] = D.z ();
	    } else {
	      // Adjust `orientation'.
	      A[l1++] = -D.x ();
	      A[l2++] = -D.y ();
	      A[l3++] = -D.z ();
	    }
	  } else {
	    sil_msg_warn ("Analyse",
			  "parallel direction aspect has non-direction feature " +
			  a->feature () -> type_id ());
	    A[l1++] = 0.0;
	    A[l2++] = 0.0;
	    A[l3++] = 0.0;
	  }
	  a = aspects.next ();
	}
	
	// SVD
	dgesvd (&JOBU, &JOBVT, &M, &N, A, &M, S, U, &M, NULL,
		&N, WORK, &LWORK, &INFO);
	
	if (INFO == 0) {
	  
	  // Compute x = V S^* U^T * e
	  
	  // Sum up the first three row of U^t (columns of U)
	  Real x1 = 0.0;
	  Real x2 = 0.0;
	  Real x3 = 0.0;
	  l1 = 0;
	  l2 = M;
	  l3 = l2 + M;
	  while (l1 < M) {
	    x1 += U[l1++];
	    x2 += U[l2++];
	    x3 += U[l3++];
	  }
	  
	  // Multiply x with S^*
	  if (fabs (S[0]) > sil_get_epsilon ())
	    x1 /= S[0];
	  else
	    x1 = 0.0;
	  if (fabs (S[1]) > sil_get_epsilon ())
	    x2 /= S[1];
	  else
	    x2 = 0.0;
	  if (fabs (S[2]) > sil_get_epsilon ())
	    x3 /= S[2];
	  else
	    x3 = 0.0;
	  
	  // Multiply x with V
	  l1 = 2 * M;
	  Real xx1 = A[0]  * x1 + A[1]      * x2 + A[2]      * x3;
	  Real xx2 = A[M]  * x1 + A[M + 1]  * x2 + A[M + 2]  * x3;
	  Real xx3 = A[l1] * x1 + A[l1 + 1] * x2 + A[l1 + 2] * x3;
	  
	  // Set direction.
	  dd->set (xx1, xx2, xx3);
	  
	} else {
	  
	  sil_msg_warn ("Analyse", 
			"parallel direction optimization failed");
	  
	}
	
      }

      if (c->tolerance () > tol_len_dist)
	tols.add_first (new Real (c->tolerance ()));

      if (c->sub_clusters ()) {
	Cluster *sc = c->sub_clusters () -> first ();
	while (sc) {
	  cls.add_first (sc);
	  sc = c->sub_clusters () -> next ();
	}
      }
      c = cls.unlink_first ();
    }

    tolerance_levels (&parallel, &tols, tol_len_dist, max_len);

    // Save main clusters to delete them later.
    SimpleList<Cluster> main_parallel;
    cn = parallel.first ();
    while (cn) {
      main_parallel.add_first (cn->data ());
      cn = cn->next ();
    }

    // Now find all lines with points on them and sort them by
    // parallel directions.
    
    List<SimpleList<PointLine> > pointlines;

    cn = parallel.unlink_first ();
    while (cn) {

      Cluster *c = cn->data ();

      // Find clusters of owners connected by lines in the parallel
      // line cluster.
      SimpleList<Aspect> aspects;
      c->get_aspects (&aspects);

      // Initially each aspect represents a separate line between two
      // points.
      SimpleList<Cluster> *lines[aspects.length ()];
      unsigned long n = 0;
      Aspect *a = aspects.first ();
      while (a) {
	lines[n] = new SimpleList<Cluster>;
	Entity *e = a->owners () . first ();
	while (e) {
	  lines[n]->add_first (static_cast<Cluster*> (e));
	  e = a->owners () . next ();
	}
	++n;
	a = aspects.next ();
      }
      
      // Combine lines as long as there is a pair which shares an owner.
      bool combine = true;
      while (combine) {
	combine = false;
	unsigned long n1 = 0;
	while (n1 < n) {
	  if (lines[n1]) {
	    unsigned long n2 = n1 + 1;
	    while (n2 < n) {
	      if (lines[n2]) {
		// Check if lines[n1] and lines[n2] share an owner.
		Cluster *c = lines[n1]->first ();
		while (c) {
		  if (lines[n2]->is_member (c)) {
		    combine = true;
		    Cluster *cc = lines[n2]->unlink_first ();
		    while (cc) {
		      lines[n1]->add_last_check (cc);
		      cc = lines[n2]->unlink_first ();
		    }
		    delete lines[n2];
		    lines[n2] = NULL;
		    break;
		  }
		  c = lines[n1]->next ();
		}
	      }
	      ++n2;
	    }
	  }
	  ++n1;
	}
      }

      // Direction of line
      ParallelLine *pl = static_cast<ParallelLine*> (c->value ());
      UnitVector3 dir (pl->x (), pl->y (), pl->z ());
      // Max tolerance to find parallel/orthogonal direction coverting
      // the distance to an angle. We have a rectangular triangle with
      // side lengths c->tolerance and pl->norm. The angle opposite to
      // c->tolerance is a measure for the angular tolerance.
      Real dir_tol = c->tolerance () / 
	sqrt (pl->norm () * pl->norm () + c->tolerance () * c->tolerance ());
      if (dir_tol >= 1.0)
	dir_tol = PI / 2.0;
      else
	dir_tol = asin (dir_tol) + tol_angle_dist;
      bool dir_par = false;
      // Parallel or orthogonal direction?
      Cluster *direction = find_parallel_orthogonal (dir, plane_normal, NULL,
						     dir_tol, dir_par);

      SimpleList<PointLine> *pls = new SimpleList<PointLine>;

      unsigned long n1 = 0;
      while (n1 < n) {
	if (lines[n1]) {
	  // Eliminate clusters in lines[n1] which are also a sub-cluster
	  // of one of the other clusters.
	  SimpleList<Cluster> check;
	  Cluster *c = lines[n1]->first ();
	  while (c) {
	    Cluster *cn = lines[n1]->next ();
	    bool found = false;
	    Cluster *cc = check.first ();
	    while (cc) {
	      if (cc->cluster_is_member (c)) {
		found = true;
		break;
	      }
	      cc = check.next ();
	    }
	    if (found) {
	      lines[n1]->del (c);
	    } else {
	      check.add_first (c);
	    }
	    c = cn;
	  }
	  // Compute the average distance to a line through 0 in direction
	  // dir considering the normal plane_normal.
	  UnitVector3 N = dir * (*plane_normal);
	  Real d = 0.0;
	  Real dmin = -NAN;
	  Real dmax = NAN;
	  unsigned long m = 0;
	  c = lines[n1]->first ();
	  while (c) {
	    AlignedAxes *aa = static_cast<AlignedAxes*> (c->value ());
	    Real dd = (*aa) % N;
	    d += dd;
	    if (dd < dmin)
	      dmin = dd;
	    if (dd > dmax)
	      dmax = dd;
	    ++m;
	    c = lines[n1]->next ();
	  }
	  d /= static_cast<Real> (m);
	  // Average position of line.
	  Vector3 pos = d * N;
	  // Compute tolerance (maximum distance from line)
	  dmin = fabs (d - dmin);
	  dmax = fabs (d - dmax);
	  if (dmin > dmax)
	    dmax = dmin; // dmax is tolerance.
	  // New PointLine object.
	  PointLine *pline = new PointLine (this, dmax, pos, dir, lines[n1]);
	  // Check regular arrangement on pline.
	  check_point_line (pline);
	  // Only keep point lines which contain regular arrangements
	  // or are parallel/orthogonal to something.
	  if (direction || (pline->regular_dist () && 
			    pline->regular_dist () -> length () > 0)) {
	    pls->add_first (pline);
	    // Orthogonal/parallel relations
	    if (direction) {
	      if (dir_par) {
		SubRegularity *sr = new ParallelDirection (this, dir_tol, 
							   direction);
		pline->add_sub_regularity (sr);
	      } else {
		SubRegularity *sr = new OrthogonalDirection (this, dir_tol, 
							     direction);
		pline->add_sub_regularity (sr);
	      }
	    }
	  } else {
	    delete pline;
	  }

	}
	++n1;
      }

      if (pls->length () > 0)
	pointlines.add_first (pls);
      else
	delete pls;

      if (c->sub_clusters ()) {
	Cluster *sc = c->sub_clusters () -> first ();
	while (sc) {
	  parallel.add_first (sc);
	  sc = c->sub_clusters () -> next ();
	}
      }
      delete cn;
      cn = parallel.unlink_first ();
    }
    
    // Delete main parallel clusters
    main_parallel.purge ();

    // Here pointlines is a list of lists where each list is a list of
    // PointLines (points approximately on a line) which are
    // approximately parallel.

    List<RegularPointPosition> c_raps;

    // Find pairs of orthogonal directions in pointlines list.
    ListNode<SimpleList<PointLine> > *pll1 = pointlines.first ();
    while (pll1) {
      UnitVector3 d1 = pll1->data () -> first () -> direction ();
      ListNode<SimpleList<PointLine> > *pll2 = pll1->next ();
      while (pll2) {	  
	UnitVector3 d2 = pll2->data () -> first () -> direction ();
	Real a = fabs (d1 % d2);
	if (a >= 1.0)
	  a = 0.0;
	else
	  a = acos (a);
	// Angle between two directions should be larger than 67.5
	// degree in order to be checked for a grid. Smaller angles
	// cannot really be used for a grid.
	if (a > 3.0 * PI / 8.0 + tol_angle_dist) 
	  check_grid (pll1->data (), pll2->data (), &c_raps);
	pll2 = pll2->next ();
      }
      pll1 = pll1->next ();
    }

    // Check if grids can be combined further (non-orthogonal
    // relations)   
    ListNode<RegularPointPosition> *rapn1 = c_raps.first ();
    while (rapn1) {
      ListNode<RegularPointPosition> *rapn1_n = rapn1->next ();
      RegularPointPosition *rap1 = rapn1->data ();
      ListNode<RegularPointPosition> *rapn2 = rapn1_n;
      while (rapn2) {
	ListNode<RegularPointPosition> *rapn2_n = rapn2->next ();
	RegularPointPosition *rap2 = rapn2->data ();
	// Try to combine grids (the one with the shorter diagonal
	// is the base grid).
	Real d1 = sqrt (rap1->distance1 () * rap1->distance1 () +
			rap1->distance2 () * rap1->distance2 ());
	Real d2 = sqrt (rap2->distance1 () * rap2->distance1 () +
			rap2->distance2 () * rap2->distance2 ());
	if (d1 < d2) {
	  Real tol = rap1->tolerance () + tol_len_dist;
	  if (rap1->match_pts (rap2, tol) &&
	      rap1->match_dirs (rap2, tol + tol_len_dist)) {
	    rap1->combine (rap2);
	    rap1->tolerance () = tol;
	    if (rapn1_n == rapn2)
	      rapn1_n = rapn1_n->next ();
	    rapn2->unlink ();
	    delete rap2;
	    delete rapn2;
	  }
	} else {
	  Real tol = rap2->tolerance () + tol_len_dist;
	  if (rap2->match_pts (rap1, tol) &&
	      rap2->match_dirs (rap1, tol + tol_len_dist)) {
	    rap2->combine (rap1);
	    rap2->tolerance () = tol;
	    rapn1->unlink ();
	    delete rap1;
	    delete rapn1;
	    break;
	  }
	}
	rapn2 = rapn2_n;
      }
      rapn1 = rapn1_n;
    }

    if (f_axis_regcirc) {
      // If we also check for circles, mark points which are on a large
      // grid such that they are noit used for the circles.

      // Large grid: at least 2 different lines in each direction and
      //             and the average of points on each line is at least
      //             75% of the maximum and the maximum is larger than 2 
      //             or the maximum is 2 and the average is 2 and there
      //             are more than 5 different points.

      ListNode<RegularPointPosition> *rapn = c_raps.first ();

      while (rapn) {

	if (rapn->data () -> points () -> length () > 5) {
	  
	  // Find occupied positions.
	  SimpleList<Pair<unsigned long> > pos;
	  Cluster *c = rapn->data () -> points () -> first ();
	  while (c) {
	    AlignedAxes *a = static_cast<AlignedAxes*> (c->value ());
	    Vector3 d = *a - rapn->data () -> position ();
	    unsigned long d1 = static_cast<unsigned long> 
	      (rint ((d % rapn->data () -> direction1 ()) /
		     rapn->data () -> distance1 ()));
	    unsigned long d2 = static_cast<unsigned long>
	      (rint ((d % rapn->data () -> direction2 ()) /
		     rapn->data () -> distance2 ()));
	    pos.add_first (new Pair<unsigned long> (d1, d2));
	    c = rapn->data () -> points () -> next ();
	  }
	  
	  // Check if there are enough positions such that the grid is
	  // considered to be large enough.
	  //
	  // First direction.
	  //
	  // Sort pairs.
	  SimpleList<SimpleList<Pair<unsigned long> > > lines;
	  Pair<unsigned long> *p = pos.first ();
	  while (p) {
	    bool found = false;
	    SimpleList<Pair<unsigned long> > *li = lines.first ();
	    while (li) {
	      if (li->first () -> first () == p->first ()) {
		found = true;
		bool found2 = false;
		Pair<unsigned long> *p2 = li->first ();
		while (p2) {
		  if (p2->second () == p->second ()) {
		    found2 = true;
		    break;
		  }
		  p2 = li->next ();
		}
		if (!found2)
		  li->add_first (p);
		break;
	      }
	      li = lines.next ();
	    }
	    if (!found) {
	      li = new SimpleList<Pair<unsigned long> >;
	      li->add_first (p);
	      lines.add_first (li);
	    }
	    p = pos.next ();
	  }
	  // Count points on lines.
	  Real avg = 0.0;
	  Real max = 0.0;
	  SimpleList<Pair<unsigned long> > *li = lines.first ();	  
	  while (li) {
	    if (li->length () > max)
	      max = static_cast<Real> (li->length ());
	    avg += static_cast<Real> (li->length ());
	    li = lines.next ();
	  }
	  avg /= static_cast<Real> (lines.length ());
	  // Check if first direction is regular.
	  bool reg = (max > 2.0 && (avg >= max * 0.75) && 
		      lines.length () >= 2) || 
	    (max == 2.0 && avg == 2.0 && lines.length () > 2);
	  lines.purge ();
	  if (reg) {
	    // Second direction.
	    //
	    // Sort pairs.
	    p = pos.first ();
	    while (p) {
	      bool found = false;
	      SimpleList<Pair<unsigned long> > *li = lines.first ();
	      while (li) {
		if (li->first () -> second () == p->second ()) {
		  found = true;
		  bool found2 = false;
		  Pair<unsigned long> *p2 = li->first ();
		  while (p2) {
		    if (p2->first () == p->first ()) {
		      found2 = true;
		      break;
		    }
		    p2 = li->next ();
		  }
		  if (!found2)
		    li->add_first (p);
		  break;
		}
		li = lines.next ();
	      }
	      if (!found) {
		li = new SimpleList<Pair<unsigned long> >;
		li->add_first (p);
		lines.add_first (li);
	      }
	      p = pos.next ();
	    }
	    // Count points on lines.
	    avg = 0.0;
	    max = 0.0;
	    SimpleList<Pair<unsigned long> > *li = lines.first ();	  
	    while (li) {
	      if (li->length () > max)
		max = static_cast<Real> (li->length ());
	      avg += static_cast<Real> (li->length ());
	      li = lines.next ();
	    }
	    avg /= static_cast<Real> (lines.length ());
	    // Check if first direction is regular.
	    reg = (max > 2.0 && (avg >= max * 0.75) && 
		   lines.length () >= 2) || 
	      (max == 2.0 && avg == 2.0 && lines.length () > 2);
	    lines.purge ();
	  }
	  pos.purge ();
	  
	  if (reg) {
	    // Grid is large enough, so mark points to avoid them to be part
	    // of a circle.
	    SimpleList<Cluster> bcl;
	    c = rapn->data () -> points () -> first ();
	    while (c) {
	      static_cast<AlignedAxes*> (c->value ()) -> flag () = true;
	      if (c->sub_clusters ()) {
		Cluster *sc = c->sub_clusters () -> first ();
		while (sc) {
		  bcl.add_first (sc);
		  sc = c->sub_clusters () -> next ();
		}
	      }
	      c = rapn->data () -> points () -> next ();
	    }
	    c = bcl.unlink_first ();
	    while (c) {
	      static_cast<AlignedAxes*> (c->value ()) -> flag () = true;
	      if (c->sub_clusters ()) {
		Cluster *sc = c->sub_clusters () -> first ();
		while (sc) {
		  bcl.add_first (sc);
		  sc = c->sub_clusters () -> next ();
		}
	      }
	      c = bcl.unlink_first ();
	    }
	  }

	}

	rapn = rapn->next ();
      }

    }

    // Add PointLine regularities.
    ListNode<SimpleList<PointLine> > *pll = pointlines.unlink_first ();
    while (pll) {
      
      PointLine *pl = pll->data () -> first ();
      while (pl) {

	// Cleanup regular point position regularities.
	if (pl->regular_dist ()) {
	  RegularPointPosition *rpp = pl->regular_dist () -> unlink_first ();
	  while (rpp) {
	    if (rpp->used () || rpp->different_points () < 3) {
	      delete rpp;
	    } else {
	      // Check if line is on a c_raps grid.
	      bool add = true;
	      rapn1 = c_raps.first ();
	      while (rapn1) {
		RegularPointPosition *rpp1 = rapn1->data ();
		Real tol = rpp1->tolerance () + tol_len_dist;
		if (rpp->match_pts (rpp, tol) &&
		    rpp->match_dir (rpp, tol + tol_len_dist)) {
		  rpp1->combine (rpp);
		  rpp1->tolerance () = tol;
		  delete rpp;
		  add = false;
		  break;
		}
		rapn1 = rapn1->next ();
	      }
	      if (add) {
		// Add regular point position on a line as regularity
		// and determine special values for the distance.
		SpecialValue *sv = check_special_scalar (rpp->distance1 (),
							 rpp->tolerance (),
							 false);
		if (sv) {
		  rpp->add_sub_regularity (sv);
		  rpp->distance1_reg () = sv;
		}
		raps->add_first (rpp);
	      }
	    }
	    rpp = pl->regular_dist () -> unlink_first ();
	  }
	  pl->delete_regular_dist ();
	}

	// Add point line.
	rpls->add_first (pl);

	pl = pll->data () -> next ();
      }

      delete pll->data ();
      delete pll;
      pll = pointlines.unlink_first ();
    }
    
    // Check for parallel directions.
    ListNode<RegularPointPosition> *rapn = c_raps.first ();
    while (rapn) {
      // Determine the angular tolerance, use tol_angle_dist to avoid too
      // small tolerances.
      Real tol = 0.5 * rapn->data () -> tolerance ();
      if (tol >= 1.0)
	tol = PI;
      else
	tol = 2.0 * fabs (asin (tol));
      tol = max (tol, tol_angle_dist);
      if (rapn->data () ->direction2 () . norm_sq () > sil_get_epsilon ()) {
	// Grid: two directions.
	Real t = tol;
	Cluster *pa = find_parallel (rapn->data () -> direction1 (), NULL,
				     t);
	if (pa) {
	  SubRegularity *sr = new ParallelDirection (this, t, pa);
	  rapn->data () -> add_sub_regularity (sr);
	  rapn->data () -> direction1_reg () = sr;
	}
	pa = find_parallel (rapn->data () -> direction2 (), NULL, tol);
	if (pa) {
	  SubRegularity *sr = new ParallelDirection (this, tol, pa);
	  rapn->data () -> add_sub_regularity (sr);
	  rapn->data () -> direction2_reg () = sr;
	}
      } else {
	// Line: one direction.
	bool par = false;
	Cluster *pa = find_parallel_orthogonal (rapn->data () -> direction1 (),
						plane_normal, NULL,  tol, par);
	if (pa) {
	  SubRegularity *sr = par ? new ParallelDirection (this, tol, pa) :
	    new OrthogonalDirection (this, tol, pa);
	  rapn->data () -> add_sub_regularity (sr);
	  rapn->data () -> direction1_reg () = sr;
	}
      }
      rapn = rapn->next ();
    }

    // Add remaining grid regularities.
    raps->combine (&c_raps);

  }

  if (f_axis_regcirc) {

    // Check for aligned axis positions on circles.

    // Get the points in the plane.
    List<Cluster> points;
    ListNode<Cluster> *cn = aligned_axes->first ();
    // Normal of the plane (should be the same for all points, so the
    // first one should be fine).
    Cluster *plane_dir = static_cast<AlignedAxes*> (cn->data () -> value ())
      -> direction ();
    Direction *plane_normal = static_cast<Direction*> (plane_dir->value ());
    while (cn) {
      cn->data () -> get_all_base_clusters (&points);
      cn = cn->next ();
    }

    // Remove points that have only plane owners or are on large grids.
    ListNode<Cluster> *p = points.first ();
    while (p) {
      bool del = true;
      if (!(static_cast<AlignedAxes*> (p->data () -> value ()) -> flag ())) {
	SimpleList<Entity> owners;
	p->data () -> get_owners (&owners);
	Entity *e = owners.first ();
	while (e) {
	  if (e->is (SIL_FACE_TYPE)) {
	    Surface *s = static_cast<Face*> (e) -> surface ();
	    if (!s->is (SIL_PLANE_TYPE)) {
	      del = false;
	      break;
	    }
	  } else {
	    del = false;
	    break;
	  }
	  e = owners.next ();
	}
      }
      ListNode<Cluster> *np = p->next ();
      if (del) {
	p->unlink ();
	delete p;
      }
      p = np;
    }

    if (points.length () < 4)
      return;

    sil_progress_bar_text ("Checking for circular arrangements of " + 
			   as_string (points.length ()) + " axes");

    // With n(n-1)(n-2)/6 circles for n points we can easily run out
    // of memory for the clustering. So we group the circles with
    // respect to their location and process the groups individually.

    // Generate the circles.
    List<List<Aspect> >radius_aspects;
    // Distance between the groups
    Real gdist = tol_len_dist * min (max_len / tol_len_dist, static_cast<Real> 
				     (max_pcircle_bucket_size));
    List<Position> ref_pos;
    ListNode<Cluster> *p1 = points.first ();
    while (p1) {
      if (!static_cast<AlignedAxes*> (p1->data () -> value ()) -> flag ()) {
	ListNode<Cluster> *p2 = p1->next ();
	Vector3 p1v = *static_cast<AlignedAxes*> (p1->data () -> value ());
	while (p2) {
	  if (!static_cast<AlignedAxes*> (p2->data () -> value ()) -> 
	      flag ()) {
	    ListNode<Cluster> *p3 = p2->next ();
	    Vector3 p2v = *static_cast<AlignedAxes*> (p2->data () -> value ());
	    while (p3) {
	      if (!static_cast<AlignedAxes*> (p3->data () -> value ()) -> 
		  flag ()) {
		Vector3 p3v = *static_cast<AlignedAxes*> (p3->data () ->
							  value ());
		UnitVector3 b1 = (p2v - p1v) . normalise ();
		Vector3 bb2 = b1 * (*plane_normal);
		if (bb2.norm_sq () > tol_eq * tol_eq) {
		  UnitVector3 b2 = bb2.normalise ();
		  Real p3y = (p3v - p1v) % b2;
		  if (fabs (p3y) > tol_len_dist) {
		    Real p2x = (p2v - p1v) % b1;
		    Real p3x = (p3v - p1v) % b1;
		    Vector3 c = p1v +
		      (p2x / 2.0) * b1 +
		      (p3x * p3x + p3y * p3y - p2x * p3x) / (2.0 * p3y) * b2;
		    Real r = (p1v - c) . norm ();
		    // Radius should be within reasonable limits.
		    if (!isnan (r) && r > tol_len_dist && r < max_len) {
		      // Section of circle should be large enough with respect
		      // to the minimum angle tolerance.
		      UnitVector3 d1 = (p1v - c) . normalise ();
		      UnitVector3 d2 = (p2v - c) . normalise ();
		      Real an1 = d1 % d2;
		      if (an1 >= 1.0)
			an1 = 0.0;
		      else if (an1 <= -1.0)
			an1 = PI;
		      else
			an1 = acos (an1);
		      if (an1 > tol_angle_dist) {
			UnitVector3 d3 = (p3v - c) . normalise ();
			an1 = d1 % d3;
			if (an1 >= 1.0)
			  an1 = 0.0;
			else if (an1 <= -1.0)
			  an1 = PI;
			else
			  an1 = acos (an1);
			if (an1 > tol_angle_dist) {
			  an1 = d2 % d3;
			  if (an1 >= 1.0)
			    an1 = 0.0;
			  else if (an1 <= -1.0)
			    an1 = PI;
			  else
			    an1 = acos (an1);
			  if (an1 > tol_angle_dist) {
			    Aspect *C =  new Aspect (new PointCircle (c, r),
						     p1->data ());
			    C->owners () . add_first (p2->data ());
			    C->owners () . add_first (p3->data ());
			    Aspect *ra = new Aspect (new Length (r), C);
			    // Try to find a radius_aspects list which
			    // is closest to the point and distance
			    // smaller than gdist.
			    ListNode<List<Aspect> > *faln = NULL;
			    ListNode<Position> *frp = NULL;
			    Real d = gdist;
			    ListNode<Position> *rp = ref_pos.first ();
			    ListNode<List<Aspect> > *aln =  radius_aspects.
			      first ();
			    while (rp) {
			      Real dd = (*rp->data () - c) . norm ();
			      if (dd < d) {
				d = dd;
				faln = aln;
				frp = rp;
			      }
			      rp = rp->next ();
			      aln = aln->next ();
			    }
			    if (faln) {
			      Vector3 avg = 
				(*frp->data () * faln->data () -> length () +
				 c) / (faln->data () -> length () + 1.0);
			      frp->data () -> set (avg.x (), avg.y (),
						   avg.z ());
			      faln->data () -> add_first (ra);
			    } else {
			      ref_pos.add_first (new Position (c));
			      List<Aspect> *al = new List<Aspect>;
			      al->add_first (ra);
			      radius_aspects.add_first (al);
			    }
			  }
			  
			}
		      }
		    }
		  }
		}
	      }
	      p3 = p3->next ();
	    }
	  }
	  p2 = p2->next ();
	}
      }
      p1 = p1->next ();
    }

    // Cluster the groups such that groups closer than gdist are
    // combined. We use a dirty trick in order for aspects to have
    // List<Aspect> owners as Entity. This does not cause trouble with
    // the current implementation.
    List<Aspect> group_aspects;
    ListNode<Position> *rp = ref_pos.first ();
    ListNode<List<Aspect> > *al =  radius_aspects.first ();
    while (rp) {
      group_aspects.add_first (new Aspect (rp->data (),
					   (Entity *)al->data ()));
      rp = rp->next ();
      al = al->next ();
    }
    radius_aspects.clear ();
    List<Cluster> group_clusters;
    cluster_aspects (&group_aspects, tol_eq, tol_eq, gdist, false, 
		     &group_clusters);

    // For each group cluster:
    cn = group_clusters.unlink_first ();
    while (cn) {
      SimpleList<Entity> owners;
      cn->data () -> get_owners (&owners);
      List<Aspect> *radius_list = (List<Aspect> *)owners.first ();
      Entity *e = owners.next ();
      while (e) {
	List<Aspect> *al = (List<Aspect> *)e;
	radius_list->combine (al);
	delete al;
	e = owners.next ();
      }

      // Cluster the circles.
      if (radius_list->length () > 0) {

	// First cluster radii.
	List<Cluster> radius_clusters;
	cluster_aspects (radius_list, tol_eq, tol_len_dist, NAN, true,
			 &radius_clusters);

	// Get tolerances and find tolerance levels for radii.
	SimpleList<Real> tols;
	tols.add_first (new Real (tol_len_dist));
	SimpleList<Cluster> cls;
	ListNode<Cluster> *cn = radius_clusters.first ();
	while (cn) {
	  cls.add_first (cn->data ());
	  cn = cn->next ();
	}
	Cluster *c = cls.unlink_first ();
	while (c) {
	  if (c->tolerance () > tol_angle_dist)
	    tols.add_first (new Real (c->tolerance ()));
	  if (c->sub_clusters ()) {
	    Cluster *sc = c->sub_clusters () -> first ();
	    while (sc) {
	      cls.add_first (sc);
	      sc = c->sub_clusters () -> next ();
	    }
	  }
	  c = cls.unlink_first ();
	}
	tolerance_levels (&radius_clusters, &tols, tol_len_dist, max_len);
	
	// Traverse the radius cluster tree depth first, cluster the
	// circles at the leaves and merge the clusters at each node.
	List<Cluster> *clusters = NULL;
	cn = radius_clusters.first ();
	while (cn) {
	  List<Cluster> *cls = cluster_merge_point_circles (cn->data ());
	  if (cls) {
	    if (clusters) {
	      clusters->combine (cls);
	      delete cls;
	    } else {
	      clusters = cls;
	    }
	  }
	  cn = cn->next ();
	}
	
	// Cleanup radius clusters.
	radius_clusters.purge ();
	
	// Determine tolerance levels for circles.
	tols.add_first (new Real (tol_len_dist));
	cn = clusters->first ();
	while (cn) {
	  cls.add_first (cn->data ());
	  cn = cn->next ();
	}
	c = cls.unlink_first ();
	while (c) {
	  if (c->tolerance () > tol_len_dist)
	    tols.add_first (new Real (c->tolerance ()));
	  if (c->sub_clusters ()) {
	    Cluster *sc = c->sub_clusters () -> first ();
	    while (sc) {
	      cls.add_first (sc);
	      sc = c->sub_clusters () -> next ();
	    }
	  }
	  c = cls.unlink_first ();
	}
	tolerance_levels (clusters, &tols, tol_len_dist, max_len);
	
	// Regularity check for each circle.
	
	List<Cluster> bcls;
	cn = clusters->first ();
	while (cn) {
	  cn->data () -> get_all_base_clusters (&bcls);
	  cn = cn->next ();
	}
	
	cn = bcls.unlink_first ();
	while (cn) {
	  
	  c = cn->data ();
	  
	  Vector3 centre = static_cast<PointCircle*> (c->value ()) ->
	    centre ();
	  Real radius = static_cast<PointCircle*> (c->value ()) -> radius ();
	  
	  SimpleList<Entity> owners;
	  c->get_owners (&owners);
	  
	  // Solve circle in least squares sense.
	  SimpleList<Vector3> pts;
	  Entity *e = owners.first ();
	  while (e) {
	    AlignedAxes *a = static_cast<AlignedAxes*>
	      (static_cast<Cluster*> (e)->value ());
	    pts.add_first (new Vector3 (a->x (), a->y (), a->z ()));
	    e = owners.next ();
	  }
	  centre = ls_circle_centre (pts, radius, *plane_normal, centre);
	  pts.purge ();
	  static_cast<PointCircle*> (c->value ()) -> centre () = centre;
	  
	  // Get radius as average distance from centre.
	  Real r = 0.0;
	  e = owners.first ();
	  while (e) {
	    AlignedAxes *a = static_cast<AlignedAxes*>
	      (static_cast<Cluster*> (e)->value ());
	    r += (*a - centre) . norm ();
	    e = owners.next ();
	  }
	  r /= static_cast<Real> (owners.length ());
	  static_cast<PointCircle*> (c->value ()) -> radius () = r;
	  
	  // Recompute tolerance.
	  Real tol = 0.0;
	  e = owners.first ();
	  while (e) {
	    AlignedAxes *a = static_cast<AlignedAxes*>
	      (static_cast<Cluster*> (e)->value ());
	    Real t = fabs ((*a - centre) . norm () - r);
	    if (t > tol)
	      tol = t;
	    e = owners.next ();
	  }
	  c->tolerance () = tol;
	  
	  if (c->sub_clusters ()) {
	    // Check if there is a sub--cluster which has the same owners
	    // than the parent cluster and collapse those clusters. Also
	    // collapse the sub-clusters that are within the tolerance
	    // range.
	    Cluster *sc = c->sub_clusters () -> first ();
	    while (sc) {
	      Cluster *nsc = c->sub_clusters () -> next ();
	      // Set parent cluster as tree guided clustering could
	      // destroy that information.
	      sc->parent_cluster () = c;
	      SimpleList<Entity> scown;
	      sc->get_owners (&scown);
	      if (scown.length () == owners.length () ||
		  sc->tolerance () >= c->tolerance () - tol_len_dist) {
		sc->collapse ();
		bcls.del (sc);
		delete sc;
	      }
	      sc = nsc;
	    }
	  }
	  
	  if (c->parent_cluster ())
	    bcls.add_last_check (c->parent_cluster ());
	  
	  delete cn;
	  cn = bcls.unlink_first ();
	}
	
	// Try to find regular arrangements on the circles.
	cn = clusters->first ();
	while (cn) {
	  cn->data () -> get_all_base_clusters (&bcls);
	  cn = cn->next ();
	}
	cn = bcls.unlink_first ();
	while (cn) {
	  check_point_circle (cn->data (),
			      *static_cast<UnitVector3*> (plane_normal), rpcs);
	  if (cn->data () -> parent_cluster ())
	    bcls.add_last_check (cn->data () -> parent_cluster ());
	  delete cn;
	  cn = bcls.unlink_first ();
	}      
	
	// Add point circles to cluster list.
	pcls->combine ((List<Regularity> *)clusters);
	delete clusters;
	
      }

      // Next group of circles.
      delete radius_list;
      delete cn->data ();
      delete cn;
      cn = group_clusters.unlink_first ();
    }

  }

  return;
}

void Analyse::check_point_line (PointLine *pl)
{
  // Find distance regularities of the points on the PointLine.

  // There should be at least two points on the line for a distance
  // regular check. For exactly two points we only note the distance.
  unsigned long n = pl->points () -> length ();
  if (n < 2)
    return;

  // Put the points in an array.
  Cluster *points[n];
  points[0] = pl->points () -> first ();
  for (unsigned long l = 1; l < n; ++l)
    points[l] = pl->points () -> next ();

  // The uncertainty in position for each point is t = pl->tolerance () + 
  // tol_len_dist. 
  //
  // To avoid that we mix actually identical positions, the minimum
  // base distance should be larger than 4 tol_len_dist. Actually even
  // 4t, or 4 max (pl->tolerance (), tol_len_dist). The precise value
  // depends on the maximum distance between the points that are in
  // the distance regularity, which we do not know.
  //
  // If m is the maximum distance in the regularity, then with N =
  // rint (m / (4 t)) for the minimum base distance d: d > m / N. m /
  // N = m / (m/(4t) + delta) where |delta| < 0.5 and thus 4t / (1+2t
  // / m) < m/N < 4t / (1-2t / m). Approximately this means (for m ->
  // infty anyway) that d > 4t.
  //
  // If we have a distance D then it is a multiple of a base distance
  // d if D = md + e for an integer m and e < 2t, i.e. rint (D / d) -
  // D / d = E such that E < 2t. This relation is also true for
  // different base distance candidates which are multiples.
  //
  // If we consider real distances and not the relations, we can check
  // if rint (D/d) * d - D is within the tolerance t, which appears to
  // be more appropraite.

  Real t = pl->tolerance () + tol_len_dist;

  // Compute all distances and for each point find possible base distances.
  HalfMatrix<Real> dist (n);
  SimpleList<Real> *valid_dists[n];
  for (unsigned long l = 0; l < n - 1; ++l) {
    valid_dists[l] = new SimpleList<Real>;
    for (unsigned long k = l + 1; k < n; ++k) {
      Real d = fabs ((*static_cast<AlignedAxes*> (points[l]->value ()) -
		      *static_cast<AlignedAxes*> (points[k]->value ())) %
		     pl->direction ());
      dist (k, l) = d;
      if (d > 4.0 * t) {
	// d is big enough for a base distance.
	bool add = true;
	Real *D = valid_dists[l]->first ();
	while (D) {
	  Real *ND = valid_dists[l]->next ();
	  if (d < *D) {
	    // Check if D is multiple of d
	    Real f = *D / d;
	    if (fabs (rint (f) * d - *D) < t) {
	      valid_dists[l]->del (D);
	      delete D;
	    }
	  } else {
	    // Check if d is multiple of D
	    Real f = d / *D;
	    if (fabs (rint (f) * (*D) - d) < t) {
	      add = false;
	      break;
	    }
	  }
	  D = ND;
	}
	if (add)
	  valid_dists[l]->add_first (new Real (d));
      }
    }
  }

  // Check for all base distances if we can find a point subset which
  // is based on the distance.
  for (unsigned long l = 0; l < n - 1; ++l) {

    Real *d = valid_dists[l]->unlink_first ();
    while (d) {

      // Find points that fit.
      //
      // We have an approximate base distance d, which covers all base
      // disstances within t tolerance, but might still be outside the
      // interval [d-t,d+t] (if the real tolerance is larger). The
      // best position for a point with distance D from the first
      // point on the line with respect to d is rint (D / d). We can
      // find such a position for any point, but we cannot tell if it
      // is within the actual tolerance. If the actual tolerance is T,
      // then the point would be on the regular arrangement if |rint
      // (D / d) * d - D| = e < T.
      //
      // We try to find an unambiguous assignment for the base
      // distances which have an uncertainty of u. Then we have e < u
      // and u >= t. As we do not know u, we first assign a regular
      // position to each point and then drop all those points whose
      // tolerance is larger than t with a distance larger than t from
      // the last tolerance level.

      SimpleList<unsigned long> rp;
      SimpleList<Real> tols, rp_tols;
      tols.add_first (new Real (t)); // Minimum tolerance
      tols.add_first (new Real (*d / 2.0)); // Maximum tolerance
      for (unsigned long k = l + 1; k < n; ++k) {
	Real f = dist (k, l) / (*d);
	f = fabs (rint (f) * (*d) - dist (k, l));
	rp.add_first (new unsigned long (k));
	// Have to store it twice, once for the assignement and once
	// for the tolerance level detection.
	rp_tols.add_first (new Real (f));
	if (f > t)
	  tols.add_first (new Real (f));
      }
      
      // Find the uncertainty level.
      tols.sort ();
      // Find largest jump
      Real *u0 = tols.first ();
      Real delta = *u0;
      Real um = *u0;
      Real *u = tols.next ();
      while (u) {
	Real dd = *u - *u0;
	if (dd > delta) {
	  um = *u0;
	  delta = dd;
	}
	u0 = u;
	u = tols.next ();
      }
      tols.purge ();

      // Remove points with a tolerance larger than um.
      u = rp_tols.first ();
      unsigned long *p = rp.first ();
      while (p) {
	unsigned long *np = rp.next ();
	if (*u > um)
	  rp.del (p);
	p = np;
	u = rp_tols.next ();
      }
      rp_tols.purge ();
      
      if (rp.length () > 0) {
	// Add first point.
	rp.add_first (new unsigned long (l));

	// Collect points.
	SimpleList<Cluster> *a = new SimpleList<Cluster>;
	unsigned long *p = rp.unlink_first ();
	while (p) {
	  a->add_first (points[*p]);
	  delete p;
	  p = rp.unlink_first ();
	}

	// Check if we should combine it with some other distance
	// regularity.
	RegularPointPosition *irap = NULL;
	if (pl->regular_dist ()) {
	  RegularPointPosition *rap = pl->regular_dist () -> first ();
	  while (rap && !irap) {
	    Real d1 = rap->distance1 ();
	    bool match = false;
	    if (d1 > *d) {
	      Real f = d1 / *d;
	      match = (fabs (rint (f) * (*d) - d1) < t);
	    } else {
	      Real f = *d / d1;
	      match = (fabs (rint (f) * d1 - (*d)) < t);
	    }
	    if (match) {
	      Cluster *c = a->first ();
	      while (c) {
		if (rap->points () -> is_member (c)) {
		  irap = rap;
		  break;
		}
		c = a->next ();
	      }
	    }
	    rap = pl->regular_dist () -> next ();
	  }
	}

	if (irap) {
	  // Combine current regular set with the one found.
	  Cluster *c = a->first ();
	  while (c) {
	    irap->points () -> add_first_check (c);
	    c = a->next ();
	  }
	  delete a;
	  if (*d < irap->distance1 ())
	    irap->distance1 () = *d;
	} else {
	  // New regularity.
	  RegularPointPosition *rap = new RegularPointPosition 
	    (this, 0.0, Vector3 (0.0, 0.0, 0.0), pl->direction (), *d,
	     UnitVector3 (0.0, 0.0, 0.0), 0.0, a);
	  pl->add_regular_dist (rap);
	}
      }

      delete d;
      d = valid_dists[l]->unlink_first ();
    }

  }

  // Check all found regular point position regularities.
  if (pl->regular_dist ()) {
    RegularPointPosition *rpp = pl->regular_dist () -> first ();
    while (rpp) {
      RegularPointPosition *rppn = pl->regular_dist () -> next ();
      
      // Recompute average base distance, reference position,
      // tolerance and check occpuied positions.

      // Reference position
      AlignedAxes *refp = static_cast<AlignedAxes*> 
	(rpp->points () -> first () -> value ());
      // Distances from reference position in line direction.
      SimpleList<Real> dists;
      Cluster *c = rpp->points () -> next ();
      while (c) {
	AlignedAxes *p = static_cast<AlignedAxes*> (c->value ());
	dists.add_first (new Real (((*p) - (*refp)) % rpp->direction1 ()));
	c = rpp->points () -> next ();
      }
      
      // For the distances we have the equation system d_i = l_i d
      // where l_i = rint (d_i / d_approx). Thus d = avg (d_i / l_i).
      // Also collect the l_i.
      SimpleList<long> pos;
      unsigned long n = 0;
      Real d = 0.0;
      Real *di = dists.unlink_first ();
      while (di) {
	++n;
	Real f = rint (*di / rpp->distance1 ());
	if (fabs (f) > 0.5) {
	  d += *di / f;
	  long li = static_cast<long> (f);
	  bool found = false;
	  long *l = pos.first ();
	  while (l) {
	    if (*l == li) {
	      found = true;
	      break;
	    }
	    l = pos.next ();
	  }
	  if (!found)
	    pos.add_first (new long (li));
	}
	delete di;
	di = dists.unlink_first ();
      }
      rpp->distance1 () = d / static_cast<Real> (n);

      // Check the occupied positions.
      bool reg = false;
      if (pos.length () > 0) {
	// Position 0 is always occupied, let's check if we can find
	// three positions with the same multiple of d distance between
	// them. 
	pos.add_first (new long (0));
	rpp->different_points () = pos.length ();
	pos.sort ();
	// Create a delta array listing the distance to the next
	// occupied position.
	unsigned long n = pos.length () - 1;
	unsigned long delta[n];
	long *p = pos.first ();
	unsigned long l = 0;
	while (p) {
	  long *np = pos.next ();
	  if (np)
	    delta[l++] = *np - *p;
	  p = np;
	}
	// We would like to check for a regular distance pattern. A
	// simple check for this is to count the number of different
	// distances between neighbours. This does not necessarily
	// detect regular patterns, but if certain distances are
	// present often we can assume that there is some regularity.
	unsigned long total = 0;
	l = 0;
	while (l < n) {
	  if (delta[l] != 0) {
	    // Count how often delta[l] are present in the remaining
	    // delta array.
	    unsigned long c = 1;
	    unsigned long k = l + 1;
	    while (k < n) {
	      if (delta[k] == delta[l]) {
		++c; 
		delta[k] = 0;
	      }
	      ++k;
	    }
	    if (c > 1)
	      total += c;
	  }
	  ++l;
	}
	// If at least half of the possible distances are present more
	// than once, we call it a regularity.
	reg = (total >= (n>>1));
	// NOTE: We might want to have a different regularity condition 
	//       here. It could either be extremly relaxed and accept it
	//       when at least one distance is present more than once
	//       or it could check the distribution of the distances more
	//       carefully.
      }     
      pos.purge ();

      if (reg) {

	// For the reference position we have the equation p = p_i - l_i
	// d direction1, thus p is the average of these.
	Vector3 p (0.0, 0.0, 0.0);
	n = 0;
	c = rpp->points () -> first ();
	while (c) {
	  AlignedAxes *pi = static_cast<AlignedAxes*> (c->value ());
	  ++n;
	  d = ((*pi) - (*refp)) % rpp->direction1 ();
	  p += (*pi) - rint (d / rpp->distance1 ()) * rpp->distance1 () *
	    rpp->direction1 ();
	  c = rpp->points () -> next ();
	}
	rpp->position () = p / static_cast<Real> (n);
	
	// Now compute the maximum tolerance || (p + l_i direction1 D) -
	// p_i ||_2
	Real tol = 0.0;
	c = rpp->points () -> first ();
	while (c) {
	  AlignedAxes *pi = static_cast<AlignedAxes*> (c->value ());
	  d = ((*pi) - rpp->position ()) % rpp->direction1 ();
	  Real t = (rpp->position () + rint (d / rpp->distance1 ()) * 
		    rpp->distance1 () * rpp->direction1 () - (*pi)) . norm ();
	  if (t > tol)
	    tol = t;
	  c = rpp->points () -> next ();
	}
	rpp->tolerance () = tol;

      } else {

	// No regular arrangments, so delete it.
	pl->regular_dist () -> del (rpp);
	delete rpp;

      }

      rpp = rppn;
    }

    if (pl->regular_dist () == 0) 
      pl->delete_regular_dist ();

  }

  return;
}

void Analyse::check_grid (SimpleList<PointLine> *lines1,
			  SimpleList<PointLine> *lines2,
			  List<RegularPointPosition> *raps)
{
  // Combine the orthogonal axis position pair lines1, lines2 to all
  // possible grids and add them to raps.

  // Find groups of compatible distances in lines1 and lines2.
  SimpleList<SimpleList<RegularPointPosition> > groups1;
  group_lines (lines1, &groups1);
  SimpleList<SimpleList<RegularPointPosition> > groups2;
  group_lines (lines2, &groups2);

  // Process all group pairs.
  SimpleList<RegularPointPosition> *group1 = groups1.first ();
  while (group1) {
    SimpleList<RegularPointPosition> *group2 = groups2.first ();
    while (group2) {

      // The step sizes in group1 and in group2 are compatible, now
      // combine the two groups to form grids.

      form_grids (group1, group2, raps);

      group2 = groups2.next ();
    }
    group1 = groups1.next ();
  }

  // Cleanup
  group1 = groups1.unlink_first ();
  while (group1) {
    delete group1;
    group1 = groups1.unlink_first ();
  }
  group1 = groups2.unlink_first ();
  while (group1) {
    delete group1;
    group1 = groups2.unlink_first ();
  }

  return;
}

void Analyse::group_lines (SimpleList<PointLine> *lines,
			   SimpleList<SimpleList<RegularPointPosition> >
			   *groups)
{
  // Group the lines in lines by compatibale step sizes in groups.

  SimpleList<Real> tols;
  SimpleList<Real> steps;

  PointLine *pl = lines->first ();
  while (pl) {
    if (pl->regular_dist ()) {
      RegularPointPosition *rpp = pl->regular_dist () -> first ();
      while (rpp) {

	Real *bstep = NULL;
	SimpleList<RegularPointPosition> *brppl = NULL;
	Real *btol = NULL;
	Real T = NAN;

	// Find best matching group.
	Real *step = steps.first ();
	Real *tol = tols.first ();
	SimpleList<RegularPointPosition> *rppl = groups->first ();
	while (step) {	  
	  Real t;
	  if (rpp->distance1 () > *step) {
	    t = fabs (*step * rint (rpp->distance1 () / *step) -
		      rpp->distance1 ());
	  } else {
	    t = fabs (rpp->distance1 () * rint (*step / rpp->distance1 ()) -
		      *step);
	  }
	  if (t < T) {
	    T = t;
	    bstep = step;
	    btol = tol;
	    brppl = rppl;
	  }
	  step = steps.next ();
	  tol = tols.next ();
	  rppl = groups->next ();
	}
	// Add it to best match, if close enough.
	if (bstep) {
	  if (T < *btol + tol_len_dist) {
	    brppl->add_first (rpp);
	    if (T > *btol)
	      *btol = T;
	    if (*bstep  > rpp->distance1 ()) {
	      Real f = rint (*bstep / rpp->distance1 ());
	      if (f > 1.0)
		*bstep = rpp->distance1 ();
	    }
	  } else {
	    bstep = NULL;
	  }
	}
	// Otherwise create a new group.
	if (!bstep) {
	  brppl = new SimpleList<RegularPointPosition>;
	  brppl->add_first (rpp);
	  bstep = new Real (rpp->distance1 ());
	  btol = new Real (rpp->tolerance ());
	  groups->add_first (brppl);
	  steps.add_first (bstep);
	  tols.add_first (btol);
	}

	rpp = pl->regular_dist () -> next ();
      }
    }
    pl = lines->next ();
  }

  // Combine gorups whose step and tolerance values lie within each
  // others range.
  SimpleList<Real> steps2;
  SimpleList<Real> tols2;
  SimpleList<SimpleList<RegularPointPosition> > groups2;
  Real *step = steps.first ();
  Real *tol = tols.first ();
  SimpleList<RegularPointPosition> *rppl = groups->first ();
  while (step) {
    Real *stepn = steps.next ();
    Real *toln = tols.next ();
    SimpleList<RegularPointPosition> *rppln = groups->next ();
    
    // Compare with previous groups.
    Real *step2 = steps2.first ();
    Real *tol2 = tols2.first ();
    SimpleList<RegularPointPosition> *rppl2 = groups2.first ();
    while (step2) {
      // We only check if the steps fit, not if they are approximatley
      // multiples. This last check here is just a consistency check.
      if (fabs (*step - *step2) < min (*tol, *tol2)) {
	rppl2->combine (rppl);
	groups->del (rppl);
	delete rppl;
	rppl = NULL;
	break;
      }
      step2 = steps2.next ();
      tol2 = tols2.next ();
      rppl2 = groups2.next ();
    }
    if (rppl) {
      steps2.add_first (step);
      tols2.add_first (tol);
      groups2.add_first (rppl);
    }

    step = stepn;
    tol = toln;
    rppl = rppln;
  }


  tols.purge ();
  steps.purge ();

  return;
}

void Analyse::form_grids (SimpleList<RegularPointPosition> *group1,
			  SimpleList<RegularPointPosition> *group2,
			  List<RegularPointPosition> *raps)
{
  // The step sizes in group1 and in group2 are compatible, now
  // combine the two groups to form grids and store them in raps.

  // Find connected components, i.e. find points that are connected by
  // lines in group1 *and* group2 and make them a grid.

  RegularPointPosition *rpp1 = group1->unlink_first ();
  while (rpp1) {

    // First find connected component by combining lines from group1
    // and group2.

    SimpleList<RegularPointPosition> lines1, lines2, common1, common2;

    // Switch between adding from group1 and group2 until there are no
    // more common points found.

    common1.add_first (rpp1);

    while (common1.length () > 0) {

      // Add lines from group2 which have a point in common with a
      // common1 line (a newly found lines1 line).
      rpp1 = common1.unlink_first ();
      while (rpp1) {
	lines1.add_first (rpp1);

	// Find group2 lines which have a point in common with rpp1.
	RegularPointPosition *rpp2 = group2->first ();
	while (rpp2) {
	  RegularPointPosition *rpp2n = group2->next ();
	  bool found = false;
	  Cluster *c = rpp2->points () -> first ();
	  while (c) {
	    if (rpp1->points () -> is_member (c)) {
	      found = true;
	      break;
	    }
	    c = rpp2->points () -> next ();
	  }
	  if (found) {
	    common2.add_first (rpp2);
	    group2->del (rpp2);
	  }
	  rpp2 = rpp2n;
	}

	rpp1 = common1.unlink_first ();
      }

      // Now find lines in group1 which have points in common with the
      // newly found lines in group2.
      RegularPointPosition *rpp2 = common2.unlink_first ();
      while (rpp2) {
	lines2.add_first (rpp2);
	
	rpp1 = group1->first ();
	while (rpp1) {
	  RegularPointPosition *rpp1n = group1->next ();
	  bool found = false;
	  Cluster *c = rpp1->points () -> first ();
	  while (c) {
	    if (rpp2->points () -> is_member (c)) {
	      found = true;
	      break;
	    }
	    c = rpp1->points () -> next ();
	  }
	  if (found) {
	    common1.add_first (rpp1);
	    group1->del (rpp1);
	  }
	  rpp1 = rpp1n;
	}
	
	rpp2 = common2.unlink_first ();
      }

    }

    if (lines1.length () > 0 && lines2.length () > 0) {

      // lines1 and lines2 contain the lines from group1 and group2
      // which form a connected part of the grid. Make this a grid
      // regularity and mark the lines as used.
      
      // Take the direciton as is (they are all the same).
      UnitVector3 dir1 = lines1.first () -> direction1 ();
      UnitVector3 dir2 = lines2.first () -> direction1 ();
      
      // Average the distances.
      Real d1 = lines1.first () -> distance1 ();
      rpp1 = lines1.next ();
      while (rpp1) {
	if (rpp1->distance1 () < d1)
	  d1 = rpp1->distance1 ();
	rpp1 = lines1.next ();
      }
      unsigned long n = 0;
      Real dp = 0.0;
      rpp1 = lines1.first ();
      while (rpp1) {
	Real f = rpp1->distance1 () / d1;
	dp += rint (f) * d1 - rpp1->distance1 ();
	++n;
	rpp1 = lines1.next ();
      }
      d1 += dp / static_cast<Real> (n);
      Real d2 = lines2.first () -> distance1 ();
      rpp1 = lines2.next ();
      while (rpp1) {
	if (rpp1->distance1 () < d2)
	  d2 = rpp1->distance1 ();
	rpp1 = lines1.next ();
      }
      n = 0;
      dp = 0.0;
      rpp1 = lines2.first ();
      while (rpp1) {
	Real f = rpp1->distance1 () / d2;
	dp += rint (f) * d2 - rpp1->distance1 ();
	++n;
	rpp1 = lines2.next ();
      }
      d2 += dp / static_cast<Real> (n);

      // Average the positions.
      Vector3 p1 = lines1.first () -> position ();
      n = 1;
      dp = 0.0;
      rpp1 = lines1.next ();
      while (rpp1) {
	Real d = (rpp1->position () - p1) % dir1;
	dp += rint (d / d1) * d1 - d;
	++n;
	rpp1 = lines1.next ();
      }
      p1 += dir1 * dp / static_cast<Real> (n);
      Vector3 p2 = lines2.first () -> position ();
      n = 1;
      dp = 0.0;
      rpp1 = lines2.next ();
      while (rpp1) {
	Real d = (rpp1->position () - p2) % dir2;
	dp += rint (d / d2) * d2 - d;
	++n;
	rpp1 = lines2.next ();
      }
      p2 += dir2 * dp / static_cast<Real> (n);
      // Average p1 and p2 to p1:
      Vector3 delta = p2 - p1;
      Real d = delta % dir1;
      Vector3 p = p1 + dir1 * (rint (d / d1) * d1 - d);
      d = delta % dir2;
      p += dir2 * (rint (d / d2) * d2 - d);
      
      // Collect points, mark lines as used and recompute the
      // tolerance.
      SimpleList<Cluster> *pts = new SimpleList<Cluster>;
      Real tol = 0.0;
      rpp1 = lines1.first ();
      while (rpp1) {
	Cluster *c = rpp1->points () -> first ();
	while (c) {
	  if (pts->add_first_check (c)) {
	    Vector3 pos = *static_cast<AlignedAxes*> (c->value ()) - p;
	    Real f1 = pos % dir1;
	    f1 = rint (f1 / d1) * d1 - f1;
	    Real f2 = pos % dir2;
	    f2 = rint (f2 / d2) * d2 - f2;
	    f1 = sqrt (f1 * f1 + f2 * f2);
	    if (f1 > tol)
	      tol = f1;
	  }
	  c = rpp1->points () -> next ();
	}
	rpp1->used () = true;
	rpp1 = lines1.next ();
      }
      rpp1 = lines2.first ();
      while (rpp1) {
	Cluster *c = rpp1->points () -> first ();
	while (c) {
	  if (pts->add_first_check (c)) {
	    Vector3 pos = *static_cast<AlignedAxes*> (c->value ()) - p;
	    Real f1 = pos % dir1;
	    f1 = rint (f1 / d1) * d1 - f1;
	    Real f2 = pos % dir2;
	    f2 = rint (f2 / d2) * d2 - f2;
	    f1 = sqrt (f1 * f1 + f2 * f2);
	    if (f1 > tol)
	      tol = f1;
	  }
	  c = rpp1->points () -> next ();
	}
	rpp1->used () = true;
	rpp1 = lines2.next ();
      }

      RegularPointPosition *rpp = new RegularPointPosition (this, tol, p, 
							    dir1, d1,
							    dir2, d2,
							    pts);
      // Check for regular distance values.
      SpecialValue *sv = check_special_scalar (d1, tol, false);
      if (sv) {
	rpp->add_sub_regularity (sv);
	rpp->distance1_reg () = sv;
      }
      sv = check_special_scalar (d2, tol, false);
      if (sv) {
	rpp->add_sub_regularity (sv);
	rpp->distance2_reg () = sv;
      }
      raps->add_first (rpp);
      
    }

    // Next grid...
    rpp1 = group1->unlink_first ();
  }

  return;
}

List<Cluster> *Analyse::cluster_merge_point_circles (Cluster *circle)
{
  // Cluster and merge circles along the radius tree in a depth first
  // manner.

  // List of top-level clusters generated for this radius tree level.
  List<Cluster> *clusters = NULL;
  // List of base level clusters on this radius tree level (either
  // real base clusters or clusters used to start clustering on this
  // level).
  List<Cluster> bcls;

  // Traverse the radius tree depth first and cluster the aspects
  // spearately on each level.
  if (circle->sub_clusters () && circle->sub_clusters () -> length () > 0) {

    // We are at a node in the tree, so collect the clusters of it's
    // children and then cluster these.
    
    SimpleList<PointCircle> reg_mark;

    // Get cluster lists from sub clusters and merge the lists.
    Cluster *c = circle->sub_clusters () -> first ();
    while (c) {
      List<Cluster> *cls = cluster_merge_point_circles (c);
      if (cls) {
	ListNode<Cluster> *cn = cls->first ();
	while (cn) {
	  bcls.add_first (cn->data ());
	  // Mark every base cluster regular to aviod that it is
	  // removed by the clustering.
	  if (!cn->data () -> value () -> regular ()) {
	    PointCircle *pc = static_cast<PointCircle*> (cn->data () ->
							 value ());
	    pc->regular () = true;
	    reg_mark.add_first (pc);
	  }
	  cn = cn->next ();
	}
	if (clusters) {
	  clusters->combine (cls);
	  delete cls;
	} else {
	  clusters = cls;
	}
      }
      c = circle->sub_clusters () -> next ();
    }

    if (circle->aspects () && circle->aspects () -> length () > 0) {

      // Add direciton cone aspects in the angle aspect at this tree
      // level.
      if (!clusters)
	clusters = new List<Cluster>;
      Aspect *a = circle->aspects () -> first ();
      while (a) {
	Entity *e = a->owners () . first ();
	while (e) {
	  clusters->add_first (new Cluster (this, static_cast<Aspect*> (e)));
	  e = a->owners () . next ();
	}
	a = circle->aspects () -> next ();
      }
    }

    if (!clusters || clusters->length () == 0) {
      delete clusters;
      return NULL;
    }

    // Continue the clustering.
    cluster_aspects (clusters, tol_eq, tol_len_dist, NAN, true);
    if (clusters->length () == 0) {
      delete clusters;
      return NULL;
    }

    // Drop temporary regular markings.
    PointCircle *pc = reg_mark.first ();
    while (pc) {
      pc->regular () = false;
      pc = reg_mark.next ();
    }

  } else {

    // We are at a leave of the angle tree, so cluster the aspects
    // which are combinations of angles and direction cone aspects.

    // Generate a new aspects list containg the direction cones whcih
    // correspond to the angle base cluster.
    SimpleList<Entity> owners;
    circle->get_owners (&owners);
    List<Aspect> aspects;
    Entity *e = owners.first ();
    while (e) {
      aspects.add_first (static_cast<Aspect*> (e));
      e = owners.next ();
    }
    if (aspects.length () == 0)
      return NULL;

    // Cluster the direction cones.
    clusters = new List<Cluster>;
    cluster_aspects (&aspects, tol_eq, tol_len_dist, NAN, true, clusters);
    if (clusters->length () == 0) {
      delete clusters;
      return NULL;
    }

    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      cn->data () -> get_all_base_clusters (&bcls);
      cn = cn->next ();
    }

  }

  // Find regular clusters.
  ListNode<Cluster> *cn = bcls.first ();
  while (cn) {
    
    Cluster *c = cn->data ();

    if (!static_cast<PointCircle*> (c->value ()) -> regular ()) {
      // Only check those that we did not already identify as regular.
      
      // Count number of different directions in the cone.
      SimpleList<Entity> owners;
      c->get_owners (&owners);

      // If we have n different points, then n(n-1)(n-2)/3! circles
      // would have to be present if we have a transitive structure in
      // the cluster. If the cluster is not transitive, it is not
      // regular.
      unsigned long n = owners.length ();
      n = (n * (n - 1) * (n - 2)) / 6;

      // NOTE: If the cluster is not regular, we could still check if
      //       there is a regular subset, but this means we'd have to find
      //       a complete hypergraph between the points present in the
      //       cluster (i.e. a subset of directions for which all cones are
      //       present). At the moment we don't do this, but collapse
      //       sub-clusters later if they have the same owners (directions)
      //       than the cluster itself. This has a similar effect.
      
      if (c->num_aspects () == n) {
	// Mark as regular, so that we do not check again.
	static_cast<PointCircle*> (c->value ()) -> regular () = true;
	// Cluster is regular.
	if (c->parent_cluster ())
	  bcls.add_last_check (c->parent_cluster ());
      } else {
	// Cluster is not regular, collapse it (i.e. move everyting to
	// parent if possible). We do not remove top-level clusters as
	// they are used for clustering on the next angle tree level.
	if (c->collapse ()) {
	  // Cluster collpase, so remove it and add it's parent to the
	  // list.
 	  bcls.add_last_check (c->parent_cluster ());
	  clusters->del (c);
	  delete c;
	} else {
	  // Mark top level cluster as regular for this radius level.
	  static_cast<PointCircle*> (c->value ()) -> regular () = true;
	}
      }

    }

    cn = cn->next ();
  }

  return clusters;
}

void Analyse::check_point_circle (Cluster *pc, UnitVector3 const &norm,
				  List<RegularPointCircle> *regs)
{
  Vector3 centre = static_cast<PointCircle*> (pc->value ()) -> centre ();
  Real r = static_cast<PointCircle*> (pc->value ()) -> radius ();

  // Determine the angular tolerance, use tol_angle_dist to avoid too
  // small tolerances.
  Real tol = 0.5 * pc->tolerance () / r;
  if (tol >= 1.0)
    tol = PI;
  else
    tol = 2.0 * fabs (asin (tol));
  Real pctol = tol;
  tol = max (tol, tol_angle_dist);

  // The owners of the point circle are the aligned axes clusters
  // lying in it. We collect the aligned axes clusters according to
  // the tolerance of the point circle.
  SimpleList<Entity> owners;
  pc->get_owners (&owners);

  // Each of the point owners is now associated with a direction in
  // the circle. Also check if we have to increase the tolerance for
  // one of the aligned axes clusters has a larger tolerance.
  SimpleList<Direction> dirs;
  Entity *e = owners.first ();
  while (e) {
    AlignedAxes *a = static_cast<AlignedAxes*> (static_cast<Cluster*> (e) -> 
						value ());
    // Project line between centre and point on plane for the
    // direciton.
    Vector3 d = *a - centre;
    d -= (d % norm) * norm;
    dirs.add_last (new Direction (d));
    // Convert positional tolernace of cluster to angular tolerance.
    // The Position is tol off the circle. This means the position on
    // the circle can move tol. So the angle variation alpha caused by
    // this movement is determined by sin (alpha/2) = (t/2) / r
    Real alpha = 0.5 * static_cast<Cluster*> (e) -> tolerance () / r;
    if (alpha >= 1.0)
      alpha = PI;
    else
      alpha = 2.0 * fabs (asin (alpha));
    if (alpha > tol)
      tol = alpha;
    e = owners.next ();
  }

  // The maximal base angle for which we can still find different
  // arrangements is 2Pi / N with N = (2Pi / (4t) where t is the
  // tolerance.
  unsigned long N;
  if (tol < tol_eq) {
    // tol is too small, so use upper limit.
    N = regular_angle_max_n;
  } else {
    N = static_cast<unsigned long> (floor (PI / (2.0 * tol)));
    if (N > (regular_angle_max_n >> 1))
      N = regular_angle_max_n >> 1;
  }
  if (N < 3) {
    dirs.purge ();
    return;
  }

  // Store the directions in an array to access them.
  unsigned long n = dirs.length ();
  // We need at least 3 directions.
  if (n < 3) {
    dirs.purge ();
    return;
  }
  Cluster **pos = new Cluster*[n];
  Direction **pdirs = new Direction*[n];
  unsigned long l = 0;
  Cluster *c = static_cast<Cluster*> (owners.first ());
  Direction *d = dirs.unlink_first ();
  while (c) {
    pos[l] = c;
    pdirs[l++] = d;
    c = static_cast<Cluster*> (owners.next ());
    d = dirs.unlink_first ();
  }

  // List of regularities for this cone.
  SimpleList<RegularPointCircle> rpc_regs;

  // Find regular angle sets between the directions, by comparing the
  // angles between an initial direction and the other directions. We
  // have to check all directions as intial directions and compare the
  // angles between the initial directions and all following
  // directions in the list.  At least three more directions should be
  // in the list or we cannot get an angular regular relation and the
  // angles between two directions are checked later.
  HalfMatrix<Real> angles (n);
  // List of regular angles for each initial direction.
  SimpleList<unsigned long> **valid_angles = new SimpleList<unsigned long>*[n];
  for (unsigned long k = 0; k < n - 2; ++k) {
    // Initialize the valid angles list for k as initial direction.
    valid_angles[k] = new SimpleList<unsigned long>;
    // Compute all angles between the initial direction and the
    // following directions in the list and collect the regular
    // angles.
    //
    // A regular direction such that we can tell on which side of the
    // initial direction we are. Note that it doesn't matter whcih
    // side is which as long as we make the distinction.
    Vector3 regdir = *pdirs[k] * norm;
    for (unsigned long l = k + 1; l < n; ++l) {
      Real a = *pdirs[k] % *pdirs[l];
      if (a >= 1.0)
	angles (l, k) = 0.0;
      else if (a <= -1.0)
	angles (l, k) = PI;
      else
	angles (l, k) = acos (a);
      if (regdir % *pdirs[l] < 0.0)
	angles (l, k) = 2.0 * PI - angles (l, k);
      // Find the regular angles suggested by this angle:
      if (angles (l, k) > tol) {
	unsigned long a_min = (unsigned long)floor (2.0 * PI / 
						    (angles (l, k) +
						     tol));
	if (a_min < N) {
	  if (a_min < 3)
	    a_min = 3;
	  unsigned long a_max = (unsigned long)ceil (2.0 * PI /
						     (angles (l, k) -
						      tol)) + 1;
	  if (a_max > N)
	    a_max = N;
	  for (unsigned long aa = a_min; aa < a_max; ++aa) {
	    // Check if the angle is alrady listed as a valid angle and
	    // also check for multiples.
	    bool add = true;
	    unsigned long *an = valid_angles[k]->first ();
	    while (an) {
	      unsigned long *next_an = valid_angles[k]->next ();
	      if (aa < *an) {
		Real a = (Real)*an / (Real)aa;
		if (fabs (a - rint (a)) < tol_eq) {
		  add = false;
		  break;
		}
	      } else {
		Real a = (Real)aa / (Real)*an;
		if (fabs (a - rint (a)) < tol_eq) {
		  valid_angles[k]->del (an);
		  delete an;
		}
	      }
	      an = next_an;
	    }
	    // If the angle is not listed as a valid angle, then add
	    // it to the list.
	    if (add)
	      valid_angles[k]->add_first (new unsigned long (aa));
	  }
	}
      }
    }
  }


  // Adjust tolerance.
  tol = pctol + tol_angle_dist;

  // Now check for all initial directions if one of the suggested
  // regular angles are actually really there.
  for (unsigned long k = 0 ; k < n - 2; ++k) {
    
    // Check all angles with the initial direction k if they are
    // multiples of the suggested regular angles.
    unsigned long *an = valid_angles[k]->unlink_first ();
    while (an) {
      
      // The value of the regular angle.
      Real beta = 2.0 * PI / (Real)*an;
      // Array keeping track which direction in the list is actually
      // occupying a direction suggested buy the regular angle.
      unsigned long directions[*an + 1];
      // Tolerances for the directions list.
      Real tols[*an + 1];
      // Number of directions which are in the current regular subset.
      unsigned long num = 1;
      // Direction 0 (0 * 2PI / n) is always occupied by the initial
      // direction k.
      directions[0] = k;
      // Mark the other regular directions as not-occupied.
      for (unsigned long j = 1; j <= *an; ++j)
	directions[j] = n;
      // Check the angles between initial direction and all following
      // directions. If the angle suggest a regular arrangement add
      // the direction to the current regular subset.
      for (unsigned long j = k + 1; j < n; ++j) {
	// Compute the tolerance between the actual angle and the
	// angle suggested by the regular angle.
	unsigned long ta = (unsigned long)rint (angles (j, k) / beta);
	Real t = fabs ((Real)ta * beta - angles (j, k));
	if (t < tol && ta <= *an) {
	  // The angle is approximately regular, check if we have to
	  // add the direction to the curent subset.
	  if (directions[ta] < n) {
	    // Direction already occupied by other direction, but
	    // parallel directions, etc.  have been determiend
	    // before, so we take the one which is closer.
	    if (t < tols[ta]) {
	      directions[ta] = j;
	      tols[ta] = t;
	    }
	  } else {
	    // Direction not occupied, add it as new one.
	    ++num;
	    directions[ta] = j;
	    tols[ta] = t;
	  }
	}
      }

      if (num > 2) {
	// The current regular subset has more than two members, so
	// it could be a regularity.
	
	// First check if any of the directions is already a member of
	// a conical angle regular set with the same base angle.
	RegularPointCircle *irpc = NULL;
	RegularPointCircle *rpc = rpc_regs.first ();
	while (rpc && !irpc) {
	  // Check if angles are compatible.
	  unsigned long angl_n = rpc->angle_n ();
	  bool match = false;
	  if (angl_n == *an) {
	    match = true;
	  } else if (angl_n < *an) {
	    // check if: integer = *an / angl_n
	    Real r = rint ((Real)*an / (Real)angl_n);
	    match = (fabs (r * (Real)angl_n - (Real)(*an)) < tol_eq);
	  } else {
	    // check if: integer =  angl_n / *an
	    Real r = rint ((Real)angl_n / (Real)*an);
	    match = (fabs (r * (Real)(*an) - (Real)angl_n) < tol_eq);
	  }
	  if (match) {
	    for (unsigned long j = 0; j <= *an; ++j) {
	      if (directions[j] < n && 
		  rpc->points () -> is_member (pos[directions[j]])) {
		irpc = rpc;
		break;
	      }
	    }
	  }
	  rpc = rpc_regs.next ();
	}
	if (irpc) {
	  // Combine current regular set with the one found.
	  for (unsigned long j = 0; j <= *an; ++j) {
	    if (directions[j] < n)
	      irpc->points () -> add_last_check (pos[directions[j]]);
	  }
	  // And set the angle.
	  if (irpc->angle_n () < *an)
	    irpc->angle_n () = *an;
	} else {

	  // New regularity. 

	  // Generate direction list and recompute tolerances
	  // considering distance from regular position.
	  SimpleList<Cluster> *poslist = new SimpleList<Cluster>;
	  for (unsigned long j = 0; j <= *an; ++j) {
	    if (directions[j] < n)
	      poslist->add_first (pos[directions[j]]);
	  }
	  // Generate and add the current subset as a new regularity.
	  RegularPointCircle *nr = 
	    new RegularPointCircle (this, tol, centre, r, *an, norm, poslist);
	  // Check if radius has a special value.
	  SpecialValue *rv = check_special_scalar (r, tol, false);
	  if (rv)
	    nr->add_sub_regularity (rv);
	  rpc_regs.add_first (nr);
	}
      }
      
      delete an;
      an = valid_angles[k]->unlink_first ();
    }
    
    delete valid_angles[k];
  }

  delete valid_angles;
  delete pos;
  for (unsigned long l = 0; l < n; ++l) 
    delete pdirs[l];
  delete pdirs;
  
  // Add regularities to reg list
  RegularPointCircle *rpc = rpc_regs.unlink_first ();
  while (rpc) {
    
    // Check for m-regular arrangements of the positions on the
    // circle, where m is a divisor of angel_n.
    
    // Base direction.
    Vector3 b = *static_cast<AlignedAxes*> (rpc->points () -> first () -> 
					    value ()) - centre;
    b -= (b % norm) * norm;
    UnitVector3 bd = b.normalise ();
    // Regular direction for orientation.
    Vector3 regdir = bd * norm;
    // Create a boolean array to list occupied directions.
    bool od[rpc->angle_n ()];
    od[0] = true;
    for (unsigned long l = 1; l < rpc->angle_n (); ++l)
      od[l] = false;
    Cluster *c = rpc->points () -> first ();
    while (c) {
      // Direction.
      Vector3 dd = *static_cast<AlignedAxes*> (c->value ()) - centre;
      dd -= (dd % norm) * norm;
      UnitVector3 d = dd.normalise ();
      // Compute angle.
      Real a = d % bd;
      if (a >= 1.0)
	a = 0.0;
      else if (a <= -1.0)
	a = PI;
      else
	a = acos (a);
      if (regdir % d < 0.0)
	a = 2.0 * PI - a;
      unsigned long l = (unsigned long) rint (a / rpc->angle ());
      if (l < rpc->angle_n ())
	od[l] = true;
      // If it's larger than angle_n, then it's not really possible
      // (>2Pi!)  If it's equal, then it's at position 0 and this one
      // is anyway true.
      c = rpc->points () -> next ();
    }
    bool all = true;
    for (unsigned long l = 1; all && l < rpc->angle_n (); ++l)
      all = (all && od[l]);
    // If all directions are set, we do not have to check anything
    // else.
    if (all) {
      rpc->all_directions () = true;
    } else if (rpc->angle_n () > 3) {
      // For angle_n == 3 we can actually only check if all directions
      // are there. The check is just for seucrity as we should
      // usually have all directions in this case anyway.
      //
      // In this half matrix we store which regularities have been
      // found.  Row l, column k is true if an l regularity starting
      // with position k has been found (- means not present, + means
      // present and * means redundant).
      unsigned long a_n = rpc->angle_n () >> 1;
      HalfMatrix<char> regs (a_n + 1);
      for (unsigned long l = 1; l < a_n; ++l) {
	for (unsigned long k = 0; k < l; ++k)
	  regs (l, k) = '-';
      }
      // Now check the divisors.
      unsigned long m = 2;
      while (m < a_n) {
	Real t = (Real)n / (Real)m;
	if (fabs (t - rint (t)) < tol_eq) {
	  // m is a divisor.
	  //
	  // Find divisors of m in order to check if regualrity has not
	  // already been verified earlier (with smaller m).
	  SimpleList<unsigned long> div;
	  unsigned long M = 2;
	  while (M < (m >> 1)) {
	    t = (Real)m / (Real)M;
	    if (fabs (t - rint (t)) < tol_eq)
	      div.add_first (new unsigned long (M));
	    ++M;
	  }
	  // Start checking if an m regularity starting with position l
	  // is there.
	  unsigned long l = 0;
	  while (l < m) {
	    // First we check if the regularity has not already been
	    // verified earlier.
	    bool reg_check = true;
	    unsigned long *MM = div.first ();
	    while (MM) {
	      unsigned long k = l % *MM;
	      if (regs (*MM, k) != '-') {
		reg_check = false;
		break;
	      }
	      MM = div.next ();
	    }
	    if (reg_check) {
	      // Check if m regularity stating with direction l is present.
	      bool reg = true;
	      unsigned long j = l;
	      while (reg && j < rpc->angle_n ()) {
		reg = (reg && od[j]);
		j += m;
	      }
	      if (reg) {
		// Regularity is present, note it.
		regs (m, l) = '+';
	      }
	    } else {
	      // Mark as redundant as already checked earlier.
	      regs (*MM, l) = '*';
	    }
	    ++l;
	  }
	  // Delete divisors of m.
	  div.purge ();
	}
	++m;
      }

      // Note the regularities found.
      SimpleList<unsigned long> *ra = new SimpleList<unsigned long>;
      for (unsigned long l = 1; l <= a_n; ++l) {
	for (unsigned long k = 0; k < l; ++k) {
	  if (regs (l, k) == '+') {
	    ra->add_first (new unsigned long (l));
	    break;
	  }
	}
      }
      if (ra->length () > 0) {
	rpc->set_regular_arrangements (ra);
      } else {
	delete ra;
      }

    }

    regs->add_first (rpc);
    rpc = rpc_regs.unlink_first ();
  }

  return;
}

Vector3 Analyse::ls_circle_centre (SimpleList<Vector3> &pts, Real r,
				   UnitVector3 const &norm,
				   Vector3 const &centre0)
{
  // Circle equation:
  //
  //   || x_i - c || - r = 0
  //
  // We want to minimize the residuals in a leasrt squares sense,
  // i.e.
  //
  //   F (c, r) = \sum (||x_i - c|| - r)^2 -> min,
  //
  //   F_r (c, r) = \sum -2 (||x_i - c|| - r).
  //
  // For a minimum it is necessary that F_r = 0. This yields
  //
  //    r = 1/n \sum ||x_i - c||.
  //
  // Thus we minimize
  //
  //    \sum (||x_i - c|| - 1/n \sum ||x_i - c||)^2.

  // First project the points on a plane and store the coordinate in
  // the first two elements of the vector.
  //
  // 2D coordinate system.
  UnitVector3 e1 = make_ortho (norm) . normalise ();
  UnitVector3 e2 = (e1 * norm) . normalise ();
  // Projection in plane through 0 with normal norm and 2D coordinate
  // system.
  //
  // First the initial centre (which will be the origin).
  Real cx = centre0 % e1;
  Real cy = centre0 % e2;
  // Now the points on the circle.
  Real maxd = 0.0;
  Vector3 *v = pts.first ();
  while (v) {
    Real x = *v % e1 - cx;
    Real y = *v % e2 - cy;
    Real d = x * x + y * y;
    if (d > maxd)
      maxd = d;
    v->x () = x;
    v->y () = y;
    v->z () = 0.0;
    v = pts.next ();
  }

  // Optimization using the downhill, simplex method by Nelder-Mead.

  // Choose initial simplex
  Real p[3][2];
  maxd = max (fabs (sqrt (maxd) - r), tol_len_dist);
  p[0][0] = maxd;
  p[0][1] = maxd;
  p[1][0] = maxd;
  p[1][1] = -maxd;
  p[2][0] = 0.0;
  p[2][1] = 0.0;

  // Array to speed up function evaluation
  Real *fd = new Real[pts.length ()];
  
  Real f[3] = {0.0, 0.0, 0.0};
  for (int l = 0; l < 3; ++l) {
    // Evaluate function.
    Real sum = 0.0;
    unsigned long k = 0;
    Vector3 *v = pts.first ();
    while (v) {
      Real d1 = v->x () - p[l][0];
      Real d2 = v->y () - p[l][1];
      sum += (fd[k++] = sqrt (d1 * d1 + d2 * d2));
      v = pts.next ();
    }
    sum /= static_cast<Real> (pts.length ());
    Real f_new = 0.0;
    for (k = 0; k < pts.length (); ++k) {
      Real h = fd[k] - sum;
      f_new += h * h;
    }
    f[l] = f_new;
  }

  // Nelder-Mead iteration.
  Real ftol = 0.0000001;
  unsigned long max_iter = 500;
  Real psum_x = p[0][0] + p[1][0] + p[2][0];
  Real psum_y = p[0][1] + p[1][1] + p[2][1];
  unsigned long iter = 0;
  while (true) {
    
    // Find the point with the smallest, the mid and the highest
    // function value.
    int n_lo = 0;
    if (f[1] < f[n_lo])
      n_lo = 1;
    if (f[2] < f[n_lo])
      n_lo = 2;
    int n_hi = (n_lo + 1) % 3;
    int n_mid = (n_hi + 1) % 3;
    if (f[n_hi] < f[n_mid]) {
      int h = n_hi;
      n_hi = n_mid;
      n_mid = h;
    }

    // Compute the fraction range between highest and lowest and stop,
    // if good enough.
    Real rtol = 2.0 * fabs (f[n_hi] - f[n_lo]) / 
      (fabs (f[n_hi]) + fabs (f[n_lo]) + sil_get_epsilon ());
    if (rtol < ftol)
      break;

    if (iter > max_iter)
      break;
    ++iter;

    // Extrapolate by a factor -1 through the face of the simplex
    // across from the high point, i.e. reflect the simplex from the
    // high point.

    Real p_new_x = psum_x - 2.0 * p[n_hi][0];
    Real p_new_y = psum_y - 2.0 * p[n_hi][1];
    // Evaluate function.
    Real sum = 0.0;
    unsigned long k = 0;
    Vector3 *v = pts.first ();
    while (v) {
      Real d1 = v->x () - p_new_x;
      Real d2 = v->y () - p_new_y;
      sum += (fd[k++] = sqrt (d1 * d1 + d2 * d2));
      v = pts.next ();
    }
    sum /= static_cast<Real> (pts.length ());
    Real f_new = 0.0;
    for (k = 0; k < pts.length (); ++k) {
      Real h = fd[k] - sum;
      f_new += h * h;
    }
    if (f_new < f[n_hi]) {
      // It's better than the highest, so replace it.
      p[n_hi][0] = p_new_x;
      p[n_hi][1] = p_new_y;
      psum_x = p[0][0] + p[1][0] + p[2][0];
      psum_y = p[0][1] + p[1][1] + p[2][1];
      f[n_hi] = f_new;
    }
    
    if (f_new <= f[n_lo]) {
      // Better result than the bext point, so try an additional extrapolation
      // by a factor 2
      
      p_new_x = -0.5 * psum_x - 2.5 * p[n_hi][0];
      p_new_y = -0.5 * psum_y - 2.5 * p[n_hi][1];
      // Evaluate function.
      sum = 0.0;
      k = 0;
      Vector3 *v = pts.first ();
      while (v) {
	Real d1 = v->x () - p_new_x;
	Real d2 = v->y () - p_new_y;
	sum += (fd[k++] = sqrt (d1 * d1 + d2 * d2));
	v = pts.next ();
      }
      sum /= static_cast<Real> (pts.length ());
      f_new = 0.0;
      for (k = 0; k < pts.length (); ++k) {
	Real h = fd[k] - sum;
	f_new += h * h;
      }
      if (f_new < f[n_hi]) {
	// It's better than the highest, so replace it.
	p[n_hi][0] = p_new_x;
	p[n_hi][1] = p_new_y;
	psum_x = p[0][0] + p[1][0] + p[2][0];
	psum_y = p[0][1] + p[1][1] + p[2][1];
	f[n_hi] = f_new;
      }

    } else if (f_new >= f[n_mid]) {
      // The reflected point is worse than the second-highest, so look
      // for an intermediate lower point, i.e. do a one-dimensional
      // contraction.

      Real f_save = f[n_hi];

      p_new_x = 0.25 * psum_x + 0.75 * p[n_hi][0];
      p_new_y = 0.25 * psum_y + 0.75 * p[n_hi][1];
      // Evaluate function.
      sum = 0.0;
      k = 0;
      Vector3 *v = pts.first ();
      while (v) {
	Real d1 = v->x () - p_new_x;
	Real d2 = v->y () - p_new_y;
	sum += (fd[k++] = sqrt (d1 * d1 + d2 * d2));
	v = pts.next ();
      }
      sum /= static_cast<Real> (pts.length ());
      f_new = 0.0;
      for (k = 0; k < pts.length (); ++k) {
	Real h = fd[k] - sum;
	f_new += h * h;
      }
      if (f_new < f[n_hi]) {
	// It's better than the highest, so replace it.
	p[n_hi][0] = p_new_x;
	p[n_hi][1] = p_new_y;
	psum_x = p[0][0] + p[1][0] + p[2][0];
	psum_y = p[0][1] + p[1][1] + p[2][1];
	f[n_hi] = f_new;
      }

      if (f_new >= f_save) {
	// The high point is still there, so contract around the
	// lowest point.
	for (int l = 0; l < 3; ++l) {
	  if (l != n_lo) {
	    p[l][0] = 0.5 * (p[l][0] + p[n_lo][0]);
	    p[l][1] = 0.5 * (p[l][1] + p[n_lo][1]);
	    // Evaluate function.
	    sum = 0.0;
	    k = 0;
	    Vector3 *v = pts.first ();
	    while (v) {
	      Real d1 = v->x () - p[l][0];
	      Real d2 = v->y () - p[l][1];
	      sum += (fd[k++] = sqrt (d1 * d1 + d2 * d2));
	      v = pts.next ();
	    }
	    sum /= static_cast<Real> (pts.length ());
	    f_new = 0.0;
	    for (k = 0; k < pts.length (); ++k) {
	      Real h = fd[k] - sum;
	      f_new += h * h;
	    }
	    f[l] = f_new;
	  }
	}
	psum_x = p[0][0] + p[1][0] + p[2][0];
	psum_y = p[0][1] + p[1][1] + p[2][1];
      }

    }

  }
  
  Real px = psum_x / 3.0;
  Real py = psum_y / 3.0;

  delete fd;
  
  return (px + cx) * e1 + (py + cy) * e2 + (centre0 % norm) * norm;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
