// analyse_cluster.cc: Analyse a SiL model / cluster
//
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_cluster.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_cluster,
       "$Id: analyse_cluster.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "kernel/matrix.hxx"
#include "kernel/analyse.hxx"

class ClusterQuadTreeIndex;

class ClusterQuadTree {
public:
  ClusterQuadTree (List<Aspect> *aspects, Real to_eq, Real to_dist,
		   Real to_approx, bool h, Analyse *ana);
  ClusterQuadTree (List<Cluster> *cls, Real to_eq, Real to_dist,
		   Real to_approx, bool h, Analyse *ana);
  ~ClusterQuadTree ();
  void cluster (List<Cluster> *clusters);
protected:
  void compute_tolerance (Cluster *c, Feature *v, Real &t) const;
  unsigned long min_index (unsigned long l, unsigned long k) const;
  bool closest_pair ();
  void combine_clusters (unsigned long l, unsigned long k);
  void update_rows (unsigned long ll, unsigned long kk);
  void update_cols (unsigned long ll, unsigned long kk);
  friend ClusterQuadTreeIndex;
private:
  Real tol_eq, tol_dist, tol_approx;
  unsigned long na, ma;
  Cluster **clustr;
  bool clustered, hier;
  HalfMatrix<Real> *dist;
  ClusterQuadTreeIndex *parent;
};

class ClusterQuadTreeIndex {
public:
  ClusterQuadTreeIndex (HalfMatrix<unsigned long> *ind, ClusterQuadTree *tr);
  ~ClusterQuadTreeIndex () {
    if (parent)
      delete parent;
    delete idx;
  }
protected:
  bool closest_pair ();
  unsigned long min_index (unsigned long l, unsigned long k) const;
  void update_rows (unsigned long ll, unsigned long kk);
  void update_cols (unsigned long ll, unsigned long kk);
  friend ClusterQuadTree;
private:
  ClusterQuadTree *t;
  HalfMatrix<unsigned long> *idx;
  ClusterQuadTreeIndex *parent;
};

ClusterQuadTree::ClusterQuadTree (List<Aspect> *aspects, Real to_eq,
				  Real to_dist, Real to_approx, 
				  bool h, Analyse *ana)
{
  // Tolerances used for hierarchical clustering.
  tol_eq = to_eq;
  if (to_dist < to_eq)
    tol_dist = to_eq;
  else
    tol_dist = to_dist;
  tol_approx = to_approx;
  hier = h;

  // Create clusters.
  // Initially we have a cluster for each aspect.
  na = aspects->length ();
  clustr = new Cluster*[na];

  // Create distance matrix.
  // It has to have an even number of rows/cols.
  ma = na;
  if ((ma & 1) == 0)
    ++ma;
  dist = new HalfMatrix<Real> (ma);

  // Setup cluster list and compute distances.
  ListNode<Aspect> *an = aspects->first ();
  clustr[0] = new Cluster (ana, an->data ());
  unsigned long l = 1;
  while (l < na) {
    an = an->next ();
    clustr[l] = new Cluster (ana, an->data ());
    unsigned long k = 0;
    while (k < l) {
      Real d = clustr[l]->sim (clustr[k]);
      if (d < tol_eq) {
	// The two values of the clusters are equal, so we combine
	// the immediately.
	Feature *avg = clustr[k]->value () ->
	  average (clustr[k]->num_aspects (),
		   clustr[l]->value (), clustr[l]->num_aspects ());
	clustr[k]->combine (clustr[l], avg);
	delete clustr[l];
	clustr[k]->tolerance () = tol_eq;
	--na;
	--l;
	break;
      } else {
	(*dist) (l, k) = d;
      }
      ++k;
    }
    ++l;
  }

  // New distrance matrix size (after combining equal clusters)
  ma = na;
  if ((ma & 1) == 0)
    ++ma;

  // Fill any additional lines in the distance matrix with tol_max
  // to avoid problems. The additional lines are just there to
  // simplify indexing.
  while (l < ma) {
    unsigned long k = 0;
    while (k < l)
      (*dist) (l, k++) = TOL_MAX;
    ++l;
  }

  if (na > 3) {

    // Create first index matrix (if na > 3, i.e. we can create a 
    // reasonable index matrix)
    unsigned long mm = ma >> 1;
    unsigned long mmm = mm + 1;
    if (mmm > 2 && (mmm & 1) == 0)
      ++mmm;
    HalfMatrix<unsigned long> *idx = new HalfMatrix<unsigned long> (mmm);
    // Find minimum indexes for index matrix.
    l = 1;
    while (l <= mm) {
      unsigned long k = 0;
      while (k < l) {
	(*idx) (l, k) = min_index (l, k);
	++k;
      }
      ++l;
    }
    
    // Also fill index matrix.
    while (l < mmm) {
      unsigned long k = 0;
      while (k < l)
	(*idx) (l, k++) = ULONG_MAX;
      ++l;
    }
    
    parent = new ClusterQuadTreeIndex (idx, this);

  } else {

    // Note enough aspects to generate an index matrix.
    parent = NULL;

  }

  clustered = false;
}

ClusterQuadTree::ClusterQuadTree (List<Cluster> *cls, Real to_eq,
				  Real to_dist, Real to_approx, bool h,
				  Analyse *ana)
{
  // Tolerances used for hierarchical clustering.
  tol_eq = to_eq;
  if (to_dist < to_eq)
    tol_dist = to_eq;
  else
    tol_dist = to_dist;
  tol_approx = to_approx;
  hier = h;

  // Create clusters.
  // Initially we have a cluster for each aspect.
  na = cls->length ();
  clustr = new Cluster*[na];

  // Create distance matrix.
  // It has to have an even number of rows/cols.
  ma = na;
  if ((ma & 1) == 0)
    ++ma;
  dist = new HalfMatrix<Real> (ma);

  // Setup cluster list and compute distances from the initial
  // clusters.
  ListNode<Cluster> *cn = cls->first ();
  clustr[0] = cn->data ();
  unsigned long l = 1;
  while (l < na) {
    cn = cn->next ();
    clustr[l] = cn->data ();
    unsigned long k = 0;
    while (k < l) {
      (*dist) (l, k) = clustr[l]->sim (clustr[k]);
      ++k;
    }
    ++l;
  }

  // New distrance matrix size (after combining equal clusters)
  ma = na;
  if ((ma & 1) == 0)
    ++ma;

  // Fill any additional lines in the distance matrix with tol_max
  // to avoid problems. The additional lines are just there to
  // simplify indexing.
  while (l < ma) {
    unsigned long k = 0;
    while (k < l)
      (*dist) (l, k++) = TOL_MAX;
    ++l;
  }

  if (na > 3) {

    // Create first index matrix (if na > 3, i.e. we can create a 
    // reasonable index matrix)
    unsigned long mm = ma >> 1;
    unsigned long mmm = mm + 1;
    if (mmm > 2 && (mmm & 1) == 0)
      ++mmm;
    HalfMatrix<unsigned long> *idx = new HalfMatrix<unsigned long> (mmm);
    // Find minimum indexes for index matrix.
    l = 1;
    while (l <= mm) {
      unsigned long k = 0;
      while (k < l) {
	(*idx) (l, k) = min_index (l, k);
	++k;
      }
      ++l;
    }
    
    // Also fill index matrix.
    while (l < mmm) {
      unsigned long k = 0;
      while (k < l)
	(*idx) (l, k++) = ULONG_MAX;
      ++l;
    }
    
    parent = new ClusterQuadTreeIndex (idx, this);

  } else {

    // Note enough aspects to generate an index matrix.
    parent = NULL;

  }

  // We started with clusters, so we are at least partially clustered.
  clustered = true;
}

ClusterQuadTree::~ClusterQuadTree ()
{
  if (parent)
    delete parent;
  delete dist;
  if (!clustered) {
    // If clustered, the clusters are handled somewhere
    // else, otherwise we can delete them (but that should
    // actually never happen as we don't want to create
    // the matrices and then not use them for clustering, this
    // is just to avoid problems).
    for (unsigned long l = 0; l < na; ++l) {
      if (clustr[l])
	delete clustr[l];
    }
  }
  delete clustr;
}

void ClusterQuadTree::cluster (List<Cluster> *clusters)
{
  // Create hierarchical cluster structure.
  while (closest_pair ());

  // Store top clusters in the cluster list.
  for (unsigned long l = 0; l < na; ++l) {
    if (clustr[l]) {
      Real t = 0.0;
      compute_tolerance (clustr[l], clustr[l]->value (), t);
      clustr[l]->tolerance () = t;
      clusters->add_first (clustr[l]);
    }
  }

  clustered = true;
  return;
}

void ClusterQuadTree::compute_tolerance (Cluster *c, Feature *v, Real &t) const
{
  if (c->aspects ()) {
    Aspect *a = c->aspects () -> first ();
    while (a) {
      Real tt = a->feature () -> sim (v);
      if (tt > t)
	t = tt;
      a = c->aspects () -> next ();
    }
  }
  if (c->sub_clusters ()) {
    Cluster *cc = c->sub_clusters () -> first ();
    while (cc) {
      compute_tolerance (cc, v, t);
      cc = c->sub_clusters () -> next ();
    }
  }
  return;
}

unsigned long ClusterQuadTree::min_index (unsigned long l, 
					  unsigned long k) const
{
  unsigned long r0 = l << 1;
  unsigned long r1 = r0 - 1;
  unsigned long c0 = k << 1;
  unsigned long c1 = c0 + 1;
  unsigned long i = dist->index (r0, c0);
  unsigned long min_i = ULONG_MAX;
  Real min = (*dist) (i);
  if (min < TOL_MAX)
    min_i = i;
  i = dist->index (r0, c1);
  Real dd = (*dist) (i);
  if (dd < min) {
    min = dd;
    min_i = i;
  }
  i = dist->index (r1, c0);
  dd = (*dist) (i);
  if (dd < min) {
    min = dd;
    min_i = i;
  }
  if (r1 > c1) {
    i = dist->index (r1, c1);
    dd = (*dist) (i);
    if (dd < min) {
      dd = min;
      min_i = i;
    }
  }
  return min_i;
}

bool ClusterQuadTree::closest_pair ()
{
  if (parent) {
    // We can use index matrices.  Find the top-most index matrix and
    // get the closest pair from it.
    ClusterQuadTreeIndex *h = parent;
    while (h->parent)
      h = h->parent;
    return h->closest_pair ();
  }

  // If there is only one or less cluster, we return immediately.
  if (na < 2)
    return false;

  // We have exaclty three or one distance values and have to choose
  // the smallest.
  
  // Find smallest value in distance matrix.
  unsigned long min_idx = min_index (1, 0);
  
  // Also stop when smallest distance too large
  if (min_idx < ULONG_MAX) {
    Real min = (*dist) (min_idx);
    if (min < tol_approx) {
      // Combine the aspects and update the distance matrix and
      // the index matrices.
      unsigned long l = ULONG_MAX;
      unsigned long k = ULONG_MAX;
      dist->rev_index (min_idx, l, k);
      // Add cluster l to cluster k and recompute average:
      combine_clusters (l, k);
      // Remove cluster l distances form distance  matrix
      // by setting them to TOL_MAX.
      unsigned long c = 0;
      while (c < l)
	(*dist) (l, c++) = TOL_MAX;
      c = l + 1;
      while (c < na)
	(*dist) (c++, l) = TOL_MAX;
      // Recompute distance values for cluster k.
      c = 0;
      while (c < k) {
	if (clustr[c])
	  (*dist) (k, c) = clustr[k]->sim (clustr[c]);
	++c;
      }
      c = k + 1;
      while (c < na) {
	if (clustr[c])
	  (*dist) (c, k) = clustr[k]->sim (clustr[c]);
	++c;
      }
      // Update minimums in index matrices:
      update_rows (l, k);
      update_cols (l, k);
      return true;
    }
  }

  // Clusters too far appart or everyting combined.
  return false;
}

void ClusterQuadTree::combine_clusters (unsigned long l, unsigned long k)
{
  // Compute new average.
  Feature *v = clustr[l]->value () -> average (clustr[l]->num_aspects (),
					       clustr[k]->value (),
					       clustr[k]->num_aspects ());

  // New tolerance.
  Real t = v->sim (clustr[l]->value ());
  Real tt = v->sim (clustr[k]->value ());
  if (tt > t)
    t = tt;

  // Only create hierarchy structure if requested. Values have to be
  // sufficiently apart to be split in two clusters and united by a
  // parent cluster. Otherwise the clusters are simply combined.
  if (hier && t >= tol_dist) {
    // Combine the two clusters using a parent cluster
    Cluster *nc = new Cluster (clustr[k]->analyse (), v);
    delete v;
    Real tol = 0.0;
    compute_tolerance (clustr[l], clustr[l]->value (), tol);
    clustr[l]->tolerance () = tol;
    nc->add_sub_cluster (clustr[l]);
    tol = 0.0;
    compute_tolerance (clustr[k], clustr[k]->value (), tol);
    clustr[k]->tolerance () = tol;
    nc->add_sub_cluster (clustr[k]);
    clustr[k] = nc;
  } else if (hier && (clustr[l]->value () -> regular () || 
		      clustr[k]->value () -> regular ())) {
    // Combine the two clusters using a parent cluster as at least one
    // of them is regular.
    Cluster *nc = new Cluster (clustr[k]->analyse (), v);
    delete v;
    if (clustr[l]->value () -> regular ()) {
      Real tol = 0.0;
      compute_tolerance (clustr[l], clustr[l]->value (), tol);
      clustr[l]->tolerance () = tol;
      nc->add_sub_cluster (clustr[l]);
    } else {
      nc->combine (clustr[l]);
      delete clustr[l];
    }
    if (clustr[k]->value () -> regular ()) {
      Real tol = 0.0;
      compute_tolerance (clustr[k], clustr[k]->value (), tol);
      clustr[k]->tolerance () = tol;
      nc->add_sub_cluster (clustr[k]);
    } else {
      nc->combine (clustr[k]);
      delete clustr[k];
    }
    clustr[k] = nc;    
  } else {
    // No hierarchical clustering or clusters very close, so simply
    // combine both clusters to a new one.
    clustr[k]->combine (clustr[l], v);
    delete clustr[l];
  }

  // Set tolerance for new cluster.
  clustr[k]->tolerance () = t;

  // Cluster l is deleted and cluster k is the new cluster in the
  // cluster list.
  clustr[l] = NULL;

  return;
}

void ClusterQuadTree::update_rows (unsigned long r0, unsigned long r1)
{
  if (!parent)
    return;

  // Update rows with comparison with cluster r0.
  unsigned long rr0 = (r0 + 1) >> 1;
  unsigned long cc = 0;
  while (cc < rr0) {
    (*parent->idx) (rr0, cc) = min_index (rr0, cc);
    ++cc;
  }

  // Update rows with comparison with cluster r1 (if it is differnt
  // from r0 comparisons).
  unsigned long rr1 = (r1 + 1) >> 1;
  if (rr1 != rr0) {
    cc = 0;
    while (cc < rr1) {
      (*parent->idx) (rr1, cc) = min_index (rr1, cc);
      ++cc;
    }
  }

  // Update rows in index matrices.
  parent->update_rows (rr0, rr1);

  return;
}

void ClusterQuadTree::update_cols (unsigned long c0, unsigned long c1)
{
  if (!parent)
    return;

  // Update cols with comparisons with cluster c0.
  unsigned long cc0 = c0 >> 1;
  unsigned long rr = cc0 + 1;
  unsigned long max_rr = (ma >> 1) + 1;
  while (rr < max_rr) {
    (*parent->idx) (rr, cc0) = min_index (rr, cc0);
    ++rr;
  }

  // Update cols with comparisons with cluster c1 (if it is different
  // from c0 comaprisions).
  unsigned long cc1 = c1 >> 1;
  if (cc1 != cc0) {
    rr = cc1 + 1;
    while (rr < max_rr) {
      (*parent->idx) (rr, cc1) = min_index (rr, cc1);
      ++rr;
    }
  }

  // Update cols in index matrices.
  parent->update_cols (cc0, cc1);
  
  return;
}

ClusterQuadTreeIndex::ClusterQuadTreeIndex (HalfMatrix<unsigned long> *ind, 
					    ClusterQuadTree *tr)
{
  t = tr;
  idx = ind;

  if (idx->size () > 3) {

    // Create  index matrix for parent index matrix.
    unsigned long nn = idx->size () >> 1;
    unsigned long mm = nn + 1;
    if (mm > 2 && (mm & 1) == 0)
      ++mm;
    HalfMatrix<unsigned long> *nidx = new HalfMatrix<unsigned long> (mm);

    // Find minimum distance between pairs.
    unsigned long l = 1;
    while (l <= nn) {
      unsigned long k = 0;
      while (k < l) {
	(*nidx) (l, k) = min_index (l, k);
	++k;
      }
      ++l;
    }

    // Fill index matrix.
    while (l < mm) {
      unsigned long k = 0;
      while (k < l)
	(*nidx) (l, k++) = ULONG_MAX;
      ++l;
    }

    parent = new ClusterQuadTreeIndex (nidx, t);

  } else {

    // Last index as we have less than 4 elements to compare
    parent = NULL;

  }

  return;
}

unsigned long ClusterQuadTreeIndex::min_index (unsigned long l, 
					       unsigned long k) const
{
  unsigned long r0 = l << 1;
  unsigned long r1 = r0 - 1;
  unsigned long c0 = k << 1;
  unsigned long c1 = c0 + 1;
  Real dd;
  unsigned long min_i = ULONG_MAX;
  Real min = TOL_MAX;
  unsigned long i = (*idx) (r0, c0);
  if (i < ULONG_MAX) {
    dd = (*t->dist) (i);
    if (dd < min) {
      min = dd;
      min_i = i;
    }
  }
  i = (*idx) (r0, c1);
  if (i < ULONG_MAX) {
    dd = (*t->dist) (i);
    if (dd < min) {
      min = dd;
      min_i = i;
    }
  }
  i = (*idx) (r1, c0);
  if (i < ULONG_MAX) {
    dd = (*t->dist) (i);
    if (dd < min) {
      min = dd;
      min_i = i;
    }
  }
  if (r1 > c1) {
    i = (*idx) (r1, c1);
    if (i < ULONG_MAX) {
      dd = (*t->dist) (i);
      if (dd < min) {
	min = dd;
	min_i = i;
      }
    }
  }
  return min_i;
}

bool ClusterQuadTreeIndex::closest_pair ()
{
  // This should only be called for the top-most index HalfMastrix.
  assert (!parent);

  // Find smallest value in index matrix.
  unsigned long min_idx = min_index (1, 0);

  if (min_idx < ULONG_MAX) {
    Real min = (*t->dist) (min_idx);
    if (min < t->tol_approx) {
      // Combine the aspects and update the distance matrix and
      // the index matrices.
      unsigned long l = ULONG_MAX;
      unsigned long k = ULONG_MAX;
      t->dist->rev_index (min_idx, l, k);
      // Add cluster l to cluster k and recompute average:
      t->combine_clusters (l, k);
      // Remove cluster l distances form distance  matrix
      // by setting them to TOL_MAX.
      unsigned long c = 0;
      while (c < l)
	(*t->dist) (l, c++) = TOL_MAX;
      c = l + 1;
      while (c < t->na)
	(*t->dist) (c++, l) = TOL_MAX;
      // Recompute distance values for cluster k.
      c = 0;
      while (c < k) {
	if (t->clustr[c])
	  (*t->dist) (k, c) = t->clustr[k]->sim (t->clustr[c]);
	++c;
      }
      c = k + 1;
      while (c < t->na) {
	if (t->clustr[c])
	  (*t->dist) (c, k) = t->clustr[k]->sim (t->clustr[c]);
	++c;
      }
      // Update minimums in index matrices:
      t->update_rows (l, k);
      t->update_cols (l, k);
      return true;
    }
  }

  // Clusters too far appart or everyting combined.
  return false;
}

void ClusterQuadTreeIndex::update_rows (unsigned long r0, unsigned long r1)
{
  if (!parent)
    return;

  // Update rows with comparisons with cluster r0.
  unsigned long rr0 = (r0 + 1) >> 1;
  unsigned long cc = 0;
  while (cc < rr0) {
    (*parent->idx) (rr0, cc) = min_index (rr0, cc);
    ++cc;
  }

  // Update rows with comparisons with cluster r1 (if it is different
  // from comaprisons with cluster r0).
  unsigned long rr1 = (r1 + 1) >> 1;
  if (rr1 != rr0) {
    cc = 0;
    while (cc < rr1) {
      (*parent->idx) (rr1, cc) = min_index (rr1, cc);
      ++cc;
    }
  }

  // Update parent index matrices.
  parent->update_rows (rr0, rr1);

  return;
}

void ClusterQuadTreeIndex::update_cols (unsigned long c0, unsigned long c1)
{
  if (!parent)
    return;

  // Update cols with comparisons with cluster c0.
  unsigned long cc0 = c0 >> 1;
  unsigned long rr = cc0 + 1;
  unsigned long max_rr = (idx->size () >> 1) + 1;
  while (rr < max_rr) {
    (*parent->idx) (rr, cc0) = min_index (rr, cc0);
    ++rr;
  }
  
  // Update cols with comparisons with cluster c1 (if it is different
  // from comparisions with cluster c0)
  unsigned long cc1 = c1 >> 1;
  if (cc1 != cc0) {
    rr = cc1 + 1;
    while (rr < max_rr) {
      (*parent->idx) (rr, cc1) = min_index (rr, cc1);
      ++rr;
    }
  }

  // Update parent index matrices.
  parent->update_cols (cc0, cc1);
  
  return;
}

void Analyse::cluster_aspects (List<Aspect> *aspects, Real tol_eq,
			       Real tol_dist, Real tol_approx,
			       bool hier, List<Cluster> *clusters)
{
  if (aspects->length () < 1)
    return;

  if (aspects->length () == 1) {
    // Only one aspect, simply create a single cluster from it.
    Aspect *a = aspects->first () -> data ();
    Cluster *c = new Cluster (this, (Feature *)a->feature ());
    c->add_aspect (a);
    clusters->add_first (c);
    return;
  }

  // Cluster the aspects using a quad tree structure
  ClusterQuadTree *qt = new ClusterQuadTree (aspects, tol_eq, tol_dist,
					     tol_approx, hier, this);
  qt->cluster (clusters);
  delete qt;

  return;
}

void Analyse::cluster_aspects (List<Cluster> *cls, Real tol_eq,
			       Real tol_dist, Real tol_approx,
			       bool hier)
{
  if (cls->length () < 2) {
    // Only one or no cluster, simply return list.
    return;
  }

  // Cluster the clusters using a quad tree structure
  ClusterQuadTree *qt = new ClusterQuadTree (cls, tol_eq, tol_dist,
					     tol_approx, hier, this);
  cls->clear ();
  qt->cluster (cls);
  delete qt;

  return;
}

void Analyse::tolerance_levels (List<Cluster> *clusters,
				SimpleList<Real> *tols, Real tol_dist,
				Real tol_max)
{
  // Reduce hierarchical clustering to basic tolerance levels.

  // Determine distinct tolerance levels. A level ist distinct from
  // it's previous level if the increase of the tolerance difference
  // is larger than in the first tolerance level step plus
  // tol_dist_angle.
  tols->sort ();
  Real *tol0 = tols->first ();
  Real *tol1 = tols->next ();
  if (tol1) {
    Real delta0 = *tol1 - *tol0 + tol_dist;
    tol0 = tol1;
    tol1 = tols->next ();
    while (tol1) {
      if (*tol1 - * tol0 < delta0) {
	tols->del (tol0);
	delete tol0;
      }
      tol0 = tol1;
      tol1 = tols->next ();
    }
  }

  // Collapse clusters with respect to tolerance levels.
  Real t0 = -1.0; 
  tol1 = tols->first ();
  while (tol1) {
    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      cn->data () -> collapse (t0, *tol1);
      cn = cn->next ();
    }
    t0 = *tol1;
    tol1 = tols->next ();
  }

  if (tol_max > 0.0 && tols->length () > 1) {

    // Determine maximum jump between tolerance levels considering maximum
    // tolerance tol_max;
    tol1 = tols->first ();
    Real *toln = tols->next ();
    Real delta = *toln - *tol1;
    Real level = *tol1;
    tol1 = toln;
    toln = tols->next ();
    while (toln) {
      Real d = *toln - *tol1;
      if (d > delta) {
	delta = d;
	level = *tol1;
      }
      tol1 = toln;
      toln = tols->next ();
    }
    if (*tols->last () < tol_max) {
      Real d = tol_max - *tols->last ();
      if (d > delta)
	level = *tols->last ();
    }

    // Remove any clusters with a tolerance larger than level.
    SimpleList<Cluster> cls;
    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      cls.add_first (cn->data ());
      cn = cn->next ();
    }
    Cluster *c = cls.unlink_first ();
    while (c) {
      if (c->tolerance () > level + sil_get_epsilon ()) {
	// Collapse tolerance level unless there are aspects involved.
	if (c->parent_cluster ()) {
	  if (c->sub_clusters ()) {
	    Cluster *sc = c->sub_clusters () -> first ();
	    while (sc) {
	      cls.add_first (sc);
	      sc = c->sub_clusters () -> next ();
	    }
	  }
	  c->collapse ();
	  delete c;
	} else {
	  if (!c->aspects () || c->aspects () -> length () == 0) {
	    if (c->sub_clusters ()) {
	      Cluster *sc = c->sub_clusters () -> unlink_first ();
	      while (sc) {
		cls.add_first (sc);
		clusters->add_last (sc);
		sc->parent_cluster () = NULL;
		sc = c->sub_clusters () -> unlink_first ();
	      }
	    }
	    clusters->del (c);
	    delete c;
	  }
	}
      }
      c = cls.unlink_first ();
    }
    
  }

  tols->purge ();

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
