// analyse_directions.cc: Analyse a SiL model / directions.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_directions.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_directions,
       "$Id: analyse_directions.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"
#include "kernel/geometry.hxx"
#include "kernel/lapack.h"

List<Cluster> *Analyse::eval_parallel_directions ()
{
  // Cluster the directions into hierarchical parallel direction
  // clusters.
  if (direction_aspects.length () > 0) {

    List<Cluster> *clusters = new List<Cluster>;
    cluster_aspects (&direction_aspects, tol_eq, tol_angle_dist, NAN, true,
		     clusters);
    reg_list->add_first ((List<Regularity> *)clusters);

    // Solve the equation system Dx = 1, i.e. d_i^t x = 1 for all
    // parallel directions d_i in a least squares sense:
    //
    //   ||Dx_0 - e||_2 -> min
    //    x = x_0 / ||x_0||_2

    sil_progress_bar_text ("Direction tolerance levels and least squares fit");

    // Also store tolreances (tol_angle_dist is always the first level)
    SimpleList<Real> tols;
    tols.add_first (new Real (tol_angle_dist));

    SimpleList<Cluster> cls;
    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      cls.add_first (cn->data ());
      cn = cn->next ();
    }
    Cluster *c = cls.unlink_first ();
    while (c) {

      // Solve system in least squares sense using SVD.
      SimpleList<Aspect> aspects;
      c->get_aspects (&aspects);

      Direction *d = static_cast<Direction*> (c->value ());
      
      if (aspects.length () > 2) {
	
	// If we have more than two aspects do a least squares fit.
	
	char JOBU = 'S'; // only compute the left singular vectors
	char JOBVT = 'O'; // only compute the right singular vectors in A
	long M = aspects.length (); // number of rows
	long N = 3; // number of columns
	Real A[M * N]; // direction matrix
	Real U[M * 3]; // U matrix (only the first three vectors)
	Real S[3]; // singular values
	long LWORK = 50 + M; // work space
	Real WORK[LWORK];
	long INFO = 0; // result
	
	// Initialize A
	Aspect *a = aspects.first ();
	long l1 = 0;
	long l2 = M;
	long l3 = l2 + M;
	while (a) {
	  Direction *D = dynamic_cast<Direction*> (a->feature ());
	  if (D) {
	    if ((*D) % (*d) >= 0.0) {
	      A[l1++] = D->x ();
	      A[l2++] = D->y ();
	      A[l3++] = D->z ();
	    } else {
	      // Adjust `orientation'.
	      A[l1++] = -D->x ();
	      A[l2++] = -D->y ();
	      A[l3++] = -D->z ();
	    }
	  } else {
	    sil_msg_warn ("Analyse",
			  "parallel direction aspect has non-direction feature " +
			  a->feature () -> type_id ());
	    A[l1++] = 0.0;
	    A[l2++] = 0.0;
	    A[l3++] = 0.0;
	  }
	  a = aspects.next ();
	}
	
	// SVD
	dgesvd (&JOBU, &JOBVT, &M, &N, A, &M, S, U, &M, NULL,
		&N, WORK, &LWORK, &INFO);
	
	if (INFO == 0) {
	  
	  // Compute x = V S^* U^T * e
	  
	  // Sum up the first three row of U^t (columns of U)
	  Real x1 = 0.0;
	  Real x2 = 0.0;
	  Real x3 = 0.0;
	  l1 = 0;
	  l2 = M;
	  l3 = l2 + M;
	  while (l1 < M) {
	    x1 += U[l1++];
	    x2 += U[l2++];
	    x3 += U[l3++];
	  }
	  
	  // Multiply x with S^*
	  if (fabs (S[0]) > sil_get_epsilon ())
	    x1 /= S[0];
	  else
	    x1 = 0.0;
	  if (fabs (S[1]) > sil_get_epsilon ())
	    x2 /= S[1];
	  else
	    x2 = 0.0;
	  if (fabs (S[2]) > sil_get_epsilon ())
	    x3 /= S[2];
	  else
	    x3 = 0.0;
	  
	  // Multiply x with V
	  l1 = 2 * M;
	  Real xx1 = A[0]  * x1 + A[1]      * x2 + A[2]      * x3;
	  Real xx2 = A[M]  * x1 + A[M + 1]  * x2 + A[M + 2]  * x3;
	  Real xx3 = A[l1] * x1 + A[l1 + 1] * x2 + A[l1 + 2] * x3;
	  
	  // Set direction.
	  d->set (xx1, xx2, xx3);
	  
	} else {
	  
	  sil_msg_warn ("Analyse", 
			"parallel direction optimization failed");
	  
	}
	
      }

      // Recompute the tolerance.
      Real t = 0.0;
      Aspect *a = aspects.first ();
      while (a) {
	Direction *D = dynamic_cast<Direction*> (a->feature ());
	if (D) {
	  Real an = fabs ((*D) % (*d));
	  if (an >= 1.0)
	    an = 0.0;
	  else
	    an = acos (an);
	  if (an > t)
	    t = an;
	}
	a = aspects.next ();
      }
      c->tolerance () = t;
      if (t > tol_angle_dist)
	tols.add_first (new Real (t));
      
      if (c->sub_clusters ()) {
	Cluster *sc = c->sub_clusters () -> first ();
	while (sc) {
	  cls.add_first (sc);
	  sc = c->sub_clusters () -> next ();
	}
      }

      c = cls.unlink_first ();
    }

    tolerance_levels (clusters, &tols, tol_angle_dist, PI / 2.0);

    return clusters;
  }
  return NULL;
}

List<Cluster> *Analyse::eval_direction_planes (List<Cluster> *directions)
{
  // Generate a normal describing a direction plane from each pair of
  // main directions.
  //
  // We are only using the main directions as we are looking for
  // planes created by different directions and not directions which
  // are close together. So we only consider the direction clusters
  // instead of the directions, but use all direction clusters which
  // are not only composed of sub-clusters.
  List<Cluster> dcl;
  ListNode<Cluster> *an1 = directions->first ();
  while (an1) {
    an1->data () -> get_all_base_clusters (&dcl);
    an1 = an1->next ();
  }

  // Create direction plane aspects.
  List<Aspect> aspects;
  an1 = dcl.first ();
  while (an1) {
    UnitVector3 d1 = *static_cast<Direction *> (an1 -> data () -> value ());
    ListNode<Cluster> *an2 = an1->next ();
    while (an2) {
      UnitVector3 d2 = *static_cast<Direction *> (an2 -> data () -> value ());
      Aspect *asp = new Aspect (new DirectionPlane (d1 * d2), an1 -> data ());
      asp->owners () . add_first (an2->data ());
      aspects.add_first (asp);
      an2 = an2->next ();
    }
    an1 = an1->next ();
  }

  // Cluster the direction planes.
  if (aspects.length () > 0) {
    sil_progress_bar_text ("Clustering " + as_string (aspects.length ()) +
			   " direction planes");
    List<Cluster> *clusters = new List<Cluster>;
    cluster_aspects (&aspects, tol_eq, tol_angle_dist, NAN, true, clusters);

    // Solve the equation system Dx = 0, i.e. d_i^t x = 0 for all
    // directions d_i in a direction plane in a least squares sense:
    //
    //   ||Dx||_2 -> min  for ||x||_2 = 1
    //
    // SVD of D is USV^T. U and V are unitary so they do not change
    // the norm, i.e. we get the smallest norm by multiplying S with
    // e_n, where n corresponds to the smallest singular value of
    // D. Thus we get V^t x = e_n, x = Ve_n.

    sil_progress_bar_text ("Direction plane tolerance levels and least squares fit");

    // Also collect the tolerances.
    SimpleList<Real> tols;
    tols.add_first (new Real (tol_angle_dist));

    SimpleList<Cluster> cls;
    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      cls.add_first (cn->data ());
      cn = cn->next ();
    }
    Cluster *c = cls.unlink_first ();
    while (c) {

      // Solve system in least squares sense using SVD.
      SimpleList<Entity> owners;
      c->get_owners (&owners);

      if (owners.length () > 2) {

	char JOBU = 'N'; // do not compute U
	char JOBVT = 'O'; // only compute the right singular vectors in A
	long M = owners.length (); // number of rows
	long N = 3; // number of columns
	Real A[M * N]; // direction matrix
	Real S[3]; // singular values
	long LWORK = 50 + M; // work space
	Real WORK[LWORK];
	long INFO = 0; // result

	// Initialize A
	Entity *e = owners.first ();
	unsigned long l1 = 0;
	unsigned long l2 = M;
	unsigned long l3 = l2 + M;
	while (e) {
	  Cluster *d = dynamic_cast<Cluster*> (e);
	  if (d) {
	    Direction *D = static_cast<Direction*> (d->value ());
	    A[l1++] = D->x ();
	    A[l2++] = D->y ();
	    A[l3++] = D->z ();
	  } else {
	    sil_msg_warn ("Analyse",
			  "direction plane has non-cluster owner " +
			  e->type_id ());
	    A[l1++] = 0.0;
	    A[l2++] = 0.0;
	    A[l3++] = 0.0;
	  }
	  e = owners.next ();
	}
	
	// SVD
	dgesvd (&JOBU, &JOBVT, &M, &N, A, &M, S, NULL, &M, NULL,
		&N, WORK, &LWORK, &INFO);

	if (INFO == 0) {

	  // Find smallest singular value.
	  unsigned long l = 2;
	  if (fabs (S[1]) < fabs (S[2]))
	    l = 1;
	  if (fabs (S[0]) < fabs (S[l]))
	    l = 0;
	  
	  // The updated normal vector is the l-th column of V which is
	  // stored as the l-th row of A.
	  l *= M;
	  static_cast<DirectionPlane*> (c->value ()) -> 
	    set (A[l], A[l + 1], A[l + 2]);
	  
	} else {
	  
	  sil_msg_warn ("Analyse", "direction plane optimization failed");
	  
	}

      }
      
      // Recompute the tolerance.
      Real t = 0.0;
      Entity *e = owners.first ();
      while (e) {
	Cluster *d = dynamic_cast<Cluster*> (e);
	if (d) {
	  Direction *D = static_cast<Direction*> (d->value ());
	  Real a = fabs ((*D) %
			 *static_cast<DirectionPlane*> (c->value ()));
	  if (a >= 1.0)
	    a = PI / 2.0;
	  else
	    a = PI / 2.0 - acos (a);
	  if (a > t)
	    t = a;
	}
	e = owners.next ();
      }
      c->tolerance () = t;
      if (t > tol_angle_dist)
	tols.add_first (new Real (t));
      
      if (c->sub_clusters ()) {
	Cluster *sc = c->sub_clusters () -> first ();
	while (sc) {
	  cls.add_first (sc);
	  sc = c->sub_clusters () -> next ();
	}
      }
      c = cls.unlink_first ();
    }

    tolerance_levels (clusters, &tols, tol_angle_dist, PI / 2.0);

    // Find and only keep regular clusters.
    List<Cluster> bcls;
    cn = clusters->first ();
    while (cn) {
      cn->data () -> get_all_base_clusters (&bcls);
      cn = cn->next ();
    }
    cn = bcls.first ();
    while (cn) {

      Cluster *c = cn->data ();

      // Count number of different directions in the plane.
      SimpleList<Entity> owners;
      c->get_owners (&owners);
      // If we have n different directions, then n(n-1)/2 direction planes
      // would have to be present if we have a transitive structure in the
      // cluster. If the cluster is not transitive, it is not regular.
      unsigned long N = owners.length ();
      unsigned long n = (N * (N - 1)) >> 1;

      if (c->num_aspects () != n) {
	// Not regular, check if we can make it regular with respect
	// to a larger direction tolerance level.

	// Find direction pairs in owners for which there is no
	// direction plane aspect in the cluster. If a missing pair
	// has the same parent cluster, count it.
	unsigned long npairs = 0;
	SimpleList<Aspect> aspects;
	c->get_aspects (&aspects);
	Entity *e = owners.unlink_first ();
	while (e) {
	  SimpleList<Aspect> asp;
	  Aspect *a = aspects.first ();
	  while (a) {
	    Aspect *an = aspects.next ();
	    if (a->owners () . is_member (e)) {
	      asp.add_first (a);
	      aspects.del (a);
	    }
	    a = an;
	  }
	  Entity *ee = owners.first ();
	  while (ee) {
	    bool miss = true;
	    a = asp.first ();
	    while (a) {
	      if (a->owners () . is_member (ee)) {
		asp.del (a);
		miss = false;;
		break;
	      }
	      a = asp.next ();
	    }
	    if (miss) {
	      // If it's not found, combine directions if they are
	      // parallel on a higher tolerance level still consistent
	      // with the tolerance of the direction plane cluster
	      // (but always check at least one tolerance level).
	      if (static_cast<Cluster*> (e) -> parent_cluster () &&
		  static_cast<Cluster*> (e) -> parent_cluster () ==
		  static_cast<Cluster*> (ee) -> parent_cluster ()) {
		++npairs;
	      } else {
		Cluster *p1 = static_cast<Cluster*> (e) -> parent_cluster ();
		Cluster *p2 = static_cast<Cluster*> (ee) -> parent_cluster ();
		while (p1 && p2) {
		  // p1 is always cluster with smaller tolerance and
		  // we move upwards at cluster with smaller
		  // tolerance.
		  if (p2->tolerance () < p1->tolerance ()) {
		    Cluster *pp = p2;
		    p2 = p1;
		    p1 = pp;
		  }
		  // Stop if we are over the tolerance level of the plane
		  // cluster.
		  if (p1->tolerance () > c->tolerance ()) 
		    break;
		  p1 = p1->parent_cluster ();
		  // Found tolerance level which is smaller than the
		  // tolerance level of the plane on which both
		  // directions are parallel.
		  if (p1 == p2) {
		    ++npairs;
		    break;
		  }
		}
	      }
	    }
	    ee = owners.next ();
	  }
	  e = owners.unlink_first ();
	}

	// Reduce the number of required aspects by number of missing
	// pairs which belong to the same parent cluster.
	n -= npairs;

      }

      if (c->num_aspects () == n) {
	// Cluster is regular.

	//Find parallel direction
	Real tol = c->tolerance () + tol_angle_dist;
	Cluster *pa = find_parallel (*static_cast<UnitVector3 const *> 
				     (static_cast<DirectionPlane*> 
				      (c->value ())), NULL, tol);
	if (pa)
	  c->add_sub_regularity (new ParallelDirection (this, tol, pa));

	if (c->parent_cluster ())
	  bcls.add_last_check (c->parent_cluster ());
	
	if (!pa && c->num_aspects () < 2) {
	  // Only a single direction plane and not parallel to
	  // anything, so collapse it.
	  if (c->collapse ()) {
	    delete c;
	  } else {
	    // c is top level cluster, so remove it, but add it's sub
	    // clusters if it only has sub-clusters. If it also has
	    // aspects, then keep the cluster to avoid losing the
	    // angles between the directions in the plane.
	    if (c->sub_clusters ()) {
	      Cluster *sc = c->sub_clusters () -> unlink_first ();
	      while (sc) {
		clusters->add_first (sc);
		sc->parent_cluster () = NULL;
		sc = c->sub_clusters () -> unlink_first ();
	      }
	    }
	    if (!c->aspects ()) {
	      clusters->del (c);
	      delete c;
	    }
	  }
	}

      } else {
 
	// Cluster is not regular, collapse it (i.e. move everyting to parent
	// if possible).
	if (c->collapse ()) {
	  // Cluster collpase, so remove it and it's parent to the list.
	  bcls.add_last_check (c->parent_cluster ());
	} else {
	  // c is top level cluster, so remove it, but add it's sub
	  // clusters.
	  clusters->del (c);
	  if (c->sub_clusters ()) {
	    Cluster *sc = c->sub_clusters () -> unlink_first ();
	    while (sc) {
	      clusters->add_first (sc);
	      sc->parent_cluster () = NULL;
	      sc = c->sub_clusters () -> unlink_first ();
	    }
	  }
	}
	delete c;

      }

      cn = cn->next ();
    }

    reg_list->add_first ((List<Regularity> *)clusters);

    return clusters;
  }

  return NULL;
}

List<Cluster> *Analyse::eval_direction_cones (List<Cluster> *directions)
{
  // From three directions we can generate a cone represented by a
  // normal and an angle. We cluster these cones. Later we check for
  // conical angle regular directions based on the cones.
  List<Cluster> dcl;
  ListNode<Cluster> *an1 = directions->first ();
  while (an1) {
    an1->data () -> get_all_base_clusters (&dcl);
    an1 = an1->next ();
  }
  an1 = dcl.first ();
  List<Aspect> angle_aspects;
  an1 = dcl.first ();
  while (an1) {
    UnitVector3 d1 = *static_cast<Direction *> (an1->data () -> value ());
    ListNode<Cluster> *an2 = an1->next ();
    while (an2) {
      UnitVector3 d2 = *static_cast<Direction *> (an2->data () -> value ());
      ListNode<Cluster> *an3 = an2->next ();
      while (an3) {
	UnitVector3 d3 = *static_cast<Direction *> (an3->data () -> value ());
	// Generate the cone.
	UnitVector3 n = solve_cone (d1, d2, d3);
	if (n.norm_sq () > tol_eq * tol_eq) {
	  Real angl = fabs (n % d1);
	  if (angl >= 1.0)
	    angl = 0.0;
	  else
	    angl = acos (angl);
	  // Get rid of approximately flat cones and parallel axis
	  // diretions forming degenerate cones.
	  if (fabs (angl) > tol_angle_dist &&
	      fabs (angl - PI / 2.0) > tol_angle_dist) {
	    // Add the direction cone.
	    Aspect *asp = new Aspect (new DirectionCone (n, angl),
				      an1->data ());
	    asp->owners () . add_first (an2->data ());
	    asp->owners () . add_first (an3->data ());
	    Aspect *aasp = new Aspect (new Angle (angl), asp);
	    angle_aspects.add_first (aasp);
	  }
	}
	an3 = an3->next ();
      }
      an2 = an2->next ();
    }
    an1 = an1->next ();
  }

  // Cluster the direction cones.
  if (angle_aspects.length () > 0) {

    // First cluster cone semi-angles.
    sil_progress_bar_text ("Clustering " + 
			   as_string (angle_aspects.length ()) +
			   " direction cone semi-angles");
    List<Cluster> angle_clusters;
    cluster_aspects (&angle_aspects, tol_eq, tol_angle_dist, NAN, true,
		     &angle_clusters);
 
    // Get tolerances.
    SimpleList<Real> tols;
    tols.add_first (new Real (tol_angle_dist));
    SimpleList<Cluster> cls;
    ListNode<Cluster> *cn = angle_clusters.first ();
    while (cn) {
      cls.add_first (cn->data ());
      cn = cn->next ();
    }
    Cluster *c = cls.unlink_first ();
    while (c) {
      if (c->tolerance () > tol_angle_dist)
	tols.add_first (new Real (c->tolerance ()));
      if (c->sub_clusters ()) {
	Cluster *sc = c->sub_clusters () -> first ();
	while (sc) {
	  cls.add_first (sc);
	  sc = c->sub_clusters () -> next ();
	}
      }
      c = cls.unlink_first ();
    }

    tolerance_levels (&angle_clusters, &tols, tol_angle_dist, PI / 2.0);

    sil_progress_bar_text ("Clustering direction cones along the semi-angle tree");

    // Traverse the angle cluster tree depth first, cluster the
    // direction cones at the leaves and merge the clusters at each
    // node.
    List<Cluster> *clusters = NULL;
    cn = angle_clusters.first ();
    while (cn) {
      List<Cluster> *cls = cluster_merge_direction_cones (cn->data ());
      if (cls) {
	if (clusters) {
	  clusters->combine (cls);
	  delete cls;
	} else {
	  clusters = cls;
	}
      }
      cn = cn->next ();
    }

    // Cleanup angle clusters.
    angle_clusters.purge ();

    // Determine tolerance levels for direction cones.
    tols.add_first (new Real (tol_angle_dist));
    cn = clusters->first ();
    while (cn) {
      cls.add_first (cn->data ());
      cn = cn->next ();
    }
    c = cls.unlink_first ();
    while (c) {
      if (c->tolerance () > tol_angle_dist)
	tols.add_first (new Real (c->tolerance ()));
      if (c->sub_clusters ()) {
	Cluster *sc = c->sub_clusters () -> first ();
	while (sc) {
	  cls.add_first (sc);
	  sc = c->sub_clusters () -> next ();
	}
      }
      c = cls.unlink_first ();
    }
    tolerance_levels (clusters, &tols, tol_angle_dist, PI / 2.0);

    // Solve the equation system Dx = cos(a), i.e. d_i^t x = cos(a)
    // for all directions d_i in a direction cone adjusted to the cone
    // direction in a least squares sense:
    //
    //   ||Dx_0 - e||_2 -> min 
    //
    //    x = x_0 / ||x_0||
    //    average angle a:  cos(a) = avg_i (d_i^t x)

    List<Cluster> bcls;
    cn = clusters->first ();
    while (cn) {
      cn->data () -> get_all_base_clusters (&bcls);
      cn = cn->next ();
    }

    cn = bcls.unlink_first ();
    while (cn) {

      c = cn->data ();

      // Solve system in least squares sense using SVD.
      SimpleList<Entity> owners;
      c->get_owners (&owners);

      DirectionCone *dc = static_cast<DirectionCone*> (c->value ());

      if (owners.length () > 2) {

	char JOBU = 'S'; // only compute the left singular vectors
	char JOBVT = 'O'; // only compute the right singular vectors in A
	long M = owners.length (); // number of rows
	long N = 3; // number of columns
	Real A[M * N]; // direction matrix
	Real U[M * 3]; // U matrix (only the first three vectors)
	Real S[3]; // singular values
	long LWORK = 50 + M; // work space
	Real WORK[LWORK];
	long INFO = 0; // result

	// Initialize A
	Entity *e = owners.first ();
	long l1 = 0;
	long l2 = M;
	long l3 = l2 + M;
	while (e) {
	  Cluster *d = dynamic_cast<Cluster*> (e);
	  if (d) {
	    Direction *D = static_cast<Direction*> (d->value ());
	    if (((*D) % *dc) > 0.0) {
	      A[l1++] = D->x ();
	      A[l2++] = D->y ();
	      A[l3++] = D->z ();
	    } else {
	      // Adjust `orientation'.
	      A[l1++] = -D->x ();
	      A[l2++] = -D->y ();
	      A[l3++] = -D->z ();
	    }
	  } else {
	    sil_msg_warn ("Analyse",
			  "direction cone has non-cluster owner " +
			  e->type_id ());
	    A[l1++] = 0.0;
	    A[l2++] = 0.0;
	    A[l3++] = 0.0;
	  }
	  e = owners.next ();
	}
	
	// SVD
	dgesvd (&JOBU, &JOBVT, &M, &N, A, &M, S, U, &M, NULL,
		&N, WORK, &LWORK, &INFO);

	if (INFO == 0) {

	  // Compute x = V S^* U^T * e
	  
	  // Sum up the first three row of U^t (columns of U)
	  Real x1 = 0.0;
	  Real x2 = 0.0;
	  Real x3 = 0.0;
	  l1 = 0;
	  l2 = M;
	  l3 = l2 + M;
	  while (l1 < M) {
	    x1 += U[l1++];
	    x2 += U[l2++];
	    x3 += U[l3++];
	  }
	  
	  // Multiply x with S^*
	  if (fabs (S[0]) > sil_get_epsilon ())
	    x1 /= S[0];
	  else
	    x1 = 0.0;
	  if (fabs (S[1]) > sil_get_epsilon ())
	    x2 /= S[1];
	  else
	    x2 = 0.0;
	  if (fabs (S[2]) > sil_get_epsilon ())
	    x3 /= S[2];
	  else
	    x3 = 0.0;
	  
	  // Multiply x with V
	  l1 = 2 * M;
	  Real xx1 = A[0]  * x1 + A[1]      * x2 + A[2]      * x3;
	  Real xx2 = A[M]  * x1 + A[M + 1]  * x2 + A[M + 2]  * x3;
	  Real xx3 = A[l1] * x1 + A[l1 + 1] * x2 + A[l1 + 2] * x3;
	  
	  // The updated normal vector is the l-th column of V which is
	  // stored as the l-th row of A.
	  dc->set (xx1, xx2, xx3);

	} else {

	  sil_msg_warn ("Analyse", "direction cone optimization failed");
	  
	}
	
	// Recompute the angle as average
	l1 = 0;
	Real angl = 0.0;
	e = owners.first ();
	while (e) {
	  Cluster *d = dynamic_cast<Cluster*> (e);
	  if (d) {
	    Direction *D = static_cast<Direction*> (d->value ());
	    Real a = fabs ((*D) % (*dc));
	    if (a >= 1.0)
	      a = 0.0;
	    else
	      a = acos (a);
	    angl += a;
	    ++l1;
	  }
	  e = owners.next ();
	}
	angl /= static_cast<Real> (l1);
	dc->angle () = angl;

      }

      // Recompute the tolerance.
      Real t = 0.0;
      Entity *e = owners.first ();
      while (e) {
	Cluster *d = dynamic_cast<Cluster*> (e);
	if (d) {
	  Direction *D = static_cast<Direction*> (d->value ());
	  Real a = fabs ((*D) % (*dc));
	  if (a >= 1.0)
	    a = dc->angle ();
	  else
	    a = fabs (dc->angle () - acos (a));
	  if (a > t)
	    t = a;
	}
	e = owners.next ();
      }
      c->tolerance () = t;      

      if (c->sub_clusters ()) {
	// Check if there is a sub--cluster which has the same owners
	// than the parent cluster and collapse those clusters. Also
	// collapse the sub-clusters that are within the tolerance
	// range.
	Cluster *sc = c->sub_clusters () -> first ();
	while (sc) {
	  Cluster *nsc = c->sub_clusters () -> next ();
	  // Set parent cluster as tree guided clustering could
	  // destroy that information.
	  sc->parent_cluster () = c;
	  SimpleList<Entity> scown;
	  sc->get_owners (&scown);
	  if (scown.length () == owners.length () ||
	      sc->tolerance () >= c->tolerance () - tol_angle_dist) {
	    sc->collapse ();
	    bcls.del (sc);
	    delete sc;
	  }
	  sc = nsc;
	}
      }

      if (c->parent_cluster ())
	bcls.add_last_check (c->parent_cluster ());

      delete cn;
      cn = bcls.unlink_first ();
    }

    reg_list->add_first ((List<Regularity> *)clusters);

    return clusters;
  }

  return NULL;
}

List<Cluster> *Analyse::cluster_merge_direction_cones (Cluster *angle)
{
  // Cluster and merge direction cones along the angle tree in a depth
  // first manner.

  // List of top-level clusters generated for this angle tree level.
  List<Cluster> *clusters = NULL;
  // List of base level clusters on this angle tree level (either real
  // base clusters or clusters used to start clustering on this level).
  List<Cluster> bcls;

  // Traverse the angle tree depth first and cluster the aspects
  // spearately on each level.
  if (angle->sub_clusters () && angle->sub_clusters () -> length () > 0) {

    // We are at a node in the tree, so collect the clusters of it's
    // children and then cluster these.
    
    SimpleList<DirectionCone> reg_mark;

    // Get cluster lists from sub clusters and merge the lists.
    Cluster *c = angle->sub_clusters () -> first ();
    while (c) {
      List<Cluster> *cls = cluster_merge_direction_cones (c);
      if (cls) {
	ListNode<Cluster> *cn = cls->first ();
	while (cn) {
	  bcls.add_first (cn->data ());
	  // Mark every base cluster regular to aviod that it is
	  // removed by the clustering.
	  if (!cn->data () -> value () -> regular ()) {
	    DirectionCone *dc = static_cast<DirectionCone*> (cn->data () ->
							     value ());
	    dc->regular () = true;
	    reg_mark.add_first (dc);
	  }
	  cn = cn->next ();
	}
	if (clusters) {
	  clusters->combine (cls);
	  delete cls;
	} else {
	  clusters = cls;
	}
      }
      c = angle->sub_clusters () -> next ();
    }

    if (angle->aspects () && angle->aspects () -> length () > 0) {

      // Add direciton cone aspects in the angle aspect at this tree
      // level.
      if (!clusters)
	clusters = new List<Cluster>;
      Aspect *a = angle->aspects () -> first ();
      while (a) {
	Entity *e = a->owners () . first ();
	while (e) {
	  clusters->add_first (new Cluster (this, static_cast<Aspect*> (e)));
	  e = a->owners () . next ();
	}
	a = angle->aspects () -> next ();
      }
    }

    if (!clusters || clusters->length () == 0) {
      delete clusters;
      return NULL;
    }

    // Continue the clustering.
    cluster_aspects (clusters, tol_eq, tol_angle_dist, NAN, true);
    if (clusters->length () == 0) {
      delete clusters;
      return NULL;
    }

    // Drop temporary regular markings.
    DirectionCone *dc = reg_mark.first ();
    while (dc) {
      dc->regular () = false;
      dc = reg_mark.next ();
    }

  } else {

    // We are at a leave of the angle tree, so cluster the aspects
    // which are combinations of angles and direction cone aspects.

    // Generate a new aspects list containg the direction cones whcih
    // correspond to the angle base cluster.
    SimpleList<Entity> owners;
    angle->get_owners (&owners);
    List<Aspect> aspects;
    Entity *e = owners.first ();
    while (e) {
      aspects.add_first (static_cast<Aspect*> (e));
      e = owners.next ();
    }
    if (aspects.length () == 0)
      return NULL;

    // Cluster the direction cones.
    clusters = new List<Cluster>;
    cluster_aspects (&aspects, tol_eq, tol_angle_dist, NAN, true, clusters);
    if (clusters->length () == 0) {
      delete clusters;
      return NULL;
    }

    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      cn->data () -> get_all_base_clusters (&bcls);
      cn = cn->next ();
    }

  }

  // Find regular clusters.
  ListNode<Cluster> *cn = bcls.first ();
  while (cn) {

    Cluster *c = cn->data ();

    if (!static_cast<DirectionCone*> (c->value ()) -> regular ()) {
      // Only check those that we did not already identify as regular.
      
      // Count number of different directions in the cone.
      SimpleList<Entity> owners;
      c->get_owners (&owners);

      // If we have n different directions, then n(n-1)(n-2)/3! direction
      // cones would have to be present if we have a transitive structure
      // in the cluster. If the cluster is not transitive, it is not
      // regular.
      unsigned long n = owners.length ();
      n = (n * (n - 1) * (n - 2)) / 6;

      // NOTE: If the cluster is not regular, we could still check if
      //       there is a regular subset, but this means we'd have to find
      //       a complete hypergraph between the directions present in the
      //       cluster (i.e. a subset of directions for which all cones are
      //       present). At the moment we don't do this, but collapse
      //       sub-clusters later if they have the same owners (directions)
      //       than the cluster itself. This has a similar effect.
      
      if (c->num_aspects () == n) {
	// Mark as regular, so that we do not check again.
	static_cast<DirectionCone*> (c->value ()) -> regular () = true;
	// Cluster is regular.
	if (c->parent_cluster ())
	  bcls.add_last_check (c->parent_cluster ());
      } else {
	// Cluster is not regular, collapse it (i.e. move everyting to
	// parent if possible). We do not remove top-level clusters as
	// they are used for clustering on the next angle tree level.
	if (c->collapse ()) {
	  // Cluster collpase, so remove it and add it's parent to the
	  // list.
 	  bcls.add_last_check (c->parent_cluster ());
	  delete c;
	} else {
	  // Keep the top level clusters for the different angle
	  // levels for now and sort it out when doing the least
	  // squares match.
	  static_cast<DirectionCone*> (c->value ()) -> regular () = true;
	}
      }

    }

    cn = cn->next ();
  }

  return clusters;
}

UnitVector3 Analyse::solve_cone (UnitVector3 const &d1, 
				 UnitVector3 const &d2, 
				 UnitVector3 const &d3) const
{
  // Pivot-Gauss algorithm to solve [d1^t;d2^t;d3^t]x = [1;...;1]

  // Initialise equation system.
  Real *R1 = new Real[3];
  Real *R2 = new Real[3];
  Real *R3 = new Real[3];
  Real B2, B3;
  *R1 = d1.x ();
  R1[1] = d1.y ();
  R1[2] = d1.z ();
  *R2 = d2.x ();
  R2[1] = d2.y ();
  R2[2] = d2.z ();
  *R3 = d3.x ();
  R3[1] = d3.y ();
  R3[2] = d3.z ();
  B2 = 1.0;
  B3 = 1.0;

  // First pivot row.
  if (fabs (*R1) < fabs (*R2)) {
    Real *HR = R2;
    R2 = R1;
    R1 = HR;
  }
  if (fabs (*R1) < fabs (*R3)) {
    Real *HR = R3;
    R3 = R1;
    R1 = HR;
  }
  if (fabs (*R1) < sil_get_epsilon ()) {
    // System is not solveable.
    delete R1;
    delete R2;
    delete R3;
    return UnitVector3 (0.0,0.0,0.0);
  }

  // First Gauss reduction.
  Real f = *R2 / *R1;
  R2[1] -=  R1[1] * f;
  R2[2] -=  R1[2] * f;
  B2    -= f;
  f = *R3 / *R1;
  R3[1] -= R1[1] * f;
  R3[2] -= R1[2] * f;
  B3    -= f;

  // Second pivot row.
  if (fabs (R2[1]) < fabs (R3[1])) {
    Real *HR = R3;
    R3 = R2;
    R2 = HR;
    Real HB = B3;
    B3 = B2;
    B2 = HB;
  }
  if (fabs (R2[1]) < sil_get_epsilon ()) {
    // System is not solveable.
    delete R1;
    delete R2;
    delete R3;
    return UnitVector3 (0.0,0.0,0.0);
  }

  // Second Gauss reduction.
  f = R3[1] / R2[1];
  R3[2] -= R2[2] * f;
  if (fabs (R3[2]) < sil_get_epsilon ()) {
    // System is not solveable.
    delete R1;
    delete R2;
    delete R3;
    return UnitVector3 (0.0,0.0,0.0);
  }
  B3 -= B2 * f;

  // Solve reduced system.
  Real v3 = B3 / R3[2];
  Real v2 = (B2 - R2[2] * v3) / R2[1];
  Real v1 = (1.0 - R1[1] * v2 - R1[2] * v3) / *R1;

  delete R1;
  delete R2;
  delete R3;
  return UnitVector3 (v1, v2, v3);
}

Cluster *Analyse::find_parallel (UnitVector3 const &n,
				 List<Cluster> *directions,
				 Real &tol) const
{
  // Find a direction in the given direction cluster (defaults to the
  // parallel directions found) which is approximately parallel to n.
  // The tolerance for the parallel relation is reported in tol. If no
  // parallel direction is found, tol is undefined.

  // Search parallel_directions by default.
  if (!directions) {
    directions = parallel_directions;
    if (!directions)
      return NULL;
  }

  // Get base direciton clusters.
  List<Cluster> dcl;
  ListNode<Cluster> *an1 = directions->first ();
  while (an1) {
    an1->data () -> get_all_base_clusters (&dcl);
    an1 = an1->next ();
  }

  // Check all directions to find the closest parallel direction.
  Cluster *direction = NULL;
  ListNode<Cluster> *an = dcl.first ();
  while (an) {
    Real t = fabs (*static_cast<Direction *> (an->data () -> value ()) % n);
    if (t >= 1.0)
      t = 0.0;
    else
      t = acos (t);
    if (t < tol) {
      tol = t;
      direction = an->data ();
    }
    an = an->next ();
  }
  return direction;
}

Cluster *Analyse::find_orthogonal (UnitVector3 const &n,
				   List<Cluster> *directions,
				   Real &tol) const
{
  // Find a direction in the given direction cluster (defaults to the
  // parallel directions found) which is approximately orthogonal to
  // n.  The tolerance for the orthogonal relation is reported in
  // tol. If no parallel direction is found, tol is undefined.

  // Search parallel_directions by default.
  if (!directions) {
    directions = parallel_directions;
    if (!directions)
      return NULL;
  }

  // Get base direciton clusters.
  List<Cluster> dcl;
  ListNode<Cluster> *an1 = directions->first ();
  while (an1) {
    an1->data () -> get_all_base_clusters (&dcl);
    an1 = an1->next ();
  }

  // Check all directions to find the closest parallel direction.
  Cluster *direction = NULL;
  ListNode<Cluster> *an = dcl.first ();
  while (an) {
    Real t = fabs (*static_cast<Direction *> (an->data () -> value ()) % n);
    if (t >= 1.0)
      t = PI / 2.0;
    else
      t = fabs (acos (t) - PI / 2.0);
    if (t < tol) {
      tol = t;
      direction = an->data ();
    }
    an = an->next ();
  }
  return direction;
}

Cluster *Analyse::find_parallel_orthogonal (UnitVector3 const &n,
					    UnitVector3 const *p,
					    List<Cluster> *directions,
					    Real &tol, bool &par) const
{
  // Find a direction in directions (default to the parallel direction
  // clusters) which is parallel or orthogonal to n, where parallel
  // directions are preferd to orthogonal directions. If p is given,
  // than an orthogonal direction has to be orthogonal to p as well
  // (to avoid detecting things twice for direction planes, etc.). The
  // tolerance for the relation is reported in tol and par indicates
  // if it's parallel (true) or orthogonal (false). If no direction is
  // found, tol and par are undefined.

  // Search parallel_directions by default.
  if (!directions) {
    directions = parallel_directions;
    if (!directions)
      return NULL;
  }

  // Get base direciton clusters.
  List<Cluster> dcl;
  ListNode<Cluster> *an1 = directions->first ();
  while (an1) {
    an1->data () -> get_all_base_clusters (&dcl);
    an1 = an1->next ();
  }

  // Find the closest parallel or orthogonal direction.
  Cluster *direction = NULL;
  Cluster *odirection = NULL;
  Real otol = tol;
  ListNode<Cluster> *ax = dcl.first ();
  while (ax) {
    // First check if it is parallel.
    Real t = fabs (*static_cast<Direction *> (ax->data () -> value ()) % n);
    if (t >= 1.0)
      t = 0.0;
    else
      t = acos (t);
    if (t < tol) {
      tol = t;
      direction = ax->data ();
    } else {
      // If not parallel, check if it is orthogonal.
      t = fabs (t - PI / 2.0);
      if (t < otol) {
	if (p) {
	  // Check if orthogonal direction is also orthogonal to p
	  // considering the tolerance of the other relation.
	  Real tt = fabs (*p % *static_cast<Direction *> (ax->data () -> 
							  value ()));
	  if (tt >= 1.0)
	    tt = 0.0;
	  else
	    tt = acos (tt);
	  if (fabs (tt - PI / 2.0) < t + tol_angle_dist) {
	    otol = t;
	    odirection = ax->data ();
	  }
	} else {
	  otol = t;
	  odirection = ax->data ();
	}
      }
    }
    ax = ax->next ();
  }
    
  // If there is no parallel direction, use the orthogonal direction.
  if (!direction) {
    if (odirection) {
      direction = odirection;
      tol = otol;
      par = false;
    }
  } else {
    par = true;
  }
  
  return direction;
}

void Analyse::add_extra_directions (List<Cluster> *extra_dirs)
{
  assert (extra_dirs);
  // Check if there are directions parallel or orthogonal to the extra
  // directions and add them to the direction list.
  ListNode<Cluster> *dir = extra_dirs->unlink_first ();
  while (dir) {
    // Find a parallel or orthogonal directions in ordinary an in extra
    // directions.
    Real t = PI;
    Cluster *pdir = find_parallel (*static_cast<Direction *> (dir->data () ->
							      value ()),
				   parallel_directions, t);
    if (pdir) {
      dir->data () -> add_sub_regularity (new ParallelDirection (this,
								 t, pdir));
      // Make sure we only find orthogonal directions which are better
      // or about as good with respect to the tolerance as the
      // parallel direction.
      t += tol_angle_dist;
    } else {
      // No parallel direction, so any orhtogonal relation will do.
      t = PI;
    }
    pdir = find_orthogonal (*static_cast<Direction *> (dir->data () ->
						       value ()),
			    extra_dirs, t);
    if (pdir)
      dir->data () -> add_sub_regularity (new OrthogonalDirection (this,
								   t, pdir));
    // Add it as another direction to the direction cluster list. It's
    // not added to the hierarchy to avoid dependency problems.
    parallel_directions->add_last (dir->data ());
    delete dir;
    dir = extra_dirs->unlink_first ();
  }
  // NOTE: We do not check for special angles (besides orthogonal and
  //       parallel) between extra directions from the direction planes
  //       and cones and between these extra directions and the other
  //       ordinary directions. In principle this could be done here
  //       by just checking for special values for all angles between
  //       them (extra directions which are parallel to an ordinary
  //       direction do not have to be checked as their parallel direction
  //       is already checked). Actually planar and conical angle 
  //       regularities could also be considered. However, this appears
  //       to generate regularities which are not very useful in general,
  //       so it is left for later consideration.
  return;
}

Cluster *Analyse::add_extra_parallel_direction (Direction *d)
{
  assert (d);
  Cluster *c = new Cluster (this, d);
  if (!extra_parallel_directions)
    extra_parallel_directions = new List<Cluster>;
  extra_parallel_directions->add_first (c);
  return c;
}

Cluster *Analyse::find_direction_pair (Direction *d1, Direction *d2,
				       List<Cluster> *directions)
{
  // Search parallel_directions by default.
  if (!directions) {
    directions = parallel_directions;
    if (!directions)
      return NULL;
  }

  // Get base direciton clusters.
  List<Cluster> dcl;
  ListNode<Cluster> *an1 = directions->first ();
  while (an1) {
    an1->data () -> get_all_base_clusters (&dcl);
    an1 = an1->next ();
  }

  // Find direction clusters.
  Cluster *c1 = NULL;
  Cluster *c2 = NULL;
  ListNode<Cluster> *c = dcl.first ();
  while (c && !(c1 && c2)) {
    if (c->data () -> aspects ()) {
      Aspect *a = c->data () -> aspects () -> first ();
      while (a) {
	if (!c1 && a->owners () . is_member (d1))
	  c1 = c->data ();
	if (!c2 && a->owners () . is_member (d2))
	  c2 = c->data ();
	a = c->data () -> aspects () -> next ();
      }
    }
    c = c->next ();
  }
  if (c1 && c2) {
    do {
      if (c1 == c2)
	return c1;
      if (c1->tolerance () < c2->tolerance ())
	c1 = c1->parent_cluster ();
      else
	c2 = c2->parent_cluster ();
    } while (c1 && c2);
  }
  return NULL;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
