// analyse_geom.cc: Analyse a SiL model / geometry types
//
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_geom.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_geom,
       "$Id: analyse_geom.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "kernel/lapack.h"
#include "kernel/kernel.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"

void Analyse::eval_geometry ()  
{
  List<Regularity> *altg = new List<Regularity>;
  
  // Check geometry types of faces.
  sil_progress_bar_text ("Checking " +
			 as_string (sld->faces () . length ()) + " surfaces");
  
  // NOTE: We use the curvature as main measure for a surface geometry
  //       to be different. However, we also have to check if the loops
  //       fit. The travesal of the faces does exactly this, but for
  //       a limited number of cases.

  // Use average length to get maximum curvature.
  Real k_max = max (1.0 / avg_len, sil_get_epsilon ());
  if (isnan (k_max))
    k_max = sil_get_epsilon ();

  ListNode<Face> *fn = sld->faces () . first ();
  while (fn) {
    if (fn->data () -> loops () . length () > 0) {
      Surface *s = fn->data () -> surface ();
      if (s) {
	
	// Check loops.
	SimpleList<Position> pos;
	SimpleList<Circle> periodic_circ;
	SimpleList<Curve> periodic;
	XListNode<Loop> *ln = fn->data () -> loops () . first ();
	while (ln) {
	  XListNode<Edge> *en = ln->data () -> edges () . first ();
	  while(en) {
	    Edge *e = en ->data ();
	    if (!(e->end_vertex () || e->start_vertex ())) {
	      if (e->curve () && e->curve () -> is (SIL_CIRCLE_TYPE))
		periodic_circ.add_first (static_cast<Circle*> (e->curve ()));
	      else
		periodic.add_first (e->curve ());
	    } else {
	      if (e->end_vertex () && e->end_vertex () -> point ())
		pos.add_first_check (e->end_vertex () -> point () -> 
				     position ());
	      if (e->start_vertex () && e->start_vertex () -> point ())
		pos.add_first_check (e->start_vertex () -> point () ->
				     position ());
	    }
	    en = en->next ();
	  }
	  ln = ln->next ();
	}
	
	if (s->is (SIL_SPHERE_TYPE)) {

	  Real k = 1.0 / static_cast<Sphere*> (s) -> radius () -> 
	    value ();

	  if (pos.length () == 0) {
	    // We do not check the curvature for this type.
	    if (periodic_circ.length () == 2 && periodic.length () == 0) {
	      // Sphere could be a cylinder.
	      Real tol = NAN;
	      Real dp = NAN;
	      Geometry *p = fit_cylinder_circ (&periodic_circ, 
					       static_cast<Sphere*> (s) -> 
					       convex (), tol, dp);
	      if (p) {
		// Error between cylinder and circle at max distance.
		Real r = static_cast<Sphere*> (s) -> radius () -> value ();
		Real R = r - sqrt (max (r * r - dp * dp / 4.0, 0.0));
		if (fabs (R - tol) < tol_len_dist)
		  altg->add_first (new AltGeometry (this, k, fn->data (),
						    p));
		else
		  delete p;
	      }
	    }
	  } else if (k < k_max &&
		     periodic.length () + periodic_circ.length () == 0 &&
		     pos.length () > 2) {
	    // Sphere could be a plane.
	    Real tol = NAN;
	    Real dist = NAN;
	    Geometry *p = fit_plane_pos (&pos, tol, dist);
	    if (p) {
	      // Error between plane and circle at max distance.
	      Real r = static_cast<Sphere*> (s) -> radius () -> value ();
	      Real R = r - sqrt (max (r * r - dist * dist / 4.0, 0.0));
	      if (fabs (R - tol) < tol_len_dist)
		altg->add_first (new AltGeometry (this, k, fn->data (), p));
	      else
		delete p;
	    }
	  }

	} else if (s->is (SIL_CYLINDER_TYPE)) {

	  Real k = 1.0 / static_cast<Cylinder*> (s) -> radius () ->
	    value ();

	  if (k < k_max &&
	      periodic.length () + periodic_circ.length () == 0 &&
	      pos.length () > 2) {
	    // Cylinder could be a plane.
	    Real tol = NAN;
	    Real dist = NAN;
	    Geometry *p = fit_plane_pos (&pos, tol, dist);
	    if (p) {
	      // Error between plane and circle at max distance.
	      Real r = static_cast<Cylinder*> (s) -> radius () -> value ();
	      Real R = r - sqrt (max (r * r - dist * dist / 4.0, 0.0));
	      if (fabs (R - tol) < tol_len_dist)
		altg->add_first (new AltGeometry (this, k, fn->data (), p));
	      else
		delete p;
	    }
	  }

	} else if (s->is (SIL_TORUS_TYPE)) {

	  Torus *T = static_cast<Torus*> (s);
	  Real k = 1.0 / T-> major_radius () -> value ();
	  if (k < k_max) {

	    if (T->major_radius () > T->minor_radius ()) {
	      
	      if (pos.length () == 0) {
		if (periodic_circ.length () == 2 && periodic.length () == 0) {
		  // Torus could be a cylinder.
		  Real tol = NAN;
		  Real dp = NAN;
		  Geometry *p = fit_cylinder_circ (&periodic_circ, 
						   static_cast<Torus*> (s) ->
						   convex (), tol, dp);
		  if (p) {
		    // Error between cylinder and circle at max distance.
		    Real r = T->major_radius () -> value ();
		    Real R = r - sqrt (max (r * r - dp * dp / 4.0, 0.0));
		    if (fabs (R - tol) < tol_len_dist)
		      altg->add_first (new AltGeometry (this, k, fn->data (),
							p));
		    else
		      delete p;
		  }
		}
	      } else if (periodic.length () + periodic_circ.length () == 0 &&
			 pos.length () > 2) {
		// Torus could be a plane.
		Real tol = NAN;
		Real dist = NAN;
		Geometry *p = fit_plane_pos (&pos, tol, dist);
		if (p) {
		  // Error between plane and circle at max distance.
		  Real r = static_cast<Torus*> (s) -> major_radius () ->
		    value ();
		  Real R = r - sqrt (max (r * r - dist * dist / 4.0, 0.0));
		  if (fabs (R - tol) < tol_len_dist) {
		    Real k = 1.0 / T->major_radius () -> value ();
		    altg->add_first (new AltGeometry (this, k, fn->data (),
						      p));
		  } else {
		    delete p;
		  }
		}
	      }

	    }
	  }

	} else if (s->is (SIL_CONE_TYPE)) {

	  if ((periodic_circ.length () + periodic.length () == 2 &&
	       pos.length () == 0) ||
	      (periodic_circ.length () + periodic.length () == 0 &&
	       pos.length () > 2)) {

	    // Find points most distant and closest to the apex.
	    Real d_close = NAN;
	    Real d_far = 0.0;
	    XListNode<Loop> *ln = fn->data () -> loops () . first ();
	    while (ln) {
	      XListNode<Edge> *en = ln->data () -> edges () . first ();
	      while(en) {
		Edge *e = en ->data ();
		if (e->start_vertex () && e->start_vertex () -> point ()) {
		  Real d = (*e->start_vertex () -> point () -> position () -
			    *static_cast<Cone*> (s) -> root_point ()) .
		    norm ();
		  if (isnan (d_close) || d < d_close)
		    d_close = d;
		  if (d > d_far)
		    d_far = d;
		}
		if (e->end_vertex () && e->end_vertex () -> point ()) {
		  Real d = (*e->end_vertex () -> point () -> position () -
			    *(static_cast<Cone*> (s) -> root_point ())) .
		    norm ();
		  if (isnan (d_close) || d < d_close)
		    d_close = d;
		  if (d > d_far)
		    d_far = d;
		}
		if (e->curve ()) {
		  if (e->curve () -> is (SIL_CIRCLE_TYPE)) {
		    Real d = (*(static_cast<Cone*> (s) -> root_point ()) -
			      *(static_cast<Circle*> (e->curve ()) 
				-> centre ())) . norm () /
		      cos (static_cast<Cone*> (s) -> angle () -> value ());
		    if (isnan (d_close) || d < d_close)
		      d_close = d;
		    if (d > d_far)
		      d_far = d;
		  } else if (e->curve () -> is (SIL_ELLIPSE_TYPE)) {
		    Ellipse *E = static_cast<Ellipse*> (e->curve ());
		    Vector3 p = *E->centre () + E->major_radius () ->
		      value () * (*E->major_direction ());
		    if (E->is_inside (p, e->start_vertex (), 
				      e->end_vertex ())) {
		      Real d =  (*(static_cast<Cone*> (s) -> root_point ()) - 
				 p) . norm ();
		      if (isnan (d_close) || d < d_close)
			d_close = d;
		      if (d > d_far)
			d_far = d;
		    }
		    p = *E->centre () - E->major_radius () -> value () *
		      *E->major_direction ();
		    if (E->is_inside (p, e->start_vertex (),
				      e->end_vertex ())) {
		      Real d =  (*(static_cast<Cone*> (s) -> root_point ()) - 
				 p) . norm ();
		      if (isnan (d_close) || d < d_close)
			d_close = d;
		      if (d > d_far)
			d_far = d;
		    }
		    UnitVector3 D = (*E->major_direction () * *E->normal ());
		    p = *E->centre () + E->minor_radius () -> value () * D;
		    if (E->is_inside (p, e->start_vertex (),
				      e->end_vertex ())) {
		      Real d =  (*(static_cast<Cone*> (s) -> root_point ()) - 
				 p) . norm ();
		      if (isnan (d_close) || d < d_close)
			d_close = d;
		      if (d > d_far)
			d_far = d;
		    }
		    p = *E->centre () - E->minor_radius () -> value () * D;
		    if (E->is_inside (p, e->start_vertex (),
				      e->end_vertex ())) {
		      Real d =  (*(static_cast<Cone*> (s) -> root_point ()) - 
				 p) . norm ();
		      if (isnan (d_close) || d < d_close)
			d_close = d;
		      if (d > d_far)
			d_far = d;
		    }
		  } else if (e->curve () -> is (SIL_SPLINECURVE_TYPE)) {
		    Vector3 *p = static_cast<SplineCurve*> (e->curve ()) ->
		      ctrlpts () -> first ();
		    while (p) {
		      Real d = fabs (((*(static_cast<Cone*> (s) -> 
					 root_point ()) - *p) %
				      *(static_cast<Cone*> (s) ->
					direction ()))) /
			cos (static_cast<Cone*> (s) -> angle () -> value ());
		      if (isnan (d_close) || d < d_close)
			d_close = d;
		      if (d > d_far)
			d_far = d;
		      p = static_cast<SplineCurve*> (e->curve ()) ->
			ctrlpts () -> next ();
		    }
		  }
		}
		en = en->next ();
	      }
	      ln = ln ->next ();
	    }

	    // Curvature range
	    if (d_far > sil_get_epsilon () && !isnan (d_far) &&
		d_close > sil_get_epsilon () && !isnan (d_close)) {
	      Real h = sin (static_cast<Cone*> (s) -> angle () -> value ());
	      Real kmin = d_far * h;
	      Real kmax = d_close * h;
	      h = sqrt (1 - h * h);
	      kmin = h / kmin;
	      kmax = h / kmax;

	      if (pos.length () == 0) {
		if (periodic_circ.length () == 2 && periodic.length () == 0) {
		  Real k = kmax - kmin;
		  if (k < k_max) {
		    // Cone could be a cylinder.
		    Real tol = NAN;
		    Real dp = NAN;
		    Geometry *p = fit_cylinder_circ (&periodic_circ,
						     static_cast<Cone*> (s) ->
						     convex (), tol, dp);
		    if (p) {
		      // Error between cylinder and cone.
		      Real a = atan (tol / dp);
		      if (a < tol_angle_dist)
			altg->add_first (new AltGeometry (this, k,
							  fn->data (), p));
		      else
			delete p;
		    }
		  }
		}
	      } else if (periodic_circ.length () +
			 periodic.length () == 0 &&
			 pos.length () > 2) {
		Real k = (kmin + kmax) / 2.0;
		if (k < k_max) {
		    // Cone could be a plane.
		  Real tol = NAN;
		  Real dist = NAN;
		  Geometry *p = fit_plane_pos (&pos, tol, dist);
		  if (p) {
		    Real d = tan (static_cast<Cone*> (s) -> angle () -> 
				  value ()) * dist;
		    if (fabs (d - tol) < tol_len_dist)
		      altg->add_first (new AltGeometry (this, k,
							fn->data (), p));
		    else
		      delete p;
		  }
		}
	      }

	    }
	  }

	}
      }
    }
    fn = fn->next ();
  }

  if (altg->length () > 0) {
    reg_list->add_first (altg);
  } else {
    delete altg;
  }
  
  // NOTE: We might also want to check if there are gaps in the model
  //       which can be closed by adding faces or combining edges.
  
  return;
}

Geometry *Analyse::fit_plane_pos (SimpleList<Position> *pos, Real &tol,
				  Real &dist)
{
  // We have the equation system p_i^t n = k, i.e. P [n/k] = e, which
  // can be solved in a least square sense with SVD: [n/k] = VS^*U^t e

  char JOBU = 'S'; // only compute the left singular vectors
  char JOBVT = 'O'; // only compute the right singular vectors in A
  long M = pos->length (); // number of rows
  long N = 3; // number of columns
  Real A[M * N]; // direction matrix
  Real U[M * 3]; // U matrix (only the first three vectors)
  Real S[3]; // singular values
  long LWORK = 50 + M; // work space
  Real WORK[LWORK];
  long INFO = 0; // result
	
  // Initialize A
  Position *p = pos->first ();
  long l1 = 0;
  long l2 = M;
  long l3 = l2 + M;
  while (p) {
    A[l1++] = p->x ();
    A[l2++] = p->y ();
    A[l3++] = p->z ();
    p = pos->next ();
  }
	
  // SVD
  dgesvd (&JOBU, &JOBVT, &M, &N, A, &M, S, U, &M, NULL,
	  &N, WORK, &LWORK, &INFO);
	
  if (INFO == 0) {
	  
    // Compute x = V S^* U^T * e
    
    // Sum up the first three row of U^t (columns of U)
    Real x1 = 0.0;
    Real x2 = 0.0;
    Real x3 = 0.0;
    l1 = 0;
    l2 = M;
    l3 = l2 + M;
    while (l1 < M) {
      x1 += U[l1++];
      x2 += U[l2++];
      x3 += U[l3++];
    }
	  
    // Multiply x with S^*
    if (fabs (S[0]) > sil_get_epsilon ())
      x1 /= S[0];
    else
      x1 = 0.0;
    if (fabs (S[1]) > sil_get_epsilon ())
      x2 /= S[1];
    else
      x2 = 0.0;
    if (fabs (S[2]) > sil_get_epsilon ())
      x3 /= S[2];
    else
      x3 = 0.0;
    
    // Multiply x with V
    l1 = 2 * M;
    Real xx1 = A[0]  * x1 + A[1]      * x2 + A[2]      * x3;
    Real xx2 = A[M]  * x1 + A[M + 1]  * x2 + A[M + 2]  * x3;
    Real xx3 = A[l1] * x1 + A[l1 + 1] * x2 + A[l1 + 2] * x3;
    
    // Create plane.
    Plane *plane = new Plane ();
    plane->direction () -> set (xx1, xx2, xx3);
    Real k = sqrt (xx1 * xx1 + xx2 * xx2 + xx3 * xx3);
    Vector3 rp = k * (*plane->direction ());
    plane->root_point () -> set (rp);

    // Tolerance and max distance between positions.
    SimpleList<Position> pos2;
    dist = 0.0;
    tol = 0.0;
    p = pos->first ();
    while (p) {
      Real d = fabs ((*p % (*plane->direction ())) - k);
      if (d > tol)
	d = tol;
      Position *p2 = pos2.first ();
      while (p2) {
	Real d = (*p2 - *p) . norm ();
	if (d > dist)
	  dist = d;
	p2 = pos2.next ();
      }
      pos2.add_first (p);
      p = pos->next ();
    }

    return plane;
  }

  return NULL;
}

Geometry *Analyse::fit_cylinder_circ (SimpleList<Circle> *circs, bool conv,
				      Real &tol, Real &dp)
{
  Circle *c = circs->first ();
  Real r1 = c->radius () -> value ();
  Position p1 (Vector3 (c->centre () -> x (), c->centre () -> y(),
			c->centre () -> z ()));
  Direction *d1 = c->normal ();
  c = circs->next ();
  Real r2 = c->radius () -> value ();
  Position p2 (Vector3 (c->centre () -> x (), c->centre () -> y(),
			c->centre () -> z ()));
  Direction *d2 = c->normal ();

  // Check if d1 and d2 are parallel at some tolerance level.
  Cluster *pair = find_direction_pair (d1, d2, NULL);
  if (pair) {
    Real atol = pair->tolerance () + tol_angle_dist;
    Direction *d = static_cast<Direction*> (pair->value ());
    Vector3 dpos = (p1 - p2);
    UnitVector3 dirp = dpos . normalise ();
    Real a = fabs (*d % dirp);
    if (a >= 1.0)
      a = 0.0;
    else
      a = acos (a);
    if (a < atol) {
      Real dr = fabs (r1 - r2);
      dp = dpos % (*d);
      if (dp > dr && dp > sil_get_epsilon ()) {
	Real a = atan (dr / dp);
	if (a < atol) {
	  // Cylinder
	  Cylinder *c = new Cylinder ();
	  c->radius () -> value () = (r1 + r2) / 2.0;
	  c->direction () -> set (*d);
	  c->root_point () -> set ((p1 + p2) / 2.0);
	  c->convex () = conv;
	  tol = max (tan (pair->tolerance ()) * dp / 2.0,
		     max (fabs (c->radius () -> value () - r1),
			  fabs (c->radius () -> value () - r2)));
	  return c;
	}
      }
    }
  }
  return NULL;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
