// analyse_positions.cc: Analyse a SiL model / positions.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_positions.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_positions,
       "$Id: analyse_positions.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"
#include "kernel/geometry.hxx"

void Analyse::eval_positions ()
{
  // Intersection positions are also positions and are considered as
  // position for all regularities considered here.
  if (intersection_positions) {
    ListNode<Cluster> *rn = intersection_positions->first ();
    while (rn) {
      position_aspects.add_first 
	(new Aspect 
	 (new Position (*static_cast<Vector3*> (static_cast<Position*> 
						(rn->data () -> value ()))),
	  rn->data ()));
      rn = rn->next ();
    }
  }

  if (position_aspects.length () < 1)
    return;

  // Cluster positions.
  sil_progress_bar_text ("Clustering " + 
			 as_string (position_aspects.length ()) +
			 " positions");
  List<Cluster> *clusters = new List<Cluster>;
  cluster_aspects (&position_aspects, tol_eq, tol_len_dist, NAN, true,
		   clusters);
  reg_list->add_first ((List<Regularity> *)clusters);

  // Find tolerances.
  SimpleList<Real> tols;
  tols.add_first (new Real (tol_len_dist));
  ListNode<Cluster> *cn = clusters->first ();
  while (cn) {
    if (cn->data () -> tolerance () > tol_len_dist)
      tols.add_first (new Real (cn->data () -> tolerance ()));
    cn = cn->next ();
  }

  tolerance_levels (clusters, &tols, tol_len_dist, max_len);

  if (f_positions_partial) {

    sil_progress_bar_text ("Checking for partial positions");
    
    // Check for partially equal positions.

    List<Cluster> dirs;

    // Collect planes and lines for projection.
    if (conical_angle_regularities) {
      // Check the orthogonal systems.
      ListNode<ConicalAngleRegular> *car = conical_angle_regularities->
	first ();
      while (car) {
	if (car->data () -> orthogonal ()) {
	  // Orthogonal system provides three planes and three lines.
	  Cluster *c= car->data () -> directions () -> first ();
	  while (c) {
	    dirs.add_last_check (c);
	    c = car->data () -> directions () -> next ();
	  }
	}
	car = car->next ();
      }
    }
    if (parallel_directions) {
      // Check for major axes. A major axis should have more than 1/3 of
      // the total faces as owners.
      List<Cluster> bcl;
      ListNode<Cluster> *c = parallel_directions->first ();
      while (c) {
	c->data () -> get_all_base_clusters (&bcl);
	c = c->next ();
      }
      c = bcl.first ();
      while (c) {
	if (!dirs.is_member (c->data ())) {
	  // Count face owners.
	  SimpleList<Entity> owners;
	  c->data () -> get_owners (&owners);
	  unsigned long n = 0;
	  Entity *e = owners.first ();
	  while (e) {
	    if (e->is (SIL_FACE_TYPE))
	      ++n;
	    e = owners.next ();
	  }
	  if (3 * n > static_cast<Real> (sld->faces () . length ()))
	    dirs.add_first (c->data ());
	}
	c = c->next ();
      }
    }

    // Remove directions for dirs whose parents are also in dirs.
    ListNode<Cluster> *c = dirs.first ();
    while (c) {
      ListNode<Cluster> *cn = c->next ();
      if (c->data () -> parent_cluster () &&
	  dirs.is_member (c->data () -> parent_cluster ())) {
	c->unlink ();
	delete c;
      }
      c = cn;
    }

    List<Cluster> *pep = new List<Cluster>;
    List<Cluster> *lep = new List<Cluster>;

    // Each direction in dirs represents a plane and a line to porject
    // the points on and check if they are equal.
    c = dirs.first ();
    while (c) {
      check_plane_partially_equal (c->data (), clusters, pep);
      check_line_partially_equal (c->data (), clusters, lep);
      c = c->next ();
    }
    
    if (pep->length () > 0)
      reg_list->add_first ((List<Regularity> *)pep);
    else
      delete pep;
    if (lep->length () > 0)
      reg_list->add_first ((List<Regularity> *)lep);
    else
      delete lep;
  }

  // NOTE: Very similarly to the axis grid and circle check, we could cehck
  //       if the position clusters are on one of the following structures:
  //       lines, 2D grids, 3D grids (with orthogonal/parallle directions
  //       and regular distances) and circles (with angle-regular arrangements)

  return;
}

void Analyse::check_plane_partially_equal (Cluster *dir, 
					   List<Cluster> *pts,
					   List<Cluster> *regs)
{
  // Project the point clusters on plane and create PartiallyPlaneEqual
  // Features for the clusters.

  // Get the points.
  List<Cluster> bcl;
  ListNode<Cluster> *c = pts->first ();
  while (c) {
    c->data () -> get_all_base_clusters (&bcl);
    c = c->next ();
  }

  sil_progress_bar_text ("Clustering " + as_string (bcl.length ()) +
			 " 2d-partial positions for " +
			 as_string (*static_cast<Direction *> (dir->
							       value ())));

  // Project them on the plane and create the aspects for clustering.
  List<Aspect> aspects;
  c = bcl.first ();
  while (c) {
    UnitVector3 d = *static_cast<Direction*> (dir->value ());
    Vector3 p = *static_cast<AlignedAxes*> (c->data () -> value ());
    p = p - (p % d) * d;
    aspects.add_first (new Aspect (new PartiallyPlaneEqual (p, dir),
				   c->data ()));
    c = c->next ();
  }

  // Cluster the PartiallyPlaneEqual aspects.
  List<Cluster> pep;
  cluster_aspects (&aspects, tol_eq, tol_len_dist, NAN, true, &pep);

  // Find tolerances.
  SimpleList<Real> tols;
  tols.add_first (new Real (tol_len_dist));
  c = pep.first ();
  while (c) {
    if (c->data () -> tolerance () > tol_len_dist)
      tols.add_first (new Real (c->data () -> tolerance ()));
    c = c->next ();
  }

  tolerance_levels (&pep, &tols, tol_len_dist, max_len);

  // Delete clusters with a single aspect.
  bcl.clear ();
  c = pep.first ();
  while (c) {
    bcl.add_first (c->data ());
    c = c->next ();
  }
  c = bcl.unlink_first ();
  while (c) {
    if (c->data () -> num_aspects () < 2) {
      if (!c->data () -> collapse ())
	pep.del (c->data ());
      delete c->data ();
    } else if (c->data () -> sub_clusters ()) {
      Cluster *sc = c->data () -> sub_clusters () -> first ();
      while (sc) {
	bcl.add_first (sc);
	sc = c->data () -> sub_clusters () -> next ();
      }
    }
    delete c;
    c = bcl.unlink_first ();
  }

  regs->combine (&pep);

  return;
}

void Analyse::check_line_partially_equal (Cluster *dir, 
					  List<Cluster> *pts,
					  List<Cluster> *regs)
{
  // Project the point clusters on line and create PartiallyLineEqual
  // Features for the clusters.

  // Get the points.
  List<Cluster> bcl;
  ListNode<Cluster> *c = pts->first ();
  while (c) {
    c->data () -> get_all_base_clusters (&bcl);
    c = c->next ();
  }

  sil_progress_bar_text ("Clustering " + as_string (bcl.length ()) +
			 " 1d-partial positions for " +
			 as_string (*static_cast<Direction *> (dir->
							       value ())));

  // Project them on the line and create the aspects for clustering.
  List<Aspect> aspects;
  c = bcl.first ();
  while (c) {
    UnitVector3 d = *static_cast<Direction*> (dir->value ());
    Vector3 p = *static_cast<AlignedAxes*> (c->data () -> value ());
    p = (p % d) * d;
    aspects.add_first (new Aspect (new PartiallyLineEqual (p, dir),
				   c->data ()));
    c = c->next ();
  }

  // Cluster the PartiallyLineEqual aspects.
  List<Cluster> lep;
  cluster_aspects (&aspects, tol_eq, tol_len_dist, NAN, true, &lep);

  // Find tolerances.
  SimpleList<Real> tols;
  tols.add_first (new Real (tol_len_dist));
  c = lep.first ();
  while (c) {
    if (c->data () -> tolerance () > tol_len_dist)
      tols.add_first (new Real (c->data () -> tolerance ()));
    c = c->next ();
  }

  tolerance_levels (&lep, &tols, tol_len_dist, max_len);

  // Delete clusters with a single aspect.
  bcl.clear ();
  c = lep.first ();
  while (c) {
    bcl.add_first (c->data ());
    c = c->next ();
  }
  c = bcl.unlink_first ();
  while (c) {
    if (c->data () -> num_aspects () < 2) {
      if (!c->data () -> collapse ())
	lep.del (c->data ());
      delete c->data ();
    } else if (c->data () -> sub_clusters ()) {
      Cluster *sc = c->data () -> sub_clusters () -> first ();
      while (sc) {
	bcl.add_first (sc);
	sc = c->data () -> sub_clusters () -> next ();
      }
    }
    delete c;
    c = bcl.unlink_first ();
  }

  regs->combine (&lep);

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
