// analyse_scalar.cc: Analyse a SiL model / scalars.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_scalar.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_scalar,
       "$Id: analyse_scalar.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "kernel/gcd.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"
#include "kernel/geometry.hxx"

void Analyse::eval_scalar ()
{
  // Cluster the scalars with respect to their type and check the
  // ratios.

  if (f_angles && angle_aspects.length () > 0) {

    sil_progress_bar_text ("Clustering " + 
			   as_string (angle_aspects.length ()) + " angles");

    // Cluster angles.
    List<Cluster> *clusters = new List<Cluster>;
    cluster_aspects (&angle_aspects, tol_eq, tol_angle_dist, NAN, true,
		     clusters);

    // Find tolerances.
    SimpleList<Real> tols;
    tols.add_first (new Real (tol_angle_dist));
    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      if (cn->data () -> tolerance () > tol_angle_dist)
	tols.add_first (new Real (cn->data () -> tolerance ()));
      cn = cn->next ();
    }

    tolerance_levels (clusters, &tols, tol_angle_dist, PI / 2.0);

    // Find special values for angles.
    if (f_angles_special) {
      sil_progress_bar_text ("Checking for special angles");
      List<Cluster> bcls;
      cn = clusters->first ();
      while (cn) {
	cn->data () -> get_all_base_clusters (&bcls);
	cn = cn->next ();
      }
      cn = bcls.first ();
      while (cn) {
	Cluster *c = cn->data ();
	SubRegularity *sr = 
	  check_special_scalar (static_cast<Angle*> (c->value ()) -> value (),
				c->tolerance (), true);
	if (sr)
	  c->add_sub_regularity (sr);
	if (c->parent_cluster ())
	  bcls.add_last_check (c->parent_cluster ());
	cn = cn->next ();
      }
    }

    // Check ratios.
    if (f_angles_ratios) {
      sil_progress_bar_text ("Checking angle ratios");
      eval_ratios (clusters);
    }

    reg_list->add_first ((List<Regularity> *)clusters);
  }

  if (f_lengths && length_aspects.length () > 0) {

    sil_progress_bar_text ("Clustering " + 
			   as_string (length_aspects.length ()) + " lengths");

    // Cluster lengths.
    List<Cluster> *clusters = new List<Cluster>;
    cluster_aspects (&length_aspects, tol_eq, tol_len_dist, NAN, true,
		     clusters);
    
    // Find tolerances.
    SimpleList<Real> tols;
    tols.add_first (new Real (tol_len_dist));
    ListNode<Cluster> *cn = clusters->first ();
    while (cn) {
      if (cn->data () -> tolerance () > tol_len_dist)
	tols.add_first (new Real (cn->data () -> tolerance ()));
      cn = cn->next ();
    }

    tolerance_levels (clusters, &tols, tol_len_dist, max_len);

    // Find special values for lengths.
    if (f_lengths_special) {
      sil_progress_bar_text ("Checking for special lengths");
      List<Cluster> bcls;
      cn = clusters->first ();
      while (cn) {
	cn->data () -> get_all_base_clusters (&bcls);
	cn = cn->next ();
      }
      cn = bcls.first ();
      while (cn) {
	Cluster *c = cn->data ();
	SubRegularity *sr = 
	  check_special_scalar (static_cast<Length*> (c->value ()) -> value (),
				c->tolerance (), false);
	if (sr)
	  c->add_sub_regularity (sr);
	if (c->parent_cluster ())
	  bcls.add_last_check (c->parent_cluster ());
	cn = cn->next ();
      }
    }

    // Check ratios.
    if (f_lengths_ratios) {
      sil_progress_bar_text ("Checking length ratios");
      eval_ratios (clusters);
    }

    reg_list->add_first ((List<Regularity> *)clusters);
  }

  return;
}

void Analyse::eval_ratios (List<Cluster> *clusters)
{
  List<SpecialRatio> *ratios = new List<SpecialRatio>;

  // Generate all ratios between all base clusters of parameters and
  // check if they are special. We add all clusters at the moment.
  // NOTE: We might choose only to use the base clusters to reduce the
  //       number fo ratios produced.
  List<Cluster> cl;
  ListNode<Cluster> *c1 = clusters->first ();
  while (c1) {
    c1->data () -> get_all_base_clusters (&cl);
    c1 = c1->next ();
  }
  c1 = cl.first ();
  while (c1) {
    if (c1->data () -> parent_cluster ())
      cl.add_last_check (c1->data () -> parent_cluster ());
    c1 = c1->next ();
  }

  sil_progress_bar_text ("Checking " + 
			 as_string (cl.length () * (cl.length () - 1) / 2) +
			 " ratios");
  c1 = cl.first ();
  while (c1) {
    Real v1 = static_cast<Scalar *> (c1->data () -> value ()) -> value ();
    if (fabs (v1) >= tol_eq) {
      ListNode<Cluster> *c2 = c1->next ();
      while (c2) {
	Real v2 = static_cast<Scalar *> (c2->data () -> value ()) -> value ();
	if (fabs (v2) >= tol_eq)
	  check_ratio (c1->data (), c2->data (), NULL, ratios);
	c2 = c2->next ();
      }
    }
    c1 = c1->next ();
  }

  // Add the special ratios found.
  if (ratios->length () > 0) {
    reg_list->add_first ((List<Regularity> *)ratios);
  } else {
    delete ratios;
  }

  return;
}

void Analyse::check_ratio (Cluster *c1, Cluster *c2,
			   SimpleList<Ratio> *taboo_ratios,
			   List<SpecialRatio> *ratios)
{
  // Only for scalar clusters.
  assert (dynamic_cast<Scalar *> (c1->value ()) &&
	  dynamic_cast<Scalar *> (c2->value ()));
  // Get the ratios between the values of cluster c1 and c2.
  Real v = static_cast<Scalar *> (c1->value ()) -> value () / 
    static_cast<Scalar *> (c2->value ()) -> value ();
  if (v < 1.0) {
    // Ensure that c1 > c2 to have a consisten input for the ratio
    // algorithm.
    v = 1.0 / v;
    Cluster *cc = c1;
    c1 = c2;
    c2 = cc;
  }
  SimpleList<Ratio> *rl = ratio (v, tol_ratio_min, tol_ratio_approx,
				 ratio_min_frac, ratio_max_frac,
				 ratio_max_power, true);
  if (rl) {
    // Delete ratios from list which are already represented by the
    // parents.
    if (taboo_ratios) {
      Ratio *r = rl->first ();
      while (r) {
	Ratio *nr = rl->next ();
	Ratio *cr = taboo_ratios->first ();
	while (cr) {
	  if (cr->is_equal (r)) {
	    rl->del (r);
	    delete r;
	    break;
	  }
	  cr = taboo_ratios->next ();
	}
	r = nr;
      }
    }
    // If some ratios are left, add them as a regularity. Otherwise
    // just delete the empy list.
    if (rl->length () > 0) {
      SpecialRatio *rr = new SpecialRatio (this, c1, c2, rl);
      ratios->add_first (rr);
    } else {
      delete rl;
      rl = NULL;
    }
  }

  // Gneerate new list of ratios which are covered by the parents and
  // should not be used by the children.
  //
  // Add the old list of taboo ratios.
  SimpleList<Ratio> newtaboo_ratios;
  if (taboo_ratios) {
    Ratio *r = taboo_ratios->first ();
    while (r) {
      newtaboo_ratios.add_first (r);
      r = taboo_ratios->next ();
    }
  }
  // Now add any new taboo ratios.
  if (rl) {
    Ratio *r = rl->first ();
    while (r) {
      newtaboo_ratios.add_first (r);
      r = rl->next ();
    }
  }

  // Check if there are ratios between the child clusters which
  // are not between the parent clusters.
  if (c1->sub_clusters ()) {
    // Ratios between c1 child clusters and c2.
    Cluster *c = c1->sub_clusters () -> first ();
    while (c) {
      if (fabs (static_cast<Scalar *> (c->value ()) -> value ()) >= tol_eq)
	check_ratio (c, c2, &newtaboo_ratios, ratios);
      c = c1->sub_clusters () -> next ();
    }
    if (c2->sub_clusters ()) {
      // Ratios between c1 child clusters and c2 child clusters and
      // between c1 and c2 child clusters.
      c = c2->sub_clusters () -> first ();
      while (c) {
	if (fabs (static_cast<Scalar *> (c->value ()) -> value ()) >= 
	    tol_eq) {
	  check_ratio (c1, c, &newtaboo_ratios, ratios);
	  Cluster *cc = c1->sub_clusters () -> first ();
	  while (cc) {
	    if (fabs (static_cast<Scalar *> (cc->value ())->value ()) >= 
		tol_eq)
	      check_ratio (cc, c, &newtaboo_ratios, ratios);
	    cc = c1->sub_clusters () -> next ();
	  }
	}
	c = c2->sub_clusters () -> next ();
      }
    }
  } else if (c2->sub_clusters ()) {
    // Ratios between c2 sub clusters and c1.
    Cluster *c = c2->sub_clusters () -> first ();
    while (c) {
      if (fabs (static_cast<Scalar *> (c->value ()) -> value ()) >= tol_eq)
	check_ratio (c1, c, &newtaboo_ratios, ratios);
      c = c2->sub_clusters () -> next ();
    }
  }

  return;
}

SpecialValue *Analyse::check_special_scalar (Real v, Real tol_approx, 
					     bool angle_flag)
{
  // Check if v is a special value with respect to the type and the
  // tolerance.

  // Find the bases for the type.
  char **base_ids;
  Real *base_vals;
  unsigned long base_nums;
  if (angle_flag) {
    if (!f_angles_special)
      return NULL;
    base_ids = angle_base_ids;
    base_vals = angle_base_vals;
    base_nums = angle_base_nums;
    if (tol_approx < tol_angle_dist)
      tol_approx = tol_angle_dist;
  } else {
    if (!f_lengths_special)
      return NULL;
    base_ids = len_base_ids;
    base_vals = len_base_vals;
    base_nums = len_base_nums;
    if (tol_approx < tol_len_dist)
      tol_approx = tol_len_dist;
  }

  // Generate the ratio lists for each base type.
  SimpleList<Ratio> **ratios = new SimpleList<Ratio>*[base_nums];
  Real reg = false;
  for (unsigned long bid = 0; bid < base_nums; bid++) {

    // Find ratios.
    if (strcmp (base_ids[bid], tan_id) == 0) {
      if (fabs (fabs (v) - PI / 2.0) > tol_angle_dist) {
	ratios[bid] = ratio (tan (v), tol_ratio_min, tol_approx,
			     para_min_frac, para_max_frac, 2, false);
      } else {
	ratios[bid] = NULL;
      }
    } else {
      ratios[bid] = ratio (v / base_vals[bid],
			   tol_ratio_min, tol_approx / base_vals[bid],
			   para_min_frac, para_max_frac, para_max_power,
			   false);
    }

    if (ratios[bid]) {
      // Delete duplicates.
      for (unsigned long bbid = 0; bbid < bid; bbid++) {
	Ratio *r = ratios[bid]->first ();
	while (r) {
	  Ratio *nr = ratios[bid]->next ();
	  if (ratios[bbid]) {
	    Real v = r->eval ();
	    if (strcmp (base_ids[bid], tan_id) == 0)
	      v = atan (v);
	    else
	      v *= base_vals[bid];
	    Ratio *rr = ratios[bbid]->first ();
	    while (rr) {
	      Real vv = rr->eval ();
	      if (strcmp (base_ids[bbid], tan_id) == 0)
		vv = atan (vv);
	      else
		vv *= base_vals[bbid];
	      if (fabs (v - vv) < tol_eq) {
		ratios[bid]->del (r);
		delete r;
		break;
	      }
	      rr = ratios[bbid]->next ();
	    }
	  }
	  r = nr;
	}
      }
      if (ratios[bid]->length () > 0) {
	reg = true;
      } else {
	delete ratios[bid];
	ratios[bid] = NULL;
      }
    }

  }

  if (reg)
    return new SpecialValue (this, ratios, base_ids, base_vals, base_nums);
  return NULL;
}

SimpleList<Ratio> *Analyse::ratio (Real v,
				   Real tol_min,
				   Real tol_approx,
				   unsigned long min_frac,
				   unsigned long max_frac,
				   unsigned long max_power,
				   bool real_ratio)
{
  SimpleList<Ratio> *rs = NULL;

  if (tol_approx < tol_ratio_approx)
    tol_approx = tol_ratio_approx;

  // Sign.
  bool pm = (v >= 0.0);
  if (!pm)
    v = -v;
  
  // Denominator limits.
  unsigned long m_f = (unsigned long)floor (1.0 / (2.0 * tol_approx));
  if (m_f < min_frac)
    m_f = min_frac;
  if (m_f >= max_frac)
    m_f = max_frac;
  
  // Find special ratios for each allowed root.
  Real w = 1.0;
  for (unsigned long p = 1; p < max_power; ++p) {
    
    w *= v;

    unsigned long n = (unsigned long)rint (w);
    
    // Check if closest integer is close enough in order to be special.
    Real t = fabs ((Real)n - w);
    if (t < tol_approx && !(real_ratio && (n == 1 || n == 0))) {
      if (!rs)
	rs = new SimpleList<Ratio>;
      bool tpm = pm;
      if (n == 0)
	tpm = true;
      Ratio *nr = new Ratio (tpm, n, 1.0, p, t);
      Real nrv = nr->eval ();
      Ratio *rr = rs->first ();
      while (rr) {
	if (fabs (rr->eval () - nrv) < tol_eq) {
	  delete nr;
	  nr = NULL;
	  break;
	}
	rr = rs->next ();
      }
      if (nr)
	rs->add_last (nr);
    }
	
    // Try to find a special integer relation for the non-integer part.
    
    Real w0 = w - (Real)n;
    
    bool neg;
    if (w0 < 0.0) {
      w0 = -w0;
      neg = true;
    } else {
      neg = false;
    }

    if (w0 > tol_min) {
      // Val = (pm?:-1:1) * (((neg?-1:1) ((w0<0.5)?1-(1-w0):w0) + n)) ^ (1/p)
      //
      // We use 1-w0 instead of w0 for the integer relation algorithm
      // if w0 < 0.5 as it is more likely to find small fractions in
      // 0.5,1 than in 0.0,0.5.
      if (w0 < 0.5) {
	integer_rel (true, 1.0 - w0, 1.0, 1.0,
		     pm, neg, n, p,
		     m_f, m_f, tol_min, tol_approx,
		     rs);
      } else {
	integer_rel (false, w0, 0.0, 1.0,
		     pm, neg, n, p,
		     m_f, m_f, tol_min, tol_approx,
		     rs);
      }
    }

  }
  
  return rs;
}

void Analyse::integer_rel (bool w_neg, Real const &w,
			   unsigned long const &n, unsigned long const &m,
			   bool const &pm, bool &N_neg, 
			   unsigned long const &N, unsigned long const &p,
			   unsigned long const &max_m,
			   unsigned long const &max_frac,
			   Real const &tol_min, Real const &tol_approx,
			   SimpleList<Ratio> *&rs)
{
  // Find a integer relations for w and combine it with the known
  // integer relation for the complete value,continue approximation
  // the errors recursively.
  //
  //  Value = (pm?:-1:1)*(((w_neg?-1:1) * w + n/m) * (N_neg?:-1:1) + N)^(1/p)

  unsigned long a, na, mb, new_n, new_m, full_n, g;
  bool r_neg;
  Real r;

  unsigned long b = 1;

  while (true) {

    // Find approximating fraction a / b for fixed b.

    a = static_cast<unsigned long> (rint (static_cast<Real> (b) / w));

    if (a > max_m || isnan (a))
      return;

    if (a > b) {

      // Difference between w and b /a.
      r = w - static_cast<Real> (b) / static_cast<Real> (a);
      if (r < 0.0) {
	r = -r;
	r_neg = true;
      } else {
	r_neg = false;
      }

      // Compute new approximating fraction.
      // Val = (w_neg?-1:1)*((r_neg?-1:1) * r + (b / a)) + n / m
      //     = (w_neg?-1:1)*(r_neg?-1:1)*r + ((w_neg?:-1:1) b m + n a) / (m a)
      //                                new: ^^^^^^^^^^^^^^^^^^^^^^^^^ / ^^^^^ 
      na = n * a;
      mb = m * b;
      if (w_neg) {
	// Subtract two fractions, if the result is negative change the
	// sign of the complete fraction (N_reg), otherwise only change the
	// sign of the remianing erorro r (r_neg).
	if (na < mb) {
	  new_n = mb - na;
	  N_neg = !N_neg;
	} else {
	  new_n = na - mb;
	  r_neg = !r_neg;
	}
      } else {
	// Add two positive fractions.
	new_n = na + mb;
      }
      new_m = m * a;

      // Reduce new fraction.
      g = gcd (new_n, new_m);
      new_n /= g;
      new_m /= g;

      bool rec = true;
      if (r < tol_approx) {

	// Add fraction as special ratio.
	if (!rs)
	  rs = new SimpleList<Ratio>;
	// Val = (r_neg?-1:1) * r + N + (N_neg?-1:1) * n / m;
	if (N_neg)
	  full_n = N * new_m - new_n;
	else
	  full_n = N * new_m + new_n;
	g = gcd (full_n, new_m);
	Ratio *nr = new Ratio (pm, full_n / g, new_m / g, p, r);

	// Check if fraction is already present. Add it if
	// not. Otherwise stop recursion for this fraction.
	Real nrv = nr->eval ();
	Ratio *rr = rs->first ();
	while (rr) {
	  if (fabs (rr->eval () - nrv) < tol_eq) {
	    rec = false;
	    delete nr;
	    nr = NULL;
	    break;
	  }
	  rr = rs->next ();
	}
	if (nr)
	  rs->add_last (nr);

      }
      
      // Recursively process errors.
      if (rec && r > tol_min)
	integer_rel (r_neg, r, new_n, new_m,
		     pm, N_neg, N, p,
		     max_m * max_frac, max_frac, tol_min, tol_approx,
		     rs);
    }

    ++b;

  }

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
