// features.cc: Features implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: features.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (features, "$Id: features.cc 1 2005-08-10 23:35:07Z frank $");

#include "xstring.hxx"
#include "xmath.hxx"
#include "kernel/topology.hxx"
#include "kernel/analyse.hxx"

string const Feature::tid = "Feature";
string const Direction::tid = "Direction";
string const Position::tid = "Position";
string const Scalar::tid = "Scalar";
string const Angle::tid = "Angle";
string const Length::tid = "Length";
string const Ratio::tid = "Ratio";
string const DirectionPlane::tid = "DirectionPlane";
string const DirectionCone::tid = "DirectionCone";
string const IntersectionPosition::tid = "IntersectionPosition";
string const AlignedAxes::tid = "AlignedAxes";
string const ParallelLine::tid = "ParallelLine";
string const PointCircle::tid = "PointCircle";
string const PartiallyPlaneEqual::tid = "PartiallyPlaneEqual";
string const PartiallyLineEqual::tid = "PartiallyLineEqual";
string const LoopPolygon::tid = "LoopPolygon";

void Feature::hilite (bool hi, string const &name, Entity *entity) const
{
  if (entity->is (SIL_FEATURE_TYPE)) {
    if (static_cast<Feature*> (entity) -> generator ()) {
      entity = static_cast<Feature*> (entity) -> generator ();
    } else {
      sil_msg_warn ("Feature", "cannot hilite " + entity->type_id ());
      return;
    }
  }
  if (entity->is (SIL_GEOMETRY_TYPE)) {
    if (static_cast<Geometry*> (entity) -> cell ()) {
      Cell *c = static_cast<Geometry*> (entity) -> cell ();
      entity = dynamic_cast<Entity *> (c);
      if (!entity) {
	sil_msg_warn ("Feature", "cannot convert cell to entity for hilite");
	return;
      }
    } else {
      sil_msg_warn ("Feature", "cannot hilite " + entity->type_id ());
      return;
    }
  }
  if (entity->is (SIL_LUMP_TYPE)) {
    if (static_cast<Lump *> (entity) -> reference ())
      sil_hilite (hi, name.c_str (),
		  static_cast<Lump *> (entity) -> reference ());
    else
      sil_hilite (hi, name.c_str (), entity);
  } else if (entity->is (SIL_FACE_TYPE)) {
    if (static_cast<Face *> (entity) -> reference ())
      sil_hilite (hi, name.c_str (),
		  static_cast<Face *> (entity) -> reference ());
    else
      sil_hilite (hi, name.c_str (), entity);
  } else if (entity->is (SIL_EDGE_TYPE)) {
    if (static_cast<Edge *> (entity) -> reference ())
      sil_hilite (hi, name.c_str (),
		  static_cast<Edge *> (entity) -> reference ());
    else
      sil_hilite (hi, name.c_str (), entity);
  } else if (entity->is (SIL_VERTEX_TYPE)) {
    if (static_cast<Vertex *> (entity) -> reference ())
      sil_hilite (hi, name.c_str (),
		  static_cast<Vertex *> (entity) -> reference ());
    else
      sil_hilite (hi, name.c_str (), entity);
  } else if (entity->is (SIL_SOLID_TYPE)) {
    XListNode<Lump> *n = static_cast<Solid *> (entity) -> lumps () . first ();
    while (n) {
      if (n->data () -> reference ())
	sil_hilite (hi, name.c_str (), n->data () -> reference ());
      else
	sil_hilite (hi, name.c_str (), n->data ());
      n = n->next ();
    }
  } else if (entity->is (SIL_SHELL_TYPE)) {
    XListNode<Face> *n = static_cast<Shell *> (entity) -> faces () . first ();
    while (n) {
      if (n->data () -> reference ())
	sil_hilite (hi, name.c_str (), n->data () -> reference ());
      else
	sil_hilite (hi, name.c_str (), n->data ());
      n = n->next ();
    }
  } else if (entity->is (SIL_LOOP_TYPE)) {
    XListNode<Edge> *n = static_cast<Loop *> (entity) -> edges () . first ();
    while (n) {
      if (n->data () -> reference ())
	sil_hilite (hi, name.c_str (), n->data () -> reference ());
      else
	sil_hilite (hi, name.c_str (), n->data ());
      n = n->next ();
    }
  } else if (entity->is (SIL_REGULARITY_TYPE)) {
    static_cast<Regularity *> (entity) -> hilite (hi);
  } else {
    sil_msg_warn ("Feature", "cannot hilite " + entity->type_id ());
  }
  return;
}

string const Direction::description () const 
{
  return as_string (*static_cast<UnitVector3 const *> (this));
}

Real Direction::sim (Feature const *f) const
{
  Direction const *d = dynamic_cast<Direction const *> (f);
  if (d) {
    Real a = fabs (*this % *d);
    if (a >= 1.0)
      return 0.0;
    return acos (a);
  }
  return NAN;
}

Feature *Direction::average (unsigned long n,
			     Feature const *f, unsigned long m) const
{
  Direction const *d = dynamic_cast<Direction const *> (f);
  if (d) {
    if (*this % *d < 0)
      return new Direction ((n * (*this) - m * (*d)) / (n + m));
    return new Direction ((n * (*this) + m * (*d)) / (n + m));
  }
  return NULL;
}

string const Position::description () const 
{
  return as_string (*static_cast<Vector3 const *> (this));
}

Real Position::sim (Feature const *f) const 
{
  Position const *p = dynamic_cast<Position const *> (f);
  if (p)
    return (*this - *p) . norm ();
  return NAN;
}

Feature *Position::average (unsigned long n,
			    Feature const *f, unsigned long m) const
{
  Position const *p = dynamic_cast<Position const *> (f);
  if (p)
    return new Position ((n * (*this) + m * (*p)) / (n + m));
  return NULL;
}

string const Angle::description () const 
{
  return as_string (value () / PI * 180.0) + " deg";
}

Real Angle::sim (Feature const *f) const
{
  Angle const *a = dynamic_cast<Angle const *> (f);
  if (a)
    return fabs (a->value () - value ());
  return NAN;
}

Feature *Angle::average (unsigned long n,
			 Feature const *f, unsigned long m) const 
{
  Angle const *a = dynamic_cast<Angle const *> (f);
  if (a)
    return new Angle ((n * value () + m * a->value ()) / (n + m));
  return NULL;
}

string const Length::description () const 
{
  return as_string (value ());
}

Real Length::sim (Feature const *f) const
{
  Length const *l = dynamic_cast<Length const *> (f);
  if (l)
    return fabs (l->value () - value ());
  return NAN;
}

Feature *Length::average (unsigned long n,
			 Feature const *f, unsigned long m) const 
{
  Length const *l = dynamic_cast<Length const *> (f);
  if (l)
    return new Length ((n * value () + m * l->value ()) / (n + m));
  return NULL;
}

string const Ratio::description () const 
{
  string str;
  if (pm)
    str = "- ";
  if (denom == 1) {
    str += as_string (num);
  } else {
    if (ro != 1)
      str += "(";
    str += as_string (num) + " / " + as_string (denom);
    if (ro != 1)
      str += ")";
  }
  if (ro != 1)
    str += " ^ (1 / " + as_string (ro) + ")";
  str += " ~" + as_string (tol);
  return str;
}

Real Ratio::sim (Feature const *f) const 
{
  Ratio const *r = dynamic_cast<Ratio const *> (f);
  if (r)
    return fabs (r->eval () - eval());
  return NAN;
}

Real Ratio::eval () const 
{
  Real v = (Real) num;
  if (denom != 1)
    v /= (Real) denom;
  if (ro != 1)
    v = pow (v, 1.0 / (Real) ro);
  if (!pm)
    v = -v;
  return v;
}

string const DirectionPlane::description () const 
{
  return as_string (*static_cast<UnitVector3 const *> (this));
}

Real DirectionPlane::sim (Feature const *f) const
{
  DirectionPlane const *d = dynamic_cast<DirectionPlane const *> (f);
  if (d) {
    Real a = fabs (*this % *d);
    if (a >= 1.0)
      return 0.0;
    return acos (a);
  }
  return NAN;
}

Feature *DirectionPlane::average (unsigned long n,
				  Feature const *f, unsigned long m) const
{
  DirectionPlane const *d = dynamic_cast<DirectionPlane const *> (f);
  if (d) {
    if (*this % *d < 0)
      return new DirectionPlane ((n * (*this) - m * (*d)) / (n + m));
    return new DirectionPlane ((n * (*this) + m * (*d)) / (n + m));
  }
  return NULL;
}

string const DirectionCone::description () const
{
  return as_string (*static_cast<UnitVector3 const *> (this)) + ", " +
    as_string (angl / PI * 180.0) + " deg";
}

Real DirectionCone::sim (Feature const *f) const 
{
  DirectionCone const *ac = dynamic_cast<DirectionCone const *> (f);
  if (ac) {
    Real d = fabs ((*this) % (*ac));
    if (d >= 1.0)
      d = 0.0;
    else
      d = acos (d);
    return d + fabs (angl - ac->angl);
  }
  return NAN;
}

Feature *DirectionCone::average (unsigned long n,
				 Feature const *f, unsigned long m) const
{
  DirectionCone const *ac = dynamic_cast<DirectionCone const *> (f);
  if (ac) {
    unsigned long nm = n + m;
    if ((*this) % (*ac) < 0)
      return new DirectionCone ((n * (*this) - m * (*ac)) / nm,
				(n * angl + m * ac->angl) / nm);
    return new DirectionCone ((n * (*this) + m * (*ac)) / nm,
			      (n * angl + m * ac->angl) / nm);
  }
  return NULL;
}

string const IntersectionPosition::description () const 
{
  if (isnan (t))
    return as_string (*static_cast<Vector3 const *> (this));
  return as_string (*static_cast<Vector3 const *> (this)) + " ~" +
    as_string (t);
}

Feature *IntersectionPosition::average (unsigned long n, Feature const *f,
					unsigned long m) const
{
  IntersectionPosition const *p =
    dynamic_cast<IntersectionPosition const *> (f);
  if (p)
    return new IntersectionPosition ((n * (*this) + m * (*p)) / (n + m),
				     (n * t + m * p->t) / (n + m));
  return NULL;
}

Feature *AlignedAxes::average (unsigned long n,
			       Feature const *f, unsigned long m) const
{
  AlignedAxes const *p = dynamic_cast<AlignedAxes const *> (f);
  if (p)
    return new AlignedAxes ((n * (*this) + m * (*p)) / (n + m), dir);
  return NULL;
}

string const AlignedAxes::description () const 
{
  if (dir)
    return as_string (*static_cast<Vector3 const *> (this)) + " / " +
      dir->description ();
  return as_string (*static_cast<Vector3 const *> (this));
}

string const ParallelLine::description () const
{
  return as_string (*static_cast<Vector3 const *> (this));
}

Real ParallelLine::sim (Feature const *f) const
{
  ParallelLine const *pl = dynamic_cast<ParallelLine const *> (f);
  if (pl) {
    if (pl->norm () > norm ()) {
      UnitVector3 u = pl->normalise ();
      return ((*this) - ((*this) % u) * u) . norm ();
    }
    UnitVector3 u = this->normalise ();
    return (*pl - (*pl % u) * u) . norm ();
  }
  return NAN;
}

Feature *ParallelLine::average (unsigned long n,
				Feature const  *f, unsigned long m) const
{
  ParallelLine const *pl = dynamic_cast<ParallelLine const *> (f);
  if (pl) {
    Real dl = this->norm ();
    Real adl = pl->norm ();
    Real w = static_cast<Real> (n + m);
    w = (dl * n + adl * m) / (w * w);
    if ((*this) % (*pl) < 0.0)
      return new ParallelLine ((this->normalise () * static_cast<Real> (n) -
				pl->normalise () * static_cast<Real> (m)) *
			       w);
    return new ParallelLine ((this->normalise () * static_cast<Real> (n) +
			      pl->normalise () * static_cast<Real> (m)) *
			     w);
  }
  return NULL;
}

string const PointCircle::description () const 
{
  return as_string (c) + ", " + as_string (r);
}

Real PointCircle::sim (Feature const *f) const 
{
  PointCircle const *ac = dynamic_cast<PointCircle const *> (f);
  if (ac)
    return (ac->c - c) . norm ();
  return NAN;
}

Feature *PointCircle::average (unsigned long n,
			       Feature const *f, unsigned long m) const
{
  PointCircle const *ac = dynamic_cast<PointCircle const *> (f);
  if (ac) {
    unsigned long nm = n + m;
    return new PointCircle ((n * c + m * ac->c) / nm,
			    (n * r + m * ac->r) / nm);
  }
  return NULL;
}

string const PartiallyPlaneEqual::description () const
{
  return as_string (*static_cast<Vector3 const *> 
		    (static_cast<Position const *> (this))) + " : " +
    d->description ();
}

Real PartiallyPlaneEqual::sim (Feature const *f) const
{
  PartiallyPlaneEqual const *p = dynamic_cast<PartiallyPlaneEqual const *> (f);
  if (p && p->d == d)
    return (*this - *p) . norm ();
  return NAN;
}

Feature *PartiallyPlaneEqual::average (unsigned long n,
				       Feature const *f, unsigned long m) const
{
  PartiallyPlaneEqual const *p = dynamic_cast<PartiallyPlaneEqual const*> (f);
  if (p && p->d == d)
    return new PartiallyPlaneEqual ((n * (*this) + m * (*p)) / (n + m), d);
  return NULL;
}

string const PartiallyLineEqual::description () const
{
  return as_string (*static_cast<Vector3 const *> 
		    (static_cast<Position const *> (this))) + " : " +
    d->description ();
}

Real PartiallyLineEqual::sim (Feature const *f) const
{
  PartiallyLineEqual const *p = dynamic_cast<PartiallyLineEqual const *> (f);
  if (p && p->d == d)
    return (*this - *p) . norm ();
  return NAN;
}

Feature *PartiallyLineEqual::average (unsigned long n,
				      Feature const *f, unsigned long m) const
{
  PartiallyLineEqual const *p = dynamic_cast<PartiallyLineEqual const *> (f);
  if (p && p->d == d)
    return new PartiallyLineEqual ((n * (*this) + m * (*p)) / (n + m), d);
  return NULL;
}

string const LoopPolygon::description () const
{
  if (a && b) {
    string str = "[" + as_string (a[0]);
    for (unsigned long k = 1; k < num_coeff; ++k) {
      if (k < num_coeff - 1)
	str += "; ";
      str += as_string (a[k]) + " + " + as_string (b[k]) + "i";
    }
    str += "]";
    return str;
  }
  return "[]";
}

Real LoopPolygon::sim (Feature const *f) const 
{
  LoopPolygon const *lp = dynamic_cast<LoopPolygon const *> (f);
  if (lp && a && lp->a && b && lp->b) {
    // Old version: sum(abs (x_l-y_l)),  considers amplitude and phase)
    // Real d = fabs (a[0] - lp->a[0]);
    // for (unsigned long k = 1; k < num_coeff; ++k) {
    //   double da = a[k] - lp->a[k];
    //   double db = b[k] - lp->b[k];
    //   d += sqrt (da * da + db * db);
    // }
    // New version: sqrt (sum_l (abs(x_l) - abs (y_l))^2), compare
    // amplidutes only.
    Real d = fabs (fabs (a[0]) - fabs (lp->a[0]));
    for (unsigned long k = 1; k < num_coeff; ++k) {
      Real dd = sqrt (a[k] * a[k] + b[k] * b[k]) -
	sqrt (lp->a[k] * lp->a[k] + lp->b[k] * lp->b[k]);
      d += dd * dd;
    }
    return sqrt (d);
  }
  return NAN;
}

Feature *LoopPolygon::average (unsigned long n,
			       Feature const *f, unsigned long m) const
{
  LoopPolygon const *lp = dynamic_cast<LoopPolygon const *> (f);
  if (lp) {
    LoopPolygon *r = new LoopPolygon ();
    r->num_coeff = num_coeff;
    r->a = new Real[num_coeff];
    r->b = new Real[num_coeff];
    Real nm = n + m;
    for (unsigned long k = 0; k < num_coeff; ++k) {
      // We do only average the amplitude, not the phase.
      // r->a[k] = (a[k] * n + lp->a[k] * m) / nm;
      // r->b[k] = (b[k] * n + lp->b[k] * m) / nm;
      Real d1 = sqrt (a[k] * a[k] + b[k] * b[k]);
      Real d2 = sqrt (lp->a[k] * lp->a[k] + lp->b[k] * lp->b[k]);
      d1 = (d1 * n + d2 * m) / (d1 * nm) ;
      r->a[k] = d1 * a[k];
      r->b[k] = d1 * b[k];
    }
    return r;
  }
  return NULL;
}

Feature *LoopPolygon::copy () const
{
  LoopPolygon *lp = new LoopPolygon ();
  lp->num_coeff = num_coeff;
  lp->a = new Real[num_coeff];
  lp->b = new Real[num_coeff];
  for (unsigned long k = 0; k < num_coeff; ++k) {
    lp->a[k] = a[k];
    lp->b[k] = b[k];
  }
  return lp;
}

void LoopPolygon::ft (unsigned long nc)
{
  if (a)
    delete a;
  if (b)
    delete b;
  num_coeff = nc;

  unsigned long m = pts->length ();
  Real *l = new Real[m + 1];
  Real *alpha = new Real[m];

  // Select an orientation for the ploygon (it's not necessarily the
  // one in the model!)
  Position *p1 = pts->first ();
  Position *p2 = pts->next ();
  Position *p3 = pts->next ();
  UnitVector3 n (0.0, 0.0, 0.0);
  Real max = 0.0;
  while (p3) {
    Vector3 N = (*p1 - *p2) * (*p3 - *p2);
    Real Nnorm = N.norm ();
    if (Nnorm > max) {
      max = Nnorm;
      n = N.normalise ();
    }
    p1 = p2;
    p2 = p3;
    p3 = pts->next ();
  }

  // Compute function.
  Position *op = pts->first ();
  Position *p = pts->next ();
  Position *np;
  Vector3 d = *p - *op;
  Vector3 nd;
  l[0] = 0;
  unsigned long k = 1;
  while (p) {
    // Length of section.
    l[k] = l[k - 1] + d.norm ();
    np = pts->next ();
    if (np) {
      nd = *np - *p;
      // Angle between d and nd to alpha[k]
      Real angl = - (nd % d) / (nd.norm () * d.norm ());
      if (angl >= 1.0)
	angl = 0.0;
      else if (angl <= -1.0)
	angl = PI;
      else
	angl = acos (angl);
      if ((d * nd) % n > 0.0)
	alpha[k] = 2 * PI - angl;
      else
	alpha[k] = angl;
      // Next.
      d = nd;
      op = p;
      p = np;
      ++k;
    } else {
      // Close the polygon.
      np = pts->first ();
      nd = *np - *p;
      l[k + 1] = l[k] + nd.norm ();
      // Angle between d and nd to alpha[k]
      Real angl = - (nd % d) / (nd.norm () * d.norm ());
      if (angl >= 1.0)
	angl = 0.0;
      else if (angl <= -1.0)
	angl = PI;
      else
	angl = acos (angl);
      if ((d * nd) % n > 0.0)
	alpha[k] = 2 * PI - angl;
      else
	alpha[k] = angl;
      // Next segment.
      d = nd;
      p = np;
      np = pts->next ();
      nd = *np - *p;
      // angle between d and nd to alpha[0]
      angl = - (nd % d) / (nd.norm () * d.norm ());
      if (angl >= 1.0)
	angl = 0.0;
      else if (angl <= -1.0)
	angl = PI;
      else
	angl = acos (angl);
      if ((d * nd) % n > 0.0)
	alpha[0] = 2 * PI - angl;
      else
	alpha[0] = angl;
      // Stop.
      p = NULL;
    }
  }

  // Orientation:
  //   The sum of all angles is smaller than m 2 *pi.
  //   Depending on the orientaiton we either have \sum alpha_k or
  //   sum (2*pi - \alpha_k) = m*2*pi - sum \alpha_k.
  //   The "natural" inside has always the smaller angle sum, so we
  //   reverse all angles (-> 2*pi - angle) if the sum of all
  //   angles is larger than m * pi.
  Real sum = alpha[0];
  for (unsigned long j = 1; j < m; ++j)
    sum += alpha[j];
  if (sum > static_cast<Real> (m) * PI) {
    for (unsigned long j = 0; j < m; ++j)
      alpha[j] = 2.0 * PI - alpha[j];
    sum = static_cast<Real> (2 * m) * PI - sum;
  }

  // Compute Fourier coefficients.
  a = new Real[num_coeff];
  b = new Real[num_coeff];
  a[0] = sum / (2.0 * PI);
  b[0] = 0.0;
  for (unsigned long j = 1; j < num_coeff; ++j) {
    a[j] = 0.0;
    b[j] = 0.0;
    for (unsigned long k = 0; k < m; ++k) {
      Real h = - static_cast<Real> (2 * j) * PI * l[k] / l[m];
      a[j] += alpha[k] * cos (h);
      b[j] += alpha[k] * sin (h);
    }
    a[j] /= 2.0 * PI;
    b[j] /= 2.0 * PI;
  }

  delete l;
  delete alpha;

  return;
}

void LoopPolygon::set (SimpleXList<Position> *points)
{
  if (pts)
    delete pts;
  if (a)
    delete a;
  if (b)
    delete b;
  pts = points;
  a = b = NULL;
  num_coeff = 0;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
