// features.hxx: solid model features.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: features.hxx,v 1.17 2001/04/17 11:02:04 scmfcl Exp $

#ifndef FEATURES_HXX
#define FEATURES_HXX 1

#include "xmath.hxx"
#include "kernel/entity.hxx"
#include "kernel/vector.hxx"

class Cluster;
class SubRegularity;

// Feature of a solid model describing the shape of a cell and its
// position and orientation in space without topological and
// inside/outside orientation related information.
class Feature : public Entity {
public:
  Feature () : Entity () {
    Entity::type (1, SIL_FEATURE_TYPE);
    gen = NULL;
    d = 0;
  }
  virtual ~Feature () {
  }
  virtual string const type_id () const {
    return tid;
  }
  // Entity responsible for generating the feature
  Entity *&generator () {
    return gen;
  }
  Entity *generator () const {
    return gen;
  }
  // Feature id for the generator.
  unsigned long &id () {
    return d;
  }
  unsigned long id () const {
    return d;
  }
  virtual void hilite (bool hi, string const &name, Entity *entity) const;
  virtual string const description () const = 0;
  virtual Real sim (Feature const *f) const = 0;
  virtual Feature *average (unsigned long n,
			   Feature const *f, unsigned long m) const = 0;
  virtual Feature *copy () const = 0;
  virtual bool regular () const {
    return false;
  }
private:
  static string const tid;
  Entity *gen;
  unsigned long d;
};

// Direction feature.
class Direction : public Feature, public UnitVector3 {
public:
  Direction () : Feature (), UnitVector3 () {
    Entity::type (2, SIL_DIRECTION_TYPE);
  }
  Direction (Vector3 const &v) : Feature (), UnitVector3 (v) {
    Entity::type (2, SIL_DIRECTION_TYPE);
  }
  Direction (UnitVector3 const &v) : Feature (), UnitVector3 (v) {
    Entity::type (2, SIL_DIRECTION_TYPE);
  }
  virtual ~Direction () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new Direction (static_cast<UnitVector3> (*this));
  }
private:
  static string const tid;
};

// Position feature.
class Position : public Feature, public Vector3 {
public:
  Position () : Feature (), Vector3 () {
    Entity::type (2, SIL_POSITION_TYPE);
  }
  Position (Vector3 const &p) : Feature (), Vector3 (p) {
    Entity::type (2, SIL_POSITION_TYPE);
  }
  virtual ~Position () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new Position (static_cast<Vector3> (*this));
  }
private:
  static string const tid;
};

// Abstract scalar feature.
class Scalar : public Feature {
public:
  explicit Scalar (Real v) : Feature () {
    Entity::type (2, SIL_SCALAR_TYPE);
    val = v;
  }
  virtual ~Scalar () {
  }
  virtual string const type_id () const {
    return tid;
  }
  Real value () const {
    return val;
  }
  Real &value () {
    return val;
  }
private:
  static string const tid;
  Real val;
};

// Angle feature.
class Angle : public Scalar {
public:
  explicit Angle (Real v = NAN) : Scalar (v) {
    Entity::type (3, SIL_ANGLE_TYPE);
  }
  virtual ~Angle () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new Angle (value ());
  }
private:
  static string const tid;
};

// Lenght feature.
class Length : public Scalar {
public:
  explicit Length (Real v = NAN) : Scalar (v) {
    Entity::type (3, SIL_LENGTH_TYPE);
  }
  virtual ~Length () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new Length (value ());
  }
private:
  static string const tid;
};

// Ratio feature.
class Ratio : public Feature {
public:
  Ratio (bool pos = true, unsigned long nu = 0, unsigned long deno = 1,
	 unsigned long r = 0, Real t = 0.0) : Feature () {
    Entity::type (2, SIL_RATIO_TYPE);
    pm = pos;
    num = nu;
    denom = deno;
    ro = r;
    tol = t;
  }
  virtual ~Ratio () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const {
    assert (false); // never use this on a ratio
  }
  virtual Feature *copy () const {
    return new Ratio (pm, num, denom, ro);
  }
  bool positive () const {
    return pm;
  }
  unsigned long numerator () const {
    return num;
  }
  unsigned long denominator () const {
    return denom;
  }
  unsigned long root () const {
    return ro;
  }
  Real tolerance () const {
    return tol;
  }
  bool is_equal (Ratio const *r) const {
    return (r->pm == pm) && (r->num == num) && (r->denom == denom) &&
      (r->ro == ro);
  }
  Real eval () const;
private:
  static string const tid;
  bool pm;
  unsigned long num, denom, ro;
  Real tol;
};

// Direction plane feature.
class DirectionPlane : public Direction {
public:
  DirectionPlane () : Direction () {
    Entity::type (3, SIL_DIRECTIONPLANE_TYPE);
  }
  DirectionPlane (Vector3 const &v) : Direction (v) {
    Entity::type (3, SIL_DIRECTIONPLANE_TYPE);
  }
  DirectionPlane (UnitVector3 const &v) : Direction (v) {
    Entity::type (3, SIL_DIRECTIONPLANE_TYPE);
  }
  virtual ~DirectionPlane () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new DirectionPlane (static_cast<UnitVector3> (*this));
  }
private:
  static string const tid;
};

// Direction cone feature.
class DirectionCone : public Direction {
public:
  DirectionCone () : Direction () {
    Entity::type (3, SIL_DIRECTIONCONE_TYPE);
    angl = NAN;
    reg = false;
  }
  DirectionCone (Vector3 const &v, Real an) : Direction (v) {
    Entity::type (3, SIL_DIRECTIONCONE_TYPE);
    angl = an;
    reg = false;
  }
  DirectionCone (UnitVector3 const &v, Real an) : Direction (v) {
    Entity::type (3, SIL_DIRECTIONCONE_TYPE);
    angl = an;
    reg = false;
  }
  virtual ~DirectionCone () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new DirectionCone (static_cast<UnitVector3> (*this), angl);
  }
  virtual bool regular () const {
    return reg;
  }
  bool &regular () {
    return reg;
  }
  Real &angle () {
    return angl;
  }
  Real angle () const {
    return angl;
  }
private:
  static string const tid;
  Real angl;
  bool reg;
};

// Intersection position feature.
class IntersectionPosition : public Position {
public:
  IntersectionPosition () : Position () {
    Entity::type (3, SIL_INTERSECTIONPOSITION_TYPE);
    t = NAN;
  }
  IntersectionPosition (Vector3 const &p, Real tol) : Position (p) {
    Entity::type (3, SIL_INTERSECTIONPOSITION_TYPE);
    t = tol;
  }
  virtual ~IntersectionPosition () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new IntersectionPosition (static_cast<Vector3> (*this), t);
  }
  Real &intersection_tolerance () {
    return t;
  }
  Real intersection_tolerance () const {
    return t;
  }
private:
  static string const tid;
  Real t;
};

// Aligned axes feature.
class AlignedAxes : public Position {
public:
  AlignedAxes () : Position () {
    Entity::type (3, SIL_ALIGNEDAXES_TYPE);
    f = false;
    dir = NULL;
  }
  AlignedAxes (Vector3 const &p, Cluster *d) : Position (p) {
    Entity::type (3, SIL_ALIGNEDAXES_TYPE);
    dir = d;
    f = false;
  }
  virtual ~AlignedAxes () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new AlignedAxes (static_cast<Vector3> (*this), dir);
  }
  Cluster *direction () {
    return dir;
  }
  bool flag () const {
    return f;
  }
  bool &flag () {
    return f;
  }
private:
  static string const tid;
  Cluster *dir;
  bool f;
};

// Parallel line feature.
class ParallelLine : public Feature, public Vector3 {
public:
  ParallelLine () : Feature (), Vector3 () {
    Entity::type (2, SIL_PARALLELLINE_TYPE);
  }
  ParallelLine (Vector3 const &d) : Feature (), Vector3 (d) {
    Entity::type (2, SIL_PARALLELLINE_TYPE);
  }
  virtual ~ParallelLine () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new ParallelLine (static_cast<Vector3> (*this));
  }
private:
  static string const tid;
};

// Axis circle feature.
class PointCircle : public Feature {
public:
  PointCircle () : Feature () {
    Entity::type (2, SIL_POINTCIRCLE_TYPE);
    c = Vector3 (0.0, 0.0, 0.0);
    r = 0.0;
    reg = false;
  }
  PointCircle (Vector3 const &cc, Real rr) : Feature () {
    Entity::type (2, SIL_POINTCIRCLE_TYPE);
    c = cc;
    r = rr;
    reg = false;
  }
  virtual ~PointCircle () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new PointCircle (c, r);
  }
  virtual bool regular () const {
    return reg;
  }
  bool &regular () {
    return reg;
  }
  Vector3 &centre () {
    return c;
  }
  Vector3 centre () const {
    return c;
  }
  Real &radius () {
    return r;
  }
  Real radius () const {
    return r;
  }
private:
  static string const tid;
  Vector3 c;
  Real r;
  bool reg;
};

// Partially equal with respect to a plane
class PartiallyPlaneEqual : public Position {
public:
  PartiallyPlaneEqual () : Position () {
    Entity::type (2, SIL_PARTIALLYPLANEEQUAL_TYPE);
    d = NULL;
  }
  PartiallyPlaneEqual (Vector3 const &p, Cluster *dir) : Position (p) {
    Entity::type (2, SIL_PARTIALLYPLANEEQUAL_TYPE);
    d = dir;
  }
  virtual ~PartiallyPlaneEqual () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new PartiallyPlaneEqual (*static_cast<Vector3 const*> 
				    (static_cast<Position const*> (this)), d);
  }
  Cluster *direction () const {
    return d;
  }
private:
  static string const tid;
  Cluster *d;
};

// Partially equal with respect to a lines
class PartiallyLineEqual : public Position {
public:
  PartiallyLineEqual () : Position () {
    Entity::type (2, SIL_PARTIALLYLINEEQUAL_TYPE);
    d = NULL;
  }
  PartiallyLineEqual (Vector3 const &p, Cluster *dir) : Position (p) {
    Entity::type (2, SIL_PARTIALLYLINEEQUAL_TYPE);
    d = dir;
  }
  virtual ~PartiallyLineEqual () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const {
    return new PartiallyLineEqual (*static_cast<Vector3 const *> 
				   (static_cast<Position const *> (this)),
				   d);
  }
  Cluster *direction () const {
    return d;
  }
private:
  static string const tid;
  Cluster *d;
};

// Loop polygon feature.
class LoopPolygon : public Feature {
public:
  LoopPolygon (SimpleXList<Position> *points) : Feature () {
    Entity::type (2, SIL_LOOPPOLYGON_TYPE);
    pts = points;
    a = b = NULL;
    num_coeff = 0;
  }
  LoopPolygon () : Feature () {
    Entity::type (2, SIL_LOOPPOLYGON_TYPE);
    pts = NULL;
    a = b = NULL;
    num_coeff = 0;
  }
  virtual ~LoopPolygon () {
    if (pts)
      delete pts;
    if (a)
      delete a;
    if (b)
      delete b;
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual string const description () const;
  virtual Real sim (Feature const *f) const;
  virtual Feature *average (unsigned long n,
			    Feature const *f, unsigned long m) const;
  virtual Feature *copy () const;
  void set (SimpleXList<Position> *points);
  void ft (unsigned long nc);
private:
  static string const tid;
  SimpleXList<Position> *pts;
  Real *a, *b;
  unsigned long num_coeff;
};

#endif // !FEATURES_HXX

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
