// geometry.hxx: cell geometries.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: geometry.hxx,v 1.37 2001/04/18 18:14:18 scmfcl Exp $

#ifndef GEOMETRY_HXX
#define GEOMETRY_HXX 1

#include "list.hxx"
#include "kernel/xlist.hxx"
#include "kernel/entity.hxx"
#include "kernel/vector.hxx"
#include "kernel/features.hxx"

class Cell;
class Vertex;

// Geometry: interior of a cell. 
class Geometry : public Entity {
public:
  explicit Geometry (unsigned short d) : Entity () {
    Entity::type (1, SIL_GEOMETRY_TYPE);
    dim = d;
    cll = NULL;
  }
  virtual ~Geometry () {
  }
  virtual string const type_id () const {
    return tid;
  }
  unsigned short dimension () const {
    return dim;
  }
  Cell *cell () const {
    return cll;
  }
  Cell *&cell () {
    return cll;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    // Returns features which are part of the geometry and not
    // generated.
    return;
  }
private:
  static string const tid;
  unsigned short dim;
  Cell *cll;
};

// 3D Geometry: R^3.
class Space : public Geometry {
public:
  explicit Space () : Geometry (3) {
    Entity::type (2, SIL_SPACE_TYPE);
  }
  virtual ~Space () {
  }
  virtual string const type_id () const {
    return tid;
  }
private:
  static string const tid;
};

// 2D Geometry: surface.
// Each surface is sided, i.e. we have an inside and an outside.
class Surface : public Geometry {
public:
  Surface () : Geometry (2) {
    Entity::type (2, SIL_SURFACE_TYPE);
    rp = new Position ();
    rp->generator () = this;
    take (rp);
  }
  virtual ~Surface () {
    drop (rp);
  }
  virtual string const type_id () const {
    return tid;
  }
  Position *root_point () const {
    return rp;
  }
  virtual Direction *direction () const {
    return NULL;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (rp);
    return;
  }
private:
  static string const tid;
  Position *rp;
};

// 2D Geometry: plane.
class Plane : public Surface {
public:
  Plane () : Surface () {
    Entity::type (3, SIL_PLANE_TYPE);
    dir = new Direction ();
    dir->generator () = this;
    take (dir);
  }
  virtual ~Plane () {
    drop (dir);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Direction *direction () const {
    return dir;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (root_point ());
    fl->add_first (dir);
    return;
  }
private:
  static string const tid;
  Direction *dir;
};

// 2D Geometry: sphere.
class Sphere : public Surface {
public:
  Sphere () : Surface () {
    Entity::type (3, SIL_SPHERE_TYPE);
    rds = new Length ();
    rds->generator () = this;
    take (rds);
    cnvx = true;
  }
  virtual ~Sphere () {
    drop (rds);
  }
  virtual string const type_id () const {
    return tid;
  }
  Length *radius () const {
    return rds;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (root_point ());
    fl->add_first (rds);
    return;
  }
  bool convex () const {
    return cnvx;
  }
  bool &convex () {
    return cnvx;
  }
private:
  static string const tid;
  Length *rds;
  bool cnvx;
};

// 2D Geometry: cylinder.
class Cylinder : public Surface {
public:
  Cylinder () : Surface () {
    Entity::type (3, SIL_CYLINDER_TYPE);
    rds = new Length ();
    rds->generator () = this;
    take (rds);
    dir = new Direction ();
    dir->generator () = this;
    take (dir);
    cnvx = true;
  }
  virtual ~Cylinder () {
    drop (rds);
    drop (dir);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Direction *direction () const {
    return dir;
  }
  Length *radius () const {
    return rds;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (root_point ());
    fl->add_first (dir);
    fl->add_first (rds);
    return;
  }
  bool convex () const {
    return cnvx;
  }
  bool &convex () {
    return cnvx;
  }
private:
  static string const tid;
  Direction *dir;
  Length *rds;
  bool cnvx;
};

// 2D Geometry: cone.
class Cone : public Surface {
public:
  Cone () : Surface () {
    Entity::type (3, SIL_CONE_TYPE);
    angl = new Angle ();
    angl->generator () = this;
    take (angl);
    dir = new Direction ();
    dir->generator () = this;
    take (dir);
    cnvx = true;
  }
  virtual ~Cone () {
    drop (angl);
    drop (dir);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Direction *direction () const {
    return dir;
  }
  Angle *angle () const {
    return angl;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (root_point ());
    fl->add_first (dir);
    fl->add_first (angl);
    return;
  }
  bool convex () const {
    return cnvx;
  }
  bool &convex () {
    return cnvx;
  }
private:
  static string const tid;
  Direction *dir;
  Angle *angl;
  bool cnvx;
};

// 2D Geometry: torus.
class Torus : public Surface {
public:
  Torus () : Surface () {
    Entity::type (3, SIL_TORUS_TYPE);
    r1 = new Length ();
    r1->generator () = this;
    r1->id () = 1;
    take (r1);
    r2 = new Length ();
    r2->generator () = this;
    r2->id () = 2;
    take (r2);
    ar = new Length ();
    ar->generator () = this;
    ar->id () = 3;
    take (ar);
    dr = new Length ();
    dr->generator () = this;
    dr->id () = 4;
    take (dr);
    dir = new Direction ();
    dir->generator () = this;
    take (dir);
    cnvx = appl = true;
  }
  virtual ~Torus () {
    drop (r1);
    drop (r2);
    drop (ar);
    drop (dr);
    drop (dir);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Direction *direction () const {
    return dir;
  }
  Length *major_radius () const {
    return r1;
  }
  Length *minor_radius () const {
    return r2;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (root_point ());
    fl->add_first (dir);
    fl->add_first (r1);
    fl->add_first (r2);
    if (r1->value () > r2->value ()) {
      // Non-degenerate torus
      ar->value () = r1->value () + r2->value ();
      fl->add_first (ar);
      dr->value () = r1->value () - r2->value ();
      fl->add_first (dr);
    } else {
      // Degenerate torus
      if (appl) {
	ar->value () = r1->value () + r2->value ();
	fl->add_first (ar);
      } else if (r1->value () < r2->value ()) {
	dr->value () = r2->value () - r1->value ();
	fl->add_first (dr);
      }
    }
    return;
  }
  bool convex () const {
    // True, if the torus is convex (including degenerate cases)
    return cnvx;
  }
  bool &convex () {
    return cnvx;
  }
  bool apple () const {
    // True if degenrate torus is an apple. False for non-degenerate case
    // and for lemon.
    return appl;
  }
  bool &apple () {
    return appl;
  }
private:
  static string const tid;
  Direction *dir;
  Length *r1, *r2, *ar, *dr;
  bool cnvx, appl;
};

// 1D Geometry: Curves.
class Curve : public Geometry {
public:
  Curve () : Geometry (1) {
    Entity::type (2, SIL_CURVE_TYPE);
  }
  virtual ~Curve () {
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Direction *normal () const {
    // If curve is flat return plane vector here (with respect to
    // orientation).
    return NULL;
  }
  Vector3 ref_point () const {
    // A point on the part of the circle that is used in the model.
    return refp;
  }
  Vector3 &ref_point () {
    return refp;
  }
private:
  static string const tid;
  Vector3 refp;
};

// Curve: straight line (totally defined by the edge).
class Straight : public Curve {
public:
  Straight () : Curve () {
    Entity::type (3, SIL_STRAIGHT_TYPE);
  }
  virtual ~Straight () {
  }
  virtual string const type_id () const {
    return tid;
  }
private:
  static string const tid;
};

// Curve: Circle.
class Circle : public Curve {
public:
  Circle () : Curve () {
    Entity::type (3, SIL_CIRCLE_TYPE);
    norm = new Direction ();
    norm->generator () = this;
    take (norm);
    cen = new Position ();
    cen->generator () = this;
    take (cen);
    rds = new Length ();
    rds->generator () = this;
    take (rds);
  }
  virtual ~Circle () {
    drop (norm);
    drop (cen);
    drop (rds);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Direction *normal () const {
    // Normal also defining circle orientation. Points in direction
    // such that circle is traversed in positive direction.
    return norm;
  }
  Position *centre () const {
    return cen;
  }
  Length *radius () const {
    return rds;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (norm);
    fl->add_first (cen); // is also be added as face root point
    fl->add_first (rds);
  }
private:
  static string const tid;
  Direction *norm;
  Position *cen;
  Length *rds;
};

// Curve: ellipse.
class Ellipse : public Curve {
public:
  Ellipse () : Curve () {
    Entity::type (3, SIL_ELLIPSE_TYPE);
    norm = new Direction ();
    norm->generator () = this;
    norm->id () = 1;
    take (norm);
    mdir = new Direction ();
    mdir->generator () = this;
    mdir->id () = 2;
    take (mdir);
    cen = new Position ();
    cen->generator () = this;
    take (cen);
    r1 = new Length ();
    r1->generator () = this;
    r1->id () = 1;
    take (r1);
    r2 = new Length ();
    r2->generator () = this;
    r2->id () = 2;
    take (r2);
  }
  virtual ~Ellipse () {
    drop (norm);
    drop (mdir);
    drop (cen);
    drop (r1);
    drop (r2);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Direction *normal () const {
    // Normal also defining ellipse orientaiton. Points in direction
    // sucht aht ellipse is traversed in positive direction.
    return norm;
  }
  Position *centre () const {
    return cen;
  }
  Direction *major_direction () const {
    return mdir;
  }
  Length *major_radius () const {
    return r1;
  }
  Length *minor_radius () const {
    return r2;
  }
  bool is_inside (Vector3 p, Vertex *s, Vertex *e) const;
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (norm);
    fl->add_first (mdir);
    fl->add_first (cen); // is also be added as face root point
    fl->add_first (r1);
    fl->add_first (r2);
    return;
  }
private:
  static string const tid;
  Direction *norm, *mdir;
  Position *cen;
  Length *r1, *r2;
};

// Curve: spline curve.
class SplineCurve : public Curve {
public:
  SplineCurve () : Curve () {
    Entity::type (3, SIL_SPLINECURVE_TYPE);
    cpts = NULL;
    ws = ks = NULL;
    rat = clo = per = false;
    deg = num_ks = 0;
    cptst = ktol = 0.0;
    plane_norm = NULL;
  }
  virtual ~SplineCurve () {
    drop (plane_norm);
    if (cpts) {
      cpts->purge ();
      delete cpts;
    }
    if (ws)
      delete ws;
    if (ks)
      delete ks;
  }
  virtual string const type_id () const {
    return tid;
  }
  UnitVector3 orientation () const {
    // Orientation vector such that curve is traversed positive with 
    // respect to this vector.
    return orv;
  }
  UnitVector3 orientation () {
    return orv;
  }
  Direction *plane_normal () const {
    return plane_norm;
  }
  void set_plane_normal (Direction const &d) {
    if (plane_norm) {
      plane_norm->set (d.x (), d.y (), d.z ());
    } else {
      plane_norm = new Direction (UnitVector3 (d.x (), d.y (), d.z ()));
      plane_norm->generator () = this;
      take (plane_norm);
    }
    return;
  }
  void unset_plane_normal () {
    drop (plane_norm);
    plane_norm = NULL;
    return;
  }
  void define (int d, bool r, bool c, bool p, SimpleList<Vector3> *cp,
	       Real *w, Real ct, int nk, Real *k, Real kt) {
    deg = d;
    rat = r;
    clo = c;
    per = p;
    if (cpts) {
      cpts->purge ();
      delete cpts;
    }
    cpts = cp;
    if (ws)
      delete ws;
    ws = w;
    cptst = ct;
    num_ks = nk;
    if (ks)
      delete ks;
    ks = k;
    ktol = kt;
  }
  int degree () const {
    return deg;
  }
  bool rational () const {
    return rat;
  }
  bool closed () const {
    return clo;
  }
  bool periodic () const {
    return per;
  }
  SimpleList<Vector3> *ctrlpts () const {
    return cpts;
  }
  Real *weights () const {
    return ws;
  }
  Real ctrlpts_tol () const {
    return cptst;
  }
  int num_knots () const {
    return num_ks;
  }
  Real *knots () const {
    return ks;
  }
  Real knot_tol () const {
    return ktol;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    if (plane_norm)
      fl->add_first (plane_norm);
    return;
  }
private:
  static string const tid;
  Direction *plane_norm;
  UnitVector3 orv;
  bool rat, clo, per;
  int deg, num_ks;
  SimpleList<Vector3> *cpts;
  Real *ws, cptst, *ks, ktol;
};

// 1D geometry.
class Point : public Geometry {
public:
  Point () : Geometry (0) {
    Entity::type (2, SIL_POINT_TYPE);
    pos = new Position ();
    pos->generator () = this;
    take (pos);
  }
  virtual ~Point () {
    drop (pos);
  }
  virtual string const type_id () const {
    return tid;
  }
  Position *position () const {
    return pos;
  }
  virtual void features (SimpleList<Feature> *fl) const {
    assert (fl);
    fl->add_first (pos);
    return;
  }
private:
  static string const tid;
  Position *pos;
};

#endif // !GEOMETRY_HXX

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
