// matrix.cxx: Matrix class implementation.
//
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: matrix.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (matrix, "$Id: matrix.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"
#include "xmath.hxx"
#include "kernel/matrix.hxx"
#include "kernel/vector.hxx"

string const RMatrix::tid = "RMatrix";

RMatrix::RMatrix (RMatrix const &m) : Entity () 
{
  Entity::type (1, SIL_RMATRIX_TYPE);
  r = m.r;
  c = m.c;
  if (m.d) {
    d = (Real *)malloc (sizeof (Real) * r * c);
    Real *P = d;
    Real *PP = d + r * c;
    Real *Q = m.d;
    while (P < PP)
      *(P++) = *(Q++);
  } else {
    d = NULL;
  }
}

void RMatrix::resize (unsigned long n, unsigned long m) 
{
  r = n;
  c = m;
  if (d) {
    if (r != 0 && c != 0) {
      d = (Real *)realloc (d, sizeof (Real) * r * c);
    } else {
      free (d);
      d = NULL;
    }
  } else {
    if (r != 0 && c != 0)
      d = (Real *)malloc (sizeof (Real) * r * c);
  }
  return;
}

void RMatrix::ones ()
{
  assert (d);
  Real *P = d;
  Real *PP = d + r * c;
  while (P < PP)
    *(P++) = 1.0;
  return;
}

void RMatrix::zeros ()
{
  assert (d);
  Real *P = d;
  Real *PP = d + r * c;
  while (P < PP)
    *(P++) = 0.0;
  return;
}

void RMatrix::unit ()
{
  assert (d);
  Real *P = d;
  Real *PP = d + r * c;
  while (P < PP)
    *(P++) = 0.0;
  P = d;
  while (P < PP) {
    *P = 1.0;
    P += c + 1;
  }
  return;
}

RMatrix RMatrix::transposed () const
{
  assert (d);
  RMatrix m (c, r);
  Real  *P = m.d;
  for (unsigned long l = 0; l < c; ++l) {
    for (unsigned long k = 0; k < r; ++k) {
      *(P++) = *(d + k * c + l);
    }
  }
  return m;
}

RMatrix const &RMatrix::operator= (RMatrix const &m1)
{
  if (d)
    free (d);
  r = m1.r;
  c = m1.c;
  if (m1.d) {
    d = new Real[r * c];
    Real *P = d;
    Real *PP = d + r * c;
    Real *Q = m1.d;
    while (P < PP)
      *(P++) = *(Q++);
  } else {
    d = NULL;
  }
  return *this;
}

RMatrix operator- (RMatrix const &m1)
{
  assert (m1.d);
  RMatrix m (m1.r, m1.c);
  Real *P = m.d;
  Real *PP = P + m.r * m.c;
  Real *Q = m1.d;
  while (P < PP)
    *(P++) *= -*(Q++);
  return m;
}

RMatrix operator+ (RMatrix const &m1, RMatrix const &m2)
{
  assert (m1.d && m2.d && m1.r == m2.r && m1.c == m2.c);
  RMatrix m (m1.r, m1.c);
  Real *P = m.d;
  Real *PP = m.d + m1.r * m1.c;
  Real *Q = m1.d;
  Real *R = m2.d;
  while (P < PP)
    *(P++) = *(Q++) + *(R++);
  return m;
}

RMatrix const &RMatrix::operator+= (RMatrix const &m1)
{
  assert (d && m1.d && m1.r == r && m1.c == c);
  Real *P = d;
  Real *PP = d + r * c;
  Real *Q = m1.d;
  while (P < PP)
    *(P++) += *(Q++);
  return *this;
}

RMatrix operator- (RMatrix const &m1, RMatrix const &m2)
{
  assert (m1.d && m2.d && m1.r == m2.r && m1.c == m2.c);
  RMatrix m (m1.r, m1.c);
  Real *P = m.d;
  Real *PP = m.d + m1.r * m1.c;
  Real *Q = m1.d;
  Real *R = m2.d;
  while (P < PP)
    *(P++) = *(Q++) - *(R++);
  return m;
}

RMatrix const &RMatrix::operator-= (RMatrix const &m1)
{
  assert (d && m1.d && m1.r == r && m1.c == c);
  Real *P = d;
  Real *PP = d + r * c;
  Real *Q = m1.d;
  while (P < PP)
    *(P++) -= *(Q++);
  return *this;
}

RMatrix operator* (Real s, RMatrix const &m1)
{
  assert (m1.d);
  RMatrix m (m1.r, m1.c);
  Real *P = m.d;
  Real *PP = P + m1.r * m1.c;
  Real *Q = m1.d;
  while (P < PP)
    *(P++) = s * (*(Q++));
  return m;
}

RMatrix operator* (RMatrix const &m1, Real s)
{
  assert (m1.d);
  RMatrix m (m1.r, m1.c);
  Real *P = m.d;
  Real *PP = P + m1.r * m1.c;
  Real *Q = m1.d;
  while (P < PP)
    *(P++) = *(Q++) * s;
  return m;
}

RMatrix const &RMatrix::operator*= (Real s)
{
  assert (d);
  Real *P = d;
  Real *PP = d + r * c;
  while (P < PP)
    *(P++) *= s;
  return *this;
}

RMatrix operator/ (RMatrix const &m1, Real s)
{
  assert (m1.d && s != 0.0);
  RMatrix m (m1.r, m1.c);
  Real *P = m.d;
  Real *PP = P + m1.r * m1.c;
  Real *Q = m1.d;
  while (P < PP)
    *(P++) = *(Q++) / s;
  return m;
}

RMatrix const &RMatrix::operator/= (Real s)
{
  assert (d && s != 0.0);
  Real *P = d;
  Real *PP = d + r * c;
  while (P < PP)
    *(P++) /= s;
  return *this;
}

RMatrix operator* (RMatrix const &m1, RMatrix const &m2)
{
  assert (m1.d && m2.d && m1.c == m2.r);
  RMatrix m (m1.r, m2.c);
  Real *P = m.d;
  Real *Q = m1.d;
  for (unsigned long l = 0; l < m1.r; ++l) {
    for (unsigned long k = 0; k < m2.c; ++k) {
      *P = 0.0;
      Real *QQ = Q;
      Real *R = m2.d + k;
      for (unsigned long j = 0; j < m1.c; ++j) {
	*P += *(QQ++) * (*R);
	R += m2.c;
      }
      P++;
    }
    Q += m1.c;
  }
  return m;
}

RMatrix RMatrix::operator*= (RMatrix const &m2)
{
  assert (d && m2.d && c == m2.r);
  Real *D = new Real[r * m2.c];
  Real *P = D;
  Real *Q = d;
  for (unsigned long l = 0; l < r; ++l) {
    for (unsigned long k = 0; k < m2.c; ++k) {
      *P = 0.0;
      Real *QQ = Q;
      Real *R = m2.d + k;
      for (unsigned long j = 0; j < c; ++j) {
	*P += *(QQ++) * (*R);
	R += m2.c;
      }
      P++;
    }
    Q += c;
  }
  delete d;
  d = D;
  c = m2.c;
  return *this;
}

Real operator% (RMatrix const &m1, RMatrix const &m2)
{
  assert (m1.d && m2.d && m1.r == m2.r && m1.c == 1 && m2.c == 1);
  Real r = 0.0;
  Real *P = m1.d;
  Real *PP = m1.d + m1.r;
  Real *Q = m2.d;
  while (P < PP)
    r += *(P++) * (*(Q++));
  return r;
}

bool operator== (RMatrix const &m1, RMatrix const &m2)
{
  if (m1.d == NULL && m2.d == NULL)
    return true;
  if (m1.r == m2.r && m1.c == m2.c) {
    Real *P = m1.d;
    Real *PP = m1.d + m1.r * m1.c;
    Real *Q = m2.d;
    while (P < PP) {
      if (*(P++) != *(Q++))
	return false;
    }
    return true;
  }
  return false;
}

bool operator!= (RMatrix const &m1, RMatrix const &m2)
{
  if ((m1.d == NULL && m2.d != NULL) ||
      (m1.d == NULL && m2.d != NULL))
    return true;
  if (m1.r == m2.r && m1.c == m2.c) {
    Real *P = m1.d;
    Real *PP = m1.d + m1.r * m1.c;
    Real *Q = m2.d;
    while (P < PP) {
      if (*(P++) != *(Q++))
	return true;
    }
    return false;
  }
  return true;
}

ostream &operator<< (ostream &out, RMatrix const &m)
{
  if (m.d) {
    out << '[';
    Real *P = m.d;
    for (unsigned long l = 0; l < m.r; ++l) {
      for (unsigned long k = 0; k < m.c; ++k) {
	out << *(P++);
	if (k < m.c - 1)
	  out << ", ";
      }
      if (l < m.r - 1)
	out << ";\n ";
    }
    out << ']';
  } else {
    out << "[]";
  }
  return out;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
