// matrix.hxx: Matrix templates.
// 
// Copyright (C) 2000-2001 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: matrix.hxx,v 1.10 2001/04/19 15:23:44 scmfcl Exp $
//

#ifndef MATRIX_HXX
#define MATRIX_HXX 1

#include "kernel/entity.hxx"

class RMatrix : public Entity {
public:
  RMatrix (unsigned long n = 0, unsigned long m = 0) : Entity () {
    Entity::type (1, SIL_RMATRIX_TYPE);
    r = n;
    c = m;
    if (n > 0 && m > 0)
      d = (Real *)malloc (sizeof (Real) * n * m);
    else
      d = NULL;
  }
  RMatrix (RMatrix const &m);
  virtual ~RMatrix () {
    if (d)
      free (d);
  }
  virtual string const type_id () const {
    return tid;
  }
  unsigned long rows () const {
    return r;
  }
  unsigned long columns () const {
    return c;
  }
  void resize (unsigned long n, unsigned long m);
  bool is_empty () const {
    return d == NULL;
  }
  Real operator () (unsigned long l, unsigned long k) const {
    assert (d && l < r && k < c);
    return *(d + l * c + k);
  }
  Real &operator () (unsigned long l, unsigned long k) {
    assert (d && l < r && k < c);
    return *(d + l * c + k);
  }
  Real operator () (unsigned long l) const {
    assert (d);
    return *(d + l);
  }
  Real &operator () (unsigned long l) {
    assert (d);
    return *(d + l);
  }
  void ones ();
  void zeros ();
  void unit ();
  RMatrix transposed () const;
  RMatrix const &RMatrix::operator= (RMatrix const &m1);
  friend RMatrix operator- (RMatrix const &m1);
  friend RMatrix operator+ (RMatrix const &m1, RMatrix const &m2);
  RMatrix const &operator+= (RMatrix const &m1);
  friend RMatrix operator- (RMatrix const &m1, RMatrix const &m2);
  RMatrix const &operator-= (RMatrix const &m1);
  friend RMatrix operator* (Real s, RMatrix const &m1);
  friend RMatrix operator* (RMatrix const &m1, Real s);
  RMatrix const &operator*= (Real s);
  friend RMatrix operator/ (RMatrix const &m1, Real s);
  RMatrix const &operator/= (Real s);
  friend RMatrix operator* (RMatrix const &m1, RMatrix const &m2);
  RMatrix operator*= (RMatrix const &m2);
  friend Real operator% (RMatrix const &m1, RMatrix const &m2);
  friend bool operator== (RMatrix const &m1, RMatrix const &m2);
  friend bool operator!= (RMatrix const &m1, RMatrix const &m2);
  friend ostream &operator<< (ostream &out, RMatrix const &m);
private:
  static string const tid;
  Real *d;
  unsigned long r, c;
};

template<class T>
class HalfMatrix {
public:
  HalfMatrix (unsigned long m) {
    assert (m > 1);
    n = m;
    dt = new T[(n * (n - 1)) >> 1];
  }
  ~HalfMatrix () {
    delete dt;
  }
  T operator () (unsigned long l, unsigned long k) const {
    assert (l > k && n > l);
    return *(dt + ((l * (l - 1)) >> 1) + k);
  }
  T &operator () (unsigned long l, unsigned long k) {
    assert (l > k && n > l);
    return *(dt + ((l * (l - 1)) >> 1) + k);
  }
  T operator () (unsigned long i) const {
    return *(dt + i);
  }
  T &operator () (unsigned long i) {
    return *(dt + i);
  }
  unsigned long size () const {
    return n;
  }
  unsigned long index (unsigned long l, unsigned long k) const {
    assert (l > k && n > l);
    return (l * (l - 1) >> 1) + k;
  }
  void rev_index (unsigned long i, unsigned long &l, unsigned long &k);
  ostream &print (ostream &out, T const max) const;
private:
  unsigned long n;
  T *dt;
};

template<class T>
void HalfMatrix<T>::rev_index (unsigned long i,
			       unsigned long &l, unsigned long &k)
{
  l = 1;
  k = i;
  while (k >= l)
    k -= l++;
  return;
}

template<class T>
ostream &HalfMatrix<T>::print (ostream &out, T const max) const
{
  out << "HalfMatrix [" << endl;
  bool c = true;
  unsigned long l = 1;
  while (c) {
    out << "  ";
    bool cc = true;
    unsigned long k = 0;
    while (cc) {
      T *d = dt + ((l * (l - 1)) >> 1) + k;
      if (*d < max)
	out << *d;
      else
	out << "*";
      ++k;
      if (k < l)
	out << ", ";
      else
	cc = false;
    }
    ++l;
    if (l < n) {
      out << ";" << endl;
    } else {
      out << "]" << endl;
      c = false;
    }
  }
  return out;
}

template<class T>
ostream &operator<< (ostream &out, HalfMatrix<T> const &m)
{
  out << "HalfMatrix [" << endl;
  bool c = true;
  unsigned long l = 1;
  while (c) {
    out << "  ";
    unsigned long k = 0;
    bool cc = true;
    while (cc) {
      out << m (l, k);
      ++k;
      if (k < l)
	out << ", ";
      else
	cc = false;
    }
    ++l;
    if (l < m.size ()) {
      out << ";" << endl;
    } else {
      out << "]" << endl;
      c = false;
    }
  }
  return out;
}

#endif // MATRIX_HXX

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
