// topology.cxx: topology classes implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: topology.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (topology, "$Id: topology.cc 1 2005-08-10 23:35:07Z frank $");

#include "xmath.hxx"
#include "kernel/topology.hxx"

string const Solid::tid = "Solid";
string const Lump::tid = "Lump";
string const Shell::tid = "Shell";
string const Face::tid = "Face";
string const Loop::tid = "Loop";
string const Edge::tid = "Edge";
string const Vertex::tid = "Vertex";

void Vertex::set_point (Point *p) 
{
  if (p == pnt)
    return;
  if (p) {
    assert (p->cell () == NULL);
    take (p);
    p->cell () = this;
  }
  if (pnt) {
    pnt->cell () = NULL;
    drop (pnt);
  }
  pnt = p;
  return;
}

void Edge::set_curve (Curve *c) 
{
  if (c == crv)
    return;
  if (c) {
    assert (c->cell () == NULL);
    take (c);
    c->cell () = this;
  }
  if (crv) {
    crv->cell () = NULL;
    drop (crv);
  }
  crv = c;
  return;
}

void Edge::set_start_vertex (Vertex *v) 
{
  if (v == start_vrtx)
    return;
  take (v);
  drop (start_vrtx);
  start_vrtx = v;
  return;
}

void Edge::set_end_vertex (Vertex *v) 
{
  if (v == end_vrtx)
    return;
  take (v);
  drop (end_vrtx);
  end_vrtx = v;
  return;
}

void Edge::features (SimpleList<Feature> *fl)
{
  if (crv) {
    // Geometry features
    crv->features (fl);
    // Boundary features considering geometry
    if (start_vrtx && end_vrtx && 
	start_vrtx->point () && end_vrtx->point ()) {
      if (crv->is (SIL_STRAIGHT_TYPE)) {
	drop (f_angl);
	f_angl = NULL;
	// Straight parameters: direction, length
	if (f_dir) {
	  f_dir->set (*end_vrtx->point () -> position () -
		      *start_vrtx->point () -> position ());
	} else {
	  f_dir = new Direction (*end_vrtx->point () -> position () -
				 *start_vrtx->point () -> position ());
	  f_dir->generator () = this;
	  take (f_dir);
	}
	fl->add_first (f_dir);
	if (f_len) {
	  f_len->value () = (*start_vrtx->point () -> position () -
			     *end_vrtx->point () -> position ()) . norm ();
	} else {
	  f_len = new Length ((*start_vrtx->point () -> position () -
			       *end_vrtx->point () -> position ()) . norm ());
	  f_len->generator () = this;
	  take (f_len);
	}
	fl->add_first (f_len);
      } else if (crv->is (SIL_CIRCLE_TYPE)) {
	drop (f_dir);
	f_dir = NULL;
	drop (f_len);
	f_len = NULL;
	// Circle parameter: arc angle.
	Circle *circ = static_cast<Circle *> (crv);
	Vector3 a = *start_vrtx->point () -> position () - *circ->centre ();
	Vector3 b = *end_vrtx->point () -> position () - *circ->centre ();
	Real angl = (a % b) / (a.norm () * b.norm ());
	if (angl >= 1.0) {
	  angl = 0.0;
	} else if (angl <= -1.0) {
	  angl = PI;
	} else {
	  angl = acos (angl);
	}
	// If normal of circle point in different direction, then we have
	// the other part of the circle.
	if (((a * b) % *circ->normal ()) < 0.0)
	  angl = 2 * PI - angl;
	if (f_angl) {
	  f_angl->value () = angl;
	} else {
	  f_angl = new Angle (angl);
	  f_angl->generator () = this;
	  take (f_angl);
	}
	fl->add_first (f_angl);
      }
    } else {
      drop (f_dir);
      f_dir = NULL;
      drop (f_len);
      f_len = NULL;
      drop (f_angl);
      f_angl = NULL;
    }
  } else {
    drop (f_dir);
    f_dir = NULL;
    drop (f_len);
    f_len = NULL;
    drop (f_angl);
    f_angl = NULL;
  }
  return;
}

Direction *Edge::get_axes (SimpleList<Position> *pos)
{
  if (crv) {
    if (crv->is (SIL_STRAIGHT_TYPE)) {
      if (f_dir)
	return f_dir;
    } else if (crv->is (SIL_CIRCLE_TYPE)) {
      pos->add_first (((Circle *)crv)->centre ());
      return ((Circle *)crv)->normal ();
    } else if (crv->is (SIL_ELLIPSE_TYPE)) {
      pos->add_first (((Ellipse *)crv)->centre ());
      return ((Ellipse *)crv)->normal ();
    }
  }
  return NULL;
}

Direction *Face::get_axes (SimpleList<Position> *pos)
{
  if (srfc) {
    if (srfc->is (SIL_PLANE_TYPE)) {
      unsigned long n = pos->length ();
      XListNode<Loop> *l = lps.first ();
      while (l) {
	l->data () -> get_root_points (pos);
	l = l->next ();
      }
      if (pos->length () > n)
	return ((Plane *)srfc)->direction ();
    } else if (srfc->is (SIL_CYLINDER_TYPE)) {
      pos->add_first (srfc->root_point ());
      return ((Cylinder *)srfc)->direction ();
    } else if (srfc->is (SIL_CONE_TYPE)) {
      pos->add_first (srfc->root_point ());
      return ((Cone *)srfc)->direction ();
    } else if (srfc->is (SIL_TORUS_TYPE)) {
      pos->add_first (srfc->root_point ());
      return ((Torus *)srfc)->direction ();
    }
  }
  return NULL;
}

void Loop::get_root_points (SimpleList<Position> *rps) const
{
  // Add root points which have been computed before (by features method).
  assert (rps);
  if (edgs.length () == 1) {
    Curve *c = edgs.first () -> data () -> curve ();
    if (c) {
      // Add circle and ellipse centre as root points.
      if (c->is (SIL_CIRCLE_TYPE)) {
	rps->add_first (static_cast<Circle *> (c) -> centre ());
      } else if (c->is (SIL_ELLIPSE_TYPE)) {
	rps->add_first (static_cast<Ellipse *> (c) -> centre ());
      }
    }
  } else if (edgs.length () > 1 && f_rp) {
    rps->add_first (f_rp);
  }
  return;
}

void Loop::features (SimpleList<Feature> *fl)
{
  assert (fl);
  if (edgs.length () == 1) {
    drop (f_rp);
    f_rp = NULL;
    drop (f_lp);
    f_lp = NULL;
    Curve *c = edgs.first () -> data () -> curve ();
    if (c) {
      // Add circle and ellipse centre as root points.
      if (c->is (SIL_CIRCLE_TYPE)) {
	fl->add_first (static_cast<Circle *> (c) -> centre ());
      } else if (c->is (SIL_ELLIPSE_TYPE)) {
	fl->add_first (static_cast<Ellipse *> (c) -> centre ());
      }
    }
  } else if (edgs.length () > 1) {
    // Add further root points for planar surface (checked by cell) with
    // straight edges: centre of polygon, convex hull, etc.
    //
    // First check if loop has only straight edges.
    XListNode<Edge> *e = edgs.first ();
    while (e) {
      if (!e->data () -> curve () ||
	  !e->data () -> curve () -> is (SIL_STRAIGHT_TYPE)) {
	drop (f_rp);
	f_rp = NULL;
	drop (f_lp);
	f_lp = NULL;
	return;
      }
      e = e->next ();
    }
    // Now compute the centre of the polygon.
    SimpleXList<Position> *pts = new SimpleXList<Position>;
    Vector3 c (0.0, 0.0, 0.0);
    e = edgs.first ();
    bool sen = sense ();
    while (e) {
      Position *p = NULL;
      XListNode<Edge> *ne = e->next ();
      if (ne) {
	if (sen) {
	  p = e->data () -> end_vertex () -> point () -> position ();
	  if (e->data () -> start_vertex () ==
	      ne->data () -> start_vertex ()) {
	    sen = false;
	  } else if (e->data () -> start_vertex () == 
		     ne->data () -> end_vertex ()) {
	    sen = true;
	  } else {
	    sil_msg_error ("Loop", "edge sequence is broken");
	    delete pts;
	    return;
	  }
	} else {
	  p = e->data () -> start_vertex () -> point () -> position ();
	  if (e->data () -> end_vertex () == ne->data () -> start_vertex ()) {
	    sen = false;
	  } else if (e->data () -> end_vertex () ==
		     ne->data () -> end_vertex ()) {
	    sen = true;
	  } else {
	    sil_msg_error ("Loop", "edge sequence is broken");
	    delete pts;
	    return;
	  }
	}
	e = ne;
      } else {
	ne = edgs.first ();
	if (sen) {
	  p = e->data () -> end_vertex () -> point () -> position ();
	  if (e->data () -> start_vertex () == 
	      ne->data () -> start_vertex ()) {
	    sen = false;
	  } else if (e->data () -> start_vertex () ==
		     ne->data () -> end_vertex ()) {
	    sen = true;
	  } else {
	    sil_msg_error ("Loop", "edge sequence is broken");
	    delete pts;
	    return;
	  }
	} else {
	  p = e->data () -> start_vertex () -> point () -> position ();
	  if (e->data () -> end_vertex () == ne->data () -> start_vertex ()) {
	    sen = false;
	  } else if (e->data () -> end_vertex () ==
		     ne->data () -> end_vertex ()) {
	    sen = true;
	  } else {
	    sil_msg_error ("Loop", "edge sequence is broken");
	    delete pts;
	    return;
	  }
	}
	e = NULL;
      }
      c += *p;
      pts->add_last (p);
    }
    c /= pts->length ();
    if (f_rp) {
      f_rp->set (c);
    } else {
      f_rp = new Position (c);
      f_rp->generator () = this;
      f_rp->id () = 1;
      take (f_rp);
    }
    fl->add_first (f_rp);
    // FIXME: add center of convex hull and/or
    //            center of completed convex hull.
    if (f_lp) {
      f_lp->set (pts);
    } else {
      f_lp = new LoopPolygon (pts);
      f_lp->generator () = this;
      take (f_lp);
    }
    fl->add_first (f_lp);
  } else {
    drop (f_rp);
    f_rp = NULL;
    drop (f_lp);
    f_lp = NULL;
  }
  return;
}

void Face::set_surface (Surface *s) 
{
  if (s == srfc)
    return;
  if (s) {
    assert (s->cell () == NULL);
    take (s);
    s->cell () = this;
  }
  if (srfc) {
    srfc->cell () = NULL;
    drop (srfc);
  }
  srfc = s;
  return;
}

void Face::features (SimpleList<Feature> *fl)
{
  assert (fl);
  if (srfc) {
    // Surface geometry features
    srfc->features (fl);
    // Root points for planar surfaces.
    if (srfc->is (SIL_PLANE_TYPE)) {
      XListNode<Loop> *n = lps.first ();
      while (n) {
	n->data () -> features (fl);
	n = n->next ();
      }
    }
  }
  return;
}

void Lump::set_space (Space *s)
{
  if (s == spc)
    return;
  if (s) {
    assert (s->cell () == NULL);
    take (s);
    s->cell () = this;
  }
  if (spc) {
    spc->cell () = NULL;
    drop (spc);
  }
  spc = s;
  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
