// topology.hxx: Solid model topology classes.
//
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: topology.hxx,v 1.30 2001/04/17 19:25:34 scmfcl Exp $

#ifndef TOPOLOGY_HXX
#define TOPOLOGY_HXX 1

#include "list.hxx"
#include "material.hxx"
#include "kernel/xlist.hxx"
#include "kernel/entity.hxx"
#include "kernel/geometry.hxx"

class Cell;
class Boundary;
class Face;

// Connected component of a cell boundary.
class Boundary {
public:
  explicit Boundary (unsigned short d) {
    dim = d;
    sns = false;
  }
  virtual ~Boundary () {
  }
  unsigned short dimension () const {
    return dim;
  }
  bool sense () const {
    // If true, reverse orientation of cells creating the boundary.
    return sns;
  }
  bool &sense () {
    return sns;
  }
private:
  unsigned short dim;
  bool sns;
};

// A (connected) cell, i.e. a set C = closure(interior(C)) represented
// by a geometry and a list of connected boundary componenets. Each
// cell is connected and it's geometry and it's boundary components
// are unique, i.e. not used by any other cell.
class Cell {
public:
  explicit Cell (unsigned short d) {
    dim = d;
    ref = NULL;
    m = NULL;
  }
  virtual ~Cell () {
  }
  unsigned short dimension () const {
    return dim;
  }
  Material *&material () {
    return m;
  }
  Material *material () const {
    return m;
  }
  virtual void *reference () const {
    // A reference to another model (SiL/ACIS conversion, etc.)
    return ref;
  }
  virtual void *&reference () {
    return ref;
  }
  virtual Geometry *geometry () const = 0;
  // Returns features of the cell, i.e. of the geometry and the
  // boundary components. Can create additional features based on
  // this.
  virtual void features (SimpleList<Feature> *fl) = 0;
private:
  unsigned short dim;
  void *ref;
  Material *m;
};

// Vertex: a combination of a 0 dimensional cell and boundary.
class Vertex : public Entity, public Cell, public Boundary {
public:
  Vertex () : Entity (), Cell (0), Boundary (0) {
    Entity::type (1, SIL_VERTEX_TYPE);
    pnt = NULL;
  }
  virtual ~Vertex () {
    drop (pnt);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Geometry *geometry () const {
    return pnt;
  }
  Point *point () const {
    return pnt;
  }
  void set_point (Point *p);
  virtual void features (SimpleList<Feature> *fl) {
    assert (fl);
    if (pnt)
      pnt->features (fl);
    return;
  }
private:
  static string const tid;
  Point *pnt;
};

// A 1 dimensional cell.
class Edge : public Entity, public Cell {
public:
  Edge () : Entity (), Cell (1) {
    Entity::type (1, SIL_EDGE_TYPE);
    start_vrtx = end_vrtx = NULL;
    crv = NULL;
    f_dir = NULL;
    f_len = NULL;
    f_angl = NULL;
    icurv = NULL;
  }
  virtual ~Edge () {    
    drop (start_vrtx);
    drop (end_vrtx);
    drop (crv);
    drop (f_dir);
    drop (f_len);
    drop (f_angl);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Geometry *geometry () const {
    return crv;
  }
  Curve *curve () const {
    return crv;
  }
  void set_curve (Curve *c);
  Vertex *start_vertex () const {
    return start_vrtx;
  }
  void set_start_vertex (Vertex *v);
  Vertex *end_vertex () const {
    return end_vrtx;
  }
  void set_end_vertex (Vertex *v);
  virtual void features (SimpleList<Feature> *fl);
  void *intersection_curve () const {
    // A reference to an intersection curve computed by some method
    // (ACIS reference).
    return icurv;
  }
  void *&intersection_curve () {
    return icurv;
  }
  Direction *direction () const {
    return f_dir;
  }
  Direction *get_axes (SimpleList<Position> *pos);
private:
  static string const tid;
  Vertex *start_vrtx, *end_vrtx;
  Curve *crv;
  Direction *f_dir;
  Length *f_len;
  Angle *f_angl;
  void *icurv;
};

// A 1 dimensional boundary component
class Loop : public Entity, public Boundary {
public:
  Loop () : Entity (), Boundary (1) {
    Entity::type (1, SIL_LOOP_TYPE);
    f_rp = NULL;
    f_lp = NULL;
  }
  virtual ~Loop () {
    drop (f_rp);
    drop (f_lp);
  }
  virtual string const type_id () const {
    return tid;
  }
  XList<Edge> edges () const {
    // Edges are traversed for in sequence the sequence they are in
    // this list with respect to sense (if true, reverse first edge
    // and sequence accordingly). If the edge is closed sense tells
    // if the geometry is traversed reversed (if sense is true).
    return edgs;
  }
  XList<Edge> &edges () {
    return edgs;
  }
  void get_root_points (SimpleList<Position> *rps) const;
protected:
  void features (SimpleList<Feature> *fl);
  friend Face;
private:
  static string const tid;
  XList<Edge> edgs;
  Position *f_rp;
  LoopPolygon *f_lp;
};

// A 2 dimensional cell.
class Face : public Entity, public Cell {
public:
  Face () : Entity (), Cell (2) {
    Entity::type (1, SIL_FACE_TYPE);
    srfc = NULL;
  }
  virtual ~Face () {
    drop (srfc);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Geometry *geometry () const {
    return srfc;
  }
  XList<Loop> loops () const {
    return lps;
  }
  XList<Loop> &loops () {
    return lps;
  }
  Surface *surface () const {
    return srfc;
  }
  void set_surface (Surface *s);
  virtual void features (SimpleList<Feature> *fl);
  Direction *get_axes (SimpleList<Position> *pos);
private:
  static string const tid;
  XList<Loop> lps;
  Surface *srfc;
};

// A 2 dimensional boundary component.
class Shell : public Entity, public Boundary {
public:
  Shell () : Entity (), Boundary (2) {
    Entity::type (1, SIL_SHELL_TYPE);
  }
  virtual ~Shell () {
  }
  virtual string const type_id () const {
    return tid;
  }
  XList<Face> faces () const {
    return fcs;
  }
  XList<Face> &faces () {
    return fcs;
  }
private:
  static string const tid;
  XList<Face> fcs;
};

// A three dimensional cell.
class Lump : public Entity, public Cell {
public:
  Lump () : Entity (), Cell (3) {
    Entity::type (1, SIL_LUMP_TYPE);
    spc = NULL;
  }
  virtual ~Lump () {
    drop (spc);
  }
  virtual string const type_id () const {
    return tid;
  }
  virtual Geometry *geometry () const {
    return spc;
  }
  Space *space () const {
    return spc;
  }
  void set_space (Space *s);
  XList<Shell> shells () const {
    return shs;
  }
  XList<Shell> &shells () {
    return shs;
  }
  virtual void features (SimpleList<Feature> *fl) {
    assert (fl);
    if (spc)
      spc->features (fl);
    return;
  }
private:
  static string const tid;
  XList<Shell> shs;
  Space *spc;
};

// A solid represented as a collection of 3 dimensional cells, which
// makes it sort of a 3 dimensional boundary component.
class Solid : public Entity, public Boundary {
public:
  Solid () : Entity (), Boundary (3) {
    Entity::type (1, SIL_SOLID_TYPE);
    m = NULL;
  }
  virtual ~Solid () {
  }
  virtual string const type_id () const {
    return tid;
  }
  XList<Lump> lumps () const {
    return lmps;
  }
  XList<Lump> &lumps () {
    return lmps;
  }
  List<Shell> shells () const {
    return shlls;
  }
  List<Shell> &shells () {
    return shlls;
  }
  List<Face> faces () const {
    return fcs;
  }
  List<Face> &faces () {
    return fcs;
  }
  List<Loop> loops () const {
    return lps;
  }
  List<Loop> &loops () {
    return lps;
  }
  List<Edge> edges () const {
    return edgs;
  }
  List<Edge> &edges () {
    return edgs;
  }
  List<Vertex> vertices () const {
    return vrtcs;
  }
  List<Vertex> &vertices () {
    return vrtcs;
  }
  Material *&material () {
    return m;
  }
  Material *material () const {
    return m;
  }
private:
  static string const tid;
  XList<Lump> lmps;
  List<Shell> shlls;
  List<Face> fcs;
  List<Loop> lps;
  List<Edge> edgs;
  List<Vertex> vrtcs;
  Material *m;
};

#endif // !TOPOLOGY_HXX

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
