// xlist.hxx: XList template.
// 
// Copyright (C) 1998-2001 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: xlist.hxx,v 1.10 2001/04/19 15:23:44 scmfcl Exp $
//

#ifndef XLIST_HXX
#define XLIST_HXX 1

#include "kernel/memory.hxx"

template<class T> class XList;

template<class T>
class XListNode {
public:
  XListNode (T *d, XList<T> *l) {
    take ((Entity *)d);
    dt = d;
    lst = l;
    prv = nxt = NULL;
  }
  virtual ~XListNode () {
    drop ((Entity *)dt);
  }
  XListNode<T> *previous () const {
    return prv;
  }
  XListNode<T> *next () const {
    return nxt;
  }
  XListNode<T> *circ_next () const {
    if (nxt)
      return nxt;
    return list->first ();
  }
  XListNode<T> *circ_previous () const {
    if (prv)
      return prv;
    return list->last ();
  }
  void unlink ();
  void add_after (T *e);
  void add_before (T *e);
  T *data () const {
    return dt;
  }
  XList<T> *list () const {
    return lst;
  }
private:
  XListNode<T> *nxt, *prv;
  XList<T> *lst;
  T *dt;
};

template<class T>
class XList {
public:
  inline XList () {
    frst = lst = NULL;
    len = 0;
  };
  virtual ~XList ();
  void add_first (T *d);
  bool add_first_check (T *d);
  void add_last (T *d);
  bool add_last_check (T *d);
  bool del (T const *D);
  XListNode<T> *first () const {
    return frst;
  }
  XListNode<T> *last () const {
    return lst;
  }
  XListNode<T> *unlink_first () {
    XListNode<T> *h = frst;
    if (h)
      frst->unlink ();
    return h;
  }
  XListNode<T> *find (T const *d) const;
  bool is_member (T const *d) {
    return find (d) != NULL;
  }
  void purge ();
  bool empty () const {
    return (frst == NULL);
  }
  unsigned long length () const {
    return len;
  }
protected:
  friend XListNode<T>;
  XListNode<T> *frst, *lst;
  unsigned long len;
};

template<class T>
XList<T>::~XList ()
{
  XListNode<T> *w = frst;
  while (w) {
    XListNode<T> *ww = w->next ();
    delete w;
    w = ww;
  }
}

template<class T>
void XList<T>::purge ()
{
  XListNode<T> *w = frst;
  while (w) {
    XListNode<T> *ww = w->next ();
    delete w;
    w = ww;
  }
  lst = frst = NULL;
  len = 0;
  return;
}

template<class T>
void XList<T>::add_first (T * e)
{
  if (!frst) {
    frst = lst = new XListNode<T> (e, this);
    ++len;
    return;
  }
  frst->add_before (e);
  return;
}

template<class T>
bool XList<T>::add_first_check (T * e)
{
  if (!frst) {
    frst = lst = new XListNode<T> (e, this);
    ++len;
    return true;
  }
  XListNode<T> *w = frst;
  while (w) {
    if (e == w->data ())
      return false;
    w = w->next ();
  }
  frst->add_before (e);
  return true;
}

template<class T>
void XList<T>::add_last (T * e)
{
  if (!frst) {
    frst = lst = new XListNode<T> (e, this);
    ++len;
    return;
  }
  lst->add_after (e);
  return;
}

template<class T>
bool XList<T>::add_last_check (T * e)
{
  if (!frst) {
    frst = lst = new XListNode<T> (e, this);
    ++len;
    return true;
  }
  XListNode<T> *w = frst;
  while (w) {
    if (e == w->data ())
      return false;
    w = w->next ();
  }
  lst->add_after (e);
  return true;
}

template<class T>
bool XList<T>::del (T const *e)
{
  XListNode<T> *w = frst;
  bool res = false;
  while (w) {
    XListNode<T> *ww = w->next ();
    if (e == w->data ()) {
      w->unlink ();
      delete w;
      res = true;
    }
    w = ww;
  }
  return res;
}

template<class T>
XListNode<T> *XList<T>::find (T const *e) const
{
  XListNode<T> *w = frst;
  while (w) {
    if (e == w->data ())
      return w;
    w = w->next ();
  }
  return NULL;
}

template<class T>
void XListNode<T>::add_after (T *e) 
{
  XListNode<T> *n = new XListNode<T>(e, lst);
  ++lst->len;
  if (nxt) {
    nxt->prv = n;
    n->nxt = nxt;
    nxt = n;
    n->prv = this;
  } else {
    nxt = n;
    n->prv = this;
    lst->lst = n;
  } 
  return;
}

template<class T>
void XListNode<T>::add_before (T *e)
{
  XListNode<T> *n = new XListNode<T>(e, lst);
  ++lst->len;
  if (prv) {
    prv->nxt = n;
    n->prv = prv;
    prv = n;
    n->nxt = this;
  } else {
    prv = n;
    n->nxt = this;
    lst->frst = n;
  }
  return;
}

template<class T>
void XListNode<T>::unlink ()
{
  --lst->len;
  if (prv) {
    prv->nxt = nxt;
  } else {
    lst->frst = nxt;
  }
  if (nxt) {
    nxt->prv = prv;
  } else {
    lst->lst = prv;
  }
  nxt = prv = NULL;
  return;
}

template<class T> class SimpleXList;

template<class T>
class SimpleXListNode {
protected:
  SimpleXListNode (T *e) : element (e) {
    take ((Entity *)e);
  }
  ~SimpleXListNode () {
    drop ((Entity *)element);
  }
  T *element;
  SimpleXListNode<T> *next;
  friend SimpleXList<T>;
};

template<class T>
class SimpleXList {
public:
  inline SimpleXList () {
    head = window = NULL;
    l = 0;
  }
  ~SimpleXList ();
  void add_first (T *e);
  bool add_first_check (T *e);
  void add_last (T *e);
  bool add_last_check (T *e);
  bool del (T const *e);
  void purge ();
  T *first ();
  T *next ();
  T *circ_next ();
  void reverse ();
  bool is_member (T const *d) const;
  void combine (SimpleXList<T> *l);
  inline bool empty () const {
    return head == NULL;
  }
  inline unsigned long length () const {
    return l;
  }
private:
  SimpleXListNode<T> *head;
  SimpleXListNode<T> *window;
  unsigned long l;
};

template<class T>
SimpleXList<T>::~SimpleXList ()
{
  while (head) {
    window = head->next;
    delete head;
    head = window;
  }
}

template<class T>
void SimpleXList<T>::add_first (T *e)
{
  SimpleXListNode<T> *w = new SimpleXListNode<T> (e);
  w->next = head;
  head = w;
  ++l;
  return;
}

template<class T>
bool SimpleXList<T>::add_first_check (T *e)
{
  SimpleXListNode<T> *w = head;
  SimpleXListNode<T> *pw = NULL;
  while (w) {
    if (e == w->element)
      return false;
    pw = w;
    w = w->next;
  }
  w = new SimpleXListNode<T> (e);
  w->next = head;
  head = w;
  ++l;
  return true;
}

template<class T>
void SimpleXList<T>::add_last (T *e)
{
  SimpleXListNode<T> *w = head;
  SimpleXListNode<T> *pw = NULL;
  while (w) {
    pw = w;
    w = w->next;
  }
  w = new SimpleXListNode<T> (e);
  w->next = NULL;
  if (pw) 
    pw->next = w;
  else
    head = w;
  ++l;
  return;
}

template<class T>
bool SimpleXList<T>::add_last_check (T *e)
{
  SimpleXListNode<T> *w = head;
  SimpleXListNode<T> *pw = NULL;
  while (w) {
    if (e == w->element)
      return false;
    pw = w;
    w = w->next;
  }
  w = new SimpleXListNode<T> (e);
  w->next = NULL;
  if (pw) 
    pw->next = w;
  else
    head = w;
  ++l;
  return true;
}

template<class T>
bool SimpleXList<T>::del (T const *e)
{
  SimpleXListNode<T> *w = head;
  SimpleXListNode<T> *p = NULL;
  while (w) {
    if (e == w->element) {
      if (p)
	p->next = w->next;
      else
	head = w->next;
      delete w;
      --l;
      return true;
    }
    p = w;
    w = w->next;
  }
  return false;
}

template<class T>
void SimpleXList<T>::purge ()
{
  SimpleXListNode<T> *w = head;
  SimpleXListNode<T> *p;
  while (w) {
    p = w->next;
    delete w;
    w = p;
  }
  head = window = NULL;
  return;
}

template <class T>
T *SimpleXList<T>::first ()
{
  window = head;
  if (!window)
    return NULL;
  return window->element;
}

template<class T>
T *SimpleXList<T>::next ()
{
  if (window && window->next) {
    window = window->next;
    return window->element;
  }
  window = NULL;
  return NULL;
}

template<class T>
T *SimpleXList<T>::circ_next ()
{
  if (windeow && window->next)
    window = window->next;
  else
    window = head;
  return window->element;
}

template<class T>
void SimpleXList<T>::reverse ()
{
  SimpleXListNode<T> *newhead = NULL;
  SimpleXListNode<T> *auxnode;
  while (head) {
    auxnode = head->next;
    head->next = newhead;
    newhead = head;
    head = auxnode;
  }
  head = newhead;
  return;
}

template<class T>
bool SimpleXList<T>::is_member (T const *e) const
{
  SimpleXListNode<T> *w = head;
  while (w) {
    if (e == w->element)
      return true;
    w = w->next;
  }
  return false;
}

template<class T>
void SimpleXList<T>::combine (SimpleXList<T> *list)
{
  SimpleXListNode<T> *w = head;
  SimpleXListNode<T> *p = NULL;
  while (w) {
    p = w;
    w = w->next;
  }
  if (p)
    p->next = list->head;
  else
    head = list->head;
  list->head = NULL;
  l += list->l;
  list->l = 0;
  return;
}

#endif /* XLIST_HXX */

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
