// main.cc: Main function for SiL.
// 
// Copyright (C) 1999-2001 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// $Id: main.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (main, "$Id: main.cc 1 2005-08-10 23:35:07Z frank $");

#include "sil_getopt.h"
#include "interface.hxx"
#include "app.hxx"

char *plugin_path = NULL;
char *module_path = NULL;
char *load_modules = NULL;

char *tmp_dir = NULL;

double EPSILON = DEFAULT_EPSILON;
double EPSILON2 = EPSILON * EPSILON;

double sens_angle = DEFAULT_SENS_ANGLE;
double sens_move = DEFAULT_SENS_MOVE;
double sens_dist = DEFAULT_SENS_DIST;

double render_box_min = DEFAULT_RENDER_BOX_MIN;
double render_box_max = DEFAULT_RENDER_BOX_MAX;

/** New error handler.
 *
 * This handler is called if a memory allocation with `new' failed.
 * It simply issues an error message and exits the program.
 *
 * Modifying this program to avoid crashes where a workaround for not
 * enough memory is possible would be highly recommended. For instance
 * those situations would be a memory fault for the OpenGL feedback
 * buffer. 
 */
void out_of_memory_handler ()
{
  sil_msg_error ("C++", "memory allocation failed");
  exit (1);
}

/** Not-a-number workaround test function.
 *
 * If we couldn't find any "not--a--number" (NaN) related code, we
 * have to emulate it. This function checks if x is NaN by comparing
 * it's bytes with the lilit_nan array, which represents the NaN code
 * according to the IEEE standard.
 *
 * @param x Double number.
 * @return A non-zero value if x is "not--a--number" (NaN), 0 otherwise.
 */
#ifdef USE_SIL_ISNAN
extern "C" {
int isnan (double x)
{
  unsigned char *y = (unsigned char *) &x;
  if (y[0] == sil_nan[0] && y[1] == sil_nan[1] &&
      y[2] == sil_nan[2] && y[3] == sil_nan[3] &&
      y[4] == sil_nan[4] && y[5] == sil_nan[5] &&
      y[6] == sil_nan[6] && y[7] == sil_nan[7]) {
    return 1;
  }
  return 0;
}
}
#endif /* USE_SIL_ISNAN */

void read_config (char *cfgfile)
{
  char buffer[512];
  FILE *f = fopen (cfgfile, "r");
  if (f) {
    while (fgets (buffer, 511, f) != NULL) {
      char *p = buffer;
      while (*p) 
	++p;
      if (p > buffer)
	*(p - 1) = '\0';
      if (*buffer) {
	char *pp, *q;
	switch (*buffer) {
	case 'e':
	  pp = buffer + 1;
	  while (*pp <= ' ')
	    ++pp;
	  if (sscanf (pp, "%lg", &EPSILON) != 1) {
	    sil_msg_error (cfgfile, "error in config file:");
	    sil_msg_error (cfgfile, buffer);
	    exit (1);
	  }
	  EPSILON2 = EPSILON * EPSILON;
	  break;
	case 'm':
	  pp = buffer + 1;
	  while (*pp <= ' ')
	    ++pp;
	  q = pp;
	  while (*pp > ' ')
	    ++pp;
	  *pp = '\0';
	  if (strlen (q) > 0) {
	    if (load_modules)
	      delete load_modules;
	    load_modules = new char[strlen (q) + 1];
	    strcpy (load_modules, q);
	  } else {
	    sil_msg_error (cfgfile, "error in config file:");
	    sil_msg_error (cfgfile, buffer);
	    exit (1);
	  }
	  break;
	case 'r':
	  pp = buffer + 1;
	  while (*pp <= ' ')
	    ++pp;
	  if (sscanf (pp, "%lg %lg",
		      &render_box_min, &render_box_max) != 2) {
	    sil_msg_error (cfgfile, "error in config file:");
	    sil_msg_error (cfgfile, buffer);
	    exit (1);
	  }
	  break;
	case 's':
	  pp = buffer + 1;
	  while (*pp <= ' ')
	    ++pp;
	  if (sscanf (pp, "%lg %lg %lg",
		      &sens_angle, &sens_move, &sens_dist) != 3) {
	    sil_msg_error (cfgfile, "error in config file:");
	    sil_msg_error (cfgfile, buffer);
	    exit (1);
	  }
	  break;
	case 't':
	  pp = buffer + 1;
	  while (*pp <= ' ')
	    ++pp;
	  q = pp;
	  while (*pp > ' ')
	    ++pp;
	  *pp = '\0';
	  if (strlen (q) > 0) {
	    if (tmp_dir)
	      delete tmp_dir;
	    tmp_dir = new char[strlen (q) + 1];
	    strcpy (tmp_dir, q);
	  } else {
	    sil_msg_error (cfgfile, "error in config file:");
	    sil_msg_error (cfgfile, buffer);
	    exit (1);
	  }
	  break;
	case 'M':
	  pp = buffer + 1;
	  while (*pp <= ' ')
	    ++pp;
	  q = pp;
	  while (*pp > ' ')
	    ++pp;
	  *pp = '\0';
	  if (strlen (q) > 0) {
	    if (module_path)
	      delete module_path;
	    module_path = new char[strlen (q) + 1];
	    strcpy (module_path, q);
	  } else {
	    sil_msg_error (cfgfile, "error in config file:");
	    sil_msg_error (cfgfile, buffer);
	    exit (1);
	  }
	  break;
	case 'P':
	  pp = buffer + 1;
	  while (*pp <= ' ')
	    ++pp;
	  q = pp;
	  while (*pp > ' ')
	    ++pp;
	  *pp = '\0';
	  if (strlen (q) > 0) {
	    if (plugin_path)
	      delete plugin_path;
	    plugin_path = new char[strlen (q) + 1];
	    strcpy (plugin_path, q);
	  } else {
	    sil_msg_error (cfgfile, "error in config file:");
	    sil_msg_error (cfgfile, buffer);
	    exit (1);
	  }
	  break;
	default:
	  if (*buffer > ' ' && *buffer != '#') {
	    sil_msg_error (cfgfile, "error in config file:");
	    sil_msg_error (cfgfile, buffer);
	    exit (1);
	  }
	}
      }
    }
    fclose (f);
  }
  return;
}

/** SiL main function.
 *
 * The main function for SiL handles command line options and
 * creates the App class object to run the GTK-- application.
 *
 * @param argc  Number of command line arguments.
 * @param argv  Pointer to command line arguments array.
 *
 * @return Exit status of the GTK-- application. */
int main (int argc, char **argv)
{
  // Set handler for failed memory allocation.
  set_new_handler (out_of_memory_handler);

  // Determine basename from program invocation name.
  char *p = strrchr (argv[0], '/');
  if (!p)
    p = argv[0];
  else
    ++p;
  char *progname = new char[strlen (p) + 1];
  strcpy (progname, p);
  // Set invocation name to basename so that getopt uses basename.
#ifndef MEM_PROFILE
  free (argv[0]);
#endif // MEM_PROFILE
  argv[0] = progname;

  module_path = new char[strlen (DEFAULT_MODULE_PATH) + 1];
  strcpy (module_path, DEFAULT_MODULE_PATH);

  plugin_path = new char[strlen (DEFAULT_PLUGIN_PATH) + 1];
  strcpy (plugin_path, DEFAULT_PLUGIN_PATH);

  load_modules = new char[strlen (DEFAULT_MODULES) + 1];
  strcpy (load_modules, DEFAULT_MODULES);

  tmp_dir = new char[strlen (DEFAULT_TMPDIR) + 1];
  strcpy (tmp_dir, DEFAULT_TMPDIR);

  char *str = new char[strlen (CONFIGDIR) + 9];
  strcpy (str, CONFIGDIR);
  strcat (str, "/sil.cfg");
  read_config (str);
  delete str;

  char *home = getenv ("HOME");
  if (home == NULL || !*home)
    home = ".";
  str = new char[strlen (home) + 10];
  strcpy (str, home);
  strcat (str, "/.sil/sil.cfg");
  read_config (str);
  delete str;

  /** Getopt long options.
   *
   * The sil_option array describing the long command line options for
   * the sil_getopt_long command which are accepted by SiL. This array
   * gives the complete string for the option and determines on which
   * short option character it should be mapped. */
  static struct sil_option long_cmdline_options[] = {
    {"help",             no_argument,       NULL, 'h'},
    {"version",          no_argument,       NULL, 'v'},
    {"epsilon",          required_argument, NULL, 'e'},
    {"modules",          required_argument, NULL, 'm'},
    {"render-box",       required_argument, NULL, 'r'},
    {"sensitivity",      required_argument, NULL, 's'},
    {"modules-path",     required_argument, NULL, 'M'},
    {"plugins-path",     required_argument, NULL, 'P'},
    {"gtk-module",       required_argument, NULL, 1},
    {"g-fatal-warnings", no_argument,       NULL, 2},
    {"gtk-debug",        required_argument, NULL, 3},
    {"gtk-no-debug",     required_argument, NULL, 4},
    {"gdk-debug",        required_argument, NULL, 5},
    {"gdk-no-debug",     required_argument, NULL, 6},
    {"sync",             no_argument,       NULL, 7},
    {"no-xshm",          no_argument,       NULL, 8},
    {"gxid_host",        required_argument, NULL, 9},
    {"gxid_port",        required_argument, NULL, 10},
    {"xim-preedit",      required_argument, NULL, 11},
    {"xim-status",       required_argument, NULL, 12},
    {"display",          required_argument, NULL, 13},
    {NULL,               no_argument,       NULL, 0}
  };

  // Initialize argument vector for GTK--
  char **gtkmm_argv = new char *[argc];
  int gtkmm_argc = 1;
  gtkmm_argv[0] = progname;
  // Process options using getopt.
  bool usage = false;
  bool error = false;
  while (true) {
    int c = sil_getopt_long (argc, argv, "hve:m:r:s:M:P:",
			     long_cmdline_options, NULL);
    if (c == -1)
      break;
    switch (c) {
    case 'h':
      // --help: Print usage information.
      usage = true;
      break;
    case 'v':
      // --version: Print version information.
      cout << "SiL (" << progname << ") Version " << VERSION << " "
	   << " (" << TARGET_TYPE << "," << OPENGLNAME << ")\n";
      return 0;
      break;
    case 'e':
      if (sscanf (sil_optarg, "%lg", &EPSILON) != 1) {
	sil_msg_error ("options", "illegal argument for -e");
	exit (1);
      }
      EPSILON2 = EPSILON * EPSILON;
      break;
    case 'm':
      if (load_modules)
	delete load_modules;
      load_modules = new char [strlen (sil_optarg) + 1];
      strcpy (load_modules, sil_optarg);
      break;
    case 'r':
      if (sscanf (sil_optarg, "%lg,%lg",
		  &render_box_min, &render_box_max) != 2) {
	sil_msg_error ("options", "illegal argument for -r");
	exit (1);
      }
      break;
    case 's':
      if (sscanf (sil_optarg, "%lg,%lg,%lg",
		  &sens_angle, &sens_move, &sens_dist) != 3) {
	sil_msg_error ("options", "illegal argument for -s");
	exit (1);
      }
      break;
    case 'M':
      if (module_path)
	delete module_path;
      module_path = new char [strlen (sil_optarg) + 1];
      strcpy (module_path, sil_optarg);
      break;
    case 'P':
      if (plugin_path)
	delete plugin_path;
      plugin_path = new char [strlen (sil_optarg) + 1];
      strcpy (module_path, sil_optarg);
      break;
    case 1:
    case 3:
    case 5:
    case 6:
    case 8:
    case 9:
    case 10:
    case 11:
    case 12:
    case 13:
      // Set GTK-- options with required arguments.
      gtkmm_argv[gtkmm_argc] = new char[strlen (long_cmdline_options[c].name)
				        + 3];
      strcpy (gtkmm_argv[gtkmm_argc], "--");
      strcpy (gtkmm_argv[gtkmm_argc++] + 2, long_cmdline_options[c].name);
      gtkmm_argv[gtkmm_argc] = new char[strlen (sil_optarg) + 1];
      strcpy (gtkmm_argv[gtkmm_argc++], sil_optarg);
      break;
    case 2:
    case 4:
    case 7:
      // Set GTK-- options with no arguments.
      gtkmm_argv[gtkmm_argc] = new char[strlen (long_cmdline_options[c].name)
				        + 3];
      strcpy (gtkmm_argv[gtkmm_argc], "--");
      strcpy (gtkmm_argv[gtkmm_argc++] + 2, long_cmdline_options[c].name);
      break;
    default:
      // Illegal option/argument.
      error = true;
    }
  }

  // Print usage information.
  if (usage) {
    cerr << "Usage: " << progname << " {<options>}\n"
	 << " Options:\n"
	 << "   --help|-h                Print usage information\n"
	 << "   --version|-v             Print version information\n"
	 << "   --epsilon|-e EPS         Machine precision\n"
	 << "   --modules|-m MODULES     Colon separated list of SiL modules\n"
	 << "   --render-box|-r MIN,MAX  Minimum and maximum for render box\n"
	 << "   --sensitivity|-s A,M,D   Angle, movement and distance sensitivity\n"
	 << "   --modules-path|-M PATH   Colon separated list of module directories\n"
	 << "   --plugins-path|-P PATH   Colon separated list of plugin directories\n"
	 << "                             where SiL searches for modules\n"
	 << "   --                       Force to stop scanning options\n"
	 << " GTK options:\n"
	 << "   --displayd DISPLAY       Use the designated X display\n"
	 << "   --gtk-module MODULE\n"
	 << "   --g-fatal-warnings\n"
	 << "   --gtk-debug DEBUG-FLAGS\n"
	 << "   --gtk-no-debug DEBUG-FLAGS\n"
	 << " GDK options:\n"
	 << "   --gdk-debug DEBUG-FLAGS\n"
	 << "   --gdk-no-debug DEBUG-FLAGS\n"
	 << "   --sync\n"
	 << "   --no-xshm\n"
	 << "   --gxid_host GXID-HOST\n"
	 << "   --gxid_port GXID_PORT\n"
	 << "   --xim-preedit none|nothing|area|position|callbacks\n"
	 << "   --xim-status none|nothing|area|position|callbacks\n";
    delete progname;
    return 0;
  }
  // Process SiL argument errors.
  if (!error && sil_optind != argc) {
    cerr << progname << ": too many arguments\n";
    error = true;
  }
  if (error) {
    cerr << progname << ": type `" << progname << " --help' for help\n";
    delete progname;
    return 1;
  }

  // Initialize the GTK-- application.
  Gtk::Main gtk (&gtkmm_argc, &gtkmm_argv);
  App *app = new App (progname, load_modules, module_path);
  int ret_code = 0;
  if (gtkmm_argc == 1) {
    // Run the GTK-- application.
    app->run ();
  } else {
    // Some of the GTK-- arguments were not accepted.
    cerr << progname << ": illegal GTK argument(s)\n"
	 << progname << ": type `" << progname << " --help' for help\n";
    for (int i = 1; i < gtkmm_argc; ++i) {
      if (gtkmm_argv[i])
	delete gtkmm_argv[i];
    }
    ret_code = 1;
  }

  // Shut down the GTK-- application and exit.
  delete app;
  delete progname;
  delete gtkmm_argv;
  if (load_modules)
    delete load_modules;
  if (module_path)
    delete module_path;
  if (plugin_path)
    delete plugin_path;

  return ret_code;
}

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
