// material.hxx: Material class.
//
// Copyright (C) 1999-2000 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: material.hxx,v 1.7 2001/04/19 15:23:39 scmfcl Exp $

#ifndef MATERIAL_HXX
#define MATERIAL_HXX 1

// General class for SiL materials. All material describtions should
// be derived from this class. The material model is based on the
// OpenGL material describtion, but there are no patterns, etc.
class Material {
public:
  // Create a SiL material setting the ambient, diffuse, specular
  // emission RGBA vectors and the shininess. The values not specified
  // default to the OpenGL defaults.
  inline Material (GLfloat ar=0.7, GLfloat ag=0.7, GLfloat ab=0.7,
		   GLfloat aa=1.0,
		   GLfloat dr=0.7, GLfloat dg=0.7, GLfloat db=0.7,
		   GLfloat da=1.0,
		   GLfloat sr=0.2, GLfloat sg=0.2, GLfloat sb=0.2,
		   GLfloat sa=0.0,
		   GLfloat er=0.0, GLfloat eg=0.0, GLfloat eb=0.0,
		   GLfloat ea=0.0,
		   GLfloat ss=1.0) {
    ambient[0] = ar;
    ambient[1] = ag;
    ambient[2] = ab;
    ambient[3] = aa;
    diffuse[0] = dr;
    diffuse[1] = dg;
    diffuse[2] = db;
    diffuse[3] = da;
    specular[0] = sr;
    specular[1] = sg;
    specular[2] = sb;
    specular[3] = sa;
    emission[0] = er;
    emission[1] = eg;
    emission[2] = eb;
    emission[3] = ea;
    shininess = ss;
  }
  // Get the ambient RGBA vector.
  inline const GLfloat *get_ambient () const {
    return ambient;
  };
  // Get the diffuse RGBA vector.
  inline const GLfloat *get_diffuse () const {
    return diffuse;
  };
  // Get the specular RGBA vector.
  inline const GLfloat *get_specular () const {
    return specular;
  };
  // Get the emission RGBA vector.
  inline const GLfloat *get_emission () const {
    return emission;
  };
  // Get the shininess RGBA vector.
  inline const GLfloat *get_shininess () const {
    return &shininess;
  };
  // Set the ambinet, diffuse, specular, emissin RGBA vectors and
  // the shininess.
  inline void set (GLfloat ar, GLfloat ag, GLfloat ab, GLfloat aa,
		   GLfloat dr, GLfloat dg, GLfloat db, GLfloat da,
		   GLfloat sr, GLfloat sg, GLfloat sb, GLfloat sa,
		   GLfloat er, GLfloat eg, GLfloat eb, GLfloat ea,
		   GLfloat ss) {
    ambient[0] = ar;
    ambient[1] = ag;
    ambient[2] = ab;
    ambient[3] = aa;
    diffuse[0] = dr;
    diffuse[1] = dg;
    diffuse[2] = db;
    diffuse[3] = da;
    specular[0] = sr;
    specular[1] = sg;
    specular[2] = sb;
    specular[3] = sa;
    emission[0] = er;
    emission[1] = eg;
    emission[2] = eb;
    emission[3] = ea;
    shininess = ss;
  };
  // Set the Material for OpenGL.
  inline void gl_material (bool inv=false) const {
    if (inv) {
      GLfloat icol[4] = {0.0, 0.0, 0.0, ambient[3]};
      int l = 0;
      if (ambient[0] > ambient[1])
	l = 1;
      if (ambient[l] > ambient[2])
	l = 2;
      icol[l] = 1.0;
      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, icol);
      icol[l] = 0.0;
      icol[3] = diffuse[3];
      l = 0;
      if (diffuse[0] > diffuse[1])
	l = 1;
      if (diffuse[l] > diffuse[2])
	l = 2;
      icol[l] = 1.0;
      glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, icol);
    } else {
      glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
    }
    glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, specular);
    glMaterialfv (GL_FRONT_AND_BACK, GL_EMISSION, emission);
    glMaterialfv (GL_FRONT_AND_BACK, GL_SHININESS, &shininess);
    return;
  };
  // Set the Material for OpenGL (for lines and vertices).
  inline void gl_material_bright (bool inv=false) const {
    if (inv) {
      GLfloat icol[4] = {0.0, 0.0, 0.0, ambient[3]};
      int l = 0;
      if (ambient[0] > ambient[1])
	l = 1;
      if (ambient[l] > ambient[2])
	l = 2;
      icol[l] = 1.0;
      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, icol);
      glMaterialfv (GL_FRONT_AND_BACK, GL_EMISSION, icol);
      icol[l] = 0.0;
      icol[3] = diffuse[3];
      l = 0;
      if (diffuse[0] > diffuse[1])
	l = 1;
      if (diffuse[l] > diffuse[2])
	l = 2;
      icol[l] = 1.0;
      glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, icol);
    } else {
      glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
      glMaterialfv (GL_FRONT_AND_BACK, GL_EMISSION, ambient);
    }
    glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, specular);
    glMaterialfv (GL_FRONT_AND_BACK, GL_SHININESS, &shininess);
    return;
  };

private:
  // RGBA vectors describing the material.
  GLfloat ambient[4];
  GLfloat diffuse[4];
  GLfloat specular[4];
  GLfloat emission[4];
  //Shininess of the material.
  GLfloat shininess;
};

#endif // !MATERIAL_HXX

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
