// modules.cxx: ModuleList implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: modules.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (modules, "$Id: modules.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"

#include "modules.hxx"
#include "module.hxx"

ModuleList::ModuleList (const char *path)
{
  head = NULL;
  window = NULL;
  formats = NULL;
  cformat = NULL;
  module_path = new char[strlen (path) + 1];
  strcpy (module_path, path);
  unkown_module = new Module;
}

ModuleList::~ModuleList ()
{
  FileFormat *f = formats;
  while (formats) {
    f = formats->next;
    delete formats;
    formats = f;
  }
  while (head) {
    ModuleElement *w = head->next;
    if (head->module)
      delete head->module;
    if (head->ptr)
      dlclose (head->ptr);
    delete head;
    head = w;
  }
  if (unkown_module)
    delete unkown_module;
  if (module_path)
    delete module_path;
}

bool ModuleList::register_module (char *name)
{
  ModuleElement *w = head;
  while (w) {
    if (strcmp(w->module->name(), name) == 0)
      return true;
    w = w->next;
  }

  char *file = new char[strlen (module_path) + strlen (name) + 13];

  char *paths = new char[strlen (module_path) + 2];
  strcpy (paths, module_path);
  paths[strlen (module_path) + 1] = '\0';
  char *path = paths;
  bool found = false;
  while (path && *path != '\0') {
    int i = 0;
    while (path[i] != '\0' && path[i] != ':')
      ++i;
    path[i] = '\0';
    strcpy (file, path);
    strcat (file, "/");
    strcat (file, name);
    strcat (file, ".sil_module");
    struct stat finfo;
    if (stat (file, &finfo) == 0 && (S_ISREG (finfo.st_mode) ||
				     S_ISLNK (finfo.st_mode))) {
      path = NULL;
      found = true;
    } else {
      path += i + 1;
    }
  }
  if (!found) {
    delete file;
    delete paths;
    return false;
  }

  void *p = dlopen (file, RTLD_NOW|RTLD_GLOBAL);
  if (!p) {
    sil_msg_error ("module open", dlerror());
    delete file;
    delete paths;
    return false;
  }

  Module *(*create_module)(void);
  create_module = (Module *(*)(void)) dlsym (p, "create_module__Fv");
  const char *err = dlerror ();
  if (err) {
    sil_msg_error ("module symbol", err);
    dlclose (p);
    delete file;
    delete paths;
    return false;
  }
  Module *module = create_module ();
  if (!module) {
    sil_msg_error ("module", "can't crate module object");
    dlclose (p);
    delete file;
    delete paths;
    return false;
  }
  w = new ModuleElement (module, p, head);
  head = w;
  // Get provided file formats
  const char **f = module->file_formats ();
  if (f) {
    int n = 0;
    while (f[n])
      ++n;
    for (int i = 0; i < n - 1; i+=2) {
      FileFormat *ff = new FileFormat (f[i], f[i + 1], module, formats);
      formats = ff;
    }
  }
  delete file;
  delete paths;
  return true;
}

Module *ModuleList::module (char *name) const
{
  ModuleElement *w = head;
  while (w) {
    if (strcmp(w->module->name (), name) == 0)
      return w->module;
    w = w->next;
  }
  if (strcmp ("unkown", name) == 0)
    return unkown_module;
  return NULL;
}

void ModuleList::facet_init ()
{
  ModuleElement *w = head;
  while (w) {
    if (w->module)
      w->module->facet_init ();
    w = w->next;
  }
  return;
}

void ModuleList::draw_init ()
{
  ModuleElement *w = head;
  while (w) {
    if (w->module)
      w->module->draw_init ();
    w = w->next;
  }
  return;
}

const char *ModuleList::first_format_name ()
{
  cformat = formats;
  if (cformat)
    return cformat->name;
  return NULL;
}

const char *ModuleList::next_format_name ()
{
  if (cformat) {
    cformat = cformat->next; 
    if (cformat)
      return cformat->name;
  }
  return NULL;
}

const char *ModuleList::current_format_desc ()
{
  if (cformat)
    return cformat->desc;
  return NULL;
}

void ModuleList::load_file (FILE *fp, const char *format, bool ext,
			    bool replace)
{
  cformat = formats;
  while (cformat) {
    if (strcmp(cformat->name, format) == 0) {
      cformat->module->load_file (fp, format, ext, replace);
      return;
    }
    cformat = cformat->next;
  }
  sil_msg_error ("load", "unkown format");
  return;
}

void ModuleList::save_file (FILE *fp, const char *format, bool ext)
{
  cformat = formats;
  while (cformat) {
    if (strcmp(cformat->name, format) == 0) {
      cformat->module->save_file (fp, format, ext);
      return;
    }
    cformat = cformat->next;
  }
  sil_msg_error ("save", "unkown format");
  return;
}

const char *ModuleList::first_name ()
{
  window = head;
  if (window)
    return window->module->name ();
  return NULL;
}

const char *ModuleList::next_name ()
{
  if (window) {
    window = window->next;
    if (window)
      return window->module->name ();
  }
  return NULL;
}

void *ModuleList::get_ptr (const char *module)
{
  ModuleElement *w = head;
  while (w) {
    if (strcmp (w->module->name (), module) == 0) {
      return w->ptr;
    }
    w = w->next;
  }
  return NULL;
}

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
