// analyse.cc: sil plugin to analyse a sil model
//
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse, "$Id: analyse.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"

#include "list.hxx"
#include "kernel/kernel.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"
#include "acis/acis.hxx"
#include "acis/silapi.hxx"

class AnalyserWindow;

Real tol_len_dist = 0.05;
Real tol_angle_dist = 0.5;
Real tol_loop_dist = 0.05;
Real tol_ratio_approx = 0.1;
Real tol_ratio_min = 0.005;

unsigned long regular_angle_max_n = 45;
unsigned long loop_fourier_coefficients = 10;
unsigned long para_max_power = 2;
unsigned long para_max_frac = 10;
unsigned long para_min_frac = 8;
unsigned long ratio_max_power = 2;
unsigned long ratio_max_frac = 10;
unsigned long ratio_min_frac = 8;
bool base_one = true;
bool base_inch = false;
bool base_deg = false;
bool base_pi = true;
bool base_tan = true;
char *str_len_base = NULL;
char *str_angle_base = NULL;
unsigned long max_pcbs = 20;

bool use_edges = false;
bool f_a = true;
bool f_as = true;
bool f_ar = true;
bool f_l = true;
bool f_ls = true;
bool f_lr = true;
bool f_d = true;
bool f_dp = true;
bool f_par = true;
bool f_car = true;
bool f_ai = true;
bool f_ap = true;
bool f_arp = true;
bool f_arc = true;
bool f_p = true;
bool f_pp = true;
bool f_los = true;
bool f_g = true;

List<AnalyserWindow> analyser_windows;

class AnalyserWindow : public Gtk::Window {
public:
  AnalyserWindow (Solid *S, Symbol const *s, Symbol const *hs);
  ~AnalyserWindow ();
  void update ();
  void close ();
private:
  void analyse ();
  void set_values (bool genbase);
  void cancel ();
  void use_pre ();
  void add_pre ();
  void remove_pre ();
  void save_pre ();
  void read_preset_file (string const &presetfile);
  void reg_toggle ();
  Solid *sld;
  char *sld_name;
  const char *hi_name;
  char **len_base, **angle_base;
  SimpleList<string> preset_strings;
  static gchar const *prelisttitles[1];
  Gtk::VBox vbox;
  Gtk::HBox hbox, titlebox;
  Gtk::Label titlelabel, title2label;
  Gtk::Notebook nb;
  Gtk::Frame f_tol, f_para, f_reg, f_preset;
  Gtk::Table t_tol, t_para, t_reg, t_preset;
  Gtk::Label l_len, l_angle, l_loop, l_ratio, l_ratiomin;
  Gtk::Adjustment a_lend;
  Gtk::SpinButton s_lend;
  Gtk::Adjustment a_angd;
  Gtk::SpinButton s_angd;
  Gtk::Adjustment a_loopd;
  Gtk::SpinButton s_loopd;
  Gtk::Adjustment a_ratioa;
  Gtk::SpinButton s_ratioa;
  Gtk::Adjustment a_ratiom;
  Gtk::SpinButton s_ratiom;
  Gtk::Label l_anglemaxn, l_fc, l_para, l_rat, l_fracmax, l_fracmin;
  Gtk::Label l_fracpower, l_lenbase, l_anglebase, l_pcbs;
  Gtk::Adjustment a_anglemaxn;
  Gtk::SpinButton s_anglemaxn;
  Gtk::Adjustment a_fc;
  Gtk::SpinButton s_fc;
  Gtk::Adjustment a_pmaxd;
  Gtk::SpinButton s_pmaxd;
  Gtk::Adjustment a_pmind;
  Gtk::SpinButton s_pmind;
  Gtk::Adjustment a_ppow;
  Gtk::SpinButton s_ppow;
  Gtk::Adjustment a_rmaxd;
  Gtk::SpinButton s_rmaxd;
  Gtk::Adjustment a_rmind;
  Gtk::SpinButton s_rmind;
  Gtk::Adjustment a_rpow;
  Gtk::SpinButton s_rpow;
  Gtk::Adjustment a_pcbs;
  Gtk::SpinButton s_pcbs;
  Gtk::CheckButton cb_one, cb_inch, cb_deg, cb_pi, cb_tan;
  Gtk::Entry e_lenbase, e_anglebase;
  Gtk::CheckButton cb_edges, cb_a, cb_as, cb_ar, cb_l, cb_ls, cb_lr, cb_d;
  Gtk::CheckButton cb_dp, cb_par, cb_car, cb_ai, cb_ap, cb_arp, cb_arc;
  Gtk::CheckButton cb_p, cb_pp, cb_los, cb_g;
  Gtk::Button b_presave, b_preadd, b_preremove, b_preuse;
  Gtk::Entry e_prename;
  Gtk::CList cl_prelist;
  Gtk::Button b_analyse, b_cancel;
};

gchar const *AnalyserWindow::prelisttitles[1] = { 
  "Analyse Configuration Sets" 
};

AnalyserWindow::AnalyserWindow (Solid *S, Symbol const *s, Symbol const *hs)
  : Gtk::Window (GTK_WINDOW_TOPLEVEL),
	 vbox (),
	 hbox (true),
	 titlebox (true),
	 titlelabel ("Analyse:"),
	 title2label (),
	 nb (),
	 f_tol ("Tolerances"),
	 f_para ("Parameters"),
	 f_reg ("Regularities"),
	 f_preset ("Configuration Sets"),
	 t_tol (11, 3, true),
	 t_para (11, 3, true),
	 t_reg (11, 4, true),
	 t_preset (11, 4, true),
	 l_len ("Minimal length tolerance:"),
	 l_angle ("Minimal angle tolerance (deg):"),
	 l_loop ("Minimal loop tolerance:"),
	 l_ratio ("Minimal fraction tolerance range:"),
	 l_ratiomin ("Closest fraction:"),
	 a_lend (tol_len_dist, 0.0, 10000.0, 0.01, 0.1, 0.0),
	 s_lend (a_lend, 0.1, 5),
	 a_angd (tol_angle_dist, 0.0, 90.0, 0.1, 1.0, 0.0),
	 s_angd (a_angd, 0.1, 5),
	 a_loopd (tol_loop_dist, 0.0, 10000.0, 0.01, 0.1, 0.0),
	 s_loopd (a_loopd, 0.1, 5),
	 a_ratioa (tol_ratio_approx, 0.0, 1.0, 0.01, 0.1, 0.0),
	 s_ratioa (a_ratioa, 0.1, 5),
	 a_ratiom (tol_ratio_min, 0.0, 1.0, 0.001, 0.01, 0.0),
	 s_ratiom (a_ratiom, 0.1, 5),
	 l_anglemaxn ("Max. n for regular angle [2]Pi/n:"),
	 l_fc ("Fourier coefficients for loops:"),
	 l_para ("Parameter"),
	 l_rat ("Ratios"),
	 l_fracmax ("Max. denominator"),
	 l_fracmin ("Min. denominator"),
	 l_fracpower ("Max. root"),
	 l_lenbase ("Length bases"),
	 l_anglebase ("Angle bases"),
	 l_pcbs ("Max. point circle bucket sizde"),
	 a_anglemaxn (regular_angle_max_n - 1, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_anglemaxn (a_anglemaxn, 1.0, 0),
	 a_fc (loop_fourier_coefficients, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_fc (a_fc, 1.0, 0),
	 a_pmaxd (para_max_frac, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_pmaxd (a_pmaxd, 1.0, 0),
	 a_pmind (para_min_frac, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_pmind (a_pmind, 1.0, 0),
	 a_ppow (para_max_power - 1, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_ppow (a_ppow, 1.0, 0),
	 a_rmaxd (ratio_max_frac, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_rmaxd (a_rmaxd, 1.0, 0),
	 a_rmind (ratio_min_frac, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_rmind (a_rmind, 1.0, 0),
	 a_rpow (ratio_max_power - 1, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_rpow (a_rpow, 1.0, 0),
	 a_pcbs (max_pcbs, 1.0, 10000.0, 1.0, 5.0, 0.0),
	 s_pcbs (a_pcbs, 1.0, 0),
	 cb_one ("One"),
	 cb_inch ("Inch"),
	 cb_deg ("Degree"),
	 cb_pi ("Pi"),
	 cb_tan ("Arctan"),
	 e_lenbase (),
	 e_anglebase (),
	 cb_edges ("Use edges"),
	 cb_a ("Similar angles"),
	 cb_as ("Special angle values"),
	 cb_ar ("Special angle ratios"),
	 cb_l ("Similar lengths"),
	 cb_ls ("Special length values"),
	 cb_lr ("Special length ratios"),
	 cb_d ("Similar directions"),
	 cb_dp ("Direction planes"),
	 cb_par ("Planar angle-regular directions"),
	 cb_car ("Conical angle-regular directions"),
	 cb_ai ("Axis intersections"),
	 cb_ap ("Aligned axes"),
	 cb_arp ("Linear axis arrangements"),
	 cb_arc ("Circular axis arrangements"),
	 cb_p ("Similar positions"),
	 cb_pp ("Partially similar positions"),
	 cb_los ("Similar polygonial loops"),
	 cb_g ("Geometry types"),
	 b_presave ("Save"),
	 b_preadd ("Add"),
	 b_preremove ("Remove"),
	 b_preuse ("Use"),
	 e_prename (),
	 cl_prelist (1, prelisttitles),
	 b_analyse ("Analyse"),
	 b_cancel ("Cancel")
{
  sld = S;
  take (sld);
  char const *n = sil_get_symbol_name (s);
  sld_name = new char [strlen(n) + 1];
  strcpy (sld_name, n);
  hi_name = sil_get_symbol_name (hs);

  len_base = NULL;
  angle_base = NULL;

  set_title ("Analyse");
  set_border_width (3);
  set_usize (500,400);

  title2label.set (sld_name);
  title2label.set_justify (GTK_JUSTIFY_LEFT);
  titlelabel.set_justify (GTK_JUSTIFY_RIGHT);

  nb.set_tab_pos (GTK_POS_TOP);

  f_tol.set_border_width (5);
  f_para.set_border_width (5);
  f_reg.set_border_width (5);
  f_preset.set_border_width (5);
  nb.pages () . push_back (Gtk::Notebook_Helpers::TabElem (f_tol, 
							   "Tolerances"));
  nb.pages () . push_back (Gtk::Notebook_Helpers::TabElem (f_para, 
							   "Parameters"));
  nb.pages () . push_back (Gtk::Notebook_Helpers::TabElem (f_reg, 
							   "Regularities"));
  nb.pages () . push_back (Gtk::Notebook_Helpers::TabElem 
			   (f_preset, "Configuration Sets"));
  t_tol.set_row_spacings (2);
  t_tol.set_col_spacings (2);
  t_tol.set_border_width (4);
  f_tol.add (t_tol);
  t_para.set_row_spacings (2);
  t_para.set_col_spacings (2);
  t_para.set_border_width (4);
  f_para.add (t_para);
  t_reg.set_row_spacings (2);
  t_reg.set_col_spacings (2);
  t_reg.set_border_width (4);
  f_reg.add (t_reg);
  t_preset.set_row_spacings (2);
  t_preset.set_col_spacings (2);
  t_preset.set_border_width (4);
  f_preset.add (t_preset);

  l_len.set_justify (GTK_JUSTIFY_RIGHT);
  l_angle.set_justify (GTK_JUSTIFY_RIGHT);
  l_loop.set_justify (GTK_JUSTIFY_RIGHT);
  l_ratio.set_justify (GTK_JUSTIFY_RIGHT);

  t_tol.attach (l_len, 
		0, 2, 1, 2,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_tol.attach (l_angle, 
		0, 2, 2, 3,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_tol.attach (l_loop, 
		0, 2, 3, 4,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_tol.attach (l_ratio, 
		0, 2, 4, 5,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_tol.attach (l_ratiomin, 
		0, 2, 5, 6,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);

  t_tol.attach (s_lend,
		2, 3, 1, 2,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_tol.attach (s_angd,
		2, 3, 2, 3,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_tol.attach (s_loopd,
		2, 3, 3, 4,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_tol.attach (s_ratioa,
		2, 3, 4, 5,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_tol.attach (s_ratiom,
		2, 3, 5, 6,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);

  l_anglemaxn.set_justify (GTK_JUSTIFY_RIGHT);
  l_fc.set_justify (GTK_JUSTIFY_RIGHT);

  t_para.attach (l_anglemaxn,
		 0, 2, 0, 1,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_fc,
		 0, 2, 1, 2,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_para,
		 1, 2, 2, 3,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_rat,
		 2, 3, 2, 3,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_fracmin,
		 0, 1, 3, 4,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_fracmax,
		 0, 1, 4, 5,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_fracpower,
		 0, 1, 5, 6,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_lenbase,
		 0, 1, 6, 7,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_anglebase,
		 0, 1, 8, 9,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_anglemaxn,
		 2, 3, 0, 1,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_fc,
		 2, 3, 1, 2,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_pmind,
		 1, 2, 3, 4,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_pmaxd,
		 1, 2, 4, 5,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_ppow,
		 1, 2, 5, 6,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_rmind,
		 2, 3, 3, 4,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_rmaxd,
		 2, 3, 4, 5,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_rpow,
		 2, 3, 5, 6,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  cb_one.set_active (base_one);
  t_para.attach (cb_one,
		 1, 2, 6, 7,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  cb_inch.set_active (base_inch);
  t_para.attach (cb_inch,
		 2, 3, 6, 7,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  if (str_len_base)
    e_lenbase.set_text (str_len_base);
  t_para.attach (e_lenbase,
		 1, 3, 7, 8,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  cb_deg.set_active (base_deg);
  t_para.attach (cb_deg,
		 1, 2, 8, 9,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  cb_pi.set_active (base_pi);
  t_para.attach (cb_pi,
		 2, 3, 8, 9,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  cb_tan.set_active (base_tan);
  t_para.attach (cb_tan,
		 1, 2, 9, 10,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  if (str_angle_base)
    e_anglebase.set_text (str_angle_base);
  t_para.attach (e_anglebase,
		 2, 3, 9, 10,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (l_pcbs,
		 0, 2, 10, 11,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);
  t_para.attach (s_pcbs,
		 2, 3, 10, 11,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		 (gint)0, (gint)0);

  cb_edges.set_active (use_edges);
  cb_a.set_active (f_a);
  cb_as.set_active (f_as);
  cb_ar.set_active (f_ar);
  cb_l.set_active (f_l);
  cb_ls.set_active (f_ls);
  cb_lr.set_active (f_lr);
  cb_d.set_active (f_d);
  cb_dp.set_active (f_dp);
  cb_par.set_active (f_par);
  cb_car.set_active (f_car);
  cb_ai.set_active (f_ai);
  cb_ap.set_active (f_ap);
  cb_arp.set_active (f_arp);
  cb_arc.set_active (f_arc);
  cb_p.set_active (f_p);
  cb_pp.set_active (f_pp);
  cb_los.set_active (f_los);
  cb_g.set_active (f_g);

  t_reg.attach (cb_edges,
		0, 2, 0, 1,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_a,
		0, 2, 1, 2,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_as,
		0, 2, 2, 3,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_ar,
		0, 2, 3, 4,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_l,
		0, 2, 4, 5,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_ls,
		0, 2, 5, 6,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_lr,
		0, 2, 6, 7,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_d,
		0, 2, 7, 8,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_dp,
		0, 2, 8, 9,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_par,
		0, 2, 9, 10,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_car,
		2, 4, 0, 1,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_ai,
		2, 4, 1, 2,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_ap,
		2, 4, 2, 3,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_arp,
		2, 4, 3, 4,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_arc,
		2, 4, 4, 5,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_p,
		2, 4, 5, 6,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_pp,
		2, 4, 6, 7,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_los,
		2, 4, 7, 8,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);
  t_reg.attach (cb_g,
		2, 4, 8, 9,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		(gint)0, (gint)0);

  reg_toggle ();
  cb_a.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_as.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_ar.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_l.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_ls.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_lr.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_d.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_dp.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_par.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_car.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_ai.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_ap.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_arp.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_arc.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_p.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_pp.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_los.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));
  cb_g.toggled.connect (slot (this, &AnalyserWindow::reg_toggle));

  cl_prelist.column_titles_passive ();
  cl_prelist.set_column_auto_resize (0, true);
  cl_prelist.set_column_justification (0, GTK_JUSTIFY_LEFT);
  cl_prelist.set_auto_sort (false);
  cl_prelist.set_selection_mode (GTK_SELECTION_SINGLE);

  char *home = getenv ("HOME");
  if (home == NULL || !*home)
    home = ".";
  string str = static_cast<string>(home) + "/.sil/analyse.cfg";
  read_preset_file (str);

  t_preset.attach (cl_prelist,
		   0, 4, 0, 9,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   (gint)0, (gint)0);
  t_preset.attach (b_preuse,
		   0, 1, 10, 11,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   (gint)0, (gint)0);
  t_preset.attach (b_preadd,
		   1, 2, 10, 11,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   (gint)0, (gint)0);
  t_preset.attach (b_preremove,
		   2, 3, 10, 11,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   (gint)0, (gint)0);
  t_preset.attach (b_presave,
		   3, 4, 10, 11,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   (gint)0, (gint)0);
  t_preset.attach (e_prename,
		   0, 4, 9, 10,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   GTK_FILL|GTK_EXPAND|GTK_SHRINK,
		   (gint)0, (gint)0);

  b_preuse.clicked.connect (slot (this, &AnalyserWindow::use_pre));
  b_preadd.clicked.connect (slot (this, &AnalyserWindow::add_pre));
  b_preremove.clicked.connect (slot (this, &AnalyserWindow::remove_pre));
  b_presave.clicked.connect (slot (this, &AnalyserWindow::save_pre));

  add (vbox);
  vbox.pack_start (titlebox, false, true, 2);
  titlebox.pack_start (titlelabel, true, true, 2);
  titlebox.pack_start (title2label, true, true, 2);
  vbox.pack_start (nb, true, true, 2);
  vbox.pack_start (hbox, false, true, 2);
  hbox.pack_start (b_analyse, true, true, 2);
  hbox.pack_start (b_cancel, true, true, 2);

  b_analyse.clicked.connect (slot (this, &AnalyserWindow::analyse));
  b_cancel.clicked.connect (slot (this, &AnalyserWindow::cancel));

  show_all ();

  return;
}

AnalyserWindow::~AnalyserWindow ()
{
  drop (sld);
  ListNode<AnalyserWindow> *n = analyser_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }
  if (sld_name)
    delete sld_name;
  preset_strings.purge ();
  return;
}

void AnalyserWindow::update ()
{
  const Symbol *s = sil_find_symbol (sld_name);
  if (!s) {
    ListNode<AnalyserWindow> *n = analyser_windows.find (this);
    if (n) {
      n->unlink ();
      delete n;
    }
    hide_all ();
    destroy ();
  }
  return;
}

void AnalyserWindow::close ()
{
  ListNode<AnalyserWindow> *n = analyser_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }
  hide_all ();
  destroy ();
  return;
}

void AnalyserWindow::analyse ()
{
  b_analyse.set_sensitive (false);
  b_cancel.set_sensitive (false);

  sil_progress_bar (true, false);

  set_values (true);

  Analyse *a = sil_analyse (sld, sld_name, hi_name,
			    sil_get_epsilon (),
			    tol_len_dist,
			    tol_angle_dist / 180.0 * PI,
			    tol_ratio_approx,
			    tol_ratio_min,
			    tol_loop_dist,
			    regular_angle_max_n,
			    loop_fourier_coefficients,
			    para_max_power,
			    para_max_frac,
			    para_min_frac,
			    ratio_max_power,
			    ratio_max_frac,
			    ratio_min_frac,
			    len_base,
			    angle_base,
			    max_pcbs,
			    use_edges,
			    f_a, f_as, f_ar, f_l, f_ls, f_lr, f_d, f_dp,
			    f_par, f_car, f_ai, f_ap, f_arp, f_arc, f_p,
			    f_pp, f_los, f_g);
 
  if (a) {
    char *n = new char [strlen (sld_name) + 9];
    strcpy (n, "ANALYSE_");
    strcat (n, sld_name);
    while (!sil_add_symbol (n, (void *)a, "kernel", false))
      sil_del_symbol (n);
    char str[1024];
    snprintf (str, 1023, "Analysed %s: %s in %0ldm %0ld.%03ds.",
	      sld_name, n,
	      a->time_secs () [0] / 60,
	      a->time_secs () [0] % 60,
	      a->time_msecs () [0]);
    sil_msg ("analyse", str);
    delete n;
    sil_updated ();
  } else {
    sil_msg_error ("analyse", "Analysing " + static_cast<string>(sld_name) +
		   " failed.");
  }

  sil_progress_bar (false);

  close ();

  return;
}

void AnalyserWindow::set_values (bool genbase)
{
  tol_len_dist = s_lend.get_value_as_float ();
  tol_angle_dist = s_angd.get_value_as_float ();
  tol_loop_dist = s_loopd.get_value_as_float ();
  tol_ratio_approx = s_ratioa.get_value_as_float ();
  tol_ratio_min = s_ratiom.get_value_as_float ();
  if (tol_ratio_min > tol_ratio_approx)
    tol_ratio_min = tol_ratio_approx;

  regular_angle_max_n = s_anglemaxn.get_value_as_int () + 1;
  loop_fourier_coefficients = s_fc.get_value_as_int ();
  para_min_frac = s_pmind.get_value_as_int ();
  para_max_frac = s_pmaxd.get_value_as_int ();
  if (para_max_frac < para_min_frac)
    para_max_frac = para_min_frac;
  para_max_power = s_ppow.get_value_as_int () + 1;
  ratio_min_frac = s_rmind.get_value_as_int ();
  ratio_max_frac = s_rmaxd.get_value_as_int ();
  if (ratio_max_frac < ratio_min_frac)
    ratio_max_frac = ratio_min_frac;
  ratio_max_power = s_rpow.get_value_as_int () + 1;
  base_one = cb_one.get_active ();
  base_inch = cb_inch.get_active ();
  base_deg = cb_deg.get_active ();
  base_pi = cb_pi.get_active ();
  base_tan = cb_tan.get_active ();
  max_pcbs = s_pcbs.get_value_as_int ();

  unsigned long len_num = 0;
  if (base_one)
    ++len_num;
  if (base_inch)
    ++len_num;
  string str = e_lenbase.get_text ();
  unsigned int pos = 0;
  while (pos < str.size ()) {
    if (str[pos] == ' ') {
      ++len_num;
      ++pos;
      while (pos < str.size () && str[pos] != ' ')
	++pos;
    } else {
      if (str[pos] == ':')
	str[pos] = ' ';
      else
	str[pos] = tolower (str[pos]);
      ++pos;
    }
  }
  if (pos > 0)
    ++len_num;
  if (genbase) {
    len_base = new char*[len_num + 1];
    unsigned long l = 0;
    if (base_one)
      len_base[l++] = one_id;
    if (base_inch)
      len_base[l++] = inch_id;
    pos = 0;
    unsigned int old_pos = 0;
    while (pos < str.size ()) {
      pos = str.find (' ', old_pos);
      string word = str.substr (old_pos, pos - old_pos);
      if (!base_one && word.compare ("one") == 0) {
	len_base[l++] = one_id;
      } else if (!base_inch && word.compare ("inch")) {
	len_base[l++] = inch_id;
      } else {
	if (atof (word.c_str ()) > 0.0) {
	  len_base[l] = new char[word.size () + 1];
	  strcpy (len_base[l++], word.c_str ());
	}
      }
      while (pos + 1 < str.size () && str[pos + 1] == ' ')
	++pos;
      old_pos = pos + 1;
    }
    len_base[l] = NULL;
  }
  if (str_len_base) {
    delete str_len_base;
    str_len_base = NULL;
  }
  if (str.size () > 0) {
    str_len_base = new char[str.size () + 1];
    strcpy (str_len_base, str.c_str ());
  }

  unsigned long angle_num = 0;
  if (base_deg)
    ++angle_num;
  if (base_pi)
    ++angle_num;
  if (base_tan)
    ++angle_num;
  str = e_anglebase.get_text ();
  pos = 0;
  while (pos < str.size ()) {
    if (str[pos] == ' ') {
      ++angle_num;
      ++pos;
      while (pos < str.size () && str[pos] != ' ')
	++pos;
    } else {
      if (str[pos] == ':')
	str[pos] = ' ';
      else
	str[pos] = tolower (str[pos]);
      ++pos;
    }
  }
  if (pos > 0)
    ++angle_num;
  if (genbase) {
    angle_base = new char*[angle_num + 1];
    unsigned long l = 0;
    if (base_deg)
      angle_base[l++] = deg_id;
    if (base_pi)
      angle_base[l++] = pi_id;
    if (base_tan)
      angle_base[l++] = tan_id;
    pos = 0;
    unsigned long old_pos = 0;
    while (pos < str.size ()) {
      pos = str.find (' ', old_pos);
      string word = str.substr (old_pos, pos - old_pos);
      if (!base_deg && (word.compare ("deg") == 0 ||
			word.compare ("degree") == 0)) {
	angle_base[l++] = deg_id;
      } else if (!base_pi && word.compare ("pi") == 0) {
	angle_base[l++] = pi_id;
      } else if (!base_tan && (word.compare ("tan") == 0 ||
			       word.compare ("atan") == 0 ||
			       word.compare ("arctan") == 0)) {
	angle_base[l++] = tan_id;
      } else {
	if (atof (word.c_str ()) > 0.0) {
	  angle_base[l] = new char[word.size () + 1];
	  strcpy (angle_base[l++], word.c_str ());
	}
      }
      while (pos + 1 < str.size () && str[pos + 1] == ' ')
	++pos;
      old_pos = pos + 1;
    }
    angle_base[l] = NULL;
  }
  if (str_angle_base) {
    delete str_angle_base;
    str_angle_base = NULL;
  }
  if (str.size () > 0) {
    str_angle_base = new char[str.size () + 1];
    strcpy (str_angle_base, str.c_str ());
  }

  use_edges = cb_edges.get_active ();

  f_a = cb_a.get_active ();
  f_as = cb_as.get_active ();
  f_ar = cb_ar.get_active ();
  f_l = cb_l.get_active ();
  f_ls = cb_ls.get_active ();
  f_lr = cb_lr.get_active ();
  f_d = cb_d.get_active ();
  f_dp = cb_dp.get_active ();
  f_par = cb_par.get_active ();
  f_car = cb_car.get_active ();
  f_ai = cb_ai.get_active ();
  f_ap = cb_ap.get_active ();
  f_arp = cb_arp.get_active ();
  f_arc = cb_arc.get_active ();
  f_p = cb_p.get_active ();
  f_pp = cb_pp.get_active ();
  f_l = cb_l.get_active ();
  f_g = cb_g.get_active ();

  return;
}

void AnalyserWindow::cancel ()
{
  close ();
  return;
}

void AnalyserWindow::use_pre ()
{
  cl_prelist.freeze ();
  for (Gtk::CList::SelectionList::iterator iter = 
	 cl_prelist.selection () . begin ();
       iter != cl_prelist.selection () . end ();
       ++iter) {
    int r = iter->get_row_num ();
    string text;
    if (cl_prelist.get_text (r, 0, text)) {
      text += ":.";
      unsigned long k = 0;
      unsigned long l = text.find (':', 0);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_lend.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_angd.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_loopd.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_ratioa.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_ratiom.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_anglemaxn.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_fc.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_pmind.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_pmaxd.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_ppow.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_rmind.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_rmaxd.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_rpow.set_value (atof (text.substr (k, l - k) . c_str ()));
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      s_pcbs.set_value (atof (text.substr (k, l - k) . c_str ()));

      cb_one.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_inch.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_deg.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_pi.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_tan.set_active (text[k] == 't');
      base_tan = cb_tan.get_active ();

      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      e_lenbase.set_text (text.substr (k, l - k) . c_str ());
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      e_anglebase.set_text (text.substr (k, l - k) . c_str ());

      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_edges.set_active (text[k] == 't');

      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_a.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_as.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_ar.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_l.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_ls.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_lr.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_d.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_dp.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_par.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_car.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_ai.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_ap.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_arp.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_arc.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_p.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_pp.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_los.set_active (text[k] == 't');
      k = l + 1;
      l = text.find (':', k);
      if (l == string::npos) {
	sil_msg_warn ("analyse", "Broken analyse configuration: " +
		      text + ".");
	break;
      }
      cb_g.set_active (text[k] == 't');

      break;
    }
  }
  cl_prelist.selection () . clear ();
  cl_prelist.thaw ();
  return;
}

void AnalyserWindow::add_pre ()
{
  string name = e_prename.get_text ();
  if (name.size () > 0) {
    e_prename.set_text ("");
    unsigned long l = 0;
    while (l < name.size ()) {
      if (!isalnum(name[l]) && !isblank(name[l])) 
	name[l] = ' ';
      ++l;
    }
    name += ":";
    string *p = preset_strings.first ();
    while (p) {
      if (name.compare (*p, 0, name.size ()) == 0)
	break;
      p = preset_strings.next ();
    }
    bool add = false;
    if (!p) {
      add = true;
      p = new string ();
    }
    set_values (false);
    char str[2048];
    snprintf (str, 2047, 
	      "%.6g:%.6g:%.6g:%.6g:%.6g:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu:",
	      tol_len_dist,
	      tol_angle_dist,
	      tol_loop_dist,
	      tol_ratio_approx,
	      tol_ratio_min,
	      regular_angle_max_n,
	      loop_fourier_coefficients,
	      para_min_frac,
	      para_max_frac,
	      para_max_power,
	      ratio_min_frac,
	      ratio_max_frac,
	      ratio_max_power,
	      max_pcbs);
    *p = name + static_cast<string>(str) +
      (base_one?"t":"f") + ":" +
      (base_inch?"t":"f") + ":" +
      (base_deg?"t":"f") + ":" +
      (base_pi?"t":"f") + ":" +
      (base_tan?"t":"f") + ":";
    if (str_len_base)
      *p += static_cast<string>(str_len_base);
    *p += ":";
    if (str_angle_base)
      *p += static_cast<string>(str_angle_base);
    *p += static_cast<string>(":") + 
      (use_edges?"t":"f") + ":" +
      (f_a?"t":"f") + ":" +
      (f_as?"t":"f") + ":" +
      (f_ar?"t":"f") + ":" +
      (f_l?"t":"f") + ":" +
      (f_ls?"t":"f") + ":" +
      (f_lr?"t":"f") + ":" +
      (f_d?"t":"f") + ":" +
      (f_dp?"t":"f") + ":" +
      (f_par?"t":"f") + ":" +
      (f_car?"t":"f") + ":" +
      (f_ai?"t":"f") + ":" +
      (f_ap?"t":"f") + ":" +
      (f_arp?"t":"f") + ":" +
      (f_arc?"t":"f") + ":" +
      (f_p?"t":"f") + ":" +
      (f_pp?"t":"f") + ":" +
      (f_los?"t":"f") + ":" +
      (f_g?"t":"f") + ":";
    if (add)
      preset_strings.add_last (p);
    cl_prelist.freeze ();
    cl_prelist.clear ();
    p = preset_strings.first ();
    while (p) {
      gchar const *text[1] = { p->c_str () };
      cl_prelist.append (text);
      p = preset_strings.next ();
    }
    cl_prelist.thaw ();
  } else {
    sil_msg_warn ("analyse",
		  "Select a name for the analyse confgiuration set before adding it.");
  }
  return;
}

void AnalyserWindow::remove_pre ()
{
  cl_prelist.freeze ();
  for (Gtk::CList::SelectionList::iterator iter = 
	 cl_prelist.selection () . begin ();
       iter != cl_prelist.selection () . end ();
       ++iter) {
    int r = iter->get_row_num ();
    string text;
    if (cl_prelist.get_text (r, 0, text)) {
      string *p = preset_strings.first ();
      while (p) {
	if (p->compare (text) == 0) {
	  preset_strings.del (p);
	  delete p;
	  break;
	}
	p = preset_strings.next ();
      }
      cl_prelist.remove_row (r);
      break;
    }
  }
  cl_prelist.selection () . clear ();
  cl_prelist.thaw ();
  return;
}

void AnalyserWindow::save_pre ()
{
  char *home = getenv ("HOME");
  if (home == NULL || !*home)
    home = ".";
  char *str = new char[strlen (home) + 18];
  strcpy (str, home);
  strcat (str, "/.sil");

  struct stat finfo;
  if (stat (str, &finfo) != 0 || !S_ISDIR (finfo.st_mode)) {
    if (mkdir (str, 0777)) {
      sil_msg_error ("analyse", "Cannot access configuration directory" + 
		     static_cast<string>(str) + ".");
      delete str;
      return;
    }
  }

  strcat (str, "/analyse.cfg");

  FILE *f = fopen (str, "w");
  if (f) {
    string *p = preset_strings.first ();
    while (p) {
      if (fputs (p->c_str (), f) == EOF ||
	  fputs ("\n", f) == EOF) {
	sil_msg_error ("analyse", "Writing to " + static_cast<string>(str) +
		       "failed.");
	break;
      }
      p = preset_strings.next ();
    }
    fclose (f);
  } else {
    sil_msg_error ("analyse", "Cannot write to " + static_cast<string>(str) +
		   ".");
  }

  delete str;

  return;
}

void AnalyserWindow::read_preset_file (string const &presetfile)
{
  char buffer[2048];
  FILE *f = fopen (presetfile.c_str (), "r");
  if (f) {
    while (fgets (buffer, 2047, f) != NULL) {
      if (isalnum(*buffer) || isblank (*buffer)) {
	unsigned long l = strlen (buffer);
	while (l > 0) {
	  --l;
	  if (!isalnum(buffer[l]) && !isblank(buffer[l]) &&
	      buffer[l] != ':' && buffer[l] != '.')
	    buffer[l] = ' ';
	}
 	string *p = new string (buffer);
	preset_strings.add_last (p);
	gchar const *text[1] = { p->c_str () };
	cl_prelist.append (text);
      }
    }
    fclose (f);
  }
  return;
}

void AnalyserWindow::reg_toggle ()
{
  cb_ar.set_sensitive (cb_a.get_active ());
  cb_lr.set_sensitive (cb_l.get_active ());
  cb_dp.set_sensitive (cb_d.get_active ());
  cb_par.set_sensitive (cb_dp.get_active () && cb_d.get_active ());
  cb_car.set_sensitive (cb_d.get_active ());
  cb_arp.set_sensitive (cb_ap.get_active ());
  cb_arc.set_sensitive (cb_ap.get_active ());
  return;
}

void analyse_symbol (SimpleList<Symbol const> *sl)
{
  Symbol const *s = sl->first ();
  while (s) {
    const char *t = sil_get_symbol_type (s);
    Symbol const *hs = NULL;
    Solid *S = NULL;
    if (strcmp (t, "kernel") == 0) {
      Entity *e = (Entity *)sil_get_symbol_data (s);
      if (e->is (SIL_SOLID_TYPE)) {
	hs = s;
	S = static_cast<Solid*> (e);
      } else if (e->is (SIL_ANALYSE_TYPE)) {
	sil_progress_bar (true, false);
	((Analyse *)e)->eval ();
	sil_progress_bar (false);
	sil_msg ("analyse", "Updated analysis of " +
		 static_cast<string> (((Analyse *)e)->name ()) + ".");
	sil_updated ();
      }
    }
#ifdef HAVE_ACIS
    else if (strcmp (t, "acis") == 0) {
      ENTITY *e = (ENTITY *)sil_get_symbol_data (s);
      if (is_BODY (e)) {
	hs = s;
	sil_progress_bar (true, false);
	unsigned long s;
	unsigned int ms;
	sil_start_time ();
	S = sil_convert ((BODY *)e);
	sil_measure_time (s, ms);
	sil_progress_bar (false);
	char str[512];
	snprintf (str, 511,
		  "Internal model conversion for analyser: %0ldm %0ld.%03ds.",
		  s / 60, s % 60, ms);
	sil_msg ("analyse", str);
      }
    }
#endif // HAVE_ACIS
    if (hs) {
      AnalyserWindow *aw = manage (new AnalyserWindow (S, s, hs));
      analyser_windows.add_first (aw);      
    }
    s = sl->next ();
  }
  return;
}

void update_handler ()
{
  ListNode<AnalyserWindow> *l = analyser_windows.first ();
  while (l) {
    ListNode<AnalyserWindow> *nn = l->next ();
    l->data () -> update ();
    l = nn;
  }
  return;
}

SIL_PLUGIN ()
{
  sil_register_tool (&analyse_symbol, "Analyse", "0.1.1",
		     "Frank C Langbein", "2000-2001 Frank C Langbein",
		     "Analyse acis solid model.", "acis", "acis:kernel");
  sil_register_tool (&analyse_symbol, "Analyse", "0.1.1",
		     "Frank C Langbein", "2000-2001 Frank C Langbein",
		     "Analyse sil solid model.", "kernel", "kernel");
  sil_register_update_handler (&update_handler);
  return;
}

SIL_CLOSE_PLUGIN ()
{
  sil_deregister_update_handler (&update_handler);
  ListNode<AnalyserWindow> *l = analyser_windows.first ();
  while (l) {
    ListNode<AnalyserWindow> *nn = l->next ();
    l->data () -> close ();
    l = nn;
  }
  if (str_len_base)
    delete str_len_base;
  if (str_angle_base)
    delete str_angle_base;
  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
