// faceter.cc: sil plugin to facet a model
//
// Copyright (C) 2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: faceter.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (faceter, "$Id: faceter.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"
#include "xstring.hxx"
#include "list.hxx"
#include "scanner/points.hxx"
#include "acis/acis.hxx"
#include "acis/silapi.hxx"

double surface_tol = 1.0;
double edge_len = 0.2;
double rand_val = 0.002;

class FaceterWindow;

List<FaceterWindow> faceter_windows;

class FaceterWindow : public Gtk::Window {
public:
  FaceterWindow (BODY *b, string const &n);
  ~FaceterWindow ();
  void update ();
  void close ();
private:
  void cancel ();
  void facet ();
  void facet_body (BODY *B);
  void facet_lump (LUMP *l, transf const *trans);
  void facet_shell (SHELL *s, transf const *trans);
  void facet_face (FACE *f, transf const *trans);
  string sld_name;
  BODY *B;
  unsigned long face_count, face_counter;
  RangePoints *Pts;
  Gtk::VBox vbox;
  Gtk::HBox titlebox, stolbox, edgelenbox, randbox, bbox;
  Gtk::Label titlelabel, title2label;
  Gtk::Label l_stol, l_edgelen, l_rand;
  Gtk::Adjustment a_stol;
  Gtk::SpinButton s_stol;
  Gtk::Adjustment a_edgelen;
  Gtk::SpinButton s_edgelen;
  Gtk::Adjustment a_rand;
  Gtk::SpinButton s_rand;
  Gtk::Button b_facet, b_cancel;
};

FaceterWindow::FaceterWindow (BODY *b, string const &n)
  : Gtk::Window (GTK_WINDOW_TOPLEVEL),
	 vbox (),
	 titlebox (true),
	 stolbox (true),
	 edgelenbox (true),
	 randbox (true),
	 bbox (true),
	 titlelabel ("Generate point cloud:"),
	 title2label (),
	 l_stol ("Tolerance:"),
	 l_edgelen ("Maximal distance:"),
	 l_rand ("Randomize:"),
	 a_stol (surface_tol, 0.0, 100.0, 0.1, 1.0, 0.0),
	 s_stol (a_stol, 0.1, 2),
	 a_edgelen (edge_len, 0.0, 1000.0, 0.1, 1.0, 0.0),
	 s_edgelen (a_edgelen, 0.1, 2),
	 a_rand (rand_val, 0.0, 1000.0, 0.01, 0.1, 0.0),
	 s_rand (a_rand, 0.1, 5),
	 b_facet ("Facet"),
	 b_cancel ("Cancel")
{
  B = b;
  sld_name = n;
  Pts = NULL;

  title2label.set (n.c_str ());
  title2label.set_justify (GTK_JUSTIFY_LEFT);
  titlelabel.set_justify (GTK_JUSTIFY_RIGHT);

  l_stol.set_justify (GTK_JUSTIFY_LEFT);
  l_edgelen.set_justify (GTK_JUSTIFY_LEFT);
  l_rand.set_justify (GTK_JUSTIFY_LEFT);

  vbox.pack_start (titlebox, false, true, 2);
  titlebox.pack_start (titlelabel, true, true, 0);
  titlebox.pack_start (title2label, true, true, 0);

  vbox.pack_start (stolbox, false, true, 2);
  stolbox.pack_start (l_stol, false, true, 0);
  stolbox.pack_start (s_stol, false, true, 0);

  vbox.pack_start (edgelenbox, false, true, 2);
  edgelenbox.pack_start (l_edgelen, false, true, 0);
  edgelenbox.pack_start (s_edgelen, false, true, 0);

  vbox.pack_start (randbox, false, true, 2);
  randbox.pack_start (l_rand, false, true, 0);
  randbox.pack_start (s_rand, false, true, 0);

  vbox.pack_start (bbox, false, true, 2);
  bbox.pack_start (b_facet, false, true, 0);
  bbox.pack_start (b_cancel, false, true, 0);
  
  add (vbox);

  b_facet.clicked.connect (slot (this, &FaceterWindow::facet));
  b_cancel.clicked.connect (slot (this, &FaceterWindow::cancel));

  show_all ();

  return;
}

FaceterWindow::~FaceterWindow ()
{
  if (Pts)
    delete Pts;
  ListNode<FaceterWindow> *n = faceter_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }
  return;
}

void FaceterWindow::update ()
{
  const Symbol *s = sil_find_symbol (sld_name.c_str ());
  if (!s) {
    ListNode<FaceterWindow> *n = faceter_windows.find (this);
    if (n) {
      n->unlink ();
      delete n;
    }
    hide_all ();
    destroy ();
  }
  return;
}

void FaceterWindow::close ()
{
  ListNode<FaceterWindow> *n = faceter_windows.find (this);
  if (n) {
    n->unlink ();
    delete n;
  }
  hide_all ();
  destroy ();
  return;
}

void FaceterWindow::cancel ()
{
  close ();
  return;
}

void FaceterWindow::facet ()
{
  b_facet.set_sensitive (false);
  b_cancel.set_sensitive (false);

  sil_progress_bar (true, false);

  sil_progress_bar_assign (0.0, 0.1);
  sil_progress_bar_text ("Initializing faceter");
  sil_progress_bar_percentage (0.0);

  sil_start_time ();

  API_BEGIN;

  double surf_tol;
  double normal_tol;
  SiL_TriangMode t_mode;
  SiL_GridMode g_mode;
  int max_l;
  int min_l;
  double max_ll;
  double ar;
  SiL_AdjustMode am;
  sil_get_refinement (surf_tol,
		      normal_tol,
		      t_mode,
		      g_mode,
		      max_l,
		      min_l,
		      max_ll,
		      ar,
		      am);

  surface_tol = s_stol.get_value_as_float ();
  edge_len = s_edgelen.get_value_as_float ();
  rand_val = s_rand.get_value_as_float ();

  surf_tol = surface_tol;
  max_ll = edge_len;
  
  REFINEMENT *ref = new REFINEMENT ();
  ref->set_surf_mode (AF_SURF_ALL);

  ref->set_surface_tol (surf_tol);
  ref->set_max_edge_length (max_ll);

  ref->set_normal_tol (normal_tol);
  ref->set_grid_aspect_ratio (ar);
  ref->set_max_grid_lines (max_l);
  ref->set_min_u_grid_lines (min_l);
  ref->set_min_v_grid_lines (min_l);

  ref->set_triang_mode (AF_TRIANG_NONE);
  ref->set_grid_mode (AF_GRID_TO_EDGES);
  ref->set_adjust_mode (AF_ADJUST_ALL);

  // ACIS doesn't explain what exactly the dynamic sufrace tolerance is.
  // Therefor we simply ignore it by setting it to 0.0 (hopefullt that's
  // what this value means).
  ref->set_dynamic_surtol (0.0);
  // Grading should adjust facets to get better parametric aspect ratio of a
  // grid cell. However if it is true, it can cause a segmentation fault
  // in the faceter. I'd call this an ACIS bug, but more information
  // is required...
  ref->set_grading_mode (false);

  sil_acis_check (api_set_default_refinement (ref),
		  "faceter", "set default refinement");

  INDEXED_MESH_MANAGER *mm = new INDEXED_MESH_MANAGER;
  mm->SetTransform (NULL);
  sil_acis_check (api_set_mesh_manager (mm),
		  "faceter", "set mesh manager");

  sil_progress_bar_assign (0.1, 0.6);
  sil_progress_bar_text ("Faceting BODY");
  sil_progress_bar_percentage (0.0);

  if (Pts)
    delete Pts;
  Pts = new RangePoints ();
  facet_body (B);

  sil_progress_bar_assign (0.7, 0.1);
  sil_progress_bar_text ("Stopping faceter");
  sil_progress_bar_percentage (0.0);

  sil_acis_check (api_set_mesh_manager (NULL),
		  "faceter", "set mesh manager");
  delete mm;
  sil_acis_check (api_set_default_refinement (NULL),
		  "faceter", "set default refinement");

  API_END;

  if (rand_val > sil_get_epsilon ()) {
    sil_progress_bar_assign (0.8, 0.2);
    sil_progress_bar_text ("Randomizing");
    sil_progress_bar_percentage (0.0);
    RangePoint *p = Pts->first ();
    while (p) {
      double dx = (static_cast<double> (rand ()) / 
		    static_cast<double> (RAND_MAX + 1.0)) - 0.5;
      double dy = (static_cast<double> (rand ()) / 
		   static_cast<double> (RAND_MAX + 1.0)) - 0.5;
      double dz = (static_cast<double> (rand ()) / 
		   static_cast<double> (RAND_MAX + 1.0)) - 0.5;
      double l = rand_val / sqrt (dx * dx + dy * dy + dz * dz);
      if (l > sil_get_epsilon ()) {
	dx *= l;
	dy *= l;
	dz *= l;
	p->set (p->x () + dx, p->y () + dy, p->z () + dz);
      }
      p = Pts->next ();
    }
  }

  unsigned long sec;
  unsigned int msec;
  sil_measure_time (sec, msec);
  sil_progress_bar_percentage (1.0);

  char *n = new char[strlen (sld_name.c_str()) + 8];
  strcpy (n, "Points_");
  strcat (n, sld_name.c_str ());
  while (!sil_add_symbol (n, (void *)Pts, "scanner", true)) {
    sil_del_symbol (n);
  }

  sil_msg ("Facet", "Faceted " + sld_name + ": " + 
	   static_cast<string const> (n) + " in " +
	   as_string (static_cast<unsigned long> (sec / 60)) + "m " + 
	   as_string (static_cast<unsigned long> (sec % 60)) + "." +
	   as_string (static_cast<unsigned long> (msec)) + "s");

  delete n;
  Pts = NULL;

  sil_refacet ();
  sil_updated ();
  sil_progress_bar (false);
  close ();

  return;
}

void FaceterWindow::facet_body (BODY *B)
{
  ENTITY_LIST face_list;
  sil_acis_check (api_get_faces (B, face_list),
		  "sil_convert", "api_get_faces");
  face_count = face_list.count ();
  face_counter = 0;

  LUMP *l = B->lump();
  LUMP *fl = l;
  while (l) {
    facet_lump (l,
		(B->transform ())?(&B->transform () -> transform ()):NULL);
    l = l->next ();
    if (fl == l)
      l = NULL;
  }
  return;
}

void FaceterWindow::facet_lump (LUMP *l, transf const *trans)
{
  SHELL *s = l->shell();
  SHELL *fs = NULL;
  while (s) {
    facet_shell (s, trans);
    s = s->next ();
    if (fs == s)
      s = NULL;
  }
  return;
}

void FaceterWindow::facet_shell (SHELL *s, transf const *trans)
{
  FACE *f =s->face_list();
  FACE *ff = f;
  while (f) {
    facet_face (f, trans);
    f = f->next_in_list ();
    if (ff == f)
      f = NULL;
  }
  return;
}

void FaceterWindow::facet_face (FACE *f, transf const *trans)
{
  sil_progress_bar_percentage (static_cast<double> (++face_counter) /
			       static_cast<double> (face_count));

  sil_acis_check(api_facet_entity (static_cast<ENTITY *> (f)),
		 "faceter", "facet entity face");

  INDEXED_MESH *m;
  sil_acis_check(api_get_indexed_mesh (f, m),
		 "faceter", "get indexed mesh");
  if (m) {
    int p_num = m->get_num_polygon ();
    int i = 0;
    while (i < p_num) {
      indexed_polygon *poly = m->get_polygon (i++);
      int vertex_num = poly->num_vertex ();
      polygon_vertex **vertex_list = poly->get_vertices ();
      int j = 0;
      while (j < vertex_num) {
	polygon_vertex *V = vertex_list[j++];
	position v = V -> get_position ();
	if (trans)
	  v *= *trans;
	Pts->add_check (v.x (), v.y (), v.z ());
      }
    }
  }

  af_delete_facets (static_cast<ENTITY *> (f));

  return;
}

void facet_symbol (SimpleList<Symbol const> *sl)
{
  Symbol const *s = sl->first ();
  while (s) {
    const char *t = sil_get_symbol_type (s);
    if (strcmp(t, "acis") == 0) {
      ENTITY *e = static_cast<ENTITY *> (sil_get_symbol_data (s));
      if (is_BODY (e)) {
	FaceterWindow *fw = manage (new FaceterWindow 
				     (static_cast<BODY *> (e),
				      static_cast<string const> 
				      (sil_get_symbol_name (s))));
	faceter_windows.add_first (fw);
      }
    }
    // FIXME: facet sil model
    s = sl->next ();
  }
  return;
}

void update_handler ()
{
  ListNode<FaceterWindow> *l = faceter_windows.first ();
  while (l) {
    ListNode<FaceterWindow> *nn = l->next ();
    l->data () -> update ();
    l = nn;
  }
  return;
}

SIL_PLUGIN ()
{
  sil_register_tool (&facet_symbol, "Facet", "0.1",
		     "Frank C Langbein", "2001 Frank C Langbein",
		     "Facet acis solid model.", "acis", "acis:scanner");
  sil_register_update_handler (&update_handler);
  return;
}

SIL_CLOSE_PLUGIN ()
{
  sil_deregister_update_handler (&update_handler);
  ListNode<FaceterWindow> *l = faceter_windows.first ();
  while (l) {
    ListNode<FaceterWindow> *nn = l->next ();
    l->data () -> close ();
    l = nn;
  }
  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
