// register.cc: register point set pairs.
//
// Copyright (C) 2000-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: register.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (register, "$Id: register.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "xmath.hxx"
#include "interface.hxx"
#include "scanner/points.hxx"

bool read_result (const char *name, 
		  double &nx, double &ny, double &nz, double &a,
		  double &dx, double &dy, double &dz)
{
  FILE *fp = fopen (name, "r");
  if (fp == NULL) {
    sil_msg_error ("read", "failed");
    return false;
  }
  double d1, d2, d3, d4, d5, d6, d7;
  fscanf (fp, "%lf %lf %lf %lf %lf %lf %lf",
	  &d1, &d2, &d3, &d4, &d5, &d6, &d7);
  int n = fscanf (fp, "%lf %lf %lf %lf %lf %lf %lf",
		  &nx, &ny, &nz, &a, &dx, &dy, &dz);
  fclose (fp);
  return (n == 7);
}

void register_data (RangePoints *M, const char *m_name, 
		    RangePoints *D, const char *d_name)
{
  sil_msg ("register", "Storing model points");
  string tmpfile1 = sil_tmp_filename ();
  if (!M->save (tmpfile1.c_str ())) {
    sil_msg_error ("register", "failed");
    return;
  }

  sil_msg ("register", "Storing data data");
  string tmpfile2 = sil_tmp_filename ();
  if (!D->save (tmpfile2.c_str ())) {
    unlink (tmpfile1.c_str ());
    sil_msg_error ("register", "failed");
    return;
  }

  sil_msg ("register", "Registering data");
  string tmpfile3 = sil_tmp_filename ();
  char cmd[256];
  strcpy (cmd, "sil_register ");
  strcat (cmd, tmpfile1.c_str ());
  strcat (cmd, " ");
  strcat (cmd, tmpfile2.c_str ());
  strcat (cmd, " ");
  strcat (cmd, tmpfile3.c_str ());
  if (system (cmd) != 0) {
    unlink (tmpfile1.c_str ());
    unlink (tmpfile2.c_str ());
    unlink (tmpfile3.c_str ());
    sil_msg_error ("register", "script failed");
    return;
  }

  sil_msg ("register", "Reading results");
  double nx, ny, nz, a, dx, dy, dz;
  if (!read_result (tmpfile3.c_str (), nx, ny, nz, a, dx, dy, dz)) {
    unlink (tmpfile1.c_str ());
    unlink (tmpfile2.c_str ());
    unlink (tmpfile3.c_str ());
    sil_msg_error ("register", "failed");
    return;
  }
  double l = nx * nx + ny * ny + nz * nz;
  if (l > sil_get_epsilon ()) {
    l = sqrt (l);
    nx /= l;
    ny /= l;
    nz /= l;
  }
  char str[512];
  snprintf (str, 511, "Transformation: %.4g %.4g %.4g %.4g %.4g %.4g %.4g",
	    nx, ny, nz, a, dx, dy, dz);
  sil_msg ("register", str);

  sil_msg ("register", "Registering data set");
  RangePoints *RP = new RangePoints ();

  // data points
  double rp, rt, rs;
  D->rotation (rp, rt, rs);
  double tx, ty, tz;
  D->translation (tx, ty, tz);
  double x,y,z;
  bool r = D->first (x, y, z);
  while (r) {
    double xx =
      ( cos (rs) * cos (rp) - cos (rt) * sin (rp) * sin (rs)) * x +
      ( cos (rs) * sin (rp) + cos (rt) * cos (rp) * sin (rs)) * y +
      ( sin (rs) * sin (rt)                                 ) * z;
    xx += tx;
    double yy =
      (-sin (rs) * cos (rp) - cos (rt) * sin (rp) * cos (rs)) * x +
      (-sin (rs) * sin (rp) + cos (rt) * cos (rp) * cos (rs)) * y +
      ( cos (rs) * sin (rt)                                 ) * z;
    yy += ty;
    double zz = 
      ( sin (rt) * sin (rp)) * x +
      (-sin (rt) * cos (rp)) * y +
      ( cos (rt))            * z;
    zz += tz;
    double f = (nx * xx +  ny * yy + nz * zz) * (1.0 - cos (a));
    double hx = yy * nz - zz * ny;
    double hy = zz * nx - xx * nz;
    double hz = xx * ny - yy * nx;
    double xxx = xx * cos (a) + f * nx + hx * sin (a);
    double yyy = yy * cos (a) + f * ny + hy * sin (a);
    double zzz = zz * cos (a) + f * nz + hz * sin (a);
    RP->add (xxx + dx, yyy + dy, zzz + dz);
    r = D->next (x, y, z);
  }
  
  // Add new set
  char *name = new char[strlen (m_name) + strlen (d_name) + 2];
  strcpy (name, d_name);
  strcat (name, "-");
  strcat (name, m_name);
  while (!sil_add_symbol (name, (void *)RP, "scanner", true)) {
    sil_del_symbol (name);
  }

  unlink (tmpfile1.c_str ());
  unlink (tmpfile2.c_str ());
  unlink (tmpfile3.c_str ());

  sil_redraw ();
  sil_updated ();

  return;
}

void register_data (SimpleList<Symbol const> *sl)
{
  if (sl->length () != 2)
    return;

  Symbol const *s1 = sl->first ();
  Symbol const *s2 = sl->next ();
  if (strcmp (sil_get_symbol_type (s1), "scanner") != 0 &&
      strcmp (sil_get_symbol_type (s2), "scanner") != 0)
    return;

  Scanner *S1 = (Scanner *)sil_get_symbol_data (s1);
  Scanner *S2 = (Scanner *)sil_get_symbol_data (s2);
  if (S1->is ("RangePoints") && 
      S1->is ("RangePoints")) {
    char str[255];
    snprintf (str, 254, "Registering %s with %s", 
	      sil_get_symbol_name (s2), 
	      sil_get_symbol_name (s1));
    sil_msg ("register", str);
    register_data ((RangePoints *)S1, sil_get_symbol_name (s1),
		   (RangePoints *)S2, sil_get_symbol_name (s2));
  }
  return;
}

void mregister_data (SimpleList<Symbol const> *sl)
{
  if (sl->length () < 2)
    return;
  SimpleList<RangePoints> views;

  Symbol const *s = sl->first ();
  while (s) {
    const char *t = sil_get_symbol_type (s);
    if (strcmp (t, "scanner") == 0) {
      Scanner *S = static_cast<Scanner *> (sil_get_symbol_data (s));
      if (S->is ("RangePoints"))
	views.add_last (static_cast<RangePoints *> (S));
    }
    s = sl->next ();
  }

  sil_progress_bar (true, false);

  sil_start_time ();

  // Save files
  sil_progress_bar_percentage (0.1);
  sil_progress_bar_text ("Writing Data Files");
  sil_progress_bar_percentage (0.1);

  unsigned long n = views.length ();
  if (n < 2)
    return;
  string files[n];
  unsigned long l = 0;
  RangePoints *v = views.first ();
  while (v) {
    files[l] = sil_tmp_filename ();
    v->apply_transformation ();
    if (!v->save (files[l])) {
      sil_msg_error ("Register", "Writing data files failed.");
      for (unsigned long k = 0; k <= l; ++k)
	unlink (files[k].c_str ());
      sil_progress_bar (false);
      return;
    }
    ++l;
    v = views.next ();
  }

  string in = sil_tmp_filename ();
  FILE *fp = fopen (in.c_str (), "w");
  if (!fp) {
    sil_msg_error ("Register", "Writing parameter file failed.");
    for (unsigned long k = 0; k <= l; ++k)
      unlink (files[k].c_str ());
    unlink (in.c_str ());
    sil_progress_bar (false);
    return;
  }

  fprintf (fp, "%lu\n", n);
  l = 0;
  v = views.first ();
  while (v) {
    fprintf (fp, "%s\n%lu 1.0 0.0 0.0 0.0 0.0 0.0 0.0\n",
	     files[l++].c_str (), v->num_points ());
    v = views.next ();
  }

  fclose (fp);

  // Run registration
  sil_progress_bar_percentage (0.5);
  sil_progress_bar_text ("Registering Views");
  sil_progress_bar_percentage (0.5);

  string cmd = "mreg <" + in;
  FILE *cp = popen (cmd.c_str (), "r");
  if (!cp) {
    sil_msg_error ("Register", "Execution of mreg failed.");
    for (unsigned long k = 0; k < n; ++k)
      unlink (files[k].c_str ());
    unlink (in.c_str ());
    sil_progress_bar (false);
    return;
  }

  double transforms[n][7];
  char str[256];
  while (fgets (str, 255, cp)) {
    if (strncmp (str, "View ", 5) == 0) {
      unsigned long k;
      float t1, t2, t3, t4, t5, t6, t7;
      sscanf (str, "View %lu transformed by ( %f %f %f %f %f %f %f)",
	      &k, &t1, &t2, &t3, &t4, &t5, &t6, &t7);
      sil_progress_bar_text ("View " + as_string (k) + " transformed by [" +
			     as_string (t1) + ";" +
			     as_string (t2) + ";" +
			     as_string (t3) + ";" +
			     as_string (t4) + ";" +
			     as_string (t5) + ";" +
			     as_string (t6) + ";" +
			     as_string (t7) + "]");
      transforms[k][0] = t1;
      transforms[k][1] = t2;
      transforms[k][2] = t3;
      transforms[k][3] = t4;
      transforms[k][4] = t5;
      transforms[k][5] = t6;
      transforms[k][6] = t7;
    } else if (str[0]) {
      str[strlen (str) - 1] = '\0';
      sil_progress_bar_text (static_cast<string const> (str));
    }
  }
  
  pclose (cp);
  
  // Set transforms
  sil_progress_bar_percentage (0.9);
  sil_progress_bar_text ("Applying Transformations");
  sil_progress_bar_percentage (0.9);
  l = 0;
  v = views.first ();
  while (v) {
    double nx = transforms[l][0];
    double ny = transforms[l][1];
    double nz = transforms[l][2];
    double a = transforms[l][3];
    double dx = transforms[l][5];
    double dy = transforms[l][6];
    double dz = transforms[l][7];
    double cosa = cos (a);
    double sina = sin (a);
    RangePoint *p = v->first ();

    while (p) {
      double xx = p->x ();
      double yy = p->y ();
      double zz = p->z ();
      double f = (nx * xx +  ny * yy + nz * zz) * (1.0 - cosa);
      double hx = yy * nz - zz * ny;
      double hy = zz * nx - xx * nz;
      double hz = xx * ny - yy * nx;
      p->x () = dx + xx * cosa + f * nx + hx * sina;
      p->y () = dy + yy * cosa + f * ny + hy * sina;
      p->z () = dz + zz * cosa + f * nz + hz * sina;
      p = v->next ();
    }
    ++l;
    v = views.next ();
  }
    
  unsigned long sec;
  unsigned int msec;
  sil_measure_time (sec, msec);

  sil_msg ("Register", "Registration finished in" +
	   as_string (static_cast<unsigned long> (sec / 60)) + "m " + 
	   as_string (static_cast<unsigned long> (sec % 60)) + "." +
	   as_string (static_cast<unsigned long> (msec)) + "s");

  for (unsigned long k = 0; k < n; ++k)
    unlink (files[k].c_str ());
  unlink (in.c_str ());
 
  sil_redraw ();
  sil_updated ();
  sil_progress_bar (false);

  return;
}

SIL_PLUGIN ()
{
  sil_register_tool (&register_data, "Register", "0.1",
		     "Frank C Langbein", "2000 Frank C Langbein",
		     "Register point set pairs.", "scanner", "scanner");

  sil_register_tool (&mregister_data, "MRegister", "0.1",
		     "Frank C Langbein", "2000 Frank C Langbein",
		     "Register multiple point sets.", "scanner", "scanner");
  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
