// cluster.cxx: Cluster point cloud.
//
// Copyright (C) 2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: cluster.cc 1 2005-08-10 23:35:07Z frank $
//
// Simple nearest neighbour clustering heuristic requiring linear
// space only.

#include "config.h"

RCSID (cluster, "$Id: cluster.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"
#include "xstring.hxx"
#include "xmath.hxx"
#include "list.hxx"
#include "scanner/scanner.hxx"
#include "scanner/points.hxx"

class ClosestPair {
public:
  ClosestPair (RangePoint **p, unsigned long num, double max_dist);
  ~ClosestPair ();
  double operator () (unsigned long &a, unsigned long &b);
  void update (RangePoint **p, unsigned long a, unsigned long b);
private:
  double sqdist (RangePoint *p, RangePoint *q, double max);
  void find_neighbour (RangePoint **p, unsigned long l);
  unsigned long n;
  unsigned long *neighbour;
  double *dist;
  double max;
};

ClosestPair::ClosestPair (RangePoint **p, unsigned long num, double max_dist)
{
  n = num;
  max = max_dist * max_dist * 4.0;
  neighbour = new unsigned long[n];
  dist = new double[n];
  unsigned long l = 0;
  while (l < n)
    dist[l++] = max;

  // Find all neighbours.
  unsigned long k = 0;
  while (k < n - 1) {
    sil_progress_bar_percentage (static_cast<double> (k) / 
				 static_cast<double> (n << 1));
    unsigned long c = neighbour[k];
    double d = dist[k];
    unsigned long j = k + 1;
    while (j < n) {
      double dd = sqdist (p[k], p[j], d);
      if (dd < d) {
	c = j;
	d = dd;
      }
      if (dd < dist[j]) {
	neighbour[j] = k;
	dist[j] = dd;
      }
      ++j;
    }
    neighbour[k] = c;
    dist[k++] = d;
  }
}

ClosestPair::~ClosestPair ()
{
  delete neighbour;
  delete dist;
}

double ClosestPair::operator () (unsigned long &a, unsigned long &b)
{
  double d = dist[0];
  unsigned long c = 0;
  unsigned long j = 1;
  while (j < n) {
    if (dist[j] < d) {
      d = dist[j];
      c = j;
    }
    ++j;
  }
  a = c;
  b = neighbour[c];
  return sqrt (d);
}

void ClosestPair::update (RangePoint **p, unsigned long a, unsigned long b)
{
  // Disable b.
  dist[b] = max;
  // Recompute distances to a.
  double d = max;
  unsigned long c = n;
  unsigned long l = 0;
  while (l < n) {
    if (l != a && p[l]) {
      double dd = sqdist (p[a], p[l], d);
      if (dd < d) {
	d = dd;
	c = l;
      }
      if (dd < dist[l]) {
	// new a closer than old neighbour, so record new neighbour
	neighbour[l] = a;
	dist[l] = dd;
      } else if (neighbour[l] == a || neighbour[l] == b) {
	// Old neighbour was a or  b, find new one.
	find_neighbour (p, l);
      }
    }
    ++l;
  }
  neighbour[a] = c;
  dist[a] = d;
  return;
}

double ClosestPair::sqdist (RangePoint *p, RangePoint *q, double max)
{
  double dx = p->x () - q->x ();
  double dy = p->y () - q->y ();
  double dz = p->z () - q->z ();
  return dx * dx + dy * dy + dz * dz;
}

void ClosestPair::find_neighbour (RangePoint **p, unsigned long l)
{
  double d = max;
  unsigned long c = n;
  unsigned long k = 0;
  while (k < n) {
    if (k != l && p[k]) {
      double dd = sqdist (p[l], p[k], d);
      if (dd < d) {
	d = dd;
	c = k;
      }
    }
    ++k;
  }
  neighbour[l] = c;
  dist[l] = d;
  return;
}

void sil_cluster (RangePoints *pts, double dist)
{
  // Points to array
  sil_progress_bar_text ("Initializing Point Array");
  unsigned long num_pts = pts->num_points ();
  if (num_pts < 1)
    return;
  RangePoint **p = new RangePoint*[num_pts + 1];
  unsigned int *n = new unsigned int[num_pts];
  unsigned long l = 0;
  RangePoint *q = pts->replace_points ();
  while (q) {
    n[l] = 1;
    p[l++] = q;
    q = q->next ();
  }
  p[num_pts] = NULL;

  // Initialize closest pair data structure
  sil_progress_bar_percentage (0.0);
  sil_progress_bar_text ("Initializing Closest Pair Data Structure");
  ClosestPair cp (p, num_pts, dist);

  // Cluster points.
  sil_progress_bar_percentage (0.5);
  sil_progress_bar_text ("Clustering");
  while (true) {
    unsigned long a, b;
    // Find closest pair.
    double d = cp (a, b);
    if (d > dist || sil_cancel_request ())
      break;
    sil_progress_bar_percentage (0.5 + d / (2.0 * dist));
    // Combine closest pair to a.
    double N = static_cast<double> (n[a] + n[b]);
    p[a]->x () = (p[a]->x () * static_cast<double> (n[a]) +
		  p[b]->x () * static_cast<double> (n[b])) / N;
    p[a]->y () = (p[a]->y () * static_cast<double> (n[a]) +
		  p[b]->y () * static_cast<double> (n[b])) / N;
    p[a]->z () = (p[a]->z () * static_cast<double> (n[a]) +
		  p[b]->z () * static_cast<double> (n[b])) / N;
    n[a] += n[b];
    // Drop b.
    delete p[b];
    p[b] = NULL;
    // Update closest pair data structure.
    cp.update (p, a, b);
  }

  // Create new point list
  sil_progress_bar_text ("Creating New Point List");
  l = 0;
  while (!p[l] && l < num_pts)
    ++l;
  RangePoint *np = p[l++];
  RangePoint *sp = np;
  while (np) {
    while (!p[l] && l < num_pts)
      ++l;
    np->next () = p[l];
    np = p[l];
    ++l;
  }
  pts->replace_points (sp);
  delete p;

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
