// icrp.cxx: ICRP multiple point cloud registration.
//
// Copyright (C) 2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: icrp.cc 1 2005-08-10 23:35:07Z frank $
//
// Based on the article "Simultaneous Registration of Multiple Range
// Views for Use in Reverse Engineering of CAD Models" by D.W. Eggert,
// A.W. Fitzgibbon and R.B. Fisher in Computer Vision and Image
// Understandig, 68(3):253-272, 1998 and their implementation. Applied
// various modifications considering our point clouds without normals,
// etc.

#include "config.h"

RCSID (icrp, "$Id: icrp.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"
#include "xstring.hxx"
#include "xmath.hxx"
#include "list.hxx"
#include "scanner/scanner.hxx"
#include "scanner/points.hxx"

class KDTree;
class RPPair;

double sn = 0.0; // Ratio between maximal bounding box and max diff
                 // between normsl (2) for x-metric.

// K-D-Trees data structures for the ICRP registration.

// A node in a KDTree.
class KDNode {
protected:
  KDNode (short c, RangePoint *s, KDNode *l, KDNode *r) {
    coord = c;
    split = s;
    left = l;
    right = r;
  }
  ~KDNode () {
    if (left)
      delete left;
    if (right)
      delete right;
  }
  RangePoint *closest_point (RangePoint *P, double &max) const;
  RangePoint *closest_normal_point (RangePoint *P, double &max) const;
  void close_points (RangePoint *P, double max, SimpleList<RangePoint> &rp);
  short coord;
  RangePoint *split;
  KDNode *left, *right;
  friend KDTree;
  friend RPPair *register_points (KDTree *t1, KDTree *t2);
};

RangePoint *KDNode::closest_point (RangePoint *P, double &max) const
{
  RangePoint *Q = NULL;
  double dx = P->x () - split->x ();
  double dy = P->y () - split->y ();
  double dz = P->z () - split->z ();
  double d = sqrt (dx * dx + dy * dy + dz * dz);
  if (d < max) {
    max = d;
    Q = split;
  }
  if (left && P->get (coord) - max < split->get (coord)) {
    RangePoint *QQ = left->closest_point (P, max);
    if (QQ)
      Q = QQ;
  }
  if (right && P->get (coord) + max > split->get (coord)) {
    RangePoint *QQ = right->closest_point (P, max);
    if (QQ)
      Q = QQ;
  }
  return Q;
}

RangePoint *KDNode::closest_normal_point (RangePoint *P, double &max) const
{
  RangePoint *Q = NULL;
  double dx = P->x () - split->x ();
  double dy = P->y () - split->y ();
  double dz = P->z () - split->z ();
  double d = sqrt (dx * dx + dy * dy + dz * dz);
  if (d < max) {
    dx = P->nx () - split->nx ();
    dy = P->ny () - split->ny ();
    dz = P->nz () - split->nz ();
    d += sn * sqrt (dx * dx + dy * dy + dz * dz);
    if (d < max) {
      max = d;
      Q = split;
    }
  }
  if (left && P->get (coord) - max < split->get (coord)) {
    RangePoint *QQ = left->closest_normal_point (P, max);
    if (QQ)
      Q = QQ;
  }
  if (right && P->get (coord) + max > split->get (coord)) {
    RangePoint *QQ = right->closest_normal_point (P, max);
    if (QQ)
      Q = QQ;
  }
  return Q;
}

void KDNode::close_points (RangePoint *P, double max,
			   SimpleList<RangePoint> &rp)
{
  if (P != split) {
    double dx = P->x () - split->x ();
    double dy = P->y () - split->y ();
    double dz = P->z () - split->z ();
    double d = sqrt (dx * dx + dy * dy + dz * dz);
    if (d < max)
      rp.add_first (split);
  }
  if (left && P->get (coord) - max < split->get (coord))
    left->close_points (P, max, rp);
  if (right && P->get (coord) + max > split->get (coord))
    right->close_points (P, max, rp);
  return;
}

// The KDTree for a view.
class KDTree {
public:
  KDTree(RangePoints *P);
  ~KDTree () {
    if (root)
      delete root;
  }
  RangePoint *closest_point (RangePoint *p, double max, bool x) const;
  void close_points (RangePoint *P, double max, SimpleList<RangePoint> &rp);
private:
  KDNode *create_tree (SimpleList<RangePoint> *pts) const;
  KDNode *root;
};

// Create a new KD Tree for a view.
KDTree::KDTree (RangePoints *P)
{
  assert (P->num_points () > 0);
  SimpleList<RangePoint> *ps = new SimpleList<RangePoint>;
  RangePoint *p = P->first ();
  while (p) {
    ps->add_last (p);
    p = P->next ();
  }
  root = create_tree (ps);
  delete ps;
};

// Recursive algorithm creating a KDTree for a list of points in the
// view.
KDNode *KDTree::create_tree (SimpleList<RangePoint> *pts) const
{
  // Last node.
  if (pts->length () == 0)
    return NULL;
  else if (pts->length () == 1)
    return new KDNode (6, pts->unlink_first (), NULL, NULL);
  
  // Compute dimension limits of point set and sum up coordinates.
  RangePoint *p = pts->first ();
  double max[3] = { p->x (), p->y (), p->z () };
  double min[3] = { p->x (), p->y (), p->z () };
  double sum[3] = { p->x (), p->y (), p->z () };
  p = pts->next ();
  while (p) {
    sum[0] += p->x ();
    sum[1] += p->y ();
    sum[2] += p->z ();
    if (p->x () > max[0])
      max[0] = p->x ();
    else if (p->x () < min[0])
      min[0] = p->x ();
    if (p->y () > max[1])
      max[1] = p->y ();
    else if (p->y () < min[1])
      min[1] = p->y ();
    if (p->z () > max[2])
      max[2] = p->z ();
    else if (p->z () < min[2])
      min[2] = p->z ();
    p = pts->next ();
  }

  // Find largest dimension
  short coord = 0;
  if (fabs (max[1] - min[1]) > fabs (max[0] - min[0]))
    coord = 1;
  if (fabs (max[2] - min[2]) > fabs (max[coord] - min[coord]))
    coord = 2;

  // Split list
  double split = sum[coord] / pts->length ();
  SimpleList<RangePoint> *left = new SimpleList<RangePoint>;
  RangePoint *s = pts->first ();
  p = pts->next ();
  if (coord == 0) {
    double sdist = fabs (s->x () - split);
    while (p) {
      RangePoint *np = pts->next ();
      double d = fabs (p->x () - split);
      if (d < sdist) {
	if (s->x () < split) {
	  left->add_last (s);
	  pts->del (s);
	}
	s = p;
	sdist = d;
      } else {
	if (p->x () < split) {
	  left->add_last (p);
	  pts->del (p);
	}
      }
      p = np;
    }
    pts->del (s);
  } else if (coord == 1) {
    double sdist = fabs (s->y () - split);
    while (p) {
      RangePoint *np = pts->next ();
      double d = fabs (p->y () - split);
      if (d < sdist) {
	if (s->y () < split) {
	  left->add_last (s);
	  pts->del (s);
	}
	s = p;
	sdist = d;
      } else {
	if (p->y () < split) {
	  left->add_last (p);
	  pts->del (p);
	}
      }
      p = np;
    }
    pts->del (s);
  } else if (coord == 2) {
    double sdist = fabs (s->z () - split);
    while (p) {
      RangePoint *np = pts->next ();
      double d = fabs (p->z () - split);
      if (d < sdist) {
	if (s->z () < split) {
	  left->add_last (s);
	  pts->del (s);
	}
	s = p;
	sdist = d;
      } else {
	if (p->z () < split) {
	  left->add_last (p);
	  pts->del (p);
	}
      }
      p = np;
    }
    pts->del (s);
  }

  // Create node.
  KDNode *n = new KDNode (coord, s, create_tree (left), create_tree (pts));
  delete left;

  return n;
}

// Find the closest point to p in the KDTree.
RangePoint *KDTree::closest_point (RangePoint *p, double max, bool x) const
{
  if (root) {
    if (x)
      return root->closest_normal_point (p, max);
    return root->closest_point (p, max);
  }
  return NULL;
}

void KDTree::close_points (RangePoint *P, double max,
			   SimpleList<RangePoint> &rp)
{
  if (root)
    root->close_points (P, max, rp);
  return;
}

// A pair of corresponding points in two different views as determined
// by the ICRP registration.
class RPPair {
public:
  RPPair (RangePoint *p1, RangePoint *p2, RPPair *n) {
    point1 = p1;
    point2 = p2;
    next = n;
  }
  RangePoint *point1;
  RangePoint *point2;
  RPPair *next;
};

// ICP normal/position registration using KDTrees.
RPPair *register_normal_points (RangePoints *v1, KDTree *t2, double max_dist)
{
  RPPair *head = NULL;
  // ICP - find pairs of closest points with respect to position and
  // normal in two views.
  RangePoint *p1 = v1->first ();
  while (p1) {
    // Closest point in t2 to p1 (closer than max_dist)
    RangePoint *p2 = t2->closest_point (p1, max_dist, true);
    if (p2)
      head = new RPPair (p1, p2, head);
    p1 = v1->next ();
  }
  return head;
}

// ICRP registration using KDTrees.
RPPair *register_points (KDTree *t1, RangePoints *v1, 
			 KDTree *t2, RangePoints *v2, 
			 double max_dist)
{
  RPPair *head = NULL;

  // Traverse the smaller list of points.
  bool rev = false;
  if (v2->num_points () < v1->num_points ()) {
    KDTree *tt = t2;
    rev = true;
    t2 = t1;
    t1 = tt;
    RangePoints *vv = v2;
    v2 = v1;
    v1 = vv;
  }

  // ICRP - find pairs of corresponding closest points in two views.
  RangePoint *p1 = v1->first ();
  while (p1) {
    // Closest point in t2 to p1 (closer than max_dist)
    RangePoint *p2 = t2->closest_point (p1, max_dist, false);
    if (p2) {
      // Closest point in t1 to p2
      RangePoint *pp = t1->closest_point (p2, max_dist, false);
      if (pp == p1) {
	// Closest points match, register pair
	if (rev) 
	  head = new RPPair (p2, p1, head);
	else
	  head = new RPPair (p1, p2, head);
      }
    }
    p1 = v1->next ();
  }
  
  return head;
}

// Use the given KDTree to compute average normals as the averae of
// normals generated by the point itself and the points close to it
// (within max) as determined by the KDTree and then regenerate a
// KDTree based on positions and normals.
void view_normals (KDTree *t, RangePoints *v, double max)
{
  // Expand kdtree using appoximate normals.
  RangePoint *p = v->first ();
  while (p) {
    SimpleList<RangePoint> rp;
    // Find points close to p.
    t->close_points (p, max, rp);
    // Generate average normals from points close to p;
    if (rp.length () > 1) {
      double dx[rp.length ()];
      double dy[rp.length ()];
      double dz[rp.length ()];
      unsigned long l = 0;
      RangePoint *q = rp.first ();
      while (q) {
	dx[l] = q->x () - p->x ();
	dy[l] = q->y () - p->y ();
	dz[l++] = q->z () - p->z ();
	q = rp.next ();
      }
      double nx = 0;
      double ny = 0;
      double nz = 0;
      unsigned long n = 0;
      l = 0;
      while (l < rp.length ()) {
	unsigned long k = l + 1;
	while (k < rp.length ()) {
	  double cx = dy[l] * dz[k] - dz[l] * dy[k];
	  double cy = dz[l] * dx[k] - dx[l] * dz[k];
	  double cz = dx[l] * dy[k] - dy[l] * dx[k];
	  double cl = sqrt (cx * cx + cy * cy + cz * cz);
	  if (cl > sil_get_epsilon ()) {
	    cx /= cl;
	    cy /= cl;
	    cz /= cl;
	    if (nx * cx + ny * cy + nz * cz < 0.0) {
	      nx = ((double)n * nx - cx) / ((double)n + 1);
	      ny = ((double)n * ny - cy) / ((double)n + 1);
	      nz = ((double)n * nz - cz) / ((double)n + 1);
	    } else {
	      nx = ((double)n * nx + cx) / ((double)n + 1);
	      ny = ((double)n * ny + cy) / ((double)n + 1);
	      nz = ((double)n * nz + cz) / ((double)n + 1);
	    }
	    double nl = sqrt (nx * nx + ny * ny + nz * nz);
	    if (nl > sil_get_epsilon ()) {
	      nx /= nl;
	      ny /= nl;
	      nz /= nl;
	    }
	  }
	  ++k;
	}
	++l;
      }
      p->set_normal (nx, ny, nz);
    } else {
      p->set_normal (0.0, 0.0, 0.0);
    }
    p = v->next ();
  }
  return;
}

// Compute the forces and from this the translational and angular
// veloctiy and acceleration for the data sets. Then update the
// transformations accordingly. This is the core of the spring
// simulation assuming that the forces are constant over a short time
// period dt. 
double compute_forces (unsigned long num_views, unsigned long l,
		       RPPair *pairs[], RangePoints *view[],
		       double cm[3], double trans[3], double rot[9],
		       double v[3], double mu[3],
		       double av[3], double amu[3],
		       double spring_constant, double accel_damp, double dt,
		       unsigned long &damped_count,
		       bool x)
{
  // Compute the forces between view l and all the other views using
  // pairs.
  double F0 = 0.0;
  double F1 = 0.0;
  double F2 = 0.0;
  double T[3] = { 0.0, 0.0, 0.0 };
  double M[3] = { 0.0, 0.0, 0.0 };
  unsigned long k = 0;
  while (k < num_views) {
    if (k != l) {
      RPPair const *p = pairs[k];
      while (p) {
	double f[3] = {
  	  p->point2->x () - p->point1->x (),
          p->point2->y () - p->point1->y (),
	  p->point2->z () - p->point1->z ()
	};
	double D[3];
	// Compute distance between point and centre of mass.
	if (x || k > l) {
	  // If preregistration or if k > l, point1 corresponds to
	  // first view in pair which we consider at the moment.
	  D[0] = p->point1->x () - cm[0];
	  D[1] = p->point1->y () - cm[1];
	  D[2] = p->point1->z () - cm[2];
	} else {
	  // Otherwise it's point2.
	  D[0] = p->point2->x () - cm[0];
	  D[1] = p->point2->y () - cm[1];
	  D[2] = p->point2->z () - cm[2];
	  // Adjust force direction for ICRP registartions (not for ICP
	  // preregistration).
	  f[0] = -f[0];
	  f[1] = -f[1];
	  f[2] = -f[2];
	}
	// Translational force.
	F0 += f[0];
	F1 += f[1];
	F2 += f[2];
	// Torsion.
	T[0] += D[1] * f[2] - D[2] * f[1];
	T[1] += D[2] * f[0] - D[0] * f[2];
	T[2] += D[0] * f[1] - D[1] * f[0];
	// Moment of inertia
	M[0] += D[1] * D[1] + D[2] * D[2];
	M[1] += D[0] * D[0] + D[2] * D[2];
	M[2] += D[0] * D[0] + D[1] * D[1];
	p = p->next;
      }
    }
    ++k;
  }

  // Compute translational acceleration.
  double a[3] = { 
    F0 * spring_constant / view[l]->num_points (),
    F1 * spring_constant / view[l]->num_points (),
    F2 * spring_constant / view[l]->num_points (),
  };

  // Compute angular acceleration.
  double aa[3] = { 0.0, 0.0, 0.0 };
  if (fabs (M[0]) > sil_get_epsilon ())
    aa[0] = T[0] * spring_constant / M[0];
  if (fabs (M[1]) > sil_get_epsilon ())
    aa[1] = T[1] * spring_constant / M[1];
  if (fabs (M[2]) > sil_get_epsilon ())
    aa[2] = T[2] * spring_constant / M[2];

  // Compute translational update.
  double dtdt2 = dt * dt / 2.0;
  trans[0] = v[0] * dt + a[0] * dtdt2;
  trans[1] = v[1] * dt + a[1] * dtdt2;
  trans[2] = v[2] * dt + a[2] * dtdt2;

  // Compute angular update.
  double au0 = av[0] * dt + aa[0] * dtdt2;
  double au1 = av[1] * dt + aa[1] * dtdt2;
  double au2 = av[2] * dt + aa[2] * dtdt2;

  // Compute rotation.
  double sin0 = sin (au0);
  double cos0 = cos (au0);
  double sin1 = sin (au1);
  double cos1 = cos (au1); 
  double sin2 = sin (au2);
  double cos2 = cos (au2);
  rot[0] = cos1 * cos2 + sin0 * sin1 * sin2;
  rot[1] = -cos1 * sin2 + sin0 * sin1 * cos2;
  rot[2] = cos0 * sin1;
  rot[3] = cos0 * sin2;
  rot[4] = cos0 * cos2;
  rot[5] = -sin0;
  rot[6] = -sin1 * cos2 + sin0 * cos1 * sin2;
  rot[7] = sin1 * sin2 + sin0 * cos1 * cos2;
  rot[8] = cos0 * cos1;

  // New translation is (I - ROT) cm + t
  trans[0] += (1.0 - rot[0]) * cm[0] - rot[1] * cm[1] - rot[2] * cm[2];
  trans[1] += -rot[3] * cm[0] + (1.0 - rot[4]) * cm[1] - rot[5] * cm[2];
  trans[2] += -rot[6] * cm[0] - rot[7] * cm[1] + (1.0 - rot[8]) * cm[2];

  // Compute velocity update.
  v[0] = mu[0] * v[0] + a[0] * dt;
  v[1] = mu[1] * v[1] + a[1] * dt;
  v[2] = mu[2] * v[2] + a[2] * dt;

  // Compute angular velocity update
  av[0] = amu[0] * av[0] + aa[0] * dt;
  av[1] = amu[1] * av[1] + aa[1] * dt;
  av[2] = amu[2] * av[2] + aa[2] * dt;

  // Adjust damping diagonal matrices
  bool flag = false;
  for (unsigned long k = 0; k < 3; ++k) {
    if (v[k] * a[k] < 0.0) {
      mu[k] *= accel_damp;
      flag = true;
    }  else {
      mu[k] = 1.0;
    }
    if (av[k] * aa[k] < 0.0) {
      amu[k] *= accel_damp;
      flag = true;
    }  else {
      amu[k] = 1.0;
    }
  }
  if (flag)
    ++damped_count;

  return F0 * F0 + F1 * F1 + F2 * F2;
}

void sil_icrp (SimpleList<RangePoints> *views,
	       double max_dist,
	       double derr,
	       double spring_constant,
	       double minf,
	       double accel_damp,
	       double time_damp,
	       double dt0,
	       unsigned int preiter,
	       unsigned int max_iter1,
	       unsigned int max_iter2,
               bool anim)
{
  unsigned int num_views = views->length ();
  if (num_views < 2)
    return;

  accel_damp = 1.0 - accel_damp;
  double time_dec = 1.0 - time_damp;
  double time_inc = 1.0 + time_damp;
  max_dist *= max_dist;
  minf *= minf;
  max_iter1 += preiter;
  preiter += 1;

  RangePoints *view[num_views];
  unsigned long l = 0; 
  RangePoints *v = views->first ();
  while (v) {
    view[l++] = v;
    v = views->next ();
  }

  sil_progress_bar_assign (0.0, 0.1);
  sil_progress_bar_text ("Initializing Views");

  // Initialize the views and their transforamtions also computing the
  // centres of mass,
  double cm[num_views][3];
  for (l = 0; l < num_views; ++l) {
    sil_progress_bar_percentage (static_cast<double> (l) /
				 static_cast<double> (num_views));
    // Apply any transformation.
    view[l]->apply_transformation ();
    // Compute centre of mass.
    cm[l][0] = cm[l][1] = cm[l][2] = 0.0;
    RangePoint *p = view[l]->first ();
    while (p) {
      cm[l][0] += p->x ();
      cm[l][1] += p->y ();
      cm[l][2] += p->z ();
      p = view[l]->next ();
    }
    if (view[l]->num_points () > 0) {
      cm[l][0] /= view[l]->num_points ();
      cm[l][1] /= view[l]->num_points ();
      cm[l][2] /= view[l]->num_points ();
    }
  }

  // Initialize x-metric constant.
  sn = max_dist / 2.0;

  KDTree *kd_tree[num_views];
  RPPair *pairs[num_views][num_views];

  sil_progress_bar_assign (0.1, 0.9);
  sil_progress_bar_text ("Registration");

  unsigned int iter1 = 0;
  double per = 0.5;
  double error = 1e50;
  double old_error = 0.0;
  double error0 = 0.0;
  double derror = 1e50;
  do {
    ++iter1;

    if (sil_cancel_request ())
      break;
    
    // Prepare new spring simulation.
    l = 0;
    while (l < num_views) {
      // Create KD Trees for registration.
      sil_progress_bar_text ("Creating KDTree for view " + as_string (l));
      kd_tree[l] = new KDTree (view[l]);
      if (iter1 == 1 && preiter > 1) {
	// In first iteration generate KDTree and then expand it
	// computing the average normals based on the first tree.
	sil_progress_bar_text ("Computing normals for view " + as_string (l));
	view_normals (kd_tree[l], view[l], max_dist);
      }
      sil_progress_bar_percentage (per);
      ++l;
    }

    // ICRP registration for each pair.
    l = 0;
    while (l < num_views) {
      unsigned int k = l + 1;
      while (k < num_views) {
	sil_progress_bar_percentage (per);
	if (iter1 < preiter) {
	  sil_progress_bar_text ("Pre-registering views " + as_string (l) +
				 " and " + as_string (k));
	  pairs[l][k] = register_normal_points (view[l], kd_tree[k], max_dist);
	  pairs[k][l] = register_normal_points (view[k], kd_tree[l], max_dist);
	} else {
	  sil_progress_bar_text ("Registering views " + as_string (l) +
				 " and " + as_string (k));
	  pairs[l][k] = pairs[k][l] = register_points (kd_tree[l], view[l],
						       kd_tree[k], view[k],
						       max_dist);
	}
	++k;
      }
      ++l;
    }
    // Delete KDTrees.
    l = 0;
    while (l < num_views)
      delete kd_tree[l++];

    // Run spring simulation.
    sil_progress_bar_percentage (per);
    sil_progress_bar_text ("Running spring simulation " + as_string (iter1));
    sil_progress_bar_percentage (per);
    double dt = dt0;
    double v[num_views][3];
    double mu[num_views][3];
    double av[num_views][3];
    double amu[num_views][3];
    double trans[num_views][3];
    double rot[num_views][9];
    l = 0;
    while (l < num_views) {
      v[l][0] = v[l][1] = v[l][2] = 
	av[l][0] = av[l][1] = av[l][2] = 0.0;
      mu[l][0] = mu[l][1] = mu[l][2] = 
	amu[l][0] = amu[l][1] = amu[l][2] = 1.0;
      ++l;
    }
    double maxf;
    unsigned int iter2 = 0;
    do {
      ++iter2;

      // Update the transformations of all views based on the spring
      // simulation.
      unsigned long damped_count = 0;
      maxf = 0.0;
      l = 0;
      while (l < num_views) {
	double f = compute_forces (num_views, l, pairs[l], view,
				   cm[l], trans[l], rot[l],
				   v[l], mu[l], av[l], amu[l],
				   spring_constant, accel_damp, dt,
				   damped_count, iter1 < preiter);
	if (f > maxf)
	  maxf = f;
	++l;
      }

      // Apply transformations
      l = 0;
      while (l < num_views) {
	// View:
	view[l]->apply_transformation (rot[l][0], rot[l][1], rot[l][2],
				       rot[l][3], rot[l][4], rot[l][5],
				       rot[l][6], rot[l][7], rot[l][8],
				       trans[l][0], trans[l][1], trans[l][2],
				       iter1 < preiter);
	// Centre of mass:
	double x = cm[l][0];
	double y = cm[l][1];
	double z = cm[l][2];
	cm[l][0] = trans[l][0] + rot[l][0] * x + rot[l][1] * y + rot[l][2] * z;
	cm[l][1] = trans[l][1] + rot[l][3] * x + rot[l][4] * y + rot[l][5] * z;
	cm[l][2] = trans[l][2] + rot[l][6] * x + rot[l][7] * y + rot[l][8] * z;
	++l;
      }

      // Adjust time step:
      if (damped_count > (num_views >> 1)) {
	// If majority of views are damped, decrease time step,
	dt *= time_dec;
      } else {
	// otherwise increase time step.
	dt *= time_inc;
      }

      if (anim && iter2 % 10 == 0)
        sil_redraw ();

    } while(maxf > minf && iter2 < max_iter2);

    if (anim)
      sil_redraw ();

    old_error = error;

    // Compute new average error of all point pairs and delete corresponding
    // pair lists.
    error = 0.0;
    unsigned long n = 0;
    l = 0;
    while (l < num_views) {
      // Error and cleanup for registration.
      if (iter1 < preiter) {
	unsigned int k = 0;
	while (k < num_views) {
	  if (l != k) {
	    RPPair *p = pairs[l][k];
	    while (p) {
	      double dx = p->point2-> x () - p->point1-> x ();
	      double dy = p->point2-> y () - p->point1-> y ();
	      double dz = p->point2-> z () - p->point1-> z ();
	      error += sqrt (dx * dx + dy * dy + dz * dz);
	      ++n;
	      RPPair *np = p->next;
	      delete p;
	      p = np;
	    }
	  }
	  ++k;
	}
      } else {
	unsigned int k = l + 1;
	while (k < num_views) {
	  RPPair *p = pairs[l][k];
	  while (p) {
	    double dx = p->point2-> x () - p->point1-> x ();
	    double dy = p->point2-> y () - p->point1-> y ();
	    double dz = p->point2-> z () - p->point1-> z ();
	    error += sqrt (dx * dx + dy * dy + dz * dz);
	    ++n;
	    RPPair *np = p->next;
	    delete p;
	    p = np;
	  }
	  ++k;
	}
      }
      ++l;
    }
    if (n > 0)
      error /= static_cast<double> (n);

    if (sil_cancel_request ())
      break;

    if (iter1 > 1) {
      per = error / error0;
      if (per > 1.0 || per < 0.1) {
	error0 = 2.0 * error;
	per = 0.5;
      } else {
	per = 1.0 - per;
      }
      sil_progress_bar_text ("Improvement: " +
                             as_string (1.0 - 2.0 * error / error0));
    } else {
      error0 = 2.0 * error;
      per = 0.5;
    }
    sil_progress_bar_percentage (per);

    derror = fabs (error - old_error);
    if (iter1 < preiter && derror < derr) {
      preiter = iter1;
      derror = 1e5;
    }

  } while (iter1 < max_iter1 && derror > derr &&
	   error > sil_get_epsilon ());

  if (iter1 > 1) 
    sil_msg ("ICRP", "Registration error " + as_string (error0) + " -> " +
	     as_string (error));

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
