// points.cxx: Points class implementation.
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: points.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (points, "$Id: points.cc 1 2005-08-10 23:35:07Z frank $");

#include "interface.hxx"
#include "scanner/points.hxx"

string const Scanner::tid ("Scanner");
string const RangePoint::tid ("RangePoint");
string const RangePoints::tid ("RangePoints");
string const RangeMesh::tid ("RangeMesh");

RangePoints::RangePoints () : Scanner ()
{
  window = head = NULL;
  tx = ty = tz = rp = rt = rs = 0.0;
  count = 0;
}

RangePoints::~RangePoints ()
{
  while (head) {
    window = head->next ();
    delete head;
    head = window;
  }
}

void RangePoints::add (double x, double y, double z)
{
  ++count;
  RangePoint *p = new RangePoint (x, y, z, head);
  head = p;
  return;
}

void RangePoints::add (RangePoint *p)
{
  ++count;
  p->next () = head;
  head = p;
  return;
}

void RangePoints::add_check (double x, double y, double z)
{
  window = head;
  while (window) {
    double dx = x - window->x ();
    double dy = y - window->y ();
    double dz = z - window->z ();
    if (fabs (dx * dx + dy * dy + dz * dz) < sil_get_epsilon2 ())
      return;
    window = window->next ();
  }
  ++count;
  RangePoint *p = new RangePoint (x, y, z, head);
  head = p;
  return;
}

bool RangePoints::del (RangePoint *p)
{
  window = head;
  RangePoint *prev = NULL;
  while (window) {
    if (window == p) {
      if (prev)
	prev->next () = window->next();
      else
	head->next () = window->next();
      delete window;
      window = NULL;
      --count;
      return true;
    }
    window = window->next ();
  }
  return false;
}

void RangePoints::unite (RangePoints *rp)
{
  rp->apply_transformation ();
  apply_transformation ();
  RangePoint *p = rp->first ();
  RangePoint *pp = p;
  p = rp->next ();
  while (p) {
    pp = p;
    p = rp->next ();
  }
  pp->next () = head;
  head = rp->first ();
  count += rp->count;
  rp->head = rp->window = NULL;
  return;
}

bool RangePoints::first (double &x, double &y, double &z)
{
  window = head;
  if (window) {
    window->get (x, y, z);
    return true;
  }
  return false;
}

RangePoint *RangePoints::first ()
{
  window = head;
  return window;
}

bool RangePoints::next (double &x, double &y, double &z)
{
  if (window) {
    window = window->next ();
    if (window) {
      window->get (x, y, z);
      return true;
    }
  }
  return false;
}

RangePoint *RangePoints::next ()
{
  if (window)
    window = window->next ();
  return window;
}

RangePoint *RangePoints::replace_points (RangePoint *p) 
{
  apply_transformation ();
  RangePoint *old = head;
  window = NULL;
  head = p;
  count = 0;
  RangePoint *q = p;
  while (q) {
    ++count;
    q = q->next ();
  }
  return old;
}

void RangePoints::apply_transformation (bool normals)
{
  RangePoint *r = first ();
  while (r) {
    double x = r->x ();
    double y = r->y ();
    double z = r->z ();
    r->x () = tx +
      ( cos (rs) * cos (rp) - cos (rt) * sin (rp) * sin (rs)) * x +
      ( cos (rs) * sin (rp) + cos (rt) * cos (rp) * sin (rs)) * y +
      ( sin (rs) * sin (rt)                                 ) * z;
    r->y () = ty +
      (-sin (rs) * cos (rp) - cos (rt) * sin (rp) * cos (rs)) * x +
      (-sin (rs) * sin (rp) + cos (rt) * cos (rp) * cos (rs)) * y +
      ( cos (rs) * sin (rt)                                 ) * z;
    r->z () = tz + 
      ( sin (rt) * sin (rp)) * x +
      (-sin (rt) * cos (rp)) * y +
      ( cos (rt))            * z;
    if (normals) {
      x = r->nx ();
      y = r->ny ();
      z = r->nz ();
      r->set_normal 
	(( cos (rs) * cos (rp) - cos (rt) * sin (rp) * sin (rs)) * x +
	 ( cos (rs) * sin (rp) + cos (rt) * cos (rp) * sin (rs)) * y +
	 ( sin (rs) * sin (rt)                                 ) * z,
	 (-sin (rs) * cos (rp) - cos (rt) * sin (rp) * cos (rs)) * x +
	 (-sin (rs) * sin (rp) + cos (rt) * cos (rp) * cos (rs)) * y +
	 ( cos (rs) * sin (rt)                                 ) * z,
	 ( sin (rt) * sin (rp)) * x +
	 (-sin (rt) * cos (rp)) * y +
	 ( cos (rt))            * z);
    }
    r = next ();
  }
  tx = ty = tz = rt = rp = rs = 0.0;
  return;
}

void RangePoints::apply_transformation (double m1, double m2, double m3,
					double m4, double m5, double m6,
					double m7, double m8, double m9,
					double t1, double t2, double t3,
					bool normals)
{
  RangePoint *r = first ();
  while (r) {
    double x = r->x ();
    double y = r->y ();
    double z = r->z ();
    r->x () = t1 + m1 * x + m2 * y + m3 * z;
    r->y () = t2 + m4 * x + m5 * y + m6 * z;
    r->z () = t3 + m7 * x + m8 * y + m9 * z;
    if (normals) {
      x = r->nx ();
      y = r->ny ();
      z = r->nz ();
      r->set_normal (m1 * x + m2 * y + m3 * z, 
		     m4 * x + m5 * y + m6 * z,
		     m7 * x + m8 * y + m9 * z);
    }
    r = next ();
  }
  tx = ty = tz = rt = rp = rs = 0.0;
  return;
}

void RangePoints::centre ()
{
  if (count == 0)
    return;
  tx = ty = tz = 0.0;
  double x = 0.0;
  double y = 0.0;
  double z = 0.0;
  RangePoint *p = head;
  while (p) {
    x += p->x ();
    y += p->y ();
    z += p->z ();
    p = p->next ();
  }
  x /= (double)count;
  y /= (double)count;
  z /= (double)count;
  p = head;
  while (p) {
    p->set (p->x () - x, p->y () - y, p->z () - z);
    p = p->next ();
  }
  return;
}

bool RangePoints::save (string const &name)
{
  FILE *fp = fopen (name.c_str (), "w");
  if (fp == NULL)
    return false;
  RangePoint *r = first ();
  while (r) {
    double x = r->x ();
    double y = r->y ();
    double z = r->z ();
    double xx =
      ( cos (rs) * cos (rp) - cos (rt) * sin (rp) * sin (rs)) * x +
      ( cos (rs) * sin (rp) + cos (rt) * cos (rp) * sin (rs)) * y +
      ( sin (rs) * sin (rt)                                 ) * z;
    double yy =
      (-sin (rs) * cos (rp) - cos (rt) * sin (rp) * cos (rs)) * x +
      (-sin (rs) * sin (rp) + cos (rt) * cos (rp) * cos (rs)) * y +
      ( cos (rs) * sin (rt)                                 ) * z;
    double zz = 
      ( sin (rt) * sin (rp)) * x +
      (-sin (rt) * cos (rp)) * y +
      ( cos (rt))            * z;
    fprintf (fp, "%f %f %f\n", xx + tx, yy + ty, zz + tz);
    r = next ();
  }
  fclose (fp);
  return true;
}

RangeMesh::RangeMesh (RangePoints *rp, unsigned long nt)
{
  if (rp)
    npts = rp->num_points ();
  else
    npts = 0;
  if (npts > 0) {
    x_pts = new double[npts];
    y_pts = new double[npts];
    z_pts = new double[npts];
    double tx, ty, tz;
    rp->translation (tx, ty, tz);
    double rph, rt, rs;
    rp->rotation (rph, rt, rs);
    unsigned long l = 0;
    RangePoint *r = rp->first ();
    while (r) {
      double x = r->x ();
      double y = r->y ();
      double z = r->z ();
      double xx =
	( cos (rs) * cos (rph) - cos (rt) * sin (rph) * sin (rs)) * x +
	( cos (rs) * sin (rph) + cos (rt) * cos (rph) * sin (rs)) * y +
	( sin (rs) * sin (rt)                                   ) * z;
      double yy =
	(-sin (rs) * cos (rph) - cos (rt) * sin (rph) * cos (rs)) * x +
	(-sin (rs) * sin (rph) + cos (rt) * cos (rph) * cos (rs)) * y +
	( cos (rs) * sin (rt)                                   ) * z;
      double zz = 
	( sin (rt) * sin (rph)) * x +
	(-sin (rt) * cos (rph)) * y +
	( cos (rt))             * z;
      x_pts[l] = xx;
      y_pts[l] = yy;
      z_pts[l++] = zz;
      r = rp->next ();
    }
  } else {
    x_pts = y_pts = z_pts = NULL;
  }
  ntrgs = nt;
  if (nt > 0) {
    trg_v1 = new unsigned long[ntrgs];
    trg_v2 = new unsigned long[ntrgs];
    trg_v3 = new unsigned long[ntrgs];
  } else {
    trg_v1 = trg_v2 = trg_v3 = NULL;
  }
  n_x = n_y = n_z = NULL;
  px_pts = py_pts = pz_pts = NULL;
  id = NULL;
  nb1 = nb2 = nb3 = NULL;
}

RangeMesh::RangeMesh (unsigned long np, unsigned long nt)
{
  npts = np;
  if (npts > 0) {
    x_pts = new double[npts];
    y_pts = new double[npts];
    z_pts = new double[npts];
  } else {
    x_pts = y_pts = z_pts = NULL;
  }
  ntrgs = nt;
  if (nt > 0) {
    trg_v1 = new unsigned long[ntrgs];
    trg_v2 = new unsigned long[ntrgs];
    trg_v3 = new unsigned long[ntrgs];
    nb1 = new unsigned long[ntrgs];
    nb2 = new unsigned long[ntrgs];
    nb3 = new unsigned long[ntrgs];
  } else {
    trg_v1 = trg_v2 = trg_v3 = NULL;
    nb1 = nb2 = nb3 = NULL;
  }
  n_x = n_y = n_z = NULL;
  px_pts = py_pts = pz_pts = NULL;
  id = NULL;
}

RangeMesh::~RangeMesh ()
{
  reg.purge ();
  clas.purge ();
  if (x_pts)
    delete x_pts;
  if (y_pts)
    delete y_pts;
  if (z_pts)
    delete z_pts;
  if (px_pts)
    delete px_pts;
  if (py_pts)
    delete py_pts;
  if (pz_pts)
    delete pz_pts;
  if (trg_v1)
    delete trg_v1;
  if (trg_v2)
    delete trg_v2;
  if (trg_v3)
    delete trg_v3;
  if (n_x)
    delete n_x;
  if (n_y)
    delete n_y; 
  if (n_z)
    delete n_z;
  if (id)
    delete id;
  if (nb1)
    delete nb1;
  if (nb2)
    delete nb2;
  if (nb3)
    delete nb3;
}

void RangeMesh::init_ids ()
{
  if (id)
    delete id;
  id = new unsigned short[ntrgs];
  unsigned long l = 0;
  while (l < ntrgs) {
    id[l] = 0;
    ++l;
  }
  return;
}

void RangeMesh::init_normals ()
{
  if (n_x)
    delete n_x;
  if (n_y)
    delete n_y; 
  if (n_z)
    delete n_z;
  n_x = new double[npts];
  n_y = new double[npts];
  n_z = new double[npts];
  unsigned long l = 0;
  while (l < npts) {
    n_x [l] = 0.0;
    n_y [l] = 0.0;
    n_z [l] = 0.0;
    ++l;
  }
  return;
}

void RangeMesh::init_project ()
{
  if (px_pts)
    delete px_pts;
  if (py_pts)
    delete py_pts;
  if (pz_pts)
    delete pz_pts;
  px_pts = new double[npts];
  py_pts = new double[npts];
  pz_pts = new double[npts];
  unsigned long l = 0;
  while (l < npts) {
    px_pts[l] = x_pts[l];
    py_pts[l] = y_pts[l];
    pz_pts[l] = z_pts[l];
    ++l;
  }
  return;
}

void RangeMesh::init_neighbours ()
{
  if (nb1)
    delete nb1;
  if (nb2)
    delete nb2;
  if (nb3)
    delete nb3;

  nb1 = new unsigned long[ntrgs];
  nb2 = new unsigned long[ntrgs];
  nb3 = new unsigned long[ntrgs];

  // Create point - triangle lists (the elements of the lists are
  // numbers, not pointers!) sorted by which vertex of the triangle
  // is at the point.
  //
  // Allocating pointer arrays to the lists and allocating space for
  // each list individually seems to avoid most memory problems (as
  // long as there is enough memory available).
  SimpleList<unsigned long> **vtrg1 = new SimpleList<unsigned long>*[npts];
  SimpleList<unsigned long> **vtrg2 = new SimpleList<unsigned long>*[npts];
  SimpleList<unsigned long> **vtrg3 = new SimpleList<unsigned long>*[npts];
  unsigned long l = 0;
  while (l < npts) {
    vtrg1[l] = new SimpleList<unsigned long>;
    vtrg2[l] = new SimpleList<unsigned long>;
    vtrg3[l++] = new SimpleList<unsigned long>;
  }
  l = 0;
  sil_progress_bar_text ("Collecting Triangles");
  while (l < ntrgs) {
    sil_progress_bar_percentage (static_cast<double> (l) /
				 static_cast<double> (ntrgs << 1));
    nb1[l] = nb2[l] = nb3[l] = ntrgs;
    unsigned long nl = l + 1;
    vtrg1[trg_v1[l]]->add_first ((unsigned long *)nl);
    vtrg2[trg_v2[l]]->add_first ((unsigned long *)nl);
    vtrg3[trg_v3[l]]->add_first ((unsigned long *)nl);
    l = nl;
  }

  // Create neighbour information - check for each vertex all vertices
  // in the triangles and assign the neighbourhood information
  // accordingly.
  sil_progress_bar_text ("Generating Neighbourhood Information");
  unsigned long n, t, k;
  l = 0;
  while (l < npts) {
    sil_progress_bar_percentage (0.5 + static_cast<double> (l) /
				 static_cast<double> (npts << 1));
    // Vertex 1 of triangle at point cases:
    k = (unsigned long)vtrg1[l]->unlink_first ();
    while (k > 0) {
      --k;
      if (nb3[k] == ntrgs) {
	t = find_f_triangle (trg_v2[k], vtrg1[l], vtrg2[l], vtrg3[l],
			     l, n);
	if (n == 1) {
	  // l/trg_v1[k] to trg_v2[k] is trg_v1[t] to l/trg_v2[t]
	  nb3[k] = t;
	  nb3[t] = k;
	} else if (n == 2) {
	  // l/trg_v1[k] to trg_v2[k] is trg_v2[t] to l/trg_v3[t]
	  nb3[k] = t;
	  nb1[t] = k;
	} else if (n == 3) {
	    // l/trg_v1[k] to trg_v2[k] is trg_v3[t] to l/trg_v1[t]
	  nb3[k] = t;
	  nb2[t] = k;
	}
      }
      if (nb2[k] == ntrgs) {
	t = find_b_triangle (trg_v3[k], vtrg1[l], vtrg2[l], vtrg3[l],
			     l, n);
	if (n == 1) {
	  // trg_v3[k] to l/trg_v1[k] is trg_v1[t] to l/trg_v3[t]
	  nb2[k] = t;
	  nb2[t] = k;
	} else if (n == 2) {
	  // trg_v3[k] to l/trg_v1[k] is trg_v2[t] to l/trg_v1[t]
	  nb2[k] = t;
	  nb3[t] = k;
	} else if (n == 3) {
	  // trg_v3[k] to l/trg_v1[k] is trg_v3[t] to l/trg_v2[t]
	  nb2[k] = t;
	  nb1[t] = k;
	}
      }
      k = (unsigned long)vtrg1[l]->unlink_first ();
    }
    // Vertex 2 of triangle at point cases:
    k = (unsigned long)vtrg2[l]->unlink_first ();
    while (k > 0) {
      --k;
      t = find_f_triangle (trg_v3[k], vtrg1[l], vtrg2[l], vtrg3[l],
			   l, n);
      if (nb1[k] == ntrgs) {
	if (n == 1) {
	  // l/trg_v2[k] to trg_v3[k] is trg_v1[t] to l/trg_v2[t]
	  nb1[k] = t;
	  nb3[t] = k;
	} else if (n == 2) {
	  // l/trg_v2[k] to trg_v3[k] is trg_v2[t] to l/trg_v3[t]
	  nb1[k] = t;
	  nb1[t] = k;
	} else if (n == 3) {
	  // l/trg_v2[k] to trg_v3[k] is trg_v3[t] to l/trg_v1[t]
	  nb1[k] = t;
	  nb2[t] = k;
	}
      }
      if (nb3[k] == ntrgs) {
	t = find_b_triangle (trg_v1[k], vtrg1[l], vtrg2[l], vtrg3[l],
			     l, n);
	if (n == 1) {
	  // trg_v1[k] to l/trg_v2[k] is trg_v1[t] to l/trg_v3[t]
	  nb3[k] = t;
	  nb2[t] = k;
	} else if (n == 2) {
	  // trg_v1[k] to l/trg_v2[k] is trg_v2[t] to l/trg_v1[t]
	  nb3[k] = t;
	  nb3[t] = k;
	} else if (n == 3) {
	  // trg_v1[k] to l/trg_v2[k] is trg_v3[t] to l/trg_v2[t]
	  nb3[k] = t;
	  nb1[t] = k;
	}
      }
      k = (unsigned long)vtrg2[l]->unlink_first ();
    }
    // Vertex 3 of triangle at point cases:
    k = (unsigned long)vtrg3[l]->unlink_first ();
    while (k > 0) {
      --k;
      if (nb2[k] == ntrgs) {
	t = find_f_triangle (trg_v1[k], vtrg1[l], vtrg2[l], vtrg3[l],
			     l, n);
	if (n == 1) {
	  // l/trg_v3[k] to trg_v1[k] is trg_v1[t] to l/trg_v2[t]
	  nb2[k] = t;
	  nb3[t] = k;
	} else if (n == 2) {
	  // l/trg_v3[k] to trg_v1[k] is trg_v2[t] to l/trg_v3[t]
	  nb2[k] = t;
	  nb1[t] = k;
	} else if (n == 3) {
	  // l/trg_v3[k] to trg_v1[k] is trg_v3[t] to l/trg_v1[t]
	  nb2[k] = t;
	  nb2[t] = k;
	}
      }
      if (nb1[k] == ntrgs) {
	t = find_b_triangle (trg_v2[k], vtrg1[l], vtrg2[l], vtrg3[l],
			     l, n);
	if (n == 1) {
	  // trg_v2[k] to l/trg_v3[k] is trg_v1[t] to l/trg_v3[t]
	  nb1[k] = t;
	  nb2[t] = k;
	} else if (n == 2) {
	  // trg_v2[k] to l/trg_v3[k] is trg_v2[t] to l/trg_v1[t]
	  nb1[k] = t;
	  nb3[t] = k;
	} else if (n == 3) {
	  // trg_v2[k] to l/trg_v3[k] is trg_v3[t] to l/trg_v2[t]
	  nb1[k] = t;
	  nb1[t] = k;
	}
      }
      k = (unsigned long)vtrg3[l]->unlink_first ();
    }

    ++l;
  }

  l = 0;
  while (l < npts) {
    delete vtrg1[l];
    delete vtrg2[l];
    delete vtrg3[l++];
  }
  delete vtrg1;
  delete vtrg2;
  delete vtrg3;

  return;
}

unsigned long RangeMesh::find_f_triangle (unsigned long v,
					  SimpleList<unsigned long> *vtrg1,
					  SimpleList<unsigned long> *vtrg2,
					  SimpleList<unsigned long> *vtrg3,
					  unsigned long l,
					  unsigned long &n) const
{
  // Find triangle in vtrg list containing vertex v such that vertex l
  // is in forward direction.
  // Check triangles with vertex 1 at vertex l:
  unsigned long t = (unsigned long)vtrg1->first ();
  while (t > 0) {
    if (trg_v3[--t] == v) {
      n = 3;
      return t;
    }
    t = (unsigned long)vtrg1->next ();
  }
  // Check triangles with vertex 2 at vertex l:
  t = (unsigned long)vtrg2->first ();
  while (t > 0) {
    if (trg_v1[--t] == v) {
      n = 1;
      return t;
    }
    t = (unsigned long)vtrg2->next ();
  }
  // Check triangles with vertex 3 at vertex l:
  t = (unsigned long)vtrg3->first ();
  while (t > 0) {
    if (trg_v2[--t] == v) {
      n = 2;
      return t;
    }
    t = (unsigned long)vtrg3->next ();
  }
  n = 0;
  return ntrgs;
}

unsigned long RangeMesh::find_b_triangle (unsigned long v,
					  SimpleList<unsigned long> *vtrg1,
					  SimpleList<unsigned long> *vtrg2,
					  SimpleList<unsigned long> *vtrg3,
					  unsigned long l,
					  unsigned long &n) const
{
  // Find triangle in vtrg list containing vertex v such that vertex l
  // is in backward direction.
  // Check triangles with vertex 1 at vertex l
  unsigned long t = (unsigned long)vtrg1->first ();
  while (t > 0) {
    if (trg_v2[--t] == v) {
      n = 2;
      return t;
    }
    t = (unsigned long)vtrg1->next ();
  }
  // Check triangles with vertex 2 at vertex l
  t = (unsigned long)vtrg2->first ();
  while (t > 0) {
    if (trg_v3[--t] == v) {
      n = 3;
      return t;
    }
    t = (unsigned long)vtrg2->next ();
  }
  // Check triangles with vertex 3 at vertex l
  t = (unsigned long)vtrg3->first ();
  while (t > 0) {
    if (trg_v1[--t] == v) {
      n = 1;
      return t;
    }
    t = (unsigned long)vtrg3->next ();
  }
  n = 0;
  return ntrgs;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
