/* sil_getopt.c: GNU's internal getopt function.
 *
 * Copyright (C) 1987, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97
 *      Free Software Foundation, Inc.
 * Copyright (C) 1998-2001 Frank Langbein <frank@langbein.org>
 *
 * This file has been modified by Frank Langbein for for the use in Sil.
 * It's a combination of the original getopt.c and getopt1.c especially
 * modified for Sil. We only use this version of getopt if libc doesn't
 * provide a getopt_long function. It has been taken from the GNU C Library.
 * Its master source is NOT part of the C library, however. The master
 * source lives in /gd/gnu/lib.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: sil_getopt.c 1 2005-08-10 23:35:07Z frank $
 */

#include "config.h"

RCSID (sil_getopt,
       "$Id: sil_getopt.c 1 2005-08-10 23:35:07Z frank $");

#include "sil_getopt.h"

/* This version of `getopt' appears to the caller like standard Unix
 * `getopt' but it behaves differently for the user, since it allows
 * the user to intersperse the options with the other arguments.
 * 
 * As `getopt' works, it permutes the elements of ARGV so that, when
 * it is done, all the options precede everything else. Thus all
 * application programs are extended to handle flexible argument
 * order.
 *
 * Setting the environment variable POSIXLY_CORRECT disables
 * permutation.  Then the behavior is completely standard.
 *
 * GNU application programs can use a third alternative mode in which
 * they can distinguish the relative order of options and other
 * arguments. */

/* Macro to test if the current command line argument is not an
 * option. */
#ifndef NONOPTION_P
#define NONOPTION_P \
    (argv[sil_optind][0] != '-' || argv[sil_optind][1] == '\0')
#endif /* !NONOPTION_P */

/* Declaration of internal functions. */
static void exchange (char **);
static const char *_sil_getopt_initialize (int, char *const *, const char *);

/* If the strchr function is available use it as my_index, otherwise
 * we have to use our own strchr function (called my_index). */
#ifdef HAVE_STRCHR
#define	my_index strchr
#else /* !HAVE_STRCHR */
static char *
my_index (str, chr)
     const char *str;
     int chr;
{
  while (*str) {
    if (*str == chr)
      return (char *) str;
    str++;
  }
  return NULL;
}
#endif /* !HAVE_STRCHR */

/* For communication from `getopt' to the caller.
 * When `getopt' finds an option that takes an argument, the argument
 * value is returned here. Also, when `ordering' is RETURN_IN_ORDER,
 * each non-option ARGV-element is returned here. */
char *sil_optarg = NULL;

/* Index in ARGV of the next element to be scanned.
 * This is used for communication to and from the caller and for
 * communication between successive calls to `sil_getopt'.
 *
 * On entry to `getopt', zero means this is the first call; initialize.
 *
 * When `sil_getopt' returns -1, this is the index of the first of
 * the non-option elements that the caller should itself scan.
 *
 * Otherwise, `sil_optind' communicates from one call to the next
 * how much of ARGV has been scanned so far.
 *
 * 1003.2 says this must be 1 before any call. */
int sil_optind = 1;

/* Formerly, initialization of getopt depended on sil_optind == 0,
 * which causes problems with re-calling getopt as programs generally
 * don't know that. */
int __sil_getopt_initialized = 0;

/* The next char to be scanned in the option-element in which the last
 * option character we returned was found. This allows us to pick up
 * the scan where we left off.
 *
 * If this is zero, or a null string, it means resume the scan by
 * advancing to the next ARGV-element. */
static char *sil_nextchar;

/* Callers store zero here to inhibit the error message for
 * unrecognized options. */
int sil_opterr = 1;

/* Set to an option character which was unrecognized. This must be
 * initialized on some systems to avoid linking in the system's own
 * sil_getopt implementation. */
int sil_optopt = '?';

/* Describe how to deal with options that follow non-option
 * ARGV-elements.
 *
 * If the caller did not specify anything, the default is
 * REQUIRE_ORDER if the environment variable POSIXLY_CORRECT is
 * defined, PERMUTE otherwise.
 *
 * REQUIRE_ORDER means don't recognize them as options; stop option
 * processing when the first non-option is seen.This is what Unix
 * does.  This mode of operation is selected by either setting the
 * environment variable POSIXLY_CORRECT, or using `+' as the first
 * character of the list of option characters.
 *
 * PERMUTE is the default.  We permute the contents of ARGV as we
 * scan, so that eventually all the non-options are at the end.  This
 * allows options to be given in any order, even with programs that
 * were not written to expect this.
 *
 * RETURN_IN_ORDER is an option available to programs that were
 * written to expect options and other ARGV-elements in any order and
 * that care about the ordering of the two.  We describe each
 * non-option ARGV-element as if it were the argument of an option
 * with character code 1. Using `-' as the first character of the list
 * of option characters selects this mode of operation.
 *
 * The special argument `--' forces an end of option-scanning
 * regardless of the value of `ordering'.  In the case of
 * RETURN_IN_ORDER, only `--' can cause `sil_getopt' to return -1
 * with `sil_optind' != ARGC. */
static enum {
  REQUIRE_ORDER, PERMUTE, RETURN_IN_ORDER
} ordering;

/* Value of POSIXLY_CORRECT environment variable. */
static char *posixly_correct;

/* Handle permutation of arguments. */

/* Describe the part of ARGV that contains non-options that have been
 * skipped. `first_nonopt' is the index in ARGV of the first of them;
 * `last_nonopt' is the index after the last of them. */
static int first_nonopt;
static int last_nonopt;

/* Exchange two adjacent subsequences of ARGV.
 * One subsequence is elements [first_nonopt,last_nonopt) which
 * contains all the non-options that have been skipped so far. The
 * other is elements [last_nonopt,sil_optind), which contains all
 * the options processed since those non-options were skipped.
 *
 * `first_nonopt' and `last_nonopt' are relocated so that they
 * describe the new indices of the non-options in ARGV after they are
 * moved. */
static void exchange (argv)
     char **argv;
{
  int bottom = first_nonopt;
  int middle = last_nonopt;
  int top = sil_optind;
  int len;
  register int i;
  char *tem;

  /* Exchange the shorter segment with the far end of the longer
   * segment. That puts the shorter segment into the right place. It
   * leaves the longer segment in the right place overall, but it
   * consists of two parts that need to be swapped next. */
  while (top > middle && middle > bottom) {
    if (top - middle > middle - bottom) {
      /* Bottom segment is the short one. */
      len = middle - bottom;
      /* Swap it with the top part of the top segment. */
      for (i = 0; i < len; i++) {
	tem = argv[bottom + i];
	argv[bottom + i] = argv[top - (middle - bottom) + i];
	argv[top - (middle - bottom) + i] = tem;
      }
      /* Exclude the moved bottom segment from further swapping. */
      top -= len;
    } else {
      /* Top segment is the short one. */
      len = top - middle;
      /* Swap it with the bottom part of the bottom segment. */
      for (i = 0; i < len; i++) {
	tem = argv[bottom + i];
	argv[bottom + i] = argv[middle + i];
	argv[middle + i] = tem;
      }
      /* Exclude the moved top segment from further swapping. */
      bottom += len;
    }
  }

  /* Update records for the slots the non-options now occupy. */
  first_nonopt += (sil_optind - last_nonopt);
  last_nonopt = sil_optind;

  return;
}

/* Initialize the internal data when the first call is made. */
static const char *_sil_getopt_initialize (argc, argv, optstring)
     int argc;
     char *const *argv;
     const char *optstring;
{
  /* Start processing options with ARGV-element 1 (since ARGV-element
   * 0 is the program name); the sequence of previously skipped
   * non-option ARGV-elements is empty. */
  first_nonopt = last_nonopt = sil_optind = 1;
  sil_nextchar = NULL;
  posixly_correct = getenv ("POSIXLY_CORRECT");

  /* Determine how to handle the ordering of options and
   * non-options. */
  if (optstring[0] == '-') {
    ordering = RETURN_IN_ORDER;
    ++optstring;
  } else if (optstring[0] == '+') {
    ordering = REQUIRE_ORDER;
    ++optstring;
  } else if (posixly_correct != NULL) {
    ordering = REQUIRE_ORDER;
  } else {
    ordering = PERMUTE;
  }

  return optstring;
}

/* Scan elements of ARGV (whose length is ARGC) for option characters
 * given in OPTSTRING.
 *
 * If an element of ARGV starts with '-', and is not exactly "-" or
 * "--", then it is an option element.  The characters of this element
 * (aside from the initial '-') are option characters.  If
 * `sil_getopt' is called repeatedly, it returns successively each
 * of the option characters from each of the option elements.
 * 
 * If `sil_getopt' finds another option character, it returns that
 * character, updating `sil_optind' and `sil_nextchar' so that the
 * next call to `getopt' can resume the scan with the following option
 * character or ARGV-element.
 *
 * If there are no more option characters, `getopt' returns -1. Then
 * `sil_optind' is the index in ARGV of the first ARGV-element that
 * is not an option. (The ARGV-elements have been permuted so that
 * those that are not options now come last.)
 *
 * OPTSTRING is a string containing the legitimate option
 * characters. If an option character is seen that is not listed in
 * OPTSTRING, return '?' after printing an error message.  If you set
 * `sil_opterr' to zero, the error message is suppressed but we
 * still return '?'.
 *
 * If a char in OPTSTRING is followed by a colon, that means it wants
 * an arg, so the following text in the same ARGV-element, or the text
 * of the following ARGV-element, is returned in `sil_optarg'.  Two
 * colons mean an option that wants an optional arg; if there is text
 * in the current ARGV-element, it is returned in `sil_optarg',
 * otherwise `sil_optarg' is set to zero.
 *
 * If OPTSTRING starts with `-' or `+', it requests different methods
 * of handling the non-option ARGV-elements. See the comments about
 * RETURN_IN_ORDER and REQUIRE_ORDER, above.
 *
 * Long-named options begin with `--' instead of `-'. Their names may
 * be abbreviated as long as the abbreviation is unique or is an exact
 * match for some defined option.  If they have an argument, it
 * follows the option name in the same ARGV-element, separated from
 * the option name by a `=', or else the in next ARGV-element. When
 * `getopt' finds along-named option, it returns 0 if that option's
 * `flag' field is nonzero, the value of the option's `val' field if
 * the `flag' field is zero.
 *
 * The elements of ARGV aren't really const, because we permute
 * them. But we pretend they're const in the prototype to be
 * compatible with other systems.
 *
 * LONGOPTS is a vector of `struct sil_option' terminated by an
 * element containing a name which is zero.
 *
 * LONGIND returns the index in LONGOPT of the long-named option
 * found. It is only valid when a long-named option has been found by
 * the most recent call.
 *
 * If LONG_ONLY is nonzero, '-' as well as '--' can introduce
 * long-named options. */
int _sil_getopt_internal (argc, argv, optstring, longopts, longind, long_only)
     int argc;
     char **argv;
     const char *optstring;
     const struct sil_option *longopts;
     int *longind;
     int long_only;
{
  sil_optarg = NULL;

  if (!__sil_getopt_initialized || sil_optind == 0) {
    optstring = _sil_getopt_initialize (argc, argv, optstring);
    sil_optind = 1; /* Don't scan ARGV[0], the program name. */
    __sil_getopt_initialized = 1;
  }
  if (sil_nextchar == NULL || *sil_nextchar == '\0') {
    /* Advance to the next ARGV-element. */

    /* Give FIRST_NONOPT & LAST_NONOPT rational values if SIL_OPTIND
     * has been moved back by the user (who may also have changed the
     * arguments). */
    if (last_nonopt > sil_optind)
      last_nonopt = sil_optind;
    if (first_nonopt > sil_optind)
      first_nonopt = sil_optind;

    if (ordering == PERMUTE) {
      /* If we have just processed some options following some
       * non-options, exchange them so that the options come first. */
      if (first_nonopt != last_nonopt && last_nonopt != sil_optind)
	exchange ((char **) argv);
      else if (last_nonopt != sil_optind)
	first_nonopt = sil_optind;
      /* Skip any additional non-options and extend the range of
       * non-options previously skipped. */
      while (sil_optind < argc && NONOPTION_P)
	sil_optind++;
      last_nonopt = sil_optind;
    }
    /* The special ARGV-element `--' means premature end of
     * options. Skip it like a null option, then exchange with
     * previous non-options as if it were an option, then skip
     * everything else like a non-option. */
    if (sil_optind != argc && !strcmp (argv[sil_optind], "--")) {
      sil_optind++;

      if (first_nonopt != last_nonopt && last_nonopt != sil_optind)
	exchange ((char **) argv);
      else if (first_nonopt == last_nonopt)
	first_nonopt = sil_optind;
      last_nonopt = argc;

      sil_optind = argc;
    }
    /* If we have done all the ARGV-elements, stop the scan and back
     * over any non-options that we skipped and permuted. */
    if (sil_optind == argc) {
      /* Set the next-arg-index to point at the non-options that we
       * previously skipped, so the caller will digest them. */
      if (first_nonopt != last_nonopt)
	sil_optind = first_nonopt;
      return -1;
    }
    /* If we have come to a non-option and did not permute it, either
     * stop the scan or describe it to the caller and pass it by. */
    if (NONOPTION_P) {
      if (ordering == REQUIRE_ORDER)
	return -1;
      sil_optarg = argv[sil_optind++];
      return 1;
    }
    /* We have found another option-ARGV-element. Skip the initial
     * punctuation. */
    sil_nextchar = (argv[sil_optind] + 1
		      + (longopts != NULL && argv[sil_optind][1] == '-'));
  }
  /* Decode the current option-ARGV-element. */

  /* Check whether the ARGV-element is a long option.
   *
   * If long_only and the ARGV-element has the form "-f", where f is a
   * valid short option, don't consider it an abbreviated form of a
   * long option that starts with f.  Otherwise there would be no way
   * to give the -f short option.
   *
   * On the other hand, if there's a long option "fubar" and the
   * ARGV-element is "-fu", do consider that an abbreviation of the
   * long option, just like "--fu", and not "-f" with argument "u".
   *
   * This distinction seems to be the most useful approach. */
  if (longopts != NULL
      && (argv[sil_optind][1] == '-'
	  || (long_only && (argv[sil_optind][2]
		       || !my_index (optstring, argv[sil_optind][1]))))) {
    char *nameend;
    const struct sil_option *p;
    const struct sil_option *pfound = NULL;
    int exact = 0;
    int ambig = 0;
    int indfound = -1;
    int option_index;

    for (nameend = sil_nextchar; *nameend && *nameend != '='; nameend++);

    /* Test all long options for either exact match or abbreviated
     * matches. */
    for (p = longopts, option_index = 0; p->name; p++, option_index++) {
      if (!strncmp (p->name, sil_nextchar,
		    (unsigned int) (nameend - sil_nextchar))) {
	if ((unsigned int) (nameend - sil_nextchar)
	    == (unsigned int) strlen (p->name)) {
	  /* Exact match found. */
	  pfound = p;
	  indfound = option_index;
	  exact = 1;
	  break;
	} else if (pfound == NULL) {
	  /* First non-exact match found. */
	  pfound = p;
	  indfound = option_index;
	} else
	  /* Second or later non-exact match found. */
	  ambig = 1;
      }
    }

    if (ambig && !exact) {
      if (sil_opterr)
	fprintf (stderr, "%s: option `%s' is ambiguous\n",
		 argv[0], argv[sil_optind]);
      sil_nextchar += strlen (sil_nextchar);
      sil_optind++;
      sil_optopt = 0;
      return '?';
    }
    if (pfound != NULL) {
      option_index = indfound;
      sil_optind++;
      if (*nameend) {
	/* Don't test has_arg with >, because some C compilers don't
	 * allow it to be used on enums. */
	if (pfound->has_arg) {
	  sil_optarg = nameend + 1;
	} else {
	  if (sil_opterr) {
	    if (argv[sil_optind - 1][1] == '-') {
	      /* --option */
	      fprintf (stderr,
		       "%s: option `--%s' doesn't allow an argument\n",
		       argv[0], pfound->name);
	    } else {
	      /* +option or -option */
	      fprintf (stderr,
		       "%s: option `%c%s' doesn't allow an argument\n",
		       argv[0], argv[sil_optind - 1][0],
		       pfound->name);
	    }
	  }
	  sil_nextchar += strlen (sil_nextchar);
	  sil_optopt = pfound->val;
	  return '?';
	}
      } else if (pfound->has_arg == 1) {
	if (sil_optind < argc) {
	  sil_optarg = argv[sil_optind++];
	} else {
	  if (sil_opterr)
	    fprintf (stderr,
		     "%s: option `%s' requires an argument\n",
		     argv[0], argv[sil_optind - 1]);
	  sil_nextchar += strlen (sil_nextchar);
	  sil_optopt = pfound->val;
	  return optstring[0] == ':' ? ':' : '?';
	}
      }
      sil_nextchar += strlen (sil_nextchar);
      if (longind != NULL)
	*longind = option_index;
      if (pfound->flag) {
	*(pfound->flag) = pfound->val;
	return 0;
      }
      return pfound->val;
    }
    /* Can't find it as a long option.  If this is not getopt_long_only,
     * or the option starts with '--' or is not a valid short option, 
     * then it's an error. Otherwise interpret it as a short option. */
    if (!long_only || argv[sil_optind][1] == '-'
	|| my_index (optstring, *sil_nextchar) == NULL) {
      if (sil_opterr) {
	if (argv[sil_optind][1] == '-') {
	  /* --option */
	  fprintf (stderr, "%s: unrecognized option `--%s'\n",
		   argv[0], sil_nextchar);
	} else {
	  /* +option or -option */
	  fprintf (stderr, "%s: unrecognized option `%c%s'\n",
		   argv[0], argv[sil_optind][0], sil_nextchar);
	}
      }
      sil_nextchar = (char *) "";
      sil_optind++;
      sil_optopt = 0;
      return '?';
    }
  }

  /* Look at and handle the next short option-character. */
  {
    char c = *sil_nextchar++;
    char *temp = my_index (optstring, c);

    /* Increment `sil_optind' when we start to process its last
     * character. */
    if (*sil_nextchar == '\0')
      ++sil_optind;

    if (temp == NULL || c == ':') {
      if (sil_opterr) {
	if (posixly_correct) {
	  /* 1003.2 specifies the format of this message. */
	  fprintf (stderr, "%s: illegal option -- %c\n",
		   argv[0], c);
	} else {
	  fprintf (stderr, "%s: invalid option -- %c\n",
		   argv[0], c);
	}
      }
      sil_optopt = c;
      return '?';
    }
    /* Convenience. Treat POSIX -W foo same as long option --foo */
    if (temp[0] == 'W' && temp[1] == ';') {
      char *nameend;
      const struct sil_option *p;
      const struct sil_option *pfound = NULL;
      int exact = 0;
      int ambig = 0;
      int indfound = 0;
      int option_index;

      /* This is an option that requires an argument. */
      if (*sil_nextchar != '\0') {
	sil_optarg = sil_nextchar;
	/* If we end this ARGV-element by taking the rest as an arg,
	 * we must advance to the next element now. */
	sil_optind++;
      } else if (sil_optind == argc) {
	if (sil_opterr) {
	  /* 1003.2 specifies the format of this message. */
	  fprintf (stderr, "%s: option requires an argument -- %c\n",
		   argv[0], c);
	}
	sil_optopt = c;
	if (optstring[0] == ':') {
	  c = ':';
	} else {
	  c = '?';
	}
	return c;
      } else {
	/* We already incremented `sil_optind' once; increment it
	 * again when taking next ARGV-element as argument. */
	sil_optarg = argv[sil_optind++];
      }

      /* sil_optarg is now the argument, see if it's in the table of
       * longopts. */
      for (sil_nextchar = nameend = sil_optarg;
	   *nameend && *nameend != '='; nameend++);

      /* Test all long options for either exact match or abbreviated
       * matches. */
      for (p = longopts, option_index = 0; p->name; p++, option_index++) {
	if (!strncmp (p->name, sil_nextchar,
		      (unsigned int) (nameend - sil_nextchar))) {
	  if ((unsigned int) (nameend - sil_nextchar)
	      == strlen (p->name)) {
	    /* Exact match found. */
	    pfound = p;
	    indfound = option_index;
	    exact = 1;
	    break;
	  } else if (pfound == NULL) {
	    /* First non-exact match found. */
	    pfound = p;
	    indfound = option_index;
	  } else {
	    /* Second or later non-exact match found. */
	    ambig = 1;
	  }
	}
      }
      if (ambig && !exact) {
	if (sil_opterr)
	  fprintf (stderr, "%s: option `-W %s' is ambiguous\n",
		   argv[0], argv[sil_optind]);
	sil_nextchar += strlen (sil_nextchar);
	sil_optind++;
	return '?';
      }
      if (pfound != NULL) {
	option_index = indfound;
	if (*nameend) {
	  /* Don't test has_arg with >, because some C compilers don't
	   * allow it to be used on enums. */
	  if (pfound->has_arg) {
	    sil_optarg = nameend + 1;
	  } else {
	    if (sil_opterr)
	      fprintf (stderr,
		       "%s: option `-W %s' doesn't allow an argument\n",
		       argv[0], pfound->name);

	    sil_nextchar += strlen (sil_nextchar);
	    return '?';
	  }
	} else if (pfound->has_arg == 1) {
	  if (sil_optind < argc) {
	    sil_optarg = argv[sil_optind++];
	  } else {
	    if (sil_opterr)
	      fprintf (stderr,
		       "%s: option `%s' requires an argument\n",
		       argv[0], argv[sil_optind - 1]);
	    sil_nextchar += strlen (sil_nextchar);
	    return optstring[0] == ':' ? ':' : '?';
	  }
	}
	sil_nextchar += strlen (sil_nextchar);
	if (longind != NULL)
	  *longind = option_index;
	if (pfound->flag) {
	  *(pfound->flag) = pfound->val;
	  return 0;
	}
	return pfound->val;
      }
      sil_nextchar = NULL;
      return 'W'; /* Let the application handle it. */
    }
    if (temp[1] == ':') {
      if (temp[2] == ':') {
	/* This is an option that accepts an argument optionally. */
	if (*sil_nextchar != '\0') {
	  sil_optarg = sil_nextchar;
	  sil_optind++;
	} else {
	  sil_optarg = NULL;
	}
	sil_nextchar = NULL;
      } else {
	/* This is an option that requires an argument. */
	if (*sil_nextchar != '\0') {
	  sil_optarg = sil_nextchar;
	  /* If we end this ARGV-element by taking the rest as an arg,
	   * we must advance to the next element now. */
	  sil_optind++;
	} else if (sil_optind == argc) {
	  if (sil_opterr) {
	    /* 1003.2 specifies the format of this message. */
	    fprintf (stderr,
		     "%s: option requires an argument -- %c\n",
		     argv[0], c);
	  }
	  sil_optopt = c;
	  if (optstring[0] == ':') {
	    c = ':';
	  } else {
	    c = '?';
	  }
	} else {
	  /* We already incremented `sil_optind' once; increment it
	   * again when taking next ARGV-element as argument. */
	  sil_optarg = argv[sil_optind++];
	}
	sil_nextchar = NULL;
      }
    }
    return c;
  }
}

/*** Local Variables:
 *** exe-comm: "./sil"
 *** compile-command: "make"
 *** End:
 ***/
