/* sil_getopt.h: Declarations for GNU's getopt.
 *
 * Copyright (C) 1989,90,91,92,93,94,96,97 Free Software Foundation, Inc.
 * Copyright (C) 1998-2001 Frank Langbein <frank@langbein.org>
 *
 * This file has been taken from the GNU C Library and has been modified
 * by Frank Langbein. Its master source is NOT part of the C library,
 * however. The master source lives in /gd/gnu/lib.
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: sil_getopt.h 1 2005-08-10 23:35:07Z frank $
 */

#ifndef SIL_GETOPT_H
#define SIL_GETOPT_H 1

/* We only use our own getopt if a GNU getopt compatible function is not
 * provided by the system. Otherwise we set the sil functions and 
 * variables for getopt to the system ones. */

#if defined(HAVE_GETOPT_H) && defined(HAVE_GETOPT_LONG)

#include <getopt.h>

#define sil_optarg optarg
#define sil_optind optind
#define sil_opterr opterr
#define sil_optopt optopt
#define sil_option option
#define sil_getopt_long getopt_long

#else /* !(defined(HAVE_GETOPT_H) && defined(HAVE_GETOPT_LONG)) */

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/** Communication from `getopt' to the caller.
 *
 * When `sil_getopt' finds an option that takes an argument, the
 * argument value is returned here. Also, when `ordering' is
 * RETURN_IN_ORDER, each non-option ARGV-element is returned here. */
extern char *sil_optarg;

/** Index in ARGV of the next element to be scanned.
 *
 * This is used for communication to and from the caller and for
 * communication between successive calls to `sil_getopt'.
 *
 * On entry to `sil_getopt', zero means this is the first call;
 * initialize.
 *
 * When `sil_getopt' returns -1, this is the index of the first of
 * the non-option elements that the caller should scan itself.
 *
 * Otherwise, `sil_optind' communicates from one call to the next
 * how much of ARGV has been scanned so far. */
extern int sil_optind;

/** Getopt error messages flag.
 *
 * Callers store zero here to inhibit the error message `sil_getopt'
 * prints for unrecognized options. */
extern int sil_opterr;

/** Unrecognized option character.
 *
 * Set to an option character which was unrecognized. */
extern int sil_optopt;

/** Sil option.
 *
 * Describe the long-named options requested by the application.  The
 * LONG_OPTIONS argument to sil_getopt_long is a vector of `struct
 * sil_option' terminated by an element containing a name which is
 * zero. */
struct sil_option {
  /** Option string.
   *
   * This is the name of the long option. */
  const char *name;
  /** Option argument.
   *
   * The field `has_arg' is:
   * `no_argument'        if the option does not take an argument,
   * `required_argument'  if the option requires an argument,
   * `optional_argument'  if the option takes an optional argument.
   *
   * has_arg can't be an enum because some compilers complain about
   * type mismatches in all the code that assumes it is an int. */
  int has_arg;
  /** Option argument value.
   *
   * If the field `flag' is not NULL, it points to a variable that is
   * set to the value given in the field `val' when the option is
   * found, but left unchanged if the option is not found.
   *
   * To have a long-named option do something other than set an `int'
   * to a compiled-in constant, such as set a value from `sil_optarg',
   * set the option's `flag' field to zero and its `val' field to a
   * nonzero value (the equivalent single-letter option character, if
   * there is one).  For long options that have a zero `flag' field,
   * `sil_getopt' returns the contents of the `val' field. */
  int *flag;
  /** Option value.
   *
   * Value returned when the option is given in the command line. */
  int val;
};

/* Names for the values of the `has_arg' field of `struct
 * sil_option'. */
#define	no_argument	  0
#define required_argument 1
#define optional_argument 2

/** Internal Sil getopt function.
 *
 * This is the internal sil_getopt function. Don't call it directly,
 * but use the sil_getopt_long function. It does all the real work to
 * parse the command line options.
 *
 * @param argc       Number of arguments.
 * @param argv       Pointer to argument vector.
 * @param shortopts  String listing the short options.
 * @param longopts   Vector of `struct sil_options' for long options.
 * @param longind    If not NULL, it points to a variable which is
 *                   set to the index of the long option relative to
 *                   longopts.
 * @param long_only  Flag to indicated getopt_long_only.
 *
 * @return The option character when a short option is recognized;
 *         for a long option, return val if flag is NULL and 0 otherwise;
 *         `:' if there was a missing parameter; `?' for an unknown
 *         option character or EOF for the end of the option list. */
int _sil_getopt_internal (int argc, char **argv, const char *optstring,
			  const struct sil_option *longopts,
			  int *longind, int long_only);

/** Parse short and long command line options.
 *
 * This is the sil_getopt_long entry point for the internal sil_getopt
 * function. It calls _sil_getopt_internal for the getopt_long version
 * of getopt.
 *
 * The sil_getopt_long function parses the command line arguments. If
 * it is called repeatedly, it returns successively each of the option
 * characters from each of the option elements. It also accepts long
 * options, started out by two dashes. Long option names may be
 * abbreviated if the abbreviation is unique or is an exact match for
 * some defined option. It can also parse parameters for the options.
 *
 * @param argc       Number of arguments.
 * @param argv       Pointer to argument vector.
 * @param shortopts  String listing the short options.
 * @param longopts   Vector of `struct sil_options' for long options.
 * @param longind    If not NULL, it points to a variable which is
 *                   set to the index of the long option relative to
 *                   longopts.
 *
 * @return The option character when a short option is recognized;
 *         for a long option, return val if flag is NULL and 0 otherwise;
 *         `:' if there was a missing parameter; `?' for an unknown
 *         option character or EOF for the end of the option list. */
#define sil_getopt_long(argc, argv, options, long_options, opt_index) \
    _sil_getopt_internal(argc, argv, options, long_options, opt_index, 0)

#ifdef	__cplusplus
}
#endif /* __cplusplus */

#endif /* !(defined(HAVE_GETOPT_H) && defined(HAVE_GETOPT_LONG)) */

#endif /* !SIL_GETOPT_H */

/*** Local Variables:
 *** exe-comm: "./sil"
 *** compile-command: "make"
 *** End:
 ***/
