// symbols.cc: Symbols class implementation.
// 
// Copyright (C) 1999-2001 Frank Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: symbols.cc 1 2005-08-10 23:35:07Z frank $
//

#include "config.h"

RCSID (symbols, "$Id: symbols.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "app.hxx"
#include "gui.hxx"
#include "interface.hxx"
#include "symbols.hxx"
#include "module.hxx"
#include "modules.hxx"

Symbol::Symbol (char *n, void *d, char *t, bool v)
{
  nm = new char[strlen(n) + 1];
  strcpy (nm, n);
  dt = d;
  visible = v;
  module = sil_get_module (t);
  if (!module)
    module = sil_get_module ("unkown");
  module->add_symbol (dt, nm, visible);
  hilite_refs = new SimpleList<void>;
}

Symbol::~Symbol ()
{
  if (dt)
    module->del_object (dt);
  if (nm)
    delete nm;
  if (hilite_refs)
    delete hilite_refs;
}

void Symbol::show ()
{
  if (!visible) {
    module->show (dt);
    visible = true;
  }
  return;
}

void Symbol::hide ()
{
  if (visible) {
    module->hide (dt);
    visible = false;
  }
  return;
}

bool Symbol::is (const char *t) const
{
  return (strcmp (module->name (), t) == 0);
}

const char *Symbol::type () const
{
  return module->name ();
}

void Symbol::facet () const
{
  module->facet_object (dt);
  return;
}

void Symbol::draw () const
{
  module->draw_object (dt, *hilite_refs);
  return;
}

void Symbol::draw_fixed () const
{
  module->draw_fixed_object (dt, *hilite_refs);
  return;
}

Symbols::Symbols ()
{
  head = window = NULL;
  handler_head = NULL;
}

Symbols::~Symbols ()
{
  while (head) {
    window = head;
    head = head->next;
    delete window;
  }
  SymbolsUpdateHandler *w;
  while (handler_head) {
    w = handler_head->next;
    delete handler_head;
    handler_head = w;
  }
}

bool Symbols::add (char *name, void *data, char *type, bool visible)
{
  window = head;
  while (window) {
    if (strcmp (window->nm, name) == 0) 
      return false;
    window = window->next;
  }
  window = new Symbol (name, data, type, visible);
  window->next = head;
  head = window;
  return true;
}

const Symbol *Symbols::find (const char *name)
{
  window = head;
  while (window) {
    if (strcmp (window->nm, name) == 0) 
      return window;
    window = window->next;
  }
  return NULL;
}

void Symbols::del (const char *name)
{
  Symbol *prev = NULL;
  window = head;
  while (window) {
    if (strcmp (window->nm, name) == 0) {
      if (prev)
	prev->next = window->next;
      else
	head = window->next;
      delete window;
      updated (); // update here as some application might
                  // rely on the existance and it definitely causes problems
                  // when replacing a symbol
      return;
    }
    prev = window;
    window = window->next;
  }
  return;
}

void Symbols::clear ()
{
  window = head;
  while (window) {
    Symbol *nxt = window->next;
    delete window;
    window = nxt;
  }
  head = NULL;
  return;
}

Symbol *Symbols::first ()
{
  window = head;
  return window;
}

Symbol *Symbols::next ()
{
  if (window)
    window = window->next;
  return window;
}

void Symbols::updated ()
{
  App::instance ()->gui->update_tool_window ();
  SymbolsUpdateHandler *w = handler_head;
  while (w) {
    (*w->handler)();
    w = w->next;
  }
  return;
}

void Symbols::add_update_handler (void (*f)(void))
{
  SymbolsUpdateHandler *w = handler_head;
  SymbolsUpdateHandler *p = NULL;
  while (w) {
    if (w->handler == f) {
      ++w->use_counter;
      return;
    }
    p = w;
    w = w->next;
  }
  w = new SymbolsUpdateHandler;
  w->use_counter = 1;
  w->handler = f;
  w->next = handler_head;
  handler_head = w;
  return;
}

void Symbols::del_update_handler (void (*f)(void))
{
  SymbolsUpdateHandler *w = handler_head;
  SymbolsUpdateHandler *p = NULL;
  while (w) {
    if (w->handler == f) {
      --w->use_counter;
      if (w->use_counter == 0) {
	if (p)
	  p->next = w->next;
	else
	  handler_head = w->next;
	delete w;
      }
      return;
    }
    p = w;
    w = w->next;
  }
  return;
}

void Symbols::hilite (bool hi, char const *name, void *ref)
{
  if (ref) {
    window = head;
    while (window) {
      if (strcmp (window->nm, name) == 0) {
	window->hilite (hi, ref);
	return;
      }
      window = window->next;
    }
  }
  return;
}

/// Local Variables:
/// exe-comm: "./sil"
/// compile-command: "make"
/// End:
