// analyse_loops.cc: Analyse a SiL model / loops
//
// Copyright (C) 1999-2001 Frank C Langbein <frank@langbein.org>
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: analyse_loops.cc 1 2005-08-10 23:35:07Z frank $

#include "config.h"

RCSID (analyse_loops,
       "$Id: analyse_loops.cc 1 2005-08-10 23:35:07Z frank $");

#include "list.hxx"
#include "xstring.hxx"
#include "kernel/analyse.hxx"
#include "kernel/topology.hxx"
#include "kernel/geometry.hxx"

void Analyse::eval_loops ()
{
  if (loop_aspects.length () < 2)
    return;
  
  // Compute the Fourier transformation for each loop polygon.
  ListNode<Aspect> *an = loop_aspects.first ();
  while (an) {
    LoopPolygon *lp = dynamic_cast<LoopPolygon *> (an->data () -> feature());
    if (lp)
      lp->ft (loop_fourier_coefficients);
    an = an->next ();
  }

  // NOTE: The loop technique could be expanded. We could determine
  //       flat loops with non-straight edges, split general ploygons
  //       into flat sub-parts and in general split loops in interesting
  //       sub-parts which might be similar.

  // NOTE: It might be possible to determine tol_loop_dist from tol_angle_dist
  //       and tol_len_dist. However, a separate value appears to be more
  //       appropriate, esp. as a derived tol_loop_dist appears to depends on
  //       the number of vertices in the polygon. We might also just
  //       use tol_eq for tol_loop_dist. There are similar problems with a
  //       maximum tolerance for the tlerance levels.

  // Cluster polygon loops.
  sil_progress_bar_text ("Clustering " + as_string (loop_aspects.length ()) +
			 " polygonial loops");
  List<Cluster> *clusters = new List<Cluster>;
  cluster_aspects (&loop_aspects, tol_eq, tol_loop_dist, NAN, true, clusters);
  reg_list->add_first ((List<Regularity> *)clusters);

  // Find tolerances.
  List<Cluster> bcl;
  ListNode<Cluster> *cn = clusters->first ();
  while (cn) {
    bcl.add_first (cn->data ());
    cn = cn->next ();
  }
  SimpleList<Real> tols;
  tols.add_first (new Real (tol_loop_dist));
  cn = bcl.unlink_first ();
  while (cn) {
    if (cn->data () -> tolerance () > tol_loop_dist)
      tols.add_first (new Real (cn->data () -> tolerance ()));
    if (cn->data () -> sub_clusters ()) {
      Cluster *sc = cn->data () -> sub_clusters () -> first ();
      while (sc) {
	bcl.add_first (sc);
	sc = cn->data () -> sub_clusters () -> next ();
      }
    }
    delete cn;
    cn = bcl.unlink_first ();
  }

  tolerance_levels (clusters, &tols, tol_loop_dist, 0.0);

  // Check for regular clusters.
  List<Cluster> bcls;
  cn = clusters->first ();
  while (cn) {
    cn->data () -> get_all_base_clusters (&bcls);
    cn = cn->next ();
  }
  cn = bcls.first ();
  while (cn) {
    Cluster *c = cn->data ();

    // NOTE: If one of the Fourier coefficients is large compared to the
    //       others this suggests that the polygon is at least based (or
    //       exactly) a regular polygon based on the frequency of the
    //       Fourier coefficient. At this point it is unlcear how to use this
    //       later, but as we keep the coefficients it can always be exploited
    //       by a subsequent step.

    // NOTE: The ratios between the length of loops which are simlar
    //       might be interesting, but this information is also
    //       part of the edge length regularities (special length values
    //       are there are well).

    if (c->num_aspects () > 1) {
      // Loop cluster is regular if there is more than one polygon in
      // the loop.
      if (c->parent_cluster ())
	bcls.add_last_check (c->parent_cluster ());
    } else {
      // Loop is not regular, so collapse it.
      if (c->collapse ()) {
	// Cluster collpase, so remove it and it's parent to the list.
	bcls.add_last_check (c->parent_cluster ());
      } else {
	// c is top level cluster, so remove it, but add it's sub
	// clusters.
	clusters->del (c);
	if (c->sub_clusters ()) {
	  Cluster *sc = c->sub_clusters () -> unlink_first ();
	  while (sc) {
	    clusters->add_first (sc);
	    sc->parent_cluster () = NULL;
	    sc = c->sub_clusters () -> unlink_first ();
	  }
	}
      }
      delete c;

    }

    cn = cn->next ();
  }

  return;
}

/// Local Variables:
/// exe-comm: "../sil"
/// compile-command: "cd .. && make"
/// End:
